//  qgu.h
//
//  Qsas - GUI Utils Header File
//
//  Tony Allen
//  Mar 2010


#ifndef QSAS_GU_H_
#define QSAS_GU_H_

#define QGU_VERSION  "QGU V_2.4 [ Mar 2010]"

#include <string.h>
#include <vector>

#include <QObject>
#include <QCursor>
#include <QMenu>
#include <QApplication>
#include <QMessageBox>
#include <QClipboard>

#include "qt_version.h"
#include "qsas_version.h"
#include "Qdos.h"
#include "qsw_helpwindow.h"
#include "qdutil.h"

enum QguListType { 
	ListTypeEntry,
	ListTypeGroup,
	ListTypeNull 
};

/// \var  static const char * shift_cursor_xpm
/// \brief Global variable that defines the layout of an icon for the cursor when in shift or control selection
///
static const char * shift_cursor_xpm[] = {
"16 16 3 1",
" 	c None",
"X	c #000000000000",
".	c #FFFFFFFF0000",
"X.              ",
"XX.             ",
"XXX.            ",
"XXXX.           ",
"XXXXX.          ",
"XXXXXX.         ",
"XXXXXXX.        ",
"XXXXXXXX.       ",
"XX.XX....  .    ",
"X. .XX.    .X   ",
".  .XX.    .X   ",
"    .XX.....X...",
"    .XX. XXXXXXX",
"     .     .X   ",
"           .X   ",
"           .X   "};




/// \class QguWindow qgu.h
/// \brief This is the QguWLinput class
///
/// This class specialises the QMainWindow class which provies a main application window.
///
/// Main windows are most often used to provide menus, toolbars and a status bar around a large central widget.
///
/// It inherits QMainWindow which is a built-in QT3 class
///
/// \remarks {
/// This class includes Q_OBJECT which is a QT3 macro that is used in all sub-classes of QObject in order to
/// benefit from the QT's Meta object system which are signals and slots mechanisms for inter-object communication,
/// runtime type information and the dynamic property system }
class QguWindow:public QMainWindow
{
  Q_OBJECT

protected:
//     void dragEnterEvent(QDragEnterEvent *event);
public:

  QMenuBar *_menu_bar;      ///<  Menu bar of the current main window object
  
  QguWindow( QWidget *parent=0, const char *name=0 );
  /// \fn ~QguWindow()
  /// \brief Destructor of the QguWindow class
  /// \param     None
  /// \exception None
  /// \return    None
  ///
  ~QguWindow();
  QStatusBar* msgBar;       ///<  Message bar of the current main window object (also qsas window)

  /// \fn void save(Vostream&)
  /// \brief Pure virtual fuction to save the contents of this window (session/working list). Overridden in sub-classes
  /// \param     Vostream&
  /// \exception None
  /// \return    None
  ///
  virtual void save(Vostream&)=0;

  /// \fn void restore(Vistream&)
  /// \brief Pure virtual function to restore the contents of this window (session/working list). Overridden in sub-classes
  /// \param     Vistream&
  /// \exception None
  /// \return    None
  ///
  virtual void restore(Vistream&,bool replace=true)=0;

  /// \fn void restore(Vistream&)
  /// \brief  Virtual function to restore the contents of this window (session/working list). Overridden in sub-classes
  /// \param     QString
  /// \exception None
  /// \return    None
  ///
  virtual bool restoreXML(QString path, bool replace){return false;}

  /// \fn void saveAsXML(QString path)
  /// \brief  Virtual function to save the contents of this window (for session save). Overridden in sub-classes
  /// \param     QString
  /// \exception None
  /// \return    None
  ///
  virtual void saveAsXML(QString path){;}

   /// \fn void status(const QString& str)
  /// \brief Function that displays the message str within the current window message bar.
  /// \param     str of const QString& type
  /// \exception None
  /// \return    None
  ///
 void status(const QString& str);

  /// \fn void status(const QString& str,int ms)
  /// \brief Function that displays the message str within the current window message bar
  /// \param     str of const QString& type
  /// \param     ms  of int            type
  /// \exception None
  /// \return    None
  ///
  void status(const QString& str,int ms);
  
  /// \fn QMenuBar* menu_bar()
  /// \brief Function that displays the message str within the current window message bar for ms milli-seconds
  /// \param     str of const QString& type
  /// \exception None
  /// \return    None
  ///
  QMenuBar* menu_bar(){ return _menu_bar;}

  /// \fn QStatusBar* status_bar()
  /// \brief Function that returns the status bar pointer of the current main window object
  /// \param     None
  /// \exception None
  /// \return    None
  ///
  QStatusBar* status_bar(){return msgBar;}

  QString _name;
  QString _filename;        ///< Filename used to save the current session or working list content
  QString SAVE_EXTENSION;   ///< File extension used to save the current session or working list content
  bool fromSession;         ///< Flag set to denote that this window save was initiated by a 'save session'

  int with;        ///<  Public member defining how main window is being saved (saved with/without data)
  QString _saveset_filter;
  static DataNamesTable_var _saved_objects; ///< Map associating item names to QD object pointers for objects being saved
  static DataNamesTable_var _restored_objects; ///< Map associating item names to QD object pointers for objects being restored

public slots:
  virtual void save();
  virtual void saveWithOut();
  virtual void saveAs();
  virtual void saveWithOutAs();
  virtual void include();
  virtual void restore(bool replace=true);
  virtual void saveSession(QString);
  virtual void restoreSession(QString);
  void bringToFront();
};



# endif // QSAS_GU_H_
