/// \file qOperationAdd.cc
///  File in which QdAdd function that calculates the sum of QdObject: \n
///  \author Nora Bounaira
///  \version 2.0          \n
///  \date    2004

#include "qOperators.h"


/// \fn QdObject_var QdAdd( QString& rqstrErrorMsg, QdObject_var qdOperand1, QdObject_var qdOperand2 )
/// \brief Function that adds two QdObject.  This function will be called from the calculator when \n
///        add operator has been selected.  This function is used as function pointer
/// \param     rqstrErrorMsg of QString       type. Any error or warning message will be returned in this output parameter
/// \param     qdOperand1    of QdObject_var  type. The first operand of the operation
/// \param     qdOperand2    of QdObject_var  type. The second operand of the operation
/// \exception None.  Any exception thrown at this level will be caught and the exception text will be returned
/// \return    Returns the result of the operation as a QdObject_var on success and a NULL object on failure
///
QdObject_var QdAdd( QString& rqstrErrorMsg, QdObject_var qdOperand1, QdObject_var qdOperand2 )
{

    QdObject_var  qdResult;
    string        attrPtr;
    int           nQarStatus;


    if ( ( qdOperand1.is_nil() ) || ( qdOperand2.is_nil() ) )
    {
        // The object will be empty - this should never happen
		// The qdResult object is NULL at this point
		// Display error message within the window status
        return ( qdResult );
    }
    
 	// Test validity of inputs...
	nQarStatus =  QarTestConformal( qdOperand1, qdOperand2 );
	if ( nQarStatus == QAR_ERR )
	{
		QcalcAppendTextDisplay( "Sequence size or join error: Notify csc_support" );
		return qdResult;
	}
	
        // some validity tests are done in isOpAllowed() called by DoOperation() before (pairwise) joining

        if( QarAreUnitsSame( qdOperand1, qdOperand2) == QAR_ERR ){
	  // try to convert into same units 
	  try{
             QcalcAppendTextDisplay( "Trying to convert to same units" );
			 std::string SIcTarget = std::string(QuGetAttrText("SI_conversion", qdOperand1) );
             qdOperand2 = QarConvertUnitsTo(qdOperand2, SIcTarget.c_str() ); 
              // this creates new converted object, copies xrefs across and sets SI_conversion
          }
          catch(Exception &e){
               QcalcAppendTextDisplay( e.msg().c_str() );
               return qdResult;  
          }  
	}
	
        //   do addition
	try
	{
    	qdResult = qdOperand1 + qdOperand2;
	}
	catch ( Exception &e )
	{
		QcalcAppendTextDisplay ( e.msg().c_str() );
		QcalcAppendTextDisplay( "This Operation Add is not yet supported between these input types");
		rqstrErrorMsg = QString( "This Operation Add is not yet supported between these input types");
		return ( qdResult );
	}

    if ( qdResult.is_nil() )
    {
		 QcalcAppendTextDisplay( "This Add operation is not supported between these input types" );
         rqstrErrorMsg = QString( "This Add operation is not supported between these input types" );
         return ( qdResult );
     }

	// Add timeline (or DEPEND_0) : Either timeline is safe if it exists
	QdObject_var qd0 = get_Depend_0( qdOperand1 );
	if ( qd0.is_nil() ) qd0 = get_Depend_0( qdOperand2 );
	if ( !qd0.is_nil() ) qdResult->change_xref(DEPEND_0 , qd0 );


    // Set QSAS essential attriutes
    // for addition they stay the same as inputs

    // SI_conversion
    attrPtr = QuGetAttrText( "SI_conversion", qdOperand1 );
    if ( !attrPtr.empty() )
    {
         QuSetTxtAttr( "SI_conversion", attrPtr.c_str(), qdResult );
    }
    else
    {
        attrPtr = QuGetAttrText( "SI_conversion", qdOperand2 );
        QuSetTxtAttr( "SI_conversion", attrPtr.c_str(), qdResult );
    }

    // Frame
    QuSetFrame( qdResult, QuGetFrameAttr(  qdOperand1, qdOperand2 ).c_str() );
   
    // UNITS
    attrPtr = QuGetAttrText( "UNITS", qdOperand1 );
    if( !attrPtr.empty() )
    {
        QuSetTxtAttr( "UNITS", attrPtr.c_str(), qdResult );
    }
    else
    {
        attrPtr = QuGetAttrText( "UNITS", qdOperand2 );
        QuSetTxtAttr( "UNITS", attrPtr.c_str(), qdResult );
    }

	// LABLAXIS
	string labl_txt = QuGetAttrText( "LABLAXIS", qdOperand1 );
	QString labl_s = QString(labl_txt.c_str());
	labl_s += " + ";
	labl_s += QuGetAttrText( "LABLAXIS", qdOperand2 ).c_str();
	QuSetTxtAttr( "LABLAXIS", labl_s.toLatin1().data(), qdResult );


 	// DEPEND_i
    if (qdOperand1->xref_exists(DEPEND_1)) 
    {
        QdObject_var d1_xref = qdOperand1->get_xref(DEPEND_1);
        qdResult->change_xref( DEPEND_1, d1_xref  );
    }
    else if (qdOperand2->xref_exists(DEPEND_1))
    {
        QdObject_var d1_xref = qdOperand2->get_xref(DEPEND_1);
        qdResult->change_xref( DEPEND_1, d1_xref  );
    }

    if (qdOperand1->xref_exists(DEPEND_2)) 
    {
        QdObject_var d1_xref = qdOperand1->get_xref(DEPEND_2);
        qdResult->change_xref( DEPEND_2, d1_xref  );
    }
    else if (qdOperand2->xref_exists(DEPEND_2))
    {
        QdObject_var d1_xref = qdOperand2->get_xref(DEPEND_2);
        qdResult->change_xref( DEPEND_2, d1_xref  );
    }

    if (qdOperand1->xref_exists(DEPEND_3)) 
    {
        QdObject_var d1_xref = qdOperand1->get_xref(DEPEND_3);
        qdResult->change_xref( DEPEND_3, d1_xref  );
    }
    else if (qdOperand2->xref_exists(DEPEND_3))
    {
        QdObject_var d1_xref = qdOperand2->get_xref(DEPEND_3);
        qdResult->change_xref( DEPEND_3, d1_xref  );
    }


    return ( qdResult );
}

