#include "qcalc.h"
#include "qwui.h"

using namespace QSAS;

extern qwuiWindow* qsasWindow;



QcalcListView::QcalcListView( QWidget     *parent,
                              QcalcWindow *QjWindow,
			                  const char  *name )
   : QTreeView(parent)
{
	setDragEnabled(true);
	setAcceptDrops(true);
	setDropIndicatorShown(true);

	viewport()->setAcceptDrops(true);
	setDragDropMode(QAbstractItemView::DragDrop);
	
	setUniformRowHeights(true);
	setSelectionMode(QAbstractItemView::ExtendedSelection);
	
	header()->setStretchLastSection(false);
#if QT_VERSION < 0x040900
    header()->setResizeMode(QHeaderView::ResizeToContents);
#else
    header()->setSectionResizeMode(QHeaderView::ResizeToContents);
#endif
	
	this->owner = QjWindow;

	RButtonMenu = new QMenu( this );
	
	setAllColumnsShowFocus( true );

	connect(qsasWindow->wlview->wlmodel(), SIGNAL(created(const QString&)), this, SLOT(validateWLitems()));
	connect(qsasWindow->wlview->wlmodel(), SIGNAL(deleted(const QString&)), this, SLOT(validateWLitems()));
	connect(qsasWindow->wlview->wlmodel(), SIGNAL(pathChanged(const QString&, const QString&)), this, SLOT(validateWLitems()));
	connect( this,  SIGNAL( listModified()     ),                 owner,  SLOT( setTimeFeedback()   ) );


	// Set the private data members
	m_strAliseName   =  QString("");
	m_chChar         = 65;

	_model=new QcalcListModel(this);
	this->setModel(_model);
	this->setMinimumWidth(400);
	dropRow = 0;
}



QcalcListView::~QcalcListView()
{


}

void QcalcListView::showTab(QcalcTab* tab)
{
	owner->tabWidget->setCurrentIndex(owner->tabWidget->indexOf(tab));

}

void QcalcListView::dragEnterEvent(QDragEnterEvent *event)
{
	if(QuSlotDebug()) printf("QcalcListView::dragEnterEvent %ld\n", (long)this);

	QTreeView::dragEnterEvent(event);

	event->acceptProposedAction();
	activateWindow();
	raise();
	
}

void QcalcListView::dragMoveEvent(QDragMoveEvent *event)
{
	QModelIndex idx = indexAt(event->pos());

	dropRow = idx.row();
	
}

void QcalcListView::displayMenu(const QPoint& p)
{
	RButtonMenu->clear();
	
	RButtonMenu->addAction( "&Copy",          this, SLOT( copy() ), 				Qt::CTRL+Qt::Key_C );
	RButtonMenu->addAction( "Paste",          this, SLOT( paste() ) );
	RButtonMenu->addAction( "select &All",    this, SLOT( selectAll() ),                 Qt::CTRL+Qt::Key_A );
	RButtonMenu->addSeparator();
	
	RButtonMenu->addAction( "Copy Join Settings",  owner,  SLOT( keepChoices() ),           Qt::CTRL+Qt::Key_Left   );
	RButtonMenu->addAction( "Apply Join Settings", owner,  SLOT( applyChoices() ),          Qt::CTRL+Qt::Key_Right  );
	RButtonMenu->addSeparator();
	
	RButtonMenu->addAction( "Delete",         this, SLOT( deleteListItems() ),             Qt::CTRL+Qt::Key_X );
	RButtonMenu->addAction( "Clear List",     this, SLOT( deleteAllItems() )                             );
	RButtonMenu->addSeparator();
	
	RButtonMenu->addAction( "Move Down",      this, SLOT( moveItemDown() ) ,                Qt::CTRL+Qt::Key_Down);
	RButtonMenu->addAction( "Move Up",        this, SLOT( moveItemUp() ) ,                  Qt::CTRL+Qt::Key_Up);
	
	RButtonMenu->exec(p);
}


void QcalcListView::mousePressEvent(QMouseEvent* event)
{
	QTreeView::mousePressEvent(event);
	
	if(event->button()==Qt::RightButton)
	{
		displayMenu(event->globalPos());
	}
}


QdObject_var QcalcListView::getAliasResultObject(QString objAlias)
{
	QdObject_var nil_obj;
	QcalcListModel* model = calcModel();
	QcalcItem *qItem = model->_root->getChildByAlias( objAlias );
	if(qItem == NULL) {
		return nil_obj;
	}
	return qItem->getResultObject();
}


QcalcAliasType QcalcListView::getAliasObject(QString objAlias, QdObject_var dobj)
{
  dobj=0;
	QcalcListModel* model = calcModel();
	QcalcItem *qItem = model->_root->getChildByAlias( objAlias );
	
  if(qItem && qItem->m_eObjType == WORKING_LIST_OBJECT)
  {
    // Is this itself a WL object, else we can't get Metadata
   dobj = qItem->m_pcalcTab->operandStack[0]->slot->getSlotObject();
   if(dobj.is_nil())
     return ALIAS_NOT_FOUND;
   else
     return ALIAS_FOUND;
  }
  else
    return ALIAS_VIRTUAL;
}

void QcalcListView::getAliasArrDim(QString objAlias,Dim_var &dim)
{
  dim=0;
	QcalcListModel* model = calcModel();
	QcalcItem *qItem = model->_root->getChildByAlias( objAlias );

  if(qItem) dim =  qItem->getVirtualArrDim();
}

QcalcAliasObjType QcalcListView::getAliasObjectType(QString objAlias)
{
	QcalcListModel* model = calcModel();
	QcalcItem *qItem = model->_root->getChildByAlias( objAlias );
	
  if(qItem)
  {
    if(qItem->getIsVector())
      return ALIAS_IS_VECTOR;
    else if (qItem->getIsArray())
      return ALIAS_IS_ARRAY;
    else
      return ALIAS_IS_OTHER;
  }
  else
    return ALIAS_IS_UNKNOWN;
}

void QcalcListView::moveItemUp()
{
    QModelIndex current=currentIndex();
	clearSelection();
	calcModel()->moveUp(current);

}

void QcalcListView::moveItemDown()
{

    QModelIndex current=currentIndex();
	clearSelection();
	calcModel()->moveDown(current);

	
}


void QcalcListView::copy()
{
	QModelIndexList indexes=selectionModel()->selectedIndexes();
	
	QMimeData* data=calcModel()->mimeData(indexes);
	
	QApplication::clipboard()->setMimeData(data);
}

void QcalcListView::paste()
{

	QModelIndex parent=currentIndex();
	if(!parent.isValid()) return;

	calcModel()->dropMimeData(QApplication::clipboard()->mimeData(), Qt::MoveAction, parent.row(), 0, parent);
}

QcalcItem * QcalcListView::findAlias(QString alias)
{
	QcalcListModel* model = calcModel();
	QcalcItem *qItem = model->_root->getChildByAlias( alias );

	return qItem;
}
									 
QcalcItem * QcalcListView::findName(QString name)
{
	QcalcListModel* model = calcModel();
	QcalcItem *qItem = model->_root->getChildByName( name );
	return qItem;
}
																		  
QList<QcalcItem*> QcalcListView::listItems()
{
	QList<QcalcItem*> items;
	items.clear();
	QList <QcalcItem*> childList = calcModel()->_root->_children;
	for(int i=0; i<childList.size(); i++)
	{
		items.push_back(dynamic_cast<QcalcItem*>(childList[i]));
	}
	return items;
}

									
QList<QcalcItem*> QcalcListView::selectedItems()
{
	QModelIndexList indexes=selectionModel()->selectedRows();
	
	QList<QcalcItem*> res;
	res.clear();
	
	foreach(QModelIndex index, indexes)
	{
		QcalcItem* item=calcModel()->itemFromIndex(index);
		if(item!=NULL && !res.contains(item))
		{
			res.push_back(item);
		}
	}
	
    QModelIndex current=currentIndex();
    QcalcItem* currentItem=calcModel()->itemFromIndex(current);
    if(!res.contains(currentItem))
        res.push_back(currentItem);
    
	return res;
}

// ----------------------------------------------------

//            QcalcListModel

//-----------------------------------------------------

QcalcListModel::QcalcListModel(QcalcListView* view):
QAbstractItemModel((QObject*)view)
{
	
	this->_view=view;
	
	_root=new QcalcItem(view, "__root__", "");		

	
}

QcalcListModel::~QcalcListModel()
{
	delete _root;
}

// Minimum set of functions


QModelIndex QcalcListModel::index(int row, int column, const QModelIndex& parent) const
{
	if (!hasIndex(row, column, parent)) return QModelIndex();
	
	QcalcItem *parentItem;
	
	if (!parent.isValid())
		parentItem = _root;
	else
		parentItem = itemFromIndex(parent);
	
	QcalcItem *childItem = parentItem->getChildByNumber(row);
	if (childItem)
		return createIndex(row, column, childItem);
	else
		return QModelIndex();
}


QModelIndex QcalcListModel::parent(const QModelIndex& index) const
{
	// calc list is only one deep, so all parents are root
	return QModelIndex();
	
}

QModelIndex QcalcListModel::indexFromItem(QcalcItem* item, int column)
{
	if(item==NULL) return QModelIndex();
	if(item==_root) return _view->rootIndex();
	return createIndex(item->childNumber(), column, item);
}	


int QcalcListModel::rowCount(const QModelIndex& parent) const
{
	if (parent.isValid()) return 0; // calc items have no children
	
	if (parent.column() > 1) return 0; // off view
		
	return _root->numChildren(); // only root item has children
}

int QcalcListModel::columnCount(const QModelIndex&) const
{
	return 2;
}

QVariant QcalcListModel::data(const QModelIndex& index, int role) const
{
	if (!index.isValid()) return QVariant();
	
	if (role == Qt::DisplayRole)
	{	
		if(index.column()==0)
		{
			QcalcItem *item = itemFromIndex(index);
			return item->_name;
		}
		else if(index.column()==1)
		{	
			QcalcItem *item = itemFromIndex(index);
			return item->_alias;
		}
	}
	return QVariant();
	
}

QVariant QcalcListModel::headerData(int section, Qt::Orientation orientation, int role) const
{
	if(role!=Qt::DisplayRole) return QVariant();
	
	if(section==0) return QVariant("Object List");
	return QVariant("Alias");
	
}

// Editable

Qt::ItemFlags QcalcListModel::flags(const QModelIndex &index) const
{	
	QcalcItem* item=itemFromIndex(index);
	
	if(item==NULL || item==_root) 
	{
		return  Qt::ItemIsDropEnabled;
	}

	
	return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;
	
	
}

bool QcalcListModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
	if (index.isValid() && role == Qt::EditRole)
	{
		QcalcItem* item=itemFromIndex(index);
		if(item)
		{
			item->setName(value.toString().trimmed());
			emit dataChanged(index, index);
			return true;
		}
	}
	return false;
}

// Add/Remove rows/cols


bool QcalcListModel::insertColumns(int, int, const QModelIndex&)
{
	return false;
}

bool QcalcListModel::removeColumns(int, int, const QModelIndex&)
{
	return false;
}

bool QcalcListModel::removeRows(int position, int rows, const QModelIndex& parent)
{
	
	if(!parent.isValid()) return false;

	QcalcItem* parentItem=itemFromIndex(parent);
	
	if(parentItem==NULL) return false;

	beginRemoveRows(parent, position, position+rows-1);

	for (int row = 0; row < rows; ++row)
	{
		parentItem->removeChild(position, true);
	}

	endRemoveRows();
	return true;
	
}

// Drag and drop tools

QStringList QcalcListModel::mimeTypes() const
{
	QStringList types;
	types << "application/wlitems";
	types << "application/calcitems";
	types << "text/plain";
	return types;
}

QMimeData* QcalcListModel::mimeData(const QModelIndexList& indexes) const
{
	QMimeData *mimeData = new QMimeData();
	
	QStringList idList;
	
	foreach (QModelIndex index, indexes) {
		if (index.isValid()) {
			idList << itemFromIndex(index)->_alias;
		}
	}
	
	QString data=idList.join("//");
	
	mimeData->setData("application/calcitems", data.toLatin1().data());
	mimeData->setText(itemFromIndex(indexes.first())->_alias);
	return mimeData;
}

bool QcalcListModel::dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent)
{
	if (action == Qt::IgnoreAction)	return true;

	if(row == -1) row = _view->dropRow; // QT seems to give -1 if drop target has same parent as dragged item
	
	// handle drag from WL
	if(data->hasFormat("application/wlitems"))
	{

		QStringList itemList=QString(data->data("application/wlitems")).split("//");
		
		foreach(QString itemName, itemList)
		{
			if(itemName.contains("__root__") ) continue;
			
			QcalcItem *newItem = _view->insertEntry(itemName);

		}
		
		qApp->processEvents();
		return true;
	}
	
	
	if (!data->hasFormat("application/calcitems")) return false;
	
	if (column > 1) return false;
	

	QStringList itemList=QString(data->data("application/calcitems")).split("//");
	
	QcalcItem* currentItem;
	
	emitLayoutAboutToBeChanged();
		
	if(action==Qt::CopyAction) // What is labelled CopyAction is a move
	{
		// drag to a sibling gives row = -1, so must find target row

		fflush(stdout);

		_view->clearSelection();
		foreach(QString itemString, itemList)
		{
			currentItem=_root->getChildByAlias(itemString);
			if(currentItem!=NULL)
			{
				currentItem->moveTo(row);
			}
		}
	}	
	emitLayoutChanged();

	return true;
}

// Utils



QModelIndex QcalcListModel::moveUp(QModelIndex& index)
{
	if(!index.isValid()) return QModelIndex();
	
	QcalcItem* item=itemFromIndex(index);
	
	if(item!=NULL )
	{
		int row=item->childNumber();
		if(row>0)
		{
			item->moveTo(row-1);
		}
	}
	
	return createIndex(item->childNumber(), 0, item);
}

QModelIndex QcalcListModel::moveDown(QModelIndex& index)
{
	if(!index.isValid()) return QModelIndex();
	
	QcalcItem* item=itemFromIndex(index);
	
	if(item!=NULL )
	{
		int row=item->childNumber();
		if(row<_root->_children.size()-1)
		{
			item->moveTo(row+1);
		}
	}
	
	return createIndex(item->childNumber(), 0, item);
}


Qt::DropActions QcalcListModel::supportedDropActions() const
{
	return Qt::CopyAction|Qt::MoveAction;
}

bool QcalcListModel::deleteItem(QcalcItem* item)
{
	if(item==NULL || item==_root) return false;

	int row=item->childNumber();

	beginRemoveRows(indexFromItem(_root, 0), row, row);
	_root->removeChild(row, true);
	endRemoveRows();

	return true;
}




