#include "qcalcContext.h"
#include "qcalcSlots.h"
#include "qcalcCalcItem.h"
#include "qcalcInputItem.h"
#include "qcalcOperatorItem.h"
#include "qcalcOutputItem.h"
#include "qcalcExtractionItem.h"

TimeLineHandler timeLineHandler;

ContextWidget::ContextWidget(QWidget* parent, CalcView* view):
    QWidget(parent)
{	
//     setMinimumWidth(300);
    pluginView=new PluginContextWidget(this);
    outView=new OutputContextWidget(this, view);
    bbView=new BlackBoxContextWidget(this, view);
	defaultView=new DefaultContextWidget(this, view);
	selectionView=new SelectionContextWidget(this, view);
	extractorView=new ExtractorContextWidget(this);
	joinView=new JoinContextWidget(this);
	timeLineView=new TimeLineWidget(this);

	connect(view, SIGNAL(componentSelected(const char *, const char *)), extractorView, SLOT(userSetComponent(const char *, const char *)));
	
    QVBoxLayout* layout=new QVBoxLayout;

	tabWidget=new QTabWidget(this);
	tabWidget->addTab(defaultView, "New Item");
	tabWidget->addTab(selectionView, "Aspect");
	tabWidget->addTab(pluginView, "Description");
	tabWidget->addTab(bbView, "Black Box");
	tabWidget->addTab(extractorView, "Extract");
	tabWidget->addTab(outView, "Output");
	tabWidget->addTab(joinView, "Input/Join");
	tabWidget->addTab(timeLineView, "Time/Scalar Sequence");
	QList<QGraphicsItem*> empty;
	setView(empty);
	layout->addWidget(tabWidget);
    setLayout(layout);
}

void ContextWidget::setView(QList<QGraphicsItem*>& items)
{
	QWidget* current=tabWidget->currentWidget();
	tabWidget->clear();

	tabWidget->addTab(defaultView, "Insertion");
	tabWidget->addTab(timeLineView, "Timeline (or Scalar Seq)");
	
	if(!items.isEmpty())
	{
		// Add plugin view if and only if there is 1 plugin in the selection
		int counter=0;
		VOperatorItem* op=NULL;
		foreach(QGraphicsItem* item, items)
		{
			OperatorItem* tempItem=qgraphicsitem_cast<OperatorItem*>(item);
			if(tempItem!=NULL)
			{
				if(tempItem->parentItem->isPlugin())
				{
					op=tempItem->parentItem;
					++counter;
				}
			}
		}
		if(counter==1 && op!=NULL)
		{
			pluginView->setPluginInfo(op);
			tabWidget->addTab(pluginView, "Description");
		}
		
		counter=0;
		op=NULL;
		foreach(QGraphicsItem* item, items)
		{
			OperatorItem* tempItem=qgraphicsitem_cast<OperatorItem*>(item);
			if(tempItem!=NULL)
			{
				if(tempItem->parentItem->isBlackBox)
				{
					op=tempItem->parentItem;
					++counter;
				}
			}
		}
		if(counter==1 && op!=NULL)
		{
			bbView->setBlackBoxInfo(op);
			tabWidget->addTab(bbView, "Black Box");
		}
		updateGeometry();
		
		QList<VOutputItem*> outList;
		foreach(QGraphicsItem* item, items)
		{
			OutputItem* tempItem=qgraphicsitem_cast<OutputItem*>(item);
			if(tempItem!=NULL)
			{
				outList.push_back(tempItem->parentItem);
			}
		}
		timeLineView->outItems = outList;
		if(!outList.isEmpty())
		{
			outView->setOutputParameters(outList);
			tabWidget->addTab(outView, "Output");
			timeLineView->setTimeLineTab(outList);

		}
		
		QList<VExtractionItem*> extList;
		foreach(QGraphicsItem* item, items)
		{
			ExtractionItem* tempItem=qgraphicsitem_cast<ExtractionItem*>(item);
			if(tempItem!=NULL)
			{
				extList.push_back(tempItem->parentItem);
			}
		}
		if(!extList.isEmpty())
		{
			extractorView->setExtractorParameters(extList);
			tabWidget->addTab(extractorView, "Extractor");
		}
		
		QList<VInputItem*> inputs;
		QList<VJoinItem*> joins;
		foreach(QGraphicsItem* item, items)
		{
			InputItem* tempItem=qgraphicsitem_cast<InputItem*>(item);
			if(tempItem!=NULL)
			{
				inputs.push_back(tempItem->parentItem);
			}
			else
			{
				JoinItem* tempItem2=qgraphicsitem_cast<JoinItem*>(item);
				if(tempItem2!=NULL)
				{
					joins.push_back(tempItem2->parentItem);
				}
			}
		}
		// if empty these are reset to empty (yes, both belong to joinView)
		joinView->inputItems=inputs;
		joinView->joinItems=joins;
		if(!inputs.isEmpty() || !joins.isEmpty())
		{
			joinView->setJoinParameters(inputs, joins);
			tabWidget->addTab(joinView, "Input/Join");

		}
		

		
		tabWidget->addTab(selectionView, "Selection");
	}
	
	int index=tabWidget->indexOf(current);
	if(index>=0)
	{
		tabWidget->setCurrentWidget(current);
	}
	updateGeometry();
}


void ContextWidget::newInput(const QString& text)
{
	TimeLine* tl=timeLineHandler.getDefaultTimeLine(); 
	if(!tl->regularTT && tl->TTString.isEmpty() && text.compare("scalar") != 0 )
	{	
		tl->TTString=text;
		timeLineView->refreshDefault();
	}
	
	joinView->checkSequence(text);
	
}

//-------------------------------

TimeLineWidget::TimeLineWidget(QWidget* parent)
{
	outItems.clear();

	QVBoxLayout * mainLayout=new QVBoxLayout;
	
	tlBox=new QComboBox(this);
	
	mainLayout->addWidget(tlBox);
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}
	connect(tlBox, SIGNAL(currentIndexChanged(const QString&)), this, SLOT(newSelection(const QString&)));
	
	QHBoxLayout* hLayout=new QHBoxLayout;
	
	deleteButton=new QPushButton("Delete");
	connect(deleteButton, SIGNAL(clicked()), this, SLOT(deleteSelected()));
	hLayout->addWidget(deleteButton);
	deleteButton->setEnabled(false);
	
	QPushButton* saveChangesButton=new QPushButton("Apply to Output");
	connect(saveChangesButton, SIGNAL(clicked()), this, SLOT(setTimeLine()));
	hLayout->addWidget(saveChangesButton);
    saveChangesButton->setToolTip("Use this timeline for selected outputs");
	
	QPushButton* saveAsButton=new QPushButton("Create new...");
	connect(saveAsButton, SIGNAL(clicked()), this, SLOT(saveAs()));
	hLayout->addWidget(saveAsButton);
	
	connect(&timeLineHandler, SIGNAL(addedTimeLine(const QString&)), this, SLOT(addedTimeLine(const QString&)));
	connect(&timeLineHandler, SIGNAL(deletedTimeLine(const QString&)), this, SLOT(deletedTimeLine(const QString&)));
	
	mainLayout->addLayout(hLayout);
	
	QGroupBox* time_widget=new QGroupBox("Sequence Dependency (Time or Scalar Sequence)");
    
    QGridLayout * time_layout=new QGridLayout(time_widget);
    time_layout->setSpacing(4);
    time_layout->setColumnStretch(1, 10);
    
    QLabel* TimeInterval_label = new QLabel("Interval ",time_widget);
    time_layout->addWidget( TimeInterval_label, 0, 0, Qt::AlignRight );
        
    _TimeIvlSelect = new QComboBox(time_widget);
    _TimeIvlSelect->insertItem(0, "from Data");
    _TimeIvlSelect->insertItem(1, "from Object");
    time_layout->addWidget( _TimeIvlSelect, 0, 1, Qt::AlignLeft );

        
    IvlSource_label = new QLabel("  ",time_widget);
    time_layout->addWidget( IvlSource_label, 1, 0, Qt::AlignRight );
        
    _IvlSource = new QguWLinput( time_widget, "IvlSource" );
	_IvlSource->setAcceptable(Qcondition(QS_NARROW, QS_TI, QS_0D, QS_NOT_SEQ, QS_ANY));
	_IvlSource->addAcceptable(Qcondition(QS_NARROW, QS_ANY, QS_ANY, QS_SEQ, QS_ANY));
    _IvlSource->active = false;
	_IvlSource->setMinimumWidth(400);
        
    _IvlSource->setToolTip("not active");
    time_layout->addWidget( _IvlSource, 1, 1, Qt::AlignLeft );
    

	msg = new QLabel(" ");
	time_layout->addWidget( msg, 2, 0, 1, 2, Qt::AlignLeft );
    
    QLabel* TimeTags_label = new QLabel("Target Tags ",time_widget);
    time_layout->addWidget( TimeTags_label, 3, 0, Qt::AlignRight );
        
    _TimeTagSelect = new QComboBox(time_widget);
    _TimeTagSelect->insertItem(0, "Create Regular");
    _TimeTagSelect->insertItem(1, "from Object");
    time_layout->addWidget( _TimeTagSelect, 3, 1, Qt::AlignLeft );
    
        
    TagSource_label = new QLabel("Target Tag Object ",time_widget);
    time_layout->addWidget( TagSource_label, 4, 0, Qt::AlignRight );
        
    _TagSource = new QguWLinput( time_widget, "TagSource" );
	_TagSource->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_ANY, QS_SEQ, QS_ANY));
    _TagSource->addAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_SEQ, QS_ANY));
	_TagSource->setMinimumWidth(400);
    
    _TagSource->setToolTip("drag and drop object here");
    time_layout->addWidget( _TagSource, 4, 1, Qt::AlignLeft );

	connect(_TimeIvlSelect, SIGNAL(currentIndexChanged(int)), this, SLOT(setTTivl(int)));
	connect(_IvlSource, SIGNAL(edited_OK()), this, SLOT(saveChanges()));

	connect(_TimeTagSelect, SIGNAL(currentIndexChanged(int)), this, SLOT(setTTsource(int)));
	connect(_TagSource, SIGNAL(edited_OK()), this, SLOT(saveChanges()));
	

    mainLayout->addWidget(time_widget);
	
	setLayout(mainLayout);
	
	newSelection("Default");
}

void TimeLineWidget::setTimeLineTab(QList<VOutputItem*>& outList){

	// just show first timeline in use
	if(outList.size() > 0) {
		int bkpIndex=tlBox->findText(outList[0]->timeLineName);
		if(bkpIndex>=0) tlBox->setCurrentIndex(bkpIndex);
	}
	newSelection(outList[0]->timeLineName);
	
	msg->setText("");

}

void TimeLineWidget::refreshDefault()
{
	QString name=tlBox->currentText();
	if(name == "Default")
	{
		// refresh with new TT object name if default
		newSelection(name);
	}
}


void TimeLineWidget::newSelection(const QString& name)
{
	deleteButton->setEnabled(name!="Default");
	
	_TimeIvlSelect->blockSignals(true);
	_TimeTagSelect->blockSignals(true);
	_TagSource->blockSignals(true);
	_IvlSource->blockSignals(true);
	
	TimeLine* tl=timeLineHandler.getTimeLine(name);
	if(tl!=NULL)
	{
		_TimeIvlSelect->setCurrentIndex((tl->fromData?0:1));
		setTTivlTab((tl->fromData?0:1));
		_TimeTagSelect->setCurrentIndex((tl->regularTT?0:1));
		setTTsourceTab((tl->regularTT?0:1));
		_IvlSource->setText(tl->fromObjString);
		_TagSource->setText(tl->TTString);

	}
	_TimeIvlSelect->blockSignals(false);
	_TimeTagSelect->blockSignals(false);
	_TagSource->blockSignals(false);
	_IvlSource->blockSignals(false);

}

void TimeLineWidget::addedTimeLine(const QString& newName)
{
	tlBox->blockSignals(true);
	//clear
	tlBox->clear();
	//fill
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}

	tlBox->blockSignals(false);
	
	int bkpIndex=tlBox->findText(newName);
	if(bkpIndex>=0)
		tlBox->setCurrentIndex(bkpIndex);
	

}


void TimeLineWidget::setTimeLine()
{
	if(!outItems.isEmpty())
	{
		QString tName = tlBox->currentText();
		for(int i=0; i<outItems.size(); i++){
			outItems[i]->timeLineName = QString(tName);
		}
		msg->setText("");

	}
	else{
		msg->setText("Warning: No Output item selected");
	}

}


void TimeLineWidget::deletedTimeLine(const QString&)
{
	QString save=tlBox->currentText();
	tlBox->blockSignals(true);
	//clear
	tlBox->clear();
	//fill
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}
	tlBox->blockSignals(false);
	int bkpIndex=tlBox->findText(save);
	if(bkpIndex>=0)
		tlBox->setCurrentIndex(bkpIndex);
}

void TimeLineWidget::deleteSelected()
{
	QString name=tlBox->currentText();
	if(name!="Default")
	{
		timeLineHandler.deleteTimeLine(name);
	}
}

void TimeLineWidget::saveChanges()
{
	QString name=tlBox->currentText();

	TimeLine* tl=timeLineHandler.getTimeLine(name);
	if(tl!=NULL)
	{
		tl->fromData=(_TimeIvlSelect->currentIndex()==0);
		tl->regularTT=(_TimeTagSelect->currentIndex()==0);
		tl->fromObjString=_IvlSource->text();
		tl->TTString=_TagSource->text();
	}
}

void TimeLineWidget::saveAs()
{
	bool ok;
	QString text = QInputDialog::getText(this, tr("Enter name for Time/Scalar Dependency"),
                                          tr("New name:"), QLineEdit::Normal,
                                          "New_Name", &ok);
	text=text.trimmed();
	if (ok && !text.isEmpty())
	{
        QList<QString> names=timeLineHandler.getNames();
		while(names.contains(text))
		{
			text.append("A");
		}
		timeLineHandler.addTimeLine(text, (_TimeIvlSelect->currentIndex()==0),
									(_TimeTagSelect->currentIndex()==0),
									 _IvlSource->text(), _TagSource->text());
									 
		qApp->processEvents();
		
		int bkpIndex=tlBox->findText(text);
		if(bkpIndex>=0)
			tlBox->setCurrentIndex(bkpIndex);
	}
}


void TimeLineWidget::setTTivl(int i)
{
	setTTivlTab(i);
	saveChanges();
}

void TimeLineWidget::setTTivlTab(int i)
{
	if(i == 0)
	{
		// from data
		IvlSource_label->setText(" ");
		_IvlSource->setToolTip("not active");
		_IvlSource->active = false;
		_IvlSource->setText(" "); // force signal so that colour updates
		_IvlSource->setVisible(false);
	}
	else if(i == 1)
	{
		IvlSource_label->setText("Interval Object");
		_IvlSource->setToolTip("drag and drop object here");
		_IvlSource->active = true;
		_IvlSource->setVisible(true);
		_IvlSource->setText("");
		_IvlSource->reValidate(true);
	}
}

void TimeLineWidget::setTTsource(int i)
{
	setTTsourceTab(i);
	
	_TagSource->reValidate(true); // updates T display if valid
	saveChanges();
}

void TimeLineWidget::setTTsourceTab(int i)
{
	if(i == 0)
	{
		// create regular
		TagSource_label->setText("Spacing");
		_TagSource->setToolTip("Spacing between regular tags (time is in seconds)");
		_TagSource->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
	}
	else if(i == 1)
	{
		TagSource_label->setText("Tag Object");
		_TagSource->setToolTip("Uses DEPEND_0 of this object as target time or scalar sequence to join onto");
		_TagSource->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_ANY, QS_SEQ, QS_ANY));
	    _TagSource->addAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_SEQ, QS_ANY));
	}
	
}

BlackBoxContextWidget::BlackBoxContextWidget(QWidget* parent, CalcView* view):
    QScrollArea(parent)
{
	
    QWidget* mainWidget=new QWidget;
    blackBox=NULL;
    inputDescriptions.clear();
    outputDescriptions.clear();

    inputWidget=new QWidget;
    outputWidget=new QWidget;
    inputLayout=new QVBoxLayout;
    outputLayout=new QVBoxLayout;
    inputWidget->setLayout(inputLayout);
    outputWidget->setLayout(outputLayout);

    QVBoxLayout* layout=new QVBoxLayout;

    QTabWidget* tab=new QTabWidget;
    
    tab->addTab(inputWidget, "Inputs");
    tab->addTab(outputWidget, "Outputs");

	QHBoxLayout* hLayout=new QHBoxLayout;
	hLayout->addWidget(new QLabel("Name:"));
	nameEditor=new QLineEdit;
	hLayout->addWidget(nameEditor);
	layout->addLayout(hLayout);
    connect(nameEditor, SIGNAL(editingFinished()), this, SLOT(updateName()));

    QPushButton * addToMenuButton=new QPushButton("Save to Library");
    connect(addToMenuButton, SIGNAL(clicked()), view, SLOT(saveBlackBoxToLibrary()));
    layout->addWidget(addToMenuButton);


    layout->addWidget(tab);
	

    mainWidget->setLayout(layout);
    setWidgetResizable(true);
    setWidget(mainWidget);
}

void BlackBoxContextWidget::setBlackBoxInfo(VOperatorItem* bb)
{

    foreach(QTextEdit* edit, inputDescriptions)
    {
        inputLayout->removeWidget(edit);
        delete edit;
    }
    foreach(QTextEdit* edit, outputDescriptions)
    {
        outputLayout->removeWidget(edit);
        delete edit;
    }
    foreach(QLabel* label, inputLabels)
    {
        inputLayout->removeWidget(label);
        delete label;
    }
    foreach(QLabel* label, outputLabels)
    {
        outputLayout->removeWidget(label);
        delete label;
    }
    foreach(QGroupBox* box, outputBoxes)
    {
        outputLayout->removeWidget(box);
        delete box;
    }
    foreach(QGroupBox* box, inputBoxes)
    {
        inputLayout->removeWidget(box);
        delete box;
    }

    inputDescriptions.clear();
    outputDescriptions.clear();
    inputLabels.clear();
    outputLabels.clear();
    inputBoxes.clear();
    outputBoxes.clear();
	
    if(bb->isBlackBox)
    {
        blackBox=bb;
        
        for(int i=0; i<blackBox->inputs.size(); ++i)
        {
            QTextEdit* text=new QTextEdit;
    		connect(text, SIGNAL(textChanged()), this, SLOT(updateInputTexts()));

            text->setText(blackBox->inputs[i]->getDescription());
            inputDescriptions.push_back(text);

			QGroupBox* box=new QGroupBox(QString("Input #%1").arg(i));
			QVBoxLayout* layout=new QVBoxLayout;
			layout->addWidget(new QLabel("Description"));
            layout->addWidget(text);
			box->setLayout(layout);
			inputLayout->addWidget(box);
			inputBoxes.push_back(box);
        }

        for(int i=0; i<blackBox->outputs.size(); ++i)
        {
           QTextEdit* text=new QTextEdit;
    		connect(text, SIGNAL(textChanged()), this, SLOT(updateOutputTexts()));

            text->setText(blackBox->outputs[i]->getDescription());
            outputDescriptions.push_back(text);

			QGroupBox* box=new QGroupBox(QString("Output #%1").arg(i));
			QVBoxLayout* layout=new QVBoxLayout;
			layout->addWidget(new QLabel("Description"));
            layout->addWidget(text);
			box->setLayout(layout);
			outputLayout->addWidget(box);
			outputBoxes.push_back(box);
        }
		nameEditor->setText(bb->getName());
    }
}


void BlackBoxContextWidget::updateName()
{
   	blackBox->setName(nameEditor->text());
	if(blackBox->hasGui())
	{
		blackBox->getGUI()->setTitle(nameEditor->text());
	}
}


void BlackBoxContextWidget::updateOutputTexts()
{
    if(outputDescriptions.size()!=blackBox->outputs.size()) return;
    
    for(int i=0; i<blackBox->outputs.size(); ++i)
    {
        blackBox->outputs[i]->setDescription(outputDescriptions[i]->toPlainText());
    }
}


void BlackBoxContextWidget::updateInputTexts()
{
    if(inputDescriptions.size()!=blackBox->inputs.size()) return;
    
    for(int i=0; i<blackBox->inputs.size(); ++i)
    {
        blackBox->inputs[i]->setDescription(inputDescriptions[i]->toPlainText());
    }

}

PluginContextWidget::PluginContextWidget(QWidget* parent):
    QTabWidget(parent)
{
//     QTabWidget* tabWidget=new QTabWidget(this);

    ioWidget=new QTextEdit; ioWidget->setReadOnly(true);
    helpWidget=new QTextEdit; helpWidget->setReadOnly(true);
    docWidget=new QTextBrowser;

    /*tabWidget->*/addTab(ioWidget, "Plugin I/O");
    /*tabWidget->*/addTab(helpWidget, "Help");
    /*tabWidget->*/addTab(docWidget, "Documentation");
}

void PluginContextWidget::setPluginInfo(VOperatorItem* plugin)
{
    ioWidget->clear();
    helpWidget->clear();
    docWidget->clear();
    if(plugin->isPlugin())
    {
        QString io, help, doc;
        plugin->getPluginInfo(io, help, doc);

        ioWidget->setHtml(io);
        helpWidget->setText(help);
//         docWidget->setText(doc);

        FILE *fp;
        fp = fopen(doc.toLatin1().data(), "r");
        if ( fp != NULL )
        {
            fclose (fp);
            docWidget->setSource(doc);
        }
    }
}

JoinContextWidget::JoinContextWidget(QWidget* parent):
	QWidget(parent)
{
	inputItems.clear();
	joinItems.clear();
	
	QGridLayout * join_layout=new QGridLayout(this);
	join_layout->setSpacing(4);
	join_layout->setColumnStretch(1, 10);
	warningLabel=new QLabel("Note: multiple Inputs/Joins selected. Values displayed are from ONE of the selected items, but changes will affect all selected items.");
	warningLabel->setWordWrap(true);
	join_layout->addWidget(warningLabel,0,0,1,2);
	_noJoinBox=new QCheckBox("No join");
    join_layout->addWidget(_noJoinBox,1,0,Qt::AlignLeft );

	_keepXrefBox=new QCheckBox("Keep Input Metadata");
    join_layout->addWidget(_keepXrefBox,1,1,Qt::AlignRight );
	
	QLabel* algorithm_label = new QLabel("Join Algorithm ");
	join_layout->addWidget( algorithm_label, 2, 0, Qt::AlignRight );
		
	_algorithm = new QComboBox(this);
	_algorithm->addItem(LINEAR);
	_algorithm->addItem(BOXCAR);
	join_layout->addWidget( _algorithm, 2, 1, Qt::AlignLeft );

				
	QLabel* width_label = new QLabel("Ave width (sec) ",this);
	join_layout->addWidget( width_label, 3, 0, Qt::AlignRight );
			
	_width_field = new QguWLinput( this, "width_field" );
	_width_field->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));

	_width_field->setToolTip("width of the boxcar (Time is in seconds)");
	join_layout->addWidget( _width_field, 3, 1 );
			
	QLabel* blank_label = new QLabel(" ",this);
	join_layout->addWidget( blank_label, 4, 0, Qt::AlignRight );
			
	QLabel* blank2_label = new QLabel(" ",this);
	join_layout->addWidget( blank2_label, 4, 1, Qt::AlignRight );
			
	QLabel* gap_label = new QLabel("Gap Handling ",this);
	join_layout->addWidget( gap_label, 5, 0, Qt::AlignRight );
			
	_gap = new QComboBox(this);
	_gap->addItem(LINEAR_INTERP_GAP);
	_gap->addItem(ZERO_FILL_GAP);
	_gap->addItem(FILL_GAP);
	_gap->addItem(NEAREST_NEIGHBOUR_GAP);
	_gap->addItem(REMOVE_GAP);
	join_layout->addWidget( _gap, 5, 1, Qt::AlignLeft );

			
	QLabel* gapSize_label = new QLabel("Gap threshold ",this);
	join_layout->addWidget( gapSize_label, 6, 0, Qt::AlignRight );
			
	_gap_field = new QguWLinput( this, "gap_field" );
	_gap_field->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));

	_gap_field->setToolTip("minimum interval to be treated as a gap if empty (time is in seconds)");
	join_layout->addWidget( _gap_field, 6, 1 );
	
	connect(_keepXrefBox, SIGNAL(clicked(bool)), this, SLOT(setKeepXrefs(bool)));
	connect(_noJoinBox, SIGNAL(clicked(bool)), this, SLOT(setUserJoinStatus()));
	connect(_noJoinBox, SIGNAL(stateChanged(int)), this, SLOT(setJoinStatus(int)));
	connect(_algorithm, SIGNAL(activated(int)), this, SLOT(setAlgorithm()));
	connect(_width_field, SIGNAL(edited_OK()), this, SLOT(setWidthField()));
	connect(_gap, SIGNAL(activated(int)), this, SLOT(setGap()));
	connect(_gap_field, SIGNAL(edited_OK()), this, SLOT(setGapField()));
	
	userJoinChecked = false;
}

void JoinContextWidget::setJoinParameters(QList<VInputItem*>& inputItems, QList<VJoinItem*>& joinItems)
{

	_noJoinBox->blockSignals(true);
	_algorithm->blockSignals(true);
	_width_field->blockSignals(true);
	_gap->blockSignals(true);
	_gap_field->blockSignals(true);
	_keepXrefBox->blockSignals(true);
	
	warningLabel->setVisible(inputItems.size()+joinItems.size()>1);
	
	if(!inputItems.isEmpty())
	{
		_keepXrefBox->setChecked(inputItems[0]->getKeepXrefs());
		_keepXrefBox->setEnabled(true);
		_noJoinBox->setChecked(inputItems[0]->getNoJoin());
		_algorithm->setCurrentIndex(inputItems[0]->getAlgorithm());
		_algorithm->setEnabled(!_noJoinBox->isChecked());
		_width_field->setText(QString("%1").arg(inputItems[0]->getWidth()));
		_width_field->setEnabled(!_noJoinBox->isChecked());
		_gap->setCurrentIndex(inputItems[0]->getGap());
		_gap->setEnabled(!_noJoinBox->isChecked());
		_gap_field->setText(QString("%1").arg(inputItems[0]->getGapSize()));
		_gap_field->setEnabled(!_noJoinBox->isChecked());
	}
	else
	{
		_keepXrefBox->setChecked(false);
		_keepXrefBox->setEnabled(false);
		_noJoinBox->setChecked(joinItems[0]->getNoJoin());
		_algorithm->setCurrentIndex(joinItems[0]->getAlgorithm());
		_algorithm->setEnabled(!_noJoinBox->isChecked());
		_width_field->setText(QString("%1").arg(joinItems[0]->getWidth()));
		_width_field->setEnabled(!_noJoinBox->isChecked());
		_gap->setCurrentIndex(joinItems[0]->getGap());
		_gap->setEnabled(!_noJoinBox->isChecked());
		_gap_field->setText(QString("%1").arg(joinItems[0]->getGapSize()));
		_gap_field->setEnabled(!_noJoinBox->isChecked());
	}
	
	_noJoinBox->blockSignals(false);
	_algorithm->blockSignals(false);
	_width_field->blockSignals(false);
	_gap->blockSignals(false);
	_gap_field->blockSignals(false);
	_keepXrefBox->blockSignals(false);

}

void JoinContextWidget::setUserJoinStatus()
{
	userJoinChecked = true;
}

void JoinContextWidget::setJoinStatus(int status)
{
	foreach(VInputItem* input, inputItems)
	{
		if(input) input->setNoJoin(status!=Qt::Unchecked);
	}
	foreach(VJoinItem* join, joinItems)
	{
		if(join) join->setNoJoin(status!=Qt::Unchecked);
	}
	_width_field->setEnabled(status==Qt::Unchecked);
	_gap->setEnabled(status==Qt::Unchecked);
	_gap_field->setEnabled(status==Qt::Unchecked);
	_algorithm->setEnabled(status==Qt::Unchecked);
}

void JoinContextWidget::setKeepXrefs(bool keep)
{
	foreach(VInputItem* input, inputItems)
	{
		if(input) input->setKeepXref(keep);
	}
	
}

void JoinContextWidget::setAlgorithm()
{
    foreach(VInputItem* input, inputItems)
	{
		input->setAlgorithm(_algorithm->currentIndex());
	}
	foreach(VJoinItem* join, joinItems)
	{
		join->setAlgorithm(_algorithm->currentIndex());
	}
}

void JoinContextWidget::setWidthField()
{
    foreach(VInputItem* input, inputItems)
	{
		input->setWidth(_width_field->text().toDouble());
	}
	foreach(VJoinItem* join, joinItems)
	{
		join->setWidth(_width_field->text().toDouble());
	}
}

void JoinContextWidget::setGap()
{
    foreach(VInputItem* input, inputItems)
	{
		input->setGap(_gap->currentIndex());
	}
	foreach(VJoinItem* join, joinItems)
	{
		join->setGap(_gap->currentIndex());
	}
}

void JoinContextWidget::setGapField()
{
    foreach(VInputItem* input, inputItems)
	{
		input->setGapSize(_gap_field->text().toDouble());
	}
	foreach(VJoinItem* join, joinItems)
	{
		join->setGapSize(_gap_field->text().toDouble());
	}
}

void JoinContextWidget::checkSequence(const QString& text)
{
	if(userJoinChecked) return;
	
	if(text.compare("scalar") == 0) _noJoinBox->setChecked(true);
	else _noJoinBox->setChecked(false);
}


OutputContextWidget::OutputContextWidget(QWidget* parent, CalcView *view):
    QWidget(parent)
{
	_view = view;
	QVBoxLayout * mainLayout=new QVBoxLayout;
	
	warningLabel=new QLabel("Note: multiple Outputs selected. Values displayed are from ONE of the selected items, but changes will affect all selected items.");
	warningLabel->setWordWrap(true);
	mainLayout->addWidget(warningLabel);
	
	tlBox=new QComboBox(this);
	
	mainLayout->addWidget(tlBox);
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}
	connect(tlBox, SIGNAL(currentIndexChanged(const QString&)), this, SLOT(newSelection(const QString&)));
	
	fromDataLabel=new QLabel("");
	fromDataLabel->setWordWrap(true);
	mainLayout->addWidget(fromDataLabel);
	
	regularTTLabel=new QLabel("");
	regularTTLabel->setWordWrap(true);
	mainLayout->addWidget(regularTTLabel);
	
	setLayout(mainLayout);
	
	connect(&timeLineHandler, SIGNAL(addedTimeLine(const QString&)), this, SLOT(addedTimeLine(const QString&)));
	connect(&timeLineHandler, SIGNAL(deletedTimeLine(const QString&)), this, SLOT(deletedTimeLine(const QString&)));

	
	newSelection("Default");

}

void OutputContextWidget::setOutputParameters(QList<VOutputItem*>& items)
{
	tlBox->blockSignals(true);

	warningLabel->setVisible(items.size()>1);
	tlBox->clear();
	//fill
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}
	int bkpIndex=tlBox->findText(items[0]->timeLineName);
	if(bkpIndex>=0)
		tlBox->setCurrentIndex(bkpIndex);
	
	TimeLine* tl=timeLineHandler.getTimeLine(items[0]->timeLineName);
	setTab(tl); 
	
	tlBox->blockSignals(false);

}


void OutputContextWidget::setTab(TimeLine* tl)
{
	if(tl!=NULL)
	{
		QString text;
		if(tl->fromData)
		{
			text="Interval from Data";
		}
		else
		{
			text=QString("Interval from Object %1").arg(tl->fromObjString);
		}
		fromDataLabel->setText(text);
		
		if(tl->regularTT)
		{
			text=QString("Regular tags, resolution %1").arg(tl->TTString);
		}
		else
		{
			text=QString("Target Tags from Object %1").arg(tl->TTString);
		}
		regularTTLabel->setText(text);
		
	}

}


void OutputContextWidget::newSelection(const QString& name)
{
	TimeLine* tl=timeLineHandler.getTimeLine(name);
	if(tl!=NULL)
	{
		setTab(tl);
		
		QList<VOutputItem*> outputItems = _view->listSelectedOutputItems();
		foreach(VOutputItem* item, outputItems)
		{
			item->timeLineName=name;
		}
	}

}

void OutputContextWidget::addedTimeLine(const QString&)
{
	qApp->processEvents();

	QString save=tlBox->currentText();

	tlBox->blockSignals(true);

	//clear
	tlBox->clear();

	//fill
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}
	int bkpIndex=tlBox->findText(save);
	if(bkpIndex>=0)
		tlBox->setCurrentIndex(bkpIndex);

	tlBox->blockSignals(false);

}

void OutputContextWidget::deletedTimeLine(const QString&)
{
	QString save=tlBox->currentText();
	tlBox->blockSignals(true);
	//clear
	tlBox->clear();
	//fill
	QList<QString> names=timeLineHandler.getNames();
	foreach(QString name, names)
	{
		tlBox->addItem(name);
	}

	int bkpIndex=tlBox->findText(save);
	
	if(bkpIndex>=0)  tlBox->setCurrentIndex(bkpIndex);
	else tlBox->setCurrentIndex(0);
	
	tlBox->blockSignals(false);


}



DefaultContextWidget::DefaultContextWidget(QWidget* parent, CalcView* view):
    QWidget(parent)
{
	this->view=view;
    QVBoxLayout * mainLayout=new QVBoxLayout;
	
	QGroupBox * addBox=new QGroupBox("Add item");
	QGridLayout * add_layout=new QGridLayout(addBox);
    add_layout->setSpacing(4);
    add_layout->setColumnStretch(1, 10);
	mainLayout->addWidget(addBox);
	QPushButton * inputButton=new QPushButton("Input");
	connect(inputButton, SIGNAL(clicked()), this, SLOT(addInput()));
	add_layout->addWidget(inputButton, 0, 0, Qt::AlignLeft );
	QPushButton * outputButton=new QPushButton("Output");
	connect(outputButton, SIGNAL(clicked()), this, SLOT(addOutput()));
	add_layout->addWidget(outputButton, 0, 1, Qt::AlignLeft );
	QPushButton * extractorButton=new QPushButton("Extractor");
	connect(extractorButton, SIGNAL(clicked()), this, SLOT(addExtractor()));
	add_layout->addWidget(extractorButton, 1, 0, Qt::AlignLeft );
	QPushButton * joinButton=new QPushButton("Join/Resample");
	connect(joinButton, SIGNAL(clicked()), this, SLOT(addJoin()));
	add_layout->addWidget(joinButton, 1, 1, Qt::AlignLeft );
	
	QPushButton* operatorButton=new QPushButton("Operator");
	add_layout->addWidget(operatorButton, 2, 0, Qt::AlignLeft );
	connect(operatorButton, SIGNAL(pressed()), view, SLOT(autoSelectPosition()));
	QMenu* operatorMenu=new QMenu(operatorButton);
	operatorButton->setMenu(operatorMenu);
	QMenu* binOp=operatorMenu->addMenu("Arith f(A,B)");
	binOp->addAction("A+B", view, SLOT(addOpAdd()));
	binOp->addAction("A-B", view, SLOT(addOpSub()));
	binOp->addAction("A*B", view, SLOT(addOpMul()));
	binOp->addAction("A/B", view, SLOT(addOpDiv()));
	binOp->addAction("A^B", view, SLOT(addOpPow()));
	binOp->addAction("A\%B", view, SLOT(addOpRemainder()));
	binOp->addAction("A(i,j) * B(i,j) -> C(i,j)", view, SLOT(addOpPiecewise()));
	binOp->addAction("A(i,j) / B(i,j) -> C(i,j)", view, SLOT(addOpPwDiv()));
	QMenu* unOp=operatorMenu->addMenu("Arith f(A)");
	unOp->addAction("mod(A)", view, SLOT(addOpMod()));
	unOp->addAction("abs(A)", view, SLOT(addOpAbs()));
	unOp->addAction("ln(A)", view, SLOT(addOpLog()));
	unOp->addAction("log(A)", view, SLOT(addOpLog10()));
	unOp->addAction("1/A", view, SLOT(addOpInverse()));
	unOp->addAction("-A", view, SLOT(addOpOpp()));
	unOp->addAction("sqrt(A)", view, SLOT(addOpSqrt()));
	unOp->addAction("exp(A)", view, SLOT(addOpExp()));
	unOp->addAction("trace(A)", view, SLOT(addOpTrace()));
	QMenu* trigOp=operatorMenu->addMenu("Trig");
	trigOp->addAction("cos(A)", view, SLOT(addOpCos()));
	trigOp->addAction("sin(A)", view, SLOT(addOpSin()));
	trigOp->addAction("tan(A)", view, SLOT(addOpTan()));
	trigOp->addAction("cot(A)", view, SLOT(addOpCot()));
	trigOp->addAction("acos(A)", view, SLOT(addOpAcos()));
	trigOp->addAction("asin(A)", view, SLOT(addOpAsin()));
	trigOp->addAction("atan(A)", view, SLOT(addOpAtan()));
	trigOp->addAction("atan2(A,B)", view, SLOT(addOpAtan2()));
	trigOp->addAction("acot(A)", view, SLOT(addOpAcot()));
	trigOp->addAction("cosh(A)", view, SLOT(addOpCosh()));
	trigOp->addAction("sinh(A)", view, SLOT(addOpSinh()));
	trigOp->addAction("tanh(A)", view, SLOT(addOpTanh()));
	trigOp->addAction("coth(A)", view, SLOT(addOpCoth()));
	trigOp->addAction("acosh(A)", view, SLOT(addOpAcosh()));
	trigOp->addAction("asinh(A)", view, SLOT(addOpAsinh()));
	trigOp->addAction("atanh(A)", view, SLOT(addOpAtanh()));
	trigOp->addAction("acoth(A)", view, SLOT(addOpAcoth()));
	trigOp->addAction("(-180,180) to (0,360)", view, SLOT(addOpAngle360()));
	trigOp->addAction("(0,360) to (-180,180)", view, SLOT(addOpAngle180()));
	QMenu* vectOp=operatorMenu->addMenu("Vector Ops");
	vectOp->addAction("A.B", view, SLOT(addOpDotProd()));
	vectOp->addAction("AxB", view, SLOT(addOpVectProd()));
	QMenu* derivOp=operatorMenu->addMenu("Derivative");
	derivOp->addAction("3pt derivative", view, SLOT(addOpDiff3()));
	derivOp->addAction("5pt derivative", view, SLOT(addOpDiff5()));
	QMenu* integOp=operatorMenu->addMenu("Integral");
	integOp->addAction("Definite Integral", view, SLOT(addOpIntDef()));
	integOp->addAction("Indefinite Integral", view, SLOT(addOpIntIndef()));
	QMenu* modOp=operatorMenu->addMenu("Filtering");
    view->addPluginMenu(modOp, "Filtering", true);
	QMenu* unitOp=operatorMenu->addMenu("Change Units");
    view->addPluginMenu(unitOp, "UnitOperations", true);
	QMenu* timeOp=operatorMenu->addMenu("Time Operations");
    view->addPluginMenu(timeOp, "TimeOperations", true);
	timeOp->addAction("Timetags to seconds from offset", view, SLOT(addOpTTtoS()));
	timeOp->addAction("Time Interval to seconds", view, SLOT(addOpTIVLtoS()));
	timeOp->addAction("Time2 - Time1 in sec", view, SLOT(addOpT2_T1()));
	timeOp->addAction("Timetags2 - Timetags1 in seconds", view, SLOT(addOpTT2_TT1()));
	QMenu* statsOp=operatorMenu->addMenu("Statistics");
    view->addPluginMenu(statsOp, "Statistics", true);
	statsOp->addAction("Linear Regression", view, SLOT(addOpLinReg()));
	QMenu* eventsOp=operatorMenu->addMenu("Event Tables");
    view->addPluginMenu(eventsOp, "EventTables", true);
    QMenu* plugOp=operatorMenu->addMenu("Plug-ins");
    QMenu* geoOp=plugOp->addMenu("Geophysics");
    view->addPluginMenu(geoOp, "Geophysics", true);
    QMenu* anaOp=plugOp->addMenu("Analysis");
    view->addPluginMenu(anaOp, "Analysis", true);
	plugOp->addSeparator();
	plugOp->addAction("Browse...", view, SLOT(addPlugInBrowse()));
	
	bbOp=operatorMenu->addMenu("Black Box from library");
	view->addBlackBoxMenu(bbOp);
	connect(bbOp, SIGNAL(triggered(QAction*)), view, SLOT(addBB(QAction*)));
	
	connect(view, SIGNAL(updateBlackBoxMenu()), this, SLOT(updateBlackBoxMenu()));

	
	QPushButton* templateButton=new QPushButton("Template");
	add_layout->addWidget(templateButton, 2, 1, Qt::AlignLeft );
	connect(templateButton, SIGNAL(pressed()), view, SLOT(autoSelectPosition()));
	QMenu* templateMenu=new QMenu(templateButton);
	templateButton->setMenu(templateMenu);
	templateMenu->addAction("A+B", view, SLOT(addTemplateAdd()));
	templateMenu->addAction("A-B", view, SLOT(addTemplateSub()));
	templateMenu->addAction("A*B", view, SLOT(addTemplateMul()));
	templateMenu->addAction("A/B", view, SLOT(addTemplateDiv()));
	templateMenu->addAction("-A", view, SLOT(addTemplateOpp()));
    
	mainLayout->addStretch();

    setLayout(mainLayout);
}

void DefaultContextWidget::updateBlackBoxMenu()
{
	bbOp->clear();
	view->addBlackBoxMenu(bbOp);
}

void DefaultContextWidget::addInput()
{
	view->autoSelectPosition();
	view->addInputItem();
}

void DefaultContextWidget::addOutput()
{
	view->autoSelectPosition();
	view->addOutputItem();
}

void DefaultContextWidget::addExtractor()
{
	view->autoSelectPosition();
	view->addExtractionItem();
}

void DefaultContextWidget::addJoin()
{
	view->autoSelectPosition();
	view->addJoinItem();
}



SelectionContextWidget::SelectionContextWidget(QWidget* parent, CalcView* view):
    QWidget(parent)
{
	this->view=view;
    QVBoxLayout * mainLayout=new QVBoxLayout;
	
	QGroupBox * bbBox=new QGroupBox("Black box");
	QGridLayout * bb_layout=new QGridLayout(bbBox);
	bb_layout->setSpacing(4);
    bb_layout->setColumnStretch(1, 10);
	mainLayout->addWidget(bbBox);
	
	QPushButton * bbButton=new QPushButton("Set as black box");
	connect(bbButton, SIGNAL(clicked()), view, SLOT(changeForBlackBox()));
	bb_layout->addWidget(bbButton, 0, 0, Qt::AlignLeft );
	
	QGroupBox * aspectBox=new QGroupBox("Aspect");
	QGridLayout * aspect_layout=new QGridLayout(aspectBox);
	aspect_layout->setSpacing(4);
    aspect_layout->setColumnStretch(1, 10);
	mainLayout->addWidget(aspectBox);
	
	QPushButton * resetButton=new QPushButton("Reset");
	connect(resetButton, SIGNAL(clicked()), view, SLOT(resetAspect()));
	aspect_layout->addWidget(resetButton, 0, 0, Qt::AlignLeft );
	
	QPushButton * hButton=new QPushButton("Horizontal");
	connect(hButton, SIGNAL(clicked()), view, SLOT(horizontalAspect()));
	aspect_layout->addWidget(hButton, 0, 1, Qt::AlignLeft );
	QPushButton * vButton=new QPushButton("Vertical");
	connect(vButton, SIGNAL(clicked()), view, SLOT(verticalAspect()));
	aspect_layout->addWidget(vButton, 0, 2, Qt::AlignLeft );
	
	QPushButton * hSizeButton=new QPushButton("1/2x");
	connect(hSizeButton, SIGNAL(clicked()), view, SLOT(halfSize()));
	aspect_layout->addWidget(hSizeButton, 1, 0, Qt::AlignLeft );
	QPushButton * nSizeButton=new QPushButton("Normal size");
	connect(nSizeButton, SIGNAL(clicked()), view, SLOT(normalSize()));
	aspect_layout->addWidget(nSizeButton, 1, 1, Qt::AlignLeft );
	QPushButton * dSizeButton=new QPushButton("2x");
	connect(dSizeButton, SIGNAL(clicked()), view, SLOT(doubleSize()));
	aspect_layout->addWidget(dSizeButton, 1, 2, Qt::AlignLeft );
	

	mainLayout->addStretch();

    setLayout(mainLayout);
}





ExtractorContextWidget::ExtractorContextWidget(ContextWidget* parent):
	QScrollArea((QWidget*)parent)
{
	QWidget* mainWidget=new QWidget;
    QVBoxLayout * mainLayout=new QVBoxLayout;
	
	ctxtWidget = parent;
	
	warningLabel=new QLabel("Note: multiple Extractors selected. Values displayed are from ONE of the selected items, but changes will affect all selected items.");
	warningLabel->setWordWrap(true);
	mainLayout->addWidget(warningLabel);
	
	QGroupBox * outGBox=new QGroupBox("Output option");
	QGridLayout * outLayout=new QGridLayout(outGBox);
	
	target=new QComboBox();
	target->addItem("Scalar");
	target->addItem("1D Array");
	target->addItem("2D Array");
	target->addItem("3D Array");
	target->addItem("Text");
	target->addItem("Time");
	target->addItem("Time Interval");
	target->addItem("Time Tags");
	target->addItem("Metadata");
	connect(target, SIGNAL(currentIndexChanged(int)), this, SLOT(setTarget(int)) );
	
	QLabel * targetLabel = new QLabel("Output choice: ");
	outLayout->addWidget(targetLabel, 0, 0, Qt::AlignRight);
	outLayout->addWidget(target, 0, 1, Qt::AlignLeft);

	outGBox->setLayout(outLayout);
	mainLayout->addWidget(outGBox);
	
	QGroupBox * compBox=new QGroupBox("Component Selection");
	QGridLayout * compLayout=new QGridLayout(compBox);
	compCheckBox=new QCheckBox("Ask user");
	compLayout->addWidget(compCheckBox, 0, 0);
	ComponentButton=new QPushButton("Component");
	EventButton=new QPushButton("Event");
	MetaButton=new QPushButton("Metadata");

	talkBack = new QLabel("Calculation will pause at unresolved Extractors");
	compLayout->addWidget(talkBack, 1, 0, 1, 2);


	QAction* action;
	
	/* QMenu* */ComponentMenu0=new QMenu;
	action=ComponentMenu0->addAction("Default"); action->setData(QVariant(""));

	action=ComponentMenu0->addSeparator();
	// vectors
	action=ComponentMenu0->addAction("r (mag)"); action->setData(QVariant("r (mag)"));
	action=ComponentMenu0->addAction("x"); action->setData(QVariant("x"));
	action=ComponentMenu0->addAction("y"); action->setData(QVariant("y"));
	action=ComponentMenu0->addAction("z"); action->setData(QVariant("z"));
	action=ComponentMenu0->addAction("Lat (deg)"); action->setData(QVariant("Lat (deg)"));
	action=ComponentMenu0->addAction("theta (deg)"); action->setData(QVariant("theta (deg)"));
	action=ComponentMenu0->addAction("phi (deg)"); action->setData(QVariant("phi (deg)"));
	action=ComponentMenu0->addAction("Lat (rad)"); action->setData(QVariant("Lat (rad)"));
	action=ComponentMenu0->addAction("theta (rad)"); action->setData(QVariant("theta (rad)"));
	action=ComponentMenu0->addAction("phi (rad)"); action->setData(QVariant("phi (rad)"));
	action=ComponentMenu0->addSeparator();
	// Array
	action=ComponentMenu0->addAction("Syntax for 3D -> 0D"); action->setData(QVariant("Sum(i,j) Sum(i,j) Sum(i,j)"));
	action=ComponentMenu0->addAction("Syntax for 2D -> 0D"); action->setData(QVariant("Sum(i,j) Sum(i,j)"));
	action=ComponentMenu0->addAction("Syntax for 1D -> 0D"); action->setData(QVariant("Sum(i,j)"));
	action=ComponentMenu0->addSeparator();
	action=ComponentMenu0->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));
	ComponentButton->setMenu(ComponentMenu0);

	/* QMenu* */ComponentMenu1=new QMenu;
	action=ComponentMenu1->addAction("Default"); action->setData(QVariant(""));
	action=ComponentMenu1->addSeparator();
	// Array
	action=ComponentMenu1->addAction("Syntax for 3D -> 1D"); action->setData(QVariant("Sum(i,j) Sum(i,j) (i,j)"));
	action=ComponentMenu1->addAction("Syntax for 2D -> 1D"); action->setData(QVariant("Sum(i,j) (i,j)"));
	action=ComponentMenu1->addAction("Syntax for 1D range"); action->setData(QVariant("(i,j)"));
	action=ComponentMenu1->addSeparator();
	action=ComponentMenu1->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	/* QMenu* */ComponentMenu2=new QMenu;
	action=ComponentMenu2->addAction("Default"); action->setData(QVariant(""));
	action=ComponentMenu2->addSeparator();
	// Array
	action=ComponentMenu2->addAction("Syntax for 3D -> 2D"); action->setData(QVariant("Sum(i,j) (i,j) (i,j)"));
	action=ComponentMenu2->addAction("Syntax for 2D range"); action->setData(QVariant("(i,j) (i,j)"));
	action=ComponentMenu2->addSeparator();
	action=ComponentMenu2->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	/* QMenu* */ComponentMenu3=new QMenu;
	action=ComponentMenu3->addAction("Default"); action->setData(QVariant(""));
	action=ComponentMenu3->addSeparator();
	// Array
	action=ComponentMenu3->addAction("Syntax for 3D range"); action->setData(QVariant("(i,j) (i,j) (i,j)"));
	action=ComponentMenu3->addSeparator();
	action=ComponentMenu3->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	/* QMenu* */ComponentMenuText=new QMenu;
	action=ComponentMenuText->addAction("Default"); action->setData(QVariant(""));
	action=ComponentMenuText->addSeparator();
	// Array
	action=ComponentMenuText->addAction("Syntax for 1D -> 0D"); action->setData(QVariant("(i)"));
	action=ComponentMenuText->addAction("Syntax for 1D range"); action->setData(QVariant("(i,j)"));
	action=ComponentMenuText->addSeparator();
	action=ComponentMenuText->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	/* QMenu* */ComponentMenuTime=new QMenu;
	
	// Time
	action=ComponentMenuTime->addAction("Start Time"); action->setData(QVariant("Start Time"));
	action=ComponentMenuTime->addAction("Centre Time"); action->setData(QVariant("Centre Time"));
	action=ComponentMenuTime->addAction("End Time"); action->setData(QVariant("End Time"));
	action=ComponentMenuTime->addSeparator();
	action=ComponentMenuTime->addAction("Time at record"); action->setData(QVariant("i"));
	action=ComponentMenuTime->addSeparator();
	action=ComponentMenuTime->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	/* QMenu* */ComponentMenuTimeIvl=new QMenu;
	
	// Time
	action=ComponentMenuTimeIvl->addAction("Object Timetag Range"); action->setData(QVariant(""));
	action=ComponentMenuTimeIvl->addAction("Interval between records"); action->setData(QVariant("i,j"));
	action=ComponentMenuTimeIvl->addSeparator();
	action=ComponentMenuTimeIvl->addAction("Event at entry"); action->setData(QVariant("i"));
	action=ComponentMenuTimeIvl->addSeparator();
	action=ComponentMenuTimeIvl->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	/* QMenu* */ComponentMenuMeta=new QMenu;
	action=ComponentMenuMeta->addAction("Time Tags"); action->setData(QVariant("=> DEPEND_0"));
	action=ComponentMenuMeta->addAction("DEPEND_1"); action->setData(QVariant("=> DEPEND_1"));
	action=ComponentMenuMeta->addAction("DEPEND_2"); action->setData(QVariant("=> DEPEND_2"));
	action=ComponentMenuMeta->addAction("DEPEND_3"); action->setData(QVariant("=> DEPEND_3"));
	action=ComponentMenuMeta->addAction("SI_CONVERSION"); action->setData(QVariant("=> DEPEND_3"));
	action=ComponentMenuMeta->addAction("COORDINATE_SYSTEM"); action->setData(QVariant("=> COORDINATE_SYSTEM"));
	action=ComponentMenuMeta->addAction("FRAME"); action->setData(QVariant("=> FRAME"));
	action=ComponentMenuMeta->addSeparator();
	action=ComponentMenuMeta->addAction("Syntax for Metadata"); action->setData(QVariant("=> META_NAME"));

	compLayout->addWidget(ComponentButton, 2, 0); // these never co-exist

	compLayout->addWidget(EventButton, 2, 0);
	EventButton->hide();
	SuffixComponent=new QLineEdit;
	SuffixEvent=new QLineEdit;

	compLayout->addWidget(MetaButton, 2, 0);
	MetaButton->hide();
	SuffixMeta=new QLineEdit;

//	QRegExp vectSuffixRegexp("r\\(mag\\)|x|y|z|Lat\\s\\(deg\\)|Lat\\s\\(rad\\)|theta\\s\\(deg\\)|theta\\s\\(rad\\)|phi\\s\\(deg\\)|phi\\s\\(rad\\)|\\([123],[123]\\)|Sum\\([123],[123]\\)|Ave\\([123],[123]\\)|\\([123]\\)|=>\\s\\w+");
//	val=new QRegExpValidator(vectSuffixRegexp, vectSuffixComponent);
//	vectSuffixComponent->setValidator(val);
	compLayout->addWidget(SuffixComponent, 2, 1);// these never co-exist
	compLayout->addWidget(SuffixEvent, 2, 1);
	compLayout->addWidget(SuffixMeta, 2, 1);
	SuffixEvent->hide();
	SuffixMeta->hide();
	
	TimeButton=new QPushButton("Time Option");
	/*QMenu* */TimeMenu=new QMenu;
	action=TimeMenu->addAction("Default"); action->setData(QVariant(""));
	action=TimeMenu->addAction("Syntax for Subset time from record i to j"); action->setData(QVariant("i,j"));
	action=TimeMenu->addAction("Syntax for Average time between records i and j"); action->setData(QVariant("Ave i,j"));
	action=TimeMenu->addAction("Syntax for time at record i"); action->setData(QVariant("i"));
	TimeButton->setMenu(TimeMenu);
	compLayout->addWidget(TimeButton, 3, 0);

	/*QMenu* */TimetagMenu=new QMenu;
	action=TimetagMenu->addAction("Default"); action->setData(QVariant(""));
	action=TimetagMenu->addAction("Syntax for Subset time from record i to j"); action->setData(QVariant("i,j"));

	SuffixTime=new QLineEdit;
//	QRegExp SuffixTimeRegexp("\\d+|\\d+,\\d+|Ave\\s\\d+,\\d+");
//	val=new QRegExpValidator(SuffixTimeRegexp, SuffixTime);
//	SuffixTime->setValidator(val);
	compLayout->addWidget(SuffixTime, 3, 1);
	Suffix=new QLabel;
	compLayout->addWidget(Suffix, 4, 0, 1, 2);

	compBox->setLayout(compLayout);
	mainLayout->addWidget(compBox);
		
	

//	QRegExp arraySuffixRegexp("(((Sum|Ave|)\\(\\d+,\\d+\\))|(\\(\\d+\\)))\\s(((Sum|Ave|)\\(\\d+,\\d+\\))|(\\(\\d+\\)))\\s(((Sum|Ave|)\\(\\d+,\\d+\\))|(\\(\\d+\\)))|=>\\s\\w+");
//	val=new QRegExpValidator(arraySuffixRegexp, arraySuffixComponent);
//	arraySuffixComponent->setValidator(val);
	
	
	mainLayout->addStretch();
	mainWidget->setLayout(mainLayout);
	setWidget(mainWidget);
	
	// for choice setting
//	QRegExp scalarSuffixRegexp("=>\\s\\w+");
//	QRegExpValidator* val=new QRegExpValidator(scalarSuffixRegexp, scalarSuffixComponent);
//	SuffixComponent->setValidator(val);

	connect(compCheckBox, SIGNAL(clicked(bool)), this, SLOT(CheckChanged(bool)));
	connect(ComponentMenu0, SIGNAL(triggered(QAction*)), this, SLOT(selectedComponent(QAction*)));
	connect(ComponentMenu1, SIGNAL(triggered(QAction*)), this, SLOT(selectedComponent(QAction*)));
	connect(ComponentMenu2, SIGNAL(triggered(QAction*)), this, SLOT(selectedComponent(QAction*)));
	connect(ComponentMenu3, SIGNAL(triggered(QAction*)), this, SLOT(selectedComponent(QAction*)));
	connect(ComponentMenuText, SIGNAL(triggered(QAction*)), this, SLOT(selectedComponent(QAction*)));
	connect(ComponentMenuTime, SIGNAL(triggered(QAction*)), this, SLOT(selectedComponent(QAction*)));
	connect(ComponentMenuTimeIvl, SIGNAL(triggered(QAction*)), this, SLOT(selectedEvent(QAction*)));
	connect(ComponentMenuMeta, SIGNAL(triggered(QAction*)), this, SLOT(selectedMeta(QAction*)));
	connect(SuffixComponent, SIGNAL(textEdited(const QString&)), this, SLOT(SuffixChanged(const QString&)));
	connect(SuffixEvent, SIGNAL(textEdited(const QString&)), this, SLOT(EventChanged(const QString&)));
	connect(SuffixMeta, SIGNAL(textEdited(const QString&)), this, SLOT(MetaChanged(const QString&)));
	connect(TimeMenu, SIGNAL(triggered(QAction*)), this, SLOT(selectedTime(QAction*)));
	connect(TimetagMenu, SIGNAL(triggered(QAction*)), this, SLOT(selectedTime(QAction*)));
	connect(SuffixTime, SIGNAL(textEdited(const QString&)), this, SLOT(SuffixChanged(const QString&)));

}

void ExtractorContextWidget::setTarget(int targetType){
	
	foreach(VExtractionItem* item, items)
	{
		ExtractionItem *gui = item->getGUI();	
		if( !gui) return;		
	
		// Set what object is acceptable and change component and time menu options as appropriate
		
		if(targetType == 0){
			ComponentButton->setMenu(ComponentMenu0);
			ComponentButton->show();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			SuffixComponent->show();
			TimeButton->show();
			SuffixTime->show();
			TimeButton->setMenu(TimeMenu);
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_ANY, QS_ANY));
		}
		else if(targetType == 1){
			ComponentButton->setMenu(ComponentMenu1);
			ComponentButton->show();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			SuffixComponent->show();
			TimeButton->show();
			SuffixTime->show();
			TimeButton->setMenu(TimeMenu);
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_ANY, QS_ANY));
		}
		else if(targetType == 2){
			ComponentButton->setMenu(ComponentMenu2);
			ComponentButton->show();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			SuffixComponent->show();
			TimeButton->show();
			SuffixTime->show();
			TimeButton->setMenu(TimeMenu);
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_2D, QS_ANY, QS_ANY));
		}
		else if(targetType == 3){
			ComponentButton->setMenu(ComponentMenu3);
			ComponentButton->show();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			SuffixComponent->show();
			TimeButton->show();
			SuffixTime->show();
			TimeButton->setMenu(TimeMenu);
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_2D, QS_ANY, QS_ANY));
		}
		else if(targetType == 4){
			ComponentButton->setMenu(ComponentMenuText);
			ComponentButton->show();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			SuffixComponent->show();
			TimeButton->show();
			SuffixTime->show();
			TimeButton->setMenu(TimeMenu);
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_TEXT, QS_ANY, QS_ANY, QS_ANY));
		}
		else if(targetType == 5){
			ComponentButton->setMenu(ComponentMenuTime);
			ComponentButton->show();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			SuffixComponent->show();
			TimeButton->hide();
			SuffixTime->hide();
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_NOT_SEQ, QS_ANY));
		}
		else if(targetType == 6){
			EventButton->setMenu(ComponentMenuTimeIvl);
			EventButton->show();
			SuffixEvent->show();
			MetaButton->hide();
			SuffixMeta->hide();
			ComponentButton->hide();
			SuffixComponent->hide();
			TimeButton->hide();
			SuffixTime->hide();
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_TI, QS_0D, QS_NOT_SEQ, QS_ANY));
		}
		else if(targetType == 7){
			ComponentButton->hide();
			SuffixComponent->hide();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->hide();
			SuffixMeta->hide();
			TimeButton->show();
			SuffixTime->show();
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_SEQ, QS_TT));
			TimeButton->setMenu(TimetagMenu);
		}
		else if(targetType == 8){
			ComponentButton->hide();
			SuffixComponent->hide();
			EventButton->hide();
			SuffixEvent->hide();
			MetaButton->setMenu(ComponentMenuMeta);
			MetaButton->show();
			SuffixMeta->show();
			TimeButton->hide();
			SuffixTime->hide();
			gui->edit->setAcceptable(Qcondition(QS_NARROW, QS_ANY, QS_ANY, QS_ANY, QS_ANY));
		}
		
	}
	
}


void ExtractorContextWidget::setExtractorParameters(QList<VExtractionItem*>& items)
{
	SuffixComponent->blockSignals(true);
	SuffixTime->blockSignals(true);
	
	bool ask = items[0]->askUser;
	compCheckBox->setChecked(ask);

	this->items=items;
	warningLabel->setVisible(items.size()>1);
	
	int i1, i2;
	
	i1=items[0]->Suffix.indexOf("[");
	if(i1>0)
	{
		i2=items[0]->Suffix.indexOf("]");
		QString text=items[0]->Suffix.mid(i1+1, i2-i1-1);
		SuffixComponent->setText(text.trimmed());
	}
	else
	{
		SuffixComponent->setText("");
	}
	
	i1=items[0]->Suffix.indexOf("<");
	if(i1>0)
	{
		i2=items[0]->Suffix.indexOf(">");
		QString text=items[0]->Suffix.mid(i1+1, i2-i1-1);
		SuffixTime->setText(text.trimmed());
	}
	else
	{
		SuffixTime->setText("");
	}

	SuffixComponent->blockSignals(false);
	SuffixTime->blockSignals(false);

}

void ExtractorContextWidget::selectedComponent(QAction* action)
{
	SuffixComponent->setText(action->data().toString());
	if(action->data().toString().contains("Sum")){
		talkBack->setText("Edit (i,j), Sum(i,j), Ave(i,j) or (i)) for each dimension");
	}
	else if(action->data().toString().contains("META_NAME")){
		talkBack->setText("Edit META_NAME to required attribute ");
	}
	else if(action->data().toString().contains("i,j")){
		talkBack->setText("Edit indices i,j for required range");
	}
	else if(action->data().toString().contains("i")){
		talkBack->setText("Edit index i for required slice");
	}
	SuffixChanged(SuffixComponent->text());
}

void ExtractorContextWidget::selectedEvent(QAction* action)
{
	SuffixEvent->setText(action->data().toString());
	if(action->data().toString().contains("META_NAME")){
		talkBack->setText("Edit META_NAME to required attribute ");
	}
	else if(action->data().toString().contains("i,j")){
		talkBack->setText("Edit indices i,j for required range");
	}
	else if(action->data().toString().contains("i")){
		talkBack->setText("Edit index i for required event");
	}
	EventChanged(SuffixEvent->text());
}

void ExtractorContextWidget::selectedMeta(QAction* action)
{
	SuffixMeta->setText(action->data().toString());
	if(action->data().toString().contains("META_NAME")){
		talkBack->setText("Edit META_NAME to required attribute ");
	}
	else {
		talkBack->setText("Edit index i for required event");
	}
	MetaChanged(SuffixMeta->text());
}

void ExtractorContextWidget::selectedTime(QAction* action)
{
	SuffixTime->setText(action->data().toString());
	SuffixChanged(SuffixTime->text());

}


void ExtractorContextWidget::SuffixChanged(const QString&)
{
	compCheckBox->setChecked(false);
	
	QString suffix=" ";
	if(!SuffixComponent->text().isEmpty())
	{
		if(SuffixComponent->text().startsWith("=>"))
		{
			suffix.append("[");
		}
		else
		{
			suffix.append("[ ");
		}
		suffix.append(SuffixComponent->text().trimmed());
		suffix.append(" ] ");
	}
	if(!SuffixTime->text().isEmpty())
	{
		suffix.append("< ");
		suffix.append(SuffixTime->text().trimmed());
		suffix.append(" >");
	}
	
	Suffix->setText(QString("Suffix: %1").arg(suffix));

	foreach(VExtractionItem* item, items)
	{
		item->Suffix=suffix;
		item->setComp();
		item->askUser=false;
		ExtractionItem *gui = item->getGUI();
		if(gui) {
			gui->edit->setEnabled(false);
		}

	}
}

void ExtractorContextWidget::EventChanged(const QString&)
{
	compCheckBox->setChecked(false);
	
	QString suffix=" ";
	if(!SuffixEvent->text().isEmpty())
	{
		if(SuffixEvent->text().startsWith("=>"))
		{
			suffix.append("[");
			suffix.append(SuffixEvent->text().trimmed());
			suffix.append(" ] ");
		}
		else
		{
			suffix.append("< ");
			suffix.append(SuffixEvent->text().trimmed());
			suffix.append(" > ");
		}
	}
	Suffix->setText(QString("Suffix: %1").arg(suffix));
	
	foreach(VExtractionItem* item, items)
	{
		item->Suffix=suffix;
		item->setComp();
		item->askUser=false;
		ExtractionItem *gui = item->getGUI();
		if(gui) {
			gui->edit->setEnabled(false);
		}

	}
}

void ExtractorContextWidget::MetaChanged(const QString&)
{
	compCheckBox->setChecked(false);
	
	QString suffix=" ";
	if(!SuffixMeta->text().isEmpty())
	{
		if(SuffixMeta->text().startsWith("=>"))
		{
			suffix.append("[");
			suffix.append(SuffixMeta->text().trimmed());
			suffix.append(" ] ");
		}
	}
	Suffix->setText(QString("Suffix: %1").arg(suffix));
	
	foreach(VExtractionItem* item, items)
	{
		item->Suffix=suffix;
		item->setComp();
		item->askUser=false;
		ExtractionItem *gui = item->getGUI();
		if(gui) {
			gui->edit->setEnabled(false);
		}

	}
}

void ExtractorContextWidget::CheckChanged(bool checked)
{
	if(checked){
		talkBack->setText("Calculation will pause at unresolved Extractors");
		SuffixComponent->blockSignals(true);
		SuffixTime->blockSignals(true);
			SuffixComponent->setText("");
			SuffixTime->setText("");
		SuffixComponent->blockSignals(false);
		SuffixTime->blockSignals(false);
		Suffix->setText("");
	
	}
	foreach(VExtractionItem* item, items)
	{
		item->askUser=checked;
		ExtractionItem *gui = item->getGUI();
		if(gui) gui->edit->setEnabled(checked);
	}
	

}

void ExtractorContextWidget::userSetComponent(const char *comp, const char *type){
	
	QString suffix=comp;
	Suffix->setText(QString("Suffix: %1").arg(suffix));
	
	int i1, i2;
	i1 = suffix.indexOf("[");
	if(i1>0)
	{
		i2 = suffix.indexOf("]");
		QString text=suffix.mid(i1+1, i2-i1-1);
		SuffixComponent->setText(text.trimmed());
	}
	else
	{
		SuffixComponent->setText("");
	}
	
	i1 = suffix.indexOf("<");
	if(i1>0)
	{
		i2 = suffix.indexOf(">");
		QString text = suffix.mid(i1+1, i2-i1-1);
		SuffixTime->setText(text.trimmed());
	}
	else
	{
		SuffixTime->setText("");
	}
	

}


