#include "qpsd_utils.h"
#include <QApplication>

// Constructor

PSD_Parameters::PSD_Parameters(QdObject_var& obj, int direction_in) : init_ok(false),  phi2rad(1.), theta2rad(1.)
{
    tetra.clear();
    _PI = 4.*atan(1.0);
    stop = false;
	
	v_sign = (double)direction_in;
    
	_x = 0;
	_y = 0;
	_z = 0;
	_vel = 0;
	
	_phi_centre = 0;
	_phi_plus = 0;
	_phi_minus = 0;
	
	_theta_centre = 0;
	_theta_plus = 0;
	_theta_minus = 0;

   if(!init_data(obj)) return;
	
    init_ok=true;
}

// ------------------

double PSD_Parameters::getData(int time, int r, int phi, int theta)
{
    int index[3];
    index[indexE]=r;
    index[indexPhi]=phi;
    index[indexTheta]=theta;

    return data[time][index[0]][index[1]][index[2]];
}


// ------------------

bool PSD_Parameters::init_PSDcentre(QdObject_var& centreOfPSD)
{
	if(centreOfPSD.is_nil()) {
		QdRMatrix_var zeroM = new QdRMatrix( Dim(3), 0.0);
		centre_PSD = new QdRMatrixSeq(data->sequence_size(), *zeroM);
		return true;
	}
	
	// check target frame
	if(rotn_obj.is_nil() ){
		message = "No rotation matrix";
		return false;
	}
	
	if(rotn_obj->xref_exists("ToFrame")){
		QdObject_var targetFrame_ptr = rotn_obj->get_xref("ToFrame");
		QdRScalar_var dummy = new QdRScalar(1.0);
		dummy->set_xref("Frame", targetFrame_ptr);
		
		if(QuAreFramesSame(centreOfPSD, dummy) == QDUTIL_ERR){
			message = "PSD bulk velocity must be in target frame";
			return false;
		}
	}
	
	// get in correct units
	
	QdObject_var temp=QarConvertUnitsTo(centreOfPSD, "1.>m s^-1"); // Work in SI units 

	

	if(temp.is_nil()) {
		message = "cannot get units (SI_CONVERSION) for slice normal";
		return false;
	}

	centre_PSD=QdRMatrixSeq_var::narrow(temp);
    if(!centre_PSD.is_nil())
    {
        
        if( QarTestJoined(centre_PSD, data) == QAR_ERR )
        {
			QdObject_var jObj =  QuLinearJoin(centre_PSD, get_timetags(data));

			if(jObj.is_nil()) {
				message = "centre of PSD and Data join failed.";
            	return false;
			}
			centre_PSD = QdRMatrixSeq_var::narrow(jObj);
			
        }
    }
    else
    {
        QdRMatrix_var cen_scalar=QdRMatrix_var::narrow(temp);
        if(!cen_scalar.is_nil())
        {
			centre_PSD = new QdRMatrixSeq(data->sequence_size(), *cen_scalar);
			
        }
        else
        {
			return false;
        }
    }
    
    return true;

}

// ------------------

bool PSD_Parameters::init_sliceNormal(QdObject_var& normal_in)
{
	if(normal_in.is_nil()) {
		message = "cannot get slice normal";
		return false;
	}
	
	// check target frame
	if(rotn_obj.is_nil() ){
		message = "No rotation matrix";
		return false;
	}
	
	if(rotn_obj->xref_exists("ToFrame")){
		QdObject_var targetFrame_ptr = rotn_obj->get_xref("ToFrame");
		QdRScalar_var dummy = new QdRScalar(1.0);
		dummy->set_xref("Frame", targetFrame_ptr);
		
		if(QuAreFramesSame(normal_in, dummy) == QDUTIL_ERR){
			// note, normal without frame does not give QDUTIL_ERR
			message = "Normal vector for slice must be in target frame";
			return false;
		}
	}
	
			
	sliceNormal = QdRMatrixSeq_var::narrow(normal_in);
    if(!sliceNormal.is_nil())
    {
       // This object is modified, so must use copy
	   // join will make a copy. 
		QdObject_var jObj =  QuLinearJoin(sliceNormal, get_timetags(data));

		if(jObj.is_nil()) {
			message = "slice normal and Data join failed.";
			return false;
		}
		sliceNormal = QdRMatrixSeq_var::narrow(jObj);
			
        
    }
    else
    {
        QdRMatrix_var n_mat=QdRMatrix_var::narrow(normal_in);
        if(!n_mat.is_nil())
        {
			sliceNormal = new QdRMatrixSeq(data->sequence_size(), *n_mat);
			
        }
        else
        {
			return false;
        }
    }
    
    return true;

}

// ------------------

bool PSD_Parameters::init_data(QdObject_var& obj)
{
    data=QdRMatrixSeq_var::narrow(obj);
    
    if(data.is_nil())
    {
        message = "Wrong data type for distribution.";
        return false;
    }

	int n_recs = data->sequence_size();
	int dimTheta = 0;
	int dimPhi = 0;
	
	if(_phi_centre) {
		delete _phi_centre; 
		_phi_centre = 0;
	}
	if(_phi_plus) {
		delete _phi_plus; 
		_phi_plus = 0;
	} 
	if(_phi_minus)  {
		delete _phi_minus; 
		_phi_minus = 0;
	} 
			
	if(_theta_centre)  {
		delete _theta_centre; 
		_theta_centre = 0;
	} 
	if(_theta_plus)  {
		delete _theta_plus; 
		_theta_plus = 0;
	} 
	if(_theta_minus)  {
		delete _theta_minus; 
		_theta_minus = 0;
	} 
   	
    QdObject_var temp;
    temp=QarConvertUnitsTo(data, "1>s^3 m^-6");
    data=QdRMatrixSeq_var::narrow(temp);
    if(data.is_nil())
    {
        message = "Failed to convert distribution to SI units.";
        return false;
    }
    
    dataFillVal=QString(QuGetAttrText("FILLVAL", data).c_str()).toDouble();

    chargeUnit=1.60217646e-19; //C
    massUnit=1.67262158e-27; // kg
  
    
    // Associating the xrefs to the coordinates
    
    bool initE(false), initPhi(false), initTheta(false);
    char strDepend[30];
    char strError[200];
    
    for(int i=1; i<=3; i++)
    {
        sprintf(strDepend, "DEPEND_%d", i);
        QdObject_var o;
        try
        {
            o=obj->get_xref(strDepend);
        }
        catch(Exception& e)
        {
            sprintf(strError, "Failed looking for %s", strDepend);
            message = strError;
            return false;
        }
        
        try
        {
			bool isDeg, isAngle;
			
			isAngle = QuIsAngle(o, isDeg);
			
            if( isAngle )
            {
                double conv2rad=1.;
                if(isDeg) conv2rad=_PI/180.;

                double fillVal=QString(QuGetAttrText("FILLVAL", o).c_str()).toDouble();

				double minVal;
				int kStart = 0, kEnd; // not used by qpsd
                bool isPhi = QuIsPhiAngle(o, isDeg, i, minVal, kStart, kEnd);
				
				int n_dim = 0;
				int i_num = 0;
				
				double *i_centre = 0;
				double *i_plus = 0;
				double *i_minus = 0;
				
				
                QdRMatrixSeq_var o_seq=QdRMatrixSeq_var::narrow(o);
				QdRMatrix_var o_const=QdRMatrix_var::narrow(o);
				
                if(!o_seq.is_nil())
                {
                    if( n_recs != o_seq->sequence_size()){
						sprintf(strError, "Depend seq not same length as data seq");
            			message = strError;
            			return false;
					}
					
                    n_dim=o_seq[0].dimdata()[0];
					i_num = n_recs*n_dim;
					i_centre = new double[i_num];
					i_plus = new double[i_num];
					i_minus = new double[i_num];
					
					int count = 0;

                    for(int j=0; j<n_recs; j++)
                    {
                        for(int k=0; k<n_dim; k++)
                        {
							double i_val = o_seq[j][k];

                            if( notFill(i_val, fillVal) )
                            {
								i_val *= conv2rad;

                            }
							i_centre[count] = i_val;
							count++;
							
                        }
                    }

                }
                else if(!o_const.is_nil())
				{
					n_dim=o_const->dimdata()[0];
					i_num = n_recs*n_dim;
					i_centre = new double[i_num];
					i_plus = new double[i_num];
					i_minus = new double[i_num];
						
					int count = 0;
						
					for(int j=0; j<n_dim; j++)
					{
						double i_val = (*o_const)[j];

						if( notFill(i_val, fillVal) )
						{
							i_val *= conv2rad;
						}
						for(int rec=0; rec < n_recs; rec++){
							// fill in sequence
							i_centre[rec*n_dim + count] = i_val;
						}
						count++;

					}

				}
				else
				{
					sprintf(strError, "Wrong data type for %s", strDepend);
					message = strError;
					return false;
				}
                
				
                // Get deltas
				
				QdObject_var oPlus=o->get_xref(DELTA_PLUS);
                QdRMatrixSeq_var o_plus_matrix_seq = QdRMatrixSeq_var::narrow(oPlus);
                QdRMatrix_var    o_plus_matrix = QdRMatrix_var::narrow(oPlus);
                QdRScalarSeq_var    o_plus_scalar_seq = QdRScalarSeq_var::narrow(oPlus);
                QdRScalar_var    o_plus_scalar = QdRScalar_var::narrow(oPlus);
                    
				if(o_plus_matrix_seq.is_nil() && o_plus_matrix.is_nil() && o_plus_scalar_seq.is_nil() && o_plus_scalar.is_nil())
				{
					message = "Invalid DELTA_PLUS for angle bin";
					return false;
				}

				if(!o_plus_matrix_seq.is_nil()) {
					for(int count=0; count < n_recs; count++) {
						for(int j=0; j<n_dim; j++){
							i_plus[count*n_dim + j] = i_centre[count*n_dim + j] + conv2rad * o_plus_matrix_seq[count](j);
						}
					}
				}
				if(!o_plus_matrix.is_nil()) {
					for(int count=0; count < n_recs; count++) {
						for(int j=0; j<n_dim; j++){
							i_plus[count*n_dim + j] = i_centre[count*n_dim + j] + conv2rad * (*o_plus_matrix)[j];
						}
					}
				}
				if(!o_plus_scalar_seq.is_nil()) {
					for(int count=0; count < n_recs; count++) {
						double delta = o_plus_scalar_seq[count] * conv2rad;
						for(int j=0; j<n_dim; j++){
							i_plus[count*n_dim + j] = i_centre[count*n_dim + j] + delta;
						}
					}
				}
				if(!o_plus_scalar.is_nil()) {
					double delta = *o_plus_scalar * conv2rad;
					for(int count=0; count < i_num; count++) {
						i_plus[count] = i_centre[count] + delta;
					}
				}
				
				
				QdObject_var oMinus=o->get_xref(DELTA_MINUS);
                QdRMatrixSeq_var o_minus_matrix_seq = QdRMatrixSeq_var::narrow(oMinus);
                QdRMatrix_var    o_minus_matrix = QdRMatrix_var::narrow(oMinus);
                QdRScalarSeq_var    o_minus_scalar_seq = QdRScalarSeq_var::narrow(oMinus);
                QdRScalar_var    o_minus_scalar = QdRScalar_var::narrow(oMinus);
                    
				if(o_minus_matrix_seq.is_nil() && o_minus_matrix.is_nil() && o_minus_scalar_seq.is_nil() && o_minus_scalar.is_nil())
				{
					message = "Invalid DELTA_MINUS for angle bin";
					return false;
				}

				if(!o_minus_matrix_seq.is_nil()) {
					for(int count=0; count < n_recs; count++) {
						for(int j=0; j<n_dim; j++){
							i_minus[count*n_dim + j] = i_centre[count*n_dim + j] - conv2rad * o_minus_matrix_seq[count](j);
						}
					}
				}
				if(!o_minus_matrix.is_nil()) {
					for(int count=0; count < n_recs; count++) {
						for(int j=0; j<n_dim; j++){
							i_minus[count*n_dim + j] = i_centre[count*n_dim + j] - conv2rad * (*o_minus_matrix)[j];
						}
					}
				}
				if(!o_minus_scalar_seq.is_nil()) {
					for(int count=0; count < n_recs; count++) {
						double delta = o_minus_scalar_seq[count] * conv2rad;
						for(int j=0; j<n_dim; j++){
							i_minus[count*n_dim + j] = i_centre[count*n_dim + j] - delta;
						}
					}
				}
				if(!o_minus_scalar.is_nil()) {
					double delta = *o_minus_scalar * conv2rad;
					for(int count=0; count < i_num; count++) {
						i_minus[count] = i_centre[count] -  delta;
					}
				}
					
				// set angle to centre of bin
				for(int count=0; count < i_num; count++) {
					i_centre[count] = 0.5 * (i_plus[count] +  i_minus[count]);
				}

                bool offsetTheta=false;
                
                if( !isPhi)
                {
                    if(minVal<0.)
                    {
                        offsetTheta=true;
                    }
                }
                
                 
                if(isPhi)
                {
					_phi_centre = i_centre;
					_phi_plus = i_plus;
					_phi_minus = i_minus;
                    
					dimPhi = n_dim;
                    
                    phiFillVal=fillVal;
                    phi2rad=conv2rad;
                    initPhi=true;
                    indexPhi=i-1;
                    
                    sprintf(strError, "Using \"%s\" as phi coordinate", QuGetAttrText("FIELDNAM", o).c_str());
                    message = strError;
                }
                else
                {
					_theta_centre = i_centre;
					_theta_plus   = i_plus;
					_theta_minus  = i_minus;
				
					dimTheta = n_dim;
                	// Theta <=> every value is between 0 and Pi
					
                   for(int count=0; count < i_num; count++)
                    {
                        if(offsetTheta)
                        {
                            _theta_centre[count] = _PI/2. - _theta_centre[count];
                            _theta_plus[count] = _PI/2. - _theta_plus[count];
                            _theta_minus[count] = _PI/2. - _theta_minus[count];
                        }
                    }
                    
                                        
                    thetaFillVal=fillVal;
                    theta2rad=conv2rad;
                    initTheta=true;
                    indexTheta=i-1;
                    
                    sprintf(strError, "Using \"%s\" as theta coordinate", QuGetAttrText("FIELDNAM", o).c_str());
                    message = strError;
                }
            }
            else
            {
                // We suppose that it is E
                e_const=QdRMatrix_var::narrow(o);
                e_seq=QdRMatrixSeq_var::narrow(o);
                
                if(!e_seq.is_nil())
                {
                    temp=QarConvertUnitsTo(e_seq, "1>J");
                    e_seq=QdRMatrixSeq_var::narrow(temp);
                    if(e_seq.is_nil())
                    {
                        message = "Wrong units for energy?";
                        return false;
                    }
                    eFillVal=QString(QuGetAttrText("FILLVAL", e_seq).c_str()).toDouble();
                }
                if(!e_const.is_nil())
                {
                    temp=QarConvertUnitsTo(e_const, "1>J");
                    e_const=QdRMatrix_var::narrow(temp);
                    if(e_const.is_nil())
                    {
                        message = "Wrong units for energy?";
                        return false;
                    }
                    eFillVal=QString(QuGetAttrText("FILLVAL", e_const).c_str()).toDouble();
                }

                QdObject_var ePlus=(e_seq.is_nil() ? e_const->get_xref("DELTA_PLUS") : e_seq->get_xref("DELTA_PLUS"));
                e_plus_seq=QdRMatrixSeq_var::narrow(ePlus);
                e_plus_const=QdRMatrix_var::narrow(ePlus);

                if(e_plus_seq.is_nil() && e_plus_const.is_nil())
                {
                    message = "Invalid DELTA_PLUS for energy";
                    return false;
                }
                
                QdObject_var rMinus=(e_seq.is_nil() ? e_const->get_xref("DELTA_MINUS") : e_seq->get_xref("DELTA_MINUS"));
                e_minus_seq=QdRMatrixSeq_var::narrow(rMinus);
                e_minus_const=QdRMatrix_var::narrow(rMinus);

                if(e_plus_seq.is_nil() && e_plus_const.is_nil())
                {
                    message = "Invalid DELTA_MINUS for energy";
                    return false;
                }

                initE=true;
                indexE=i-1;
                
                sprintf(strError, "Using \"%s\" as energy coordinate", QuGetAttrText("FIELDNAM", o).c_str());
                message = strError;
            }
            
        }
        catch(Exception& e)
        {
            cout << e.msg() << endl;
            return false;
        }
    }
    
    if(!initE || !initPhi || !initTheta)
    {
        message = "Failed finding the 3 coordinates";
        return false;
    }
    
    Dim mDim=data[0].dimdata();
    if(mDim[indexE]!=(e_const.is_nil() ? e_seq[0].dimdata()[0] : e_const->dimdata()[0]))
    { 
        message = "Dimensions for data and energy coordinates do not match";
        return false;
    }
    if(mDim[indexTheta] != dimTheta)
    {
        message = "Dimensions for data and theta coordinates do not match";
        return false;
    }
    if(mDim[indexPhi] != dimPhi)
    {
        message = "Dimensions for data and Phi coordinates do not match";
        return false;
    }
    
    maxTime=data->sequence_size()-1;
    maxE=mDim[indexE]-1;
    maxTheta=mDim[indexTheta]-1;
    maxPhi=mDim[indexPhi]-1;
   
    return true;
}



// ------------------

bool PSD_Parameters::init_cartesian()
{
	// set up space for cartesian coords and other versions of values used often

	
	if(_x) {delete _x; _x=NULL;}
	if(_y) {delete _y; _y=NULL;}
	if(_z) {delete _z; _z=NULL;}
	if(_vel) {delete _vel; _vel=NULL;}

	try {
		_x = new double[(maxE+1) * (maxTheta+1) * (maxPhi+1)];
		_y = new double[(maxE+1) * (maxTheta+1) * (maxPhi+1)];
		_z = new double[(maxE+1) * (maxTheta+1) * (maxPhi+1)];
		_vel = new double[(maxE+1)];
	}
	catch (Exception& e) {
		message = "Cannot allocate memory for cartesian array";
		return false;
	}	
    return true;

}
// ------------------

bool PSD_Parameters::fill_cartesian(int time)
{
	// set up cartesian coords and other versions of values used often
	if(data.is_nil()) return false;
	
	int n_seq = data->sequence_size();
	double _theta, _phi;
	double x, y, z;
	
	// loops must be in this order
	int j=0;
	
	for(int e=0; e<=maxE; e++){
		_vel[e] = getV(time, e);
	}
	
	for(int phi=0; phi<=maxPhi; phi++){
		_phi = getPhi(time, phi);
				
		for(int theta=0; theta<=maxTheta; theta++){
			_theta = getTheta(time, theta);

			for(int e=0; e<=maxE; e++){
				double r = _vel[e];
				if(r >= 0.)
				{
					r *= v_sign;
					x = r*sin(_theta)*cos(_phi);
					y = r*sin(_theta)*sin(_phi);
					z = r*cos(_theta);
					// rotate
/* post-mult mtx
					if(!rotn_const.is_nil()) {
						_x[j] = x*(*rotn_const)(0,0) + y*(*rotn_const)(1,0) + z*(*rotn_const)(2,0);
						_y[j] = x*(*rotn_const)(0,1) + y*(*rotn_const)(1,1) + z*(*rotn_const)(2,1);
						_z[j] = x*(*rotn_const)(0,2) + y*(*rotn_const)(1,2) + z*(*rotn_const)(2,2);
					}
					else {
						_x[j] = x*(*rotn_seq)[time](0,0) + y*(*rotn_seq)[time](1,0) + z*(*rotn_seq)[time](2,0);
						_y[j] = x*(*rotn_seq)[time](0,1) + y*(*rotn_seq)[time](1,1) + z*(*rotn_seq)[time](2,1);
						_z[j] = x*(*rotn_seq)[time](0,2) + y*(*rotn_seq)[time](1,2) + z*(*rotn_seq)[time](2,2);
					}
*/
					if(!rotn_const.is_nil()) {
						_x[j] = x*(*rotn_const)(0,0) + y*(*rotn_const)(0,1) + z*(*rotn_const)(0,2);
						_y[j] = x*(*rotn_const)(1,0) + y*(*rotn_const)(1,1) + z*(*rotn_const)(1,2);
						_z[j] = x*(*rotn_const)(2,0) + y*(*rotn_const)(2,1) + z*(*rotn_const)(2,2);
					}
					else {
						_x[j] = x*(*rotn_seq)[time](0,0) + y*(*rotn_seq)[time](0,1) + z*(*rotn_seq)[time](0,2);
						_y[j] = x*(*rotn_seq)[time](1,0) + y*(*rotn_seq)[time](1,1) + z*(*rotn_seq)[time](1,2);
						_z[j] = x*(*rotn_seq)[time](2,0) + y*(*rotn_seq)[time](2,1) + z*(*rotn_seq)[time](2,2);
					}
				}
				else{
					_x[j] = 0;
					_y[j] = 0;
					_z[j] = 0;
				}
				j++;
			}
		}
	}
    return true;
}


// ------------------

bool PSD_Parameters::init_potential(QdObject_var & obj)
{
    potential=QdRScalarSeq_var::narrow(obj);
    if(!potential.is_nil())
    {
        QdObject_var temp=QarConvertUnitsTo(potential, "1>V");
        if(temp.is_nil())
        {
            message = "Potential (sequence) conversion to Volts failed - Assuming values are in Volts";
        }
		else  potential=QdRScalarSeq_var::narrow(temp);
        
        if( QarTestJoined(potential, data) == QAR_ERR )
        {
			QdObject_var jObj =  QuLinearJoin(potential, get_timetags(data));

			if(jObj.is_nil()) {
				message = "Potential and Data join failed.";
            	return false;
			}
			potential = QdRScalarSeq_var::narrow(jObj);
			
        }
        //TODO potFillVal
    }
    else
    {
        QdRScalar_var pot_scalar=QdRScalar_var::narrow(obj);
        if(!pot_scalar.is_nil())
        {
       		QdObject_var temp=QarConvertUnitsTo(pot_scalar, "1>V");
			
			if(!temp.is_nil())  pot_scalar=QdRScalar_var::narrow(temp);
			
            potential=new QdRScalarSeq(data->sequence_size(), *pot_scalar);
			
        }
        else
        {
            potential=new QdRScalarSeq(data->sequence_size(), 0.0);
        }
    }
    
    return true;
}

// ------------------

bool PSD_Parameters::init_threshold(QdObject_var & obj)
{
    threshold=QdRScalarSeq_var::narrow(obj);
    if(!threshold.is_nil())
    {
        QdObject_var temp=QarConvertUnitsTo(threshold, "1>J");
        if(temp.is_nil())
        {
            message = "Threshold conversion to Joules failed, assuming eV";
			(*threshold) *=  1.60217646e-19;  // convert eV to J
        }
        else threshold=QdRScalarSeq_var::narrow(temp);
        
        if( QarTestJoined(threshold, data) == QAR_ERR )
        {
			QdObject_var jObj =  QuLinearJoin(threshold, get_timetags(data));

			if(jObj.is_nil()) {
				message = "Potential and Data join failed.";
            	return false;
			}
			threshold = QdRScalarSeq_var::narrow(jObj);
			
        }
    }
    else
    {
        QdRScalar_var thr_scalar=QdRScalar_var::narrow(obj);
        if(!thr_scalar.is_nil())
        {
			QdObject_var temp=QarConvertUnitsTo(thr_scalar, "1>J");
			
			if(!temp.is_nil())  thr_scalar=QdRScalar_var::narrow(temp);
			else (*thr_scalar) *= 1.60217646e-19; // assume eV and convert to J
			
            threshold=new QdRScalarSeq(data->sequence_size(), *thr_scalar);
        }
        else
        {
            threshold=new QdRScalarSeq(data->sequence_size(), 0.);
        }
    }
    
    return true;
}

// ------------------


bool PSD_Parameters::init_rotn(QdObject_var& rotMatx)
{
	rotn_obj = rotMatx;

    rotn_seq=QdRMatrixSeq_var::narrow(rotMatx);
    if(!rotn_seq.is_nil())
    {
        int qar_status=QarTestJoined(rotn_seq, data);
        if(qar_status==QAR_ERR)
        {
			// join rotnMatrix to data tt
			QdObject_var jObj =  QuLinearJoin(rotn_obj, get_timetags(data));

			if(jObj.is_nil() ) {
				message = "rotation sequence and Data not joined.";
            	return false;
			}
	
			rotn_obj   = jObj;
			rotn_seq   = QdRMatrixSeq_var::narrow(rotn_obj);
			if(rotn_seq.is_nil()) {
				message = "rotation matrix not sequence after join to Data.";
            	return false;
			}
        }
        
        if(rotn_seq[0].rank()!=2)
        {
            message = "Wrong dimensions for rotation matrix.";
            return false;
        }
        
        if(rotn_seq[0].dimdata()[0]!=3 || rotn_seq[0].dimdata()[1]!=3)
        {
            message = "Wrong dimensions for rotation matrix.";
            return false;
        }
        
        type_rotn=MATRIX_SEQ;
        return true;
    }
    
    rotn_const=QdRMatrix_var::narrow(rotMatx);
    if(!rotn_const.is_nil())
    {
        if(rotn_const->rank()!=2)
        {
            message = "Wrong dimensions for rotation matrix.";
            return false;
        }
        
        if(rotn_const->dimdata()[0]!=3 || rotn_const->dimdata()[1]!=3)
        {
            message = "Wrong dimensions for rotation matrix.";
            return false;
        }
        
        type_rotn=MATRIX;
        return true;
    }
    
    type_rotn=UNKNOWN;
    
    return true;
}


// ------------------


QdRMatrix_var PSD_Parameters::computeValuesAlongVector(int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double v_min, double v_max, int nb_values)
{
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "MomentsParameters::computeValuesAlongVector: nil centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    if(CoM->dimdata()[0]!=3 || direction->dimdata()[0]!=3)
    {
        cout << "MomentsParameters::computeValuesAlongVector: wrong dimension for centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    double val;
    

	buildTetra(time);
    
    
    QdRMatrix_var res=new QdRMatrix(Dim(nb_values));
    double element_size=(v_max-v_min)/nb_values;
    
    QdRMatrix_var rotation=new QdRMatrix(Dim(3,3));
//    getRotationMatrix(*direction, rotation);
    
    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<nb_values; i++)
        {
            double _z=v_min+(i+0.5)*element_size;
            double x=(*rotation)[0][2]*_z+(*CoM)[0];
            double y=(*rotation)[1][2]*_z+(*CoM)[1];
            double z=(*rotation)[2][2]*_z+(*CoM)[2];
            
            Point3D temp(x, y, z, -1.);
            temp.orig_x=i;
            points.push_front(temp);
        }
        
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
        
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            (*res)[(int)(iter->orig_x)]=iter->val;
        }
        
        // Creation of the metadata
		QdRScalar_var unitFill = new QdRScalar(-1.);
        res->change_xref(FILL_VALUE, unitFill );
        QuSetTxtAttr("UNITS", "s^3/m^6", (QdObject_var) res);
        string fieldnam = QuGetAttrText(FIELDNAM, data);
        fieldnam += " - Values along vector";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) res);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) res);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) res);
        QuSetTxtAttr("SI_CONVERSION", "1>s^3 m^-6", (QdObject_var) res);
//         QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) res);
        
        QdRMatrix_var axis=new QdRMatrix(Dim(nb_values));
        for(int i=0; i<nb_values; i++)
        {
            (*axis)[i]=(i+0.5)*element_size+v_min;
        }
		QdRScalar_var doubleFill = new QdRScalar(-1e+31);
        axis->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) axis);
        fieldnam = QuGetAttrText(FIELDNAM, data);
		fieldnam += " - Values along vector - Axis";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) axis);
        
        QdRScalar_var delta_axis=new QdRScalar(element_size/2.);
        delta_axis->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) delta_axis);
        fieldnam =  QuGetAttrText(FIELDNAM, data);
		fieldnam +=  " - Values along vector - Axis Delta";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) delta_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) delta_axis);
        
        axis->set_xref("DELTA_PLUS", (QdObject_var)delta_axis);
        axis->set_xref("DELTA_MINUS", (QdObject_var)delta_axis);
        res->set_xref("DEPEND_1", (QdObject_var)axis);
    }
    else
    {
        return QdRMatrix_var();
    }

    return res;
}

//----------------------------

QdRMatrix_var PSD_Parameters::computePAD(int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double v_min, double v_max, int nb_v, int nb_theta)
{
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "PSD_Parameters::computePAD: nil centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    if(CoM->dimdata()[0]!=3 || direction->dimdata()[0]!=3)
    {
        cout << "PSD_Parameters::computePAD: wrong dimension for centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    

	buildTetra(time);
    
    
    QdRMatrix_var res=new QdRMatrix(Dim(nb_v, nb_theta));
    *res=-1.;
    double element_size_v=(v_max-v_min)/nb_v;
    
    // Find rotation matrix
    
    QdRMatrix_var rotation=new QdRMatrix(Dim(3,3));
 //   getRotationMatrix(*direction, rotation);
    
    // End find rotation matrix
    
    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<nb_v; i++)
        {
            double v=v_min+(i+0.5)*element_size_v;
            for(int j=0; j<nb_theta; j++)
            {
                double theta=(_PI/nb_theta)*(j+0.5);
                double total=0.;
                bool foundOnce=false;
                
                for(int k=0; k<360; k+=10)
                {
                    double phi=_PI/180.*(k+5.);
                    double x=v*sin(theta)*cos(phi);
                    double y=v*sin(theta)*sin(phi);
                    double z=v*cos(theta);
                    
                    Point3D temp(x, y, z, -1.);
                    temp.orig_x=i;
                    temp.orig_y=j;
                    points.push_front(temp);
                }
               
            }
        }
        
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
                
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            if(iter->val>=0.)
            {
                if((*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]>=0.)
                    (*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]+=iter->val*10.;
                else
                    (*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]=iter->val*10.;
            }
        }
        
        // Creation of the metadata
		QdRScalar_var unitFill = new QdRScalar(-1.);
        res->change_xref(FILL_VALUE, unitFill );
        QuSetTxtAttr("UNITS", "s^3 rad/m^6", (QdObject_var) res);
        string fieldnam = QuGetAttrText(FIELDNAM, data);
		fieldnam += " - PAD";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) res);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) res);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) res);
        QuSetTxtAttr("SI_CONVERSION", "1>s^3 rad m^-6", (QdObject_var) res);
        QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) res);
        
        QdRMatrix_var v_axis=new QdRMatrix(Dim(nb_v));
        for(int i=0; i<nb_v; i++)
        {
            (*v_axis)[i]=v_min+(i+0.5)*element_size_v;
        }
		QdRScalar_var doubleFill = new QdRScalar(-1.);
        v_axis->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) v_axis);
        fieldnam = QuGetAttrText(FIELDNAM, data);
		fieldnam += " - PAD - V Axis";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) v_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) v_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) v_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) v_axis);
        
        QdRScalar_var delta_v_axis=new QdRScalar(element_size_v/2.);
        delta_v_axis->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) delta_v_axis);
        fieldnam = QuGetAttrText(FIELDNAM, data);
		fieldnam += " - PAD - V Axis Delta";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) delta_v_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_v_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_v_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) delta_v_axis);
        
        QdRMatrix_var t_axis=new QdRMatrix(Dim(nb_theta));
        for(int i=0; i<nb_theta; i++)
        {
            (*t_axis)[i]=(_PI/nb_theta)*(i+0.5);
        }
        t_axis->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr("UNITS", "rad", (QdObject_var) t_axis);
        fieldnam = QuGetAttrText(FIELDNAM, data);
		fieldnam += " - PAD - Theta Axis";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) t_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) t_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) t_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>rad", (QdObject_var) t_axis);
        
        QdRScalar_var delta_t_axis=new QdRScalar(5./180.*_PI);
        delta_t_axis->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr("UNITS", "rad", (QdObject_var) delta_t_axis);
        fieldnam = QuGetAttrText(FIELDNAM, data);
		fieldnam += " - PAD - Theta Axis Delta";
        QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) delta_t_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_t_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_t_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>rad", (QdObject_var) delta_t_axis);
        
        v_axis->set_xref("DELTA_PLUS", (QdObject_var)delta_v_axis);
        v_axis->set_xref("DELTA_MINUS", (QdObject_var)delta_v_axis);
        t_axis->set_xref("DELTA_PLUS", (QdObject_var)delta_t_axis);
        t_axis->set_xref("DELTA_MINUS", (QdObject_var)delta_t_axis);
        res->set_xref("DEPEND_1", (QdObject_var)v_axis);
        res->set_xref("DEPEND_2", (QdObject_var)t_axis);
    }
    else
    {
        return QdRMatrix_var();
    }
    
    return res;
}



// ------------------

void PSD_Parameters::buildTetra(int time)
{

    tetra.clear();
    list<Point3D> points;
    double val, vol, ePThis, eMThis;
	double x, y, z;
	
	// for finding Vmax and min
	
	int e_of_max = -1;
	int e_of_min = -1;	
	
	double max_E = 0;
	double min_E = 1.e31;
	for(int e=0; e<=maxE; e++)
	{
		if(_vel[e] >= 0.){
			ePThis = getEPlus(time, e);	
			eMThis = getEMinus(time, e);
			if(ePThis > max_E)
			{
				max_E=ePThis;
				e_of_max = e;
			}
			if(eMThis > 0. && eMThis < min_E)
			{
				min_E=eMThis;
				e_of_min = e;
			}
		}
	}
	
	// find points
	
	int j = 0;
	
    for(int phi=0; phi<=maxPhi; phi++)
    {		
        for(int theta=0; theta<=maxTheta; theta++)
        {
            for(int e=0; e<=maxE; e++)
            {
                val = getData(time, e, phi, theta);
				
                if( notFill(val, dataFillVal) )
                {
					vol = getElementVolume(time, e, phi, theta);

                    if(_vel[e] >= 0. && vol > 0.)
                    {
						// set point of bin centre
						Point3D temp(_x[j], _y[j], _z[j], val, vol);
                        points.push_front(temp);
						
						// Max and min points...
						// direction is same for E and EPlus/minus, so just renormalise
						if(e == e_of_max)
                        {
							double max_v = E2V(max_E)/_vel[e];
			
							x = max_v * _x[j];
            				y = max_v * _y[j];
            				z = max_v * _z[j];
               
							Point3D temp_max(x, y, z, val, vol);
							points.push_front(temp_max);
							
                        }
                        if(e == e_of_min)
                        {
							double min_v = E2V(min_E)/_vel[e];
			
							x = min_v * _x[j];
            				y = min_v * _y[j];
            				z = min_v * _z[j];
               
							Point3D temp_min(x, y, z, val, vol);
							points.push_front(temp_min);
                        }
						
						
                    }
                }
				j++;
            }
			
        }
    }
    if(points.size()>4)
    {
         triangulation(points, tetra);
    }

}

//-----------------------------

void PSD_Parameters::getRotationMatrix(double direction[], double rotation[][3])
{
	// NOTE this modifies the direction matrix, so it must be a copy of WL obj
    double norm=sqrt(direction[0]*direction[0]+direction[1]*direction[1]+direction[2]*direction[2]);
    direction[0]/=norm;
    direction[1]/=norm;
    direction[2]/=norm;
    
    if(fabs(direction[0])<0.01 && fabs(direction[1])<0.01)
    {
        if(direction[2]>0.)
        {
            rotation[0][0]=1.;
            rotation[1][1]=1.;
            rotation[2][2]=1.;
            
            rotation[0][1]=0.;
            rotation[0][2]=0.;
            rotation[1][0]=0.;
            rotation[1][2]=0.;
            rotation[2][0]=0.;
            rotation[2][1]=0.;
        }
        else
        {
            rotation[0][0]=-1.;
            rotation[1][1]=-1.;
            rotation[2][2]=-1.;
            
            rotation[0][1]=0.;
            rotation[0][2]=0.;
            rotation[1][0]=0.;
            rotation[1][2]=0.;
            rotation[2][0]=0.;
            rotation[2][1]=0.;
        }
    }
    else
    {
        double normal[3]; // Normal to z axis and direction vector
        normal[0]=-(double)direction[1];
        normal[1]=direction[0];
        normal[2]=0.;
        double normal_norm=sqrt(normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);
        normal[0]/=normal_norm;
        normal[1]/=normal_norm;
        normal[2]/=normal_norm;
        double angle=acos(direction[2]);// Rotation angle around the normal so that z matches the direction 
        // Change angle sign if necessary
        //if ((normal X z).direction >0, sin>0, no sign change)
        if(normal[1]*direction[0]-normal[0]*direction[1] <0.) angle=-angle;

        rotation[0][0]=1. + (1.-cos(angle))*(normal[0]*normal[0]-1);
        rotation[0][1]=-normal[2]*sin(angle)+(1.-cos(angle))*normal[0]*normal[1];
        rotation[0][2]=normal[1]*sin(angle)+(1.-cos(angle))*normal[0]*normal[2];
        rotation[1][0]=normal[2]*sin(angle)+(1.-cos(angle))*normal[0]*normal[1];
        rotation[1][1]=1. + (1.-cos(angle))*(normal[1]*normal[1]-1);
        rotation[1][2]=-normal[0]*sin(angle)+(1.-cos(angle))*normal[1]*normal[2];
        rotation[2][0]=-normal[1]*sin(angle)+(1.-cos(angle))*normal[0]*normal[2];
        rotation[2][1]=normal[0]*sin(angle)+(1.-cos(angle))*normal[1]*normal[2];
        rotation[2][2]=1. + (1.-cos(angle))*(normal[2]*normal[2]-1);
    }
}

// ------------------

void PSD_Parameters::getSlice(QdRMatrixSeq_var& PSD_slice, QLabel *percLabl)
{
    double val, vol, x, y, z, w;
    double scalemin(1.e30), scalemax(-1.);
		
    int seq_size=data->sequence_size();

	PSD_slice = new QdRMatrixSeq (seq_size, Dim(resolution, resolution));
	
	// get V scale
	double max_v=0.;
   	for(int time=0; time<seq_size; time++){
		for(int e=0; e<=maxE; e++)
    	{
        	double v=getVPlus(time, e);
        	if(v>max_v) max_v=v;
    	}
    }
	double element_size=2.*max_v/resolution;
	
	// get slices
    for(int time=0; time<seq_size; time++)
    {
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}

		
		// get points in rotated cartesian frame
		
		fill_cartesian(time);
        
		// build the teselating tetrahedra
		
		buildTetra(time);
		// get the actual slice
		
		
    
    	// Find rotation matrix into normal
    
    	double rotation[3][3];
		
		double dirn[3];
		dirn[0] = sliceNormal[time][0];
		dirn[1] = sliceNormal[time][1];
		dirn[2] = sliceNormal[time][2];
		
    	getRotationMatrix(dirn, rotation);

    	if(!tetra.empty())
    	{
        	list<Point3D> points;
			for(int i=0; i<resolution; i++)
        	{
            	for(int j=0; j<resolution; j++)
            	{
                	double _x=(i+0.5)*element_size-max_v;
                	double _y=(j+0.5)*element_size-max_v;
                	double _z=0.;
					
                	double x=rotation[0][0]*_x+rotation[0][1]*_y+rotation[0][2]*_z+(*centre_PSD)[time](0);
                	double y=rotation[1][0]*_x+rotation[1][1]*_y+rotation[1][2]*_z+(*centre_PSD)[time](1);
                	double z=rotation[2][0]*_x+rotation[2][1]*_y+rotation[2][2]*_z+(*centre_PSD)[time](2);
                
                	Point3D temp(x, y, z, -1.);
                	temp.orig_x=i;
                	temp.orig_y=j;
                	points.push_front(temp);
                
            	}
        	}
        	interpolation(points, tetra);
        
        	for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        	{
            	PSD_slice[time][(int)(iter->orig_x)][(int)(iter->orig_y)] = iter->val;
				
				// update scale min/max
				if( iter->val > 0) // not fill
				{
					if( iter->val < scalemin) scalemin= iter->val;
					if( iter->val > scalemax) scalemax= iter->val;					
				}
				
        	}
        }		


	}
	
	//  get xrefs and modify DEPENDS
	
	PSD_slice->copy_xrefs_from(data);

	QdRScalar_var unitFill = new QdRScalar(-1.);
	PSD_slice->change_xref(FILL_VALUE, unitFill );
	QuSetTxtAttr("UNITS", "s^3/m^6", (QdObject_var) PSD_slice);
	string fieldnam = QuGetAttrText(FIELDNAM, data);
	fieldnam +=  " - Slice";
	QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) PSD_slice);
	QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) PSD_slice);
	QuSetTxtAttr("SI_CONVERSION", "1>s^3 m^-6", (QdObject_var) PSD_slice);
	QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) PSD_slice);
        
	QdRMatrix_var axis=new QdRMatrix(Dim(resolution));
	for(int i=0; i<resolution; i++)
	{
		(*axis)[i]=(i+0.5)*element_size-max_v;
	}
	QdRScalar_var doubleFill = new QdRScalar(-1.);
	axis->change_xref(FILL_VALUE, doubleFill );
	QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) axis);
	fieldnam = QuGetAttrText(FIELDNAM, data);
	fieldnam +=   " - Slice - Axis";
	QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) axis);
	QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) axis);
	QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) axis);
	QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) axis);
        
	QdRScalar_var delta_axis=new QdRScalar(element_size/2.);
	delta_axis->change_xref(FILL_VALUE, doubleFill );
	QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) delta_axis);
	fieldnam = QuGetAttrText(FIELDNAM, data);
	fieldnam +=   " - Slice - Axis Delta";
	QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), (QdObject_var) delta_axis);
	QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_axis);
	QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_axis);
	QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) delta_axis);
        
	axis->set_xref("DELTA_PLUS", (QdObject_var)delta_axis);
	axis->set_xref("DELTA_MINUS", (QdObject_var)delta_axis);
	PSD_slice->change_xref("DEPEND_1", (QdObject_var)axis);
	PSD_slice->change_xref("DEPEND_2", (QdObject_var)axis);
	if (PSD_slice->xref_exists("DEPEND_3")) PSD_slice->delete_xref("DEPEND_3");
     
	
    PSD_slice->change_xref("SCALEMIN", new QdRScalar(scalemin));
    PSD_slice->change_xref("SCALEMAX", new QdRScalar(scalemax));
	
	

}

// ------------------

void PSD_Parameters::getMomentsSummed(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature, QLabel *percLabl)
{
    density=new QdRScalarSeq(data->sequence_size());
    velocity=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    pressure=new QdRMatrixSeq(data->sequence_size(), Dim(3, 3));
    heat=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    temperature=new QdRScalarSeq(data->sequence_size());
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v_centre[3], v_centre_norm;
    double val, vol, _phi, _theta;
    bool density_ok, velocity_ok, pressure_ok, heat_ok;
    double w;
	
    int seq_size=data->sequence_size();

    struct timeval t1;
    struct timeval t2;
    
    for(int time=0; time<seq_size; time++)
    {
		int j=0;
		
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}

		fill_cartesian(time);
		
		tot_density=0.; density_ok=false;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; velocity_ok=false;
		
        // Computation of density and velocity
		// loops must be in this order
        for(int phi=0; phi<=maxPhi; phi++)
        {
            for(int theta=0; theta<=maxTheta; theta++)
            {
                for(int e=0; e<=maxE; e++)
                {
                    val = getData(time, e, phi, theta);
                    
                    if(val>=0.)
                    {
						vol = getElementVolume(time, e, phi, theta);
						
						if(vol > 0.){
							w = val*vol;
                        	tot_density += w;

                       	 density_ok=true;
                        
                       	 if(flags & MOM_VELOCITY)
                       	 {
								tot_numberflux[0] += _x[j]*w;
								tot_numberflux[1] += _y[j]*w;
								tot_numberflux[2] += _z[j]*w;
								velocity_ok=true;
                       
                        	}
						}
                    }
					j++;
                }
            }
        }
		
        // re-use j counter
		j=0;

        if(density_ok) density[time]=tot_density;
        else density[time]=-1;
        
        if(flags & MOM_VELOCITY)
        {
            if(velocity_ok && density_ok)
            {
                velocity[time][0]=tot_numberflux[0]/(density[time]);
                velocity[time][1]=tot_numberflux[1]/(density[time]);
                velocity[time][2]=tot_numberflux[2]/(density[time]);
            }
            else
            {
                velocity[time]=-1e+31;
            }
        }

        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && (double)velocity[time][0]>-.9e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            pressure_ok=false;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            heat_ok=false;

            for(int phi=0; phi<=maxPhi; phi++)
            {
                for(int theta=0; theta<=maxTheta; theta++)
                {
                    for(int e=0; e<=maxE; e++)
                    {
                        val = getData(time, e, phi, theta);

                        if(val>=0.)
                        {
                        	vol = getElementVolume(time, e, phi, theta);
							if (vol > 0.){ 
								w = vol*val;
						
								v_centre[0] = _x[j] - (double)velocity[time][0];
								v_centre[1] = _y[j] - (double)velocity[time][1];
								v_centre[2] = _z[j] - (double)velocity[time][2];
                                
								v_centre_norm=v_centre[0]*v_centre[0]+v_centre[1]*v_centre[1]+v_centre[2]*v_centre[2];
    
								if(flags & MOM_PRESSURE)
								{
									for(int i=0; i<3; i++)
									{
										for(int ii=0; ii<3; ii++)
										{
											tot_pressure[i][ii] += v_centre[i]*v_centre[ii]*w;
										}
									}
									pressure_ok=true;
								}
                                
								if(flags & MOM_HEATFLUX)
								{
									for(int i=0; i<3; i++)
									{
										tot_heat[i] += v_centre_norm*v_centre[i]*w;
									}
									heat_ok=true;
								}
							}
						}
						j++;

                    } // end for e
                } // end for theta
            } // end for phi
            
        } // end computation of pressure and heat

        if(flags & MOM_PRESSURE)
        {
            if(pressure_ok)
            {
                pressure[time][0][0]=tot_pressure[0][0]*mass;
                pressure[time][1][0]=tot_pressure[1][0]*mass;
                pressure[time][2][0]=tot_pressure[2][0]*mass;
                pressure[time][0][1]=tot_pressure[0][1]*mass;
                pressure[time][1][1]=tot_pressure[1][1]*mass;
                pressure[time][2][1]=tot_pressure[2][1]*mass;
                pressure[time][0][2]=tot_pressure[0][2]*mass;
                pressure[time][1][2]=tot_pressure[1][2]*mass;
                pressure[time][2][2]=tot_pressure[2][2]*mass;
                
                temperature[time]=(pressure[time][0][0]+pressure[time][1][1]+pressure[time][2][2])/(3.*density[time]*1.3806503e-23);
            }
            else
            {
                pressure[time][0][0]=-1e+31;
                pressure[time][1][0]=-1e+31;
                pressure[time][2][0]=-1e+31;
                pressure[time][0][1]=-1e+31;
                pressure[time][1][1]=-1e+31;
                pressure[time][2][1]=-1e+31;
                pressure[time][0][2]=-1e+31;
                pressure[time][1][2]=-1e+31;
                pressure[time][2][2]=-1e+31;
                temperature[time]=-1e+31;
            }
        }
        
        if(flags & MOM_HEATFLUX)
        {
            if(heat_ok)
            {
                heat[time][0]=tot_heat[0]*mass/2.;
                heat[time][1]=tot_heat[1]*mass/2.;
                heat[time][2]=tot_heat[2]*mass/2.;
            }
            else
            {
				heat[time][0]=-1e+31;
                heat[time][1]=-1e+31;
                heat[time][2]=-1e+31;
            }
        }

    } // end loop over records (time)

    // Metadata
	QdRScalar_var unitFill = new QdRScalar(-1.);
	QdRScalar_var doubleFill = new QdRScalar(-1e31);
    
    if(flags & MOM_DENSITY)
    {
		set_timetags(density, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(density, 0, seq_size-1);
			density = QdRScalarSeq_var::narrow(sub);
		}

		if(!density.is_nil()){
       		for(int i=0; i<seq_size; i++)
        	{
            	if((double)density[i]!=-1) density[i]=density[i]/1000000.;
        	}
		}
		
		QdRScalar_var unitFill = new QdRScalar(-1.);
		density->change_xref(FILL_VALUE, unitFill );
        QuSetTxtAttr(UNITS, "cm^-3", (QdObject_var) density);
        QuSetTxtAttr(FIELDNAM, "Density - Sum", (QdObject_var) density);
        QuSetTxtAttr(FRAME, "scalar>na", (QdObject_var) density);
        QuSetTxtAttr(PARAMETER_TYPE, "Data", (QdObject_var) density);
        QuSetTxtAttr(SI_CONVERSION, "1.0e6>m^-3", (QdObject_var) density);
        QuSetTxtAttr(DISPLAYTYPE, "Time_Series", (QdObject_var) density);
    }
    
    if(flags & MOM_VELOCITY)
    {
		set_timetags(velocity, get_timetags(density));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(velocity, 0, seq_size-1);
			velocity = QdRMatrixSeq_var::narrow(sub);
		}
		
		if(!velocity.is_nil()){
        	for(int i=0; i<seq_size; i++)
       		{
            	if((double)velocity[i][0] > -.9e+31)
            	{
                	velocity[i][0]=velocity[i][0]/1000.;
                	velocity[i][1]=velocity[i][1]/1000.;
                	velocity[i][2]=velocity[i][2]/1000.;
            	}
        	}
		}
        QuSetTxtAttr(UNITS, "km s^-1", (QdObject_var) velocity);
        QuSetTxtAttr(FIELDNAM, "Bulk Velocity", (QdObject_var) velocity);
		velocity->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr(SI_CONVERSION, "1.0e3>m s^-1", (QdObject_var) velocity);
        QuSetTxtAttr(DISPLAYTYPE, "Time_Series", (QdObject_var) velocity);
        QuSetTxtAttr(FRAME, "vector>gse_xyz", (QdObject_var) velocity);
    }
    
    if(flags & MOM_PRESSURE)
    {
		set_timetags(pressure, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(pressure, 0, seq_size-1);
			pressure = QdRMatrixSeq_var::narrow(sub);
		}

		if(!pressure.is_nil()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if((double)pressure[i][0][0] > -.9e+31)
            	{
                	for(int j=0; j<3; j++)
                	{
                    	for(int k=0; k<3; k++)
                    	{
                        	pressure[i][j][k]=pressure[i][j][k]*1000000000.;
                    	}
                	}
            	}
			}
        }
        
		QdSMatrix_var rep=new QdSMatrix(Dim(3));
		(*rep)[0]="x";
		(*rep)[1]="y";
		(*rep)[2]="z";
		pressure->set_xref(REPRESENTATION_1, rep);
		pressure->set_xref(REPRESENTATION_2, rep);
		pressure->set_xref(LABEL_1, rep);
		pressure->set_xref(LABEL_2, rep);
		
		QdRMatrix_var dep=new QdRMatrix(Dim(3));
		(*dep)[0]=1;
		(*dep)[1]=2;
		(*dep)[2]=3;
		pressure->set_xref(DEPEND_1, dep);
		pressure->set_xref(DEPEND_2, dep);
		
        QuSetTxtAttr(UNITS, "nPa", (QdObject_var) pressure);
        QuSetTxtAttr(FIELDNAM, "Pressure Tensor", (QdObject_var) pressure);
		pressure->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr(SI_CONVERSION, "1.0e-9>Pa", (QdObject_var) pressure);
        QuSetTxtAttr(DISPLAYTYPE, "Time_Series", (QdObject_var) pressure);
        QuSetTxtAttr(FRAME, "vector>gse_xyz", (QdObject_var) pressure);
		QuSetTxtAttr(TENSOR_ORDER, "2", (QdObject_var) pressure);
    }
    
    if((flags & MOM_HEATFLUX) || (flags & MOM_PRESSURE))
    {
        set_timetags(heat, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(heat, 0, seq_size-1);
			heat = QdRMatrixSeq_var::narrow(sub);
		}
		if(!heat.is_nil()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if((double)heat[i][0] > -.9e+31)
            	{
                	heat[i][0]=heat[i][0]*1000000.;
                	heat[i][1]=heat[i][1]*1000000.;
                	heat[i][2]=heat[i][2]*1000000.;
            	}
        	}
		}
		QuSetTxtAttr(UNITS, "microW m^-2", (QdObject_var) heat);
        QuSetTxtAttr(FIELDNAM, "Heat Flux", (QdObject_var) heat);
		heat->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr(SI_CONVERSION, "1.0e-6>W m^-2", (QdObject_var) heat);
        QuSetTxtAttr(DISPLAYTYPE, "Time_Series", (QdObject_var) heat);
        QuSetTxtAttr(FRAME, "vector>gse_xyz", (QdObject_var) heat);
        
        set_timetags(temperature, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(temperature, 0, seq_size-1);
			temperature = QdRScalarSeq_var::narrow(sub);
		}
		if(!temperature.is_nil()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if((double)temperature[i] > -.9e+31) temperature[i]=temperature[i]/1000000.;
        	}
		}
		temperature->change_xref(FILL_VALUE, doubleFill );
        QuSetTxtAttr(UNITS, "MK", (QdObject_var) temperature);
        QuSetTxtAttr(FIELDNAM, "Temperature", (QdObject_var) temperature);
        QuSetTxtAttr(FRAME, "scalar>na", (QdObject_var) temperature);
        QuSetTxtAttr(PARAMETER_TYPE, "Data", (QdObject_var) temperature);
        QuSetTxtAttr(SI_CONVERSION, "1e+6>K", (QdObject_var) temperature);
        QuSetTxtAttr(DISPLAYTYPE, "Time_Series", (QdObject_var) temperature);
    }

}

// ------------------

void PSD_Parameters::getMomentsIntegrated(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature, QLabel *percLabl)
{
    density=new QdRScalarSeq(data->sequence_size());
    velocity=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    pressure=new QdRMatrixSeq(data->sequence_size(), Dim(3, 3));
    heat=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    temperature=new QdRScalarSeq(data->sequence_size());
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v_centre[3], v_centre_norm;
    double val, vol, x, y, z, w;
    
    int seq_size=data->sequence_size();

    
    for(int time=0; time<seq_size; time++)
    {
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}
		fill_cartesian(time);
        
        tot_density=0.;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; 
        
        // Computation of density and velocity
       

		buildTetra(time);
    		
		// Minimum velocity, to avoid integrating in the (undefined) centre of the space
		// Ignore tetra that have only VMinus bin edges at apexes
		
        double min_v=-1.;
        
        for(int e=0; e<=maxE; e++)
        {
            if(getVMinus(time, e)>0.)
            {
                if(min_v < 0.) min_v = _vel[e];
                else
                {
                    if(min_v > _vel[e]) min_v=_vel[e];
                }
            }
        }
        
        min_v=min_v*min_v*0.999; // slightly below data point for > test (avoids using == on doubles)


        if(!tetra.empty())
        {
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
				// At least one apex of tetrahedron must be above min_v (otherwise all on inner VMinus)
                if(	iter->dist1 > min_v ||
                	iter->dist2 > min_v ||
                	iter->dist3 > min_v ||
                	iter->dist4 > min_v  ) 
                {
				
				val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3]) * 0.25;
				vol=iter->volume();
				w = val*vol;
					
				tot_density += w;

				x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0]);

				y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1]);

				z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2]);

				tot_numberflux[0] += x*w;
				tot_numberflux[1] += y*w;
				tot_numberflux[2] += z*w;
              
				}
				
            }

			tot_numberflux[0] *= 0.25;
			tot_numberflux[1] *= 0.25;
			tot_numberflux[2] *= 0.25;
			
            
            density[time]=tot_density;
            
            if(flags & MOM_VELOCITY)
            {
                velocity[time][0]=tot_numberflux[0]/(density[time]);
                velocity[time][1]=tot_numberflux[1]/(density[time]);
                velocity[time][2]=tot_numberflux[2]/(density[time]);
            }
        }
        else
        {
            density[time]=-1;
            velocity[time][0]=-1e+31;
            velocity[time][1]=-1e+31;
            velocity[time][2]=-1e+31;
        }
        
        
        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && (double)velocity[time][0]!=-1e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
                // At least one apex of tetrahedron must be above min_v (otherwise all on inner VMinus)
                if(	iter->dist1 > min_v ||
                	iter->dist2 > min_v ||
                	iter->dist3 > min_v ||
                	iter->dist4 > min_v  ) 
                {

					val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3]) * 0.25;
					vol=iter->volume();
					w = vol*val;
					
					x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0]) * 0.25;

					y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1]) * 0.25;

					z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2]) * 0.25;
                    
					v_centre[0]=x-(double)velocity[time][0];
					v_centre[1]=y-(double)velocity[time][1];
					v_centre[2]=z-(double)velocity[time][2];
                    
					v_centre_norm=v_centre[0]*v_centre[0]+v_centre[1]*v_centre[1]+v_centre[2]*v_centre[2];
                    
					if(flags & MOM_PRESSURE)
					{
						for(int i=0; i<3; i++)
						{
							for(int j=0; j<3; j++)
							{
								tot_pressure[i][j] += v_centre[i]*v_centre[j]*w;
							}
						}
					}
                            
					if(flags & MOM_HEATFLUX)
					{
						for(int i=0; i<3; i++)
						{
							tot_heat[i] += v_centre_norm*v_centre[i]*w;
						}
					}
				} // end if on min tetra
                
            }
        } // end computation of pressure and heat
        
        if(flags & MOM_PRESSURE)
        {
            pressure[time][0][0]=tot_pressure[0][0]*mass;
            pressure[time][1][0]=tot_pressure[1][0]*mass;
            pressure[time][2][0]=tot_pressure[2][0]*mass;
            pressure[time][0][1]=tot_pressure[0][1]*mass;
            pressure[time][1][1]=tot_pressure[1][1]*mass;
            pressure[time][2][1]=tot_pressure[2][1]*mass;
            pressure[time][0][2]=tot_pressure[0][2]*mass;
            pressure[time][1][2]=tot_pressure[1][2]*mass;
            pressure[time][2][2]=tot_pressure[2][2]*mass;
            
            temperature[time]=(pressure[time][0][0]+pressure[time][1][1]+pressure[time][2][2])/(3.*density[time]*1.3806503e-23);
        }
        else
        {
            pressure[time]=-1e+31;
            temperature[time]=-1e+31;
        }
        
        if(flags & MOM_HEATFLUX)
        {
            heat[time][0]=tot_heat[0]*mass*0.5;
            heat[time][1]=tot_heat[1]*mass*0.5;
            heat[time][2]=tot_heat[2]*mass*0.5;
        }
        else
        {
            heat[time]=-1e+31;
        }
    }

    // Metadata
	QdRScalar_var unitFill = new QdRScalar(-1.);
	QdRScalar_var doubleFill = new QdRScalar(-1e31);
    
    if(flags & MOM_DENSITY)
    {
		set_timetags(density, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(density, 0, seq_size-1);
			density = QdRScalarSeq_var::narrow(sub);
		}
		if(!density.is_nil()){
			for(int i=0; i<seq_size; i++)
        	{
            	if((double)density[i]!=-1) density[i]=density[i]/1000000.;
        	}
			density->change_xref(FILL_VALUE, unitFill );
        	QuSetTxtAttr("UNITS", "cm^-3", (QdObject_var) density);
        	QuSetTxtAttr("FIELDNAM", "Density - Tetra", (QdObject_var) density);
        	QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) density);
        	QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) density);
        	QuSetTxtAttr("SI_CONVERSION", "1.0e6>m^-3", (QdObject_var) density);
        	QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) density);
		}
    }
    
    if(flags & MOM_VELOCITY)
    {
		set_timetags(velocity, get_timetags(density));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(velocity, 0, seq_size-1);
			velocity = QdRMatrixSeq_var::narrow(sub);
		}
		if(!velocity.is_nil()){

        	for(int i=0; i<seq_size; i++)
			{
            	if((double)velocity[i][0] > -.9e+31)
            	{
                	velocity[i][0]=velocity[i][0]/1000.;
                	velocity[i][1]=velocity[i][1]/1000.;
                	velocity[i][2]=velocity[i][2]/1000.;
            	}
        	}
        	QuSetTxtAttr("UNITS", "km s^-1", (QdObject_var) velocity);
        	QuSetTxtAttr("FIELDNAM", "Bulk Velocity", (QdObject_var) velocity);
			velocity->change_xref(FILL_VALUE, doubleFill );
        	QuSetTxtAttr("SI_CONVERSION", "1.0e3>m s^-1", (QdObject_var) velocity);
        	QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) velocity);
        	QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) velocity);
		}
    }
    
    if(flags & MOM_PRESSURE)
    {
		set_timetags(pressure, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(pressure, 0, seq_size-1);
			pressure = QdRMatrixSeq_var::narrow(sub);
		}
		if(!pressure.is_nil()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if((double)pressure[i][0][0] > -.9e+31)
            	{
                	for(int j=0; j<3; j++)
                	{
                    	for(int k=0; k<3; k++)
                    	{
                       		pressure[i][j][k]=pressure[i][j][k]*1000000000.;
                    	}
                	}
            	}
        	}
        
			QdSMatrix_var rep=new QdSMatrix(Dim(3));
			(*rep)[0]="x";
			(*rep)[1]="y";
			(*rep)[2]="z";
			pressure->set_xref(REPRESENTATION_1, rep);
			pressure->set_xref(REPRESENTATION_2, rep);
			pressure->set_xref(LABEL_1, rep);
			pressure->set_xref(LABEL_2, rep);
			
			QdRMatrix_var dep=new QdRMatrix(Dim(3));
			(*dep)[0]=1;
			(*dep)[1]=2;
			(*dep)[2]=3;
			pressure->set_xref(DEPEND_1, dep);
			pressure->set_xref(DEPEND_2, dep);
			
			QuSetTxtAttr(UNITS, "nPa", (QdObject_var) pressure);
			QuSetTxtAttr(FIELDNAM, "Pressure Tensor", (QdObject_var) pressure);
			pressure->change_xref(FILL_VALUE, doubleFill );
			QuSetTxtAttr(SI_CONVERSION, "1.eE-9>Pa", (QdObject_var) pressure);
			QuSetTxtAttr(DISPLAYTYPE, "Time_Series", (QdObject_var) pressure);
			QuSetTxtAttr(FRAME, "vector>gse_xyz", (QdObject_var) pressure);
			QuSetTxtAttr(TENSOR_ORDER, "2", (QdObject_var) pressure);
		}
    }
    
    if(flags & MOM_HEATFLUX)
    {
        set_timetags(heat, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(heat, 0, seq_size-1);
			heat = QdRMatrixSeq_var::narrow(sub);
		}
		if(!heat.is_nil()){
        	for(int i=0; i<seq_size; i++)
        	{
           		if((double)heat[i][0] > -.9e+31)
            	{
                	heat[i][0]=heat[i][0]*1000000.;
                	heat[i][1]=heat[i][1]*1000000.;
                	heat[i][2]=heat[i][2]*1000000.;
            	}
        	}
        	QuSetTxtAttr("UNITS", "microW m^-2", (QdObject_var) heat);
        	QuSetTxtAttr("FIELDNAM", "Heat Flux", (QdObject_var) heat);
			heat->change_xref(FILL_VALUE, doubleFill );
        	QuSetTxtAttr("SI_CONVERSION", "1.0e-6>W m^-2", (QdObject_var) heat);
        	QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) heat);
        	QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) heat);
        }
		
        set_timetags(temperature, get_timetags(data));
		if(data->sequence_size() != seq_size) {
			QdObject_var sub = QuGetSubset(temperature, 0, seq_size-1);
			temperature = QdRScalarSeq_var::narrow(sub);
		}
		if(!temperature.is_nil()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if((double)temperature[i] > -.9e+31) temperature[i]=temperature[i]/1000000.;
        	}
			temperature->change_xref(FILL_VALUE, doubleFill );
        	QuSetTxtAttr("UNITS", "MK", (QdObject_var) temperature);
        	QuSetTxtAttr("FIELDNAM", "Temperature", (QdObject_var) temperature);
        	QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) temperature);
        	QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) temperature);
        	QuSetTxtAttr("SI_CONVERSION", "1e+6>K", (QdObject_var) temperature);
        	QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) temperature);
		}
    }


}


