/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	BoundaryNormalCoords.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Boundary Normal Coordinates Plugin wrapper
*
* PURPOSE:      Calculate Rotation Matrix to transform into Boundary
*               Normal Coordinates lmn. n is normal to a model magnetopause
*               or bow shock, the l-n plane is parallel the Earth's dipole
*               axis, and lmn is a RH system.
* 
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*             pos_in      Constant vector or Vector TS
*             time_c      Crossing time
*             Vsw         Approx solar wind speed to calculate aberration
*             imodel      Model number from Ch10 of ISSI Book
*             frame_out   GSE->lmn (0) or GSM->lmn (1)
*
*   OUT:
*             frametolmn  Rotation Matrix R. 
*                          R dot vector(frame_out) = vector(lmn)		
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: built on qrtn and qbowshock modules
*
* HISTORY: July 2001 Steve Schwartz 
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"

#include "qdutil.h"
#include "Qdos.h"
#include "qrtn.h"
#include "qar.h"
#include "qbowshock.h"

const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

extern "C" QplugReturnStatus BoundaryNormalCoords(QplugArgList *);
QplugReturnStatus QBNPosition(QrVector *vec_gsep,
                              QdObject_var pos_in,
							  QdTime_var time_c);
QplugReturnStatus QBNConstPosition(QrVector *vec_gsep, 
								   QdRMatrix_var pos_in, 
				                   QdTime_var time_c);
QplugReturnStatus QBNCentrePosition(QrVector *vec_gsep, 
                                    QdRMatrixSeq_var pos_in_seq, 
				                    QdTime_var time_c);
QplugReturnStatus QBNMakeGSERe (QrVector *vecp);
double SIConv_factor(QdObject_var in1_ptr);
void QrVector_diag_print(QrVector *vecp);
  
/**********************************************************************/
QplugReturnStatus BoundaryNormalCoords(  QplugArgList *call_list){
/**********************************************************************/

// define some C variables

long  check_rtn;
float Vsw, r_vec[3], n_gse[3];
int   imodel;
int   i, qbs_rtn, qar_status;

double modm;

// call-list arguments
 
QdObject_var         pos_in;       // 0
QdTime_var			 time_c;     // 1
QdRScalar_var        Vsw_var, imodel_var, frame_out_var; // 2, 3, 4
QdRMatrix_var        frametolmn;                         // 5

// define Qdos data objects

QrVector  vec_gse, z, l, m, n;

QplugReturnStatus ret_status;

// ...first 2 items are input position vector/vector TS and time interval

pos_in = QdObject_var::narrow((*call_list)[0]);
if (pos_in.is_nil()){
  QplugAppendTextDisplay("BNCoords>> Error getting pos_in object\n");
  return QPLUG_FAILURE;
}

time_c = QdTime_var::narrow((*call_list)[1]);
if( time_c.is_nil() ) {
  QplugAppendTextDisplay ("BNCoords>> Time object invalid.\n");
  return QPLUG_FAILURE;
}

vec_gse.IsPositionVector = 0;     // force so QrDoRotation will accept

ret_status = QBNPosition(&vec_gse, pos_in, time_c);
if (ret_status != QPLUG_SUCCESS){
  QplugAppendTextDisplay("BNCoords>> QBNPosition failed\n");
  return QPLUG_FAILURE;
}

// Whew! vec_gse now has a sequence with 1 element, the gse_xyz position
// of the crossing in Re with one timetag. 

Vsw_var = QdRScalar_var::narrow((*call_list)[2]);
if (Vsw_var.is_nil()){
   QplugAppendTextDisplay("BNCoords>> Input arg Vsw not Scalar\n");
   return QPLUG_FAILURE;
}
imodel_var = QdRScalar_var::narrow((*call_list)[3]);
if(imodel_var.is_nil()){
   QplugAppendTextDisplay("BNCoords>> Input arg imodel not Scalar\n");
   return QPLUG_FAILURE;
}

// unpack inputs for QBOWSHOCK routines and find gse normal;
Vsw = (float) (*Vsw_var);
imodel = (int) (*imodel_var);
for (i=0; i<3; i++){r_vec[i] = (float)((vec_gse.ptr)[0][i]);}

qbs_rtn = QBSNormalAtSurfaceGSE(r_vec,Vsw,(QBSMP_models)imodel,n_gse);
if (qbs_rtn != QBS_OK){
   QplugAppendTextDisplay("BNCoords>> QBSNormal returned error\n");
   return QPLUG_FAILURE;
}

// put normal in n, set z to (0,0,1)GSM

n          = vec_gse;   // copy over n_recs = 1, timetag, ...
n.in_frame = GSE;
n.ptr      = new QdRMatrixSeq(1,RMatrix(Dim(3)));
for (i=0; i<3; i++){(n.ptr)[0][i] = n_gse[i];}      // must be a better way
QuSetFrame(n.ptr, "vector", "gse", "xyz" );

z           = n;
z.in_frame  = GSM;
z.ptr       = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(z.ptr)[0][0] = 0.0; (z.ptr)[0][1] = 0.0; (z.ptr)[0][2] = 1.0;   // ditto
QuSetFrame(z.ptr, "vector", "gsm", "xyz" );

frame_out_var = QdRScalar_var::narrow((*call_list)[4]);
if(frame_out_var.is_nil()){
   QplugAppendTextDisplay("BNCoords>> Frame_out not a scalar\n");
   return QPLUG_FAILURE;
}

frametolmn = new QdRMatrix(Dim(3,3));

switch ( (int)(*frame_out_var) ){
   case 1:      // gsm
      n.out_frame = GSM;
      z.out_frame = GSM;
      QuSetFrame(frametolmn, "rot", "gsm", "xyz");
      break;
   case 0: default:    // gse
      n.out_frame = GSE;
      z.out_frame = GSE;
      QuSetFrame(frametolmn, "rot", "gse", "xyz");
      break;
}

check_rtn = n.QrDoRotation();  // put n in correct frame
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay( "BNCoords>> Failed to put n into out frame\n");
   return QPLUG_FAILURE;
}
n.in_frame = n.out_frame;      // make in_frame consistent with holding
check_rtn = z.QrDoRotation();  // put z in correct frame
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay( "BNCoords>> Failed to put z into out frame\n");
   return QPLUG_FAILURE;
}
z.in_frame = z.out_frame;      // make in_frame consistent with holding


// find n x z ...  [this took me a long time; I couldn't get .cross to work]

//lots of error checking; could replace with try...catch around whole plugin?

m.ptr = QdRMatrixSeq_var::narrow(outer(n.ptr, z.ptr));
if ((m.ptr).is_nil()) {
    QplugAppendTextDisplay ("BNCoords>> n cross z failed\n");
    return QPLUG_FAILURE;
}

modm = norm(((m.ptr)[0])) ;
if (modm<=0){
    QplugAppendTextDisplay ("BNCoords>> |m| returned invalid number\n");
    return QPLUG_FAILURE;
}

*(m.ptr) /=  modm;
if ((m.ptr).is_nil()) {
    QplugAppendTextDisplay ("BNCoords>> m/|m| failed\n");
    return QPLUG_FAILURE;
}

l.ptr = QdRMatrixSeq_var::narrow(outer(m.ptr,n.ptr));
if ((l.ptr).is_nil()){
    QplugAppendTextDisplay ("BNCoords>> l = m x n failed\n");
    return QPLUG_FAILURE;
}

for (i=0; i<3; i++){
    (*frametolmn)[0][i] = (l.ptr)[0][i];
    (*frametolmn)[1][i] = (m.ptr)[0][i];
    (*frametolmn)[2][i] = (n.ptr)[0][i];
}

// set the attributes for the output rotation matrix

QuSetTxtAttr("FIELDNAM", "RMatrix_to_lmn", frametolmn);
QuSetTxtAttr("SI_conversion", "1>()", frametolmn);
QuSetTxtAttr("UNITS", " ", frametolmn);
QuSetTxtAttr("ToFrame", "vector>lmn_xyz", frametolmn);
char model_name[200];
sprintf(model_name,"Model %d: %s",imodel,BSMP_name[imodel]);
QuSetTxtAttr("Boundary Model", model_name, frametolmn);

QdRMatrixSeq_var crossing = vec_gse.ptr;
char crossing_str[200];
sprintf(crossing_str,"( %9.3g, %9.3g, %9.3g )GSE Re",
   (double)crossing[0][0],(double)crossing[0][1],(double)crossing[0][2]);
QuSetTxtAttr("Crossing",crossing_str,frametolmn);
char crossingt_str[200];
sprintf(crossingt_str,"%s", (vec_gse.t)[0].iso_srep().c_str());
QuSetTxtAttr("Crossing Time",crossingt_str,frametolmn);


call_list->push_back( (QdObject_var) frametolmn);     

// put some results to plug-in text window
char outstr[200];
sprintf(outstr,"BNCoords>> Crossing %s at %s\n", crossing_str, crossingt_str);
QplugAppendTextDisplay(outstr);
sprintf(outstr,"BNCoords>> Normal is (%9.3g, %9.3g, %9.3g) Frame = %s\n",
   (double)(n.ptr)[0][0], (double)(n.ptr)[0][1], (double)(n.ptr)[0][2],
   QuGetAttrText("Frame",frametolmn).c_str());
QplugAppendTextDisplay(outstr);
QplugAppendTextDisplay("BNCoords>> finished successfully\n");

return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus QBNPosition(QrVector *vec_gsep,
                              QdObject_var pos_in,
			      QdTime_var time_c){
/**********************************************************************/

long  check_rtn;
QdRMatrixSeq_var pos_in_seq;
QdRMatrix_var    pos_in_const;
QplugReturnStatus ret_status;

// sort out position of crossing. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vec_gse a sequence of 1 cartesian vector whose timetag
// is the selected time, and whose frame is
// that of pos_in. This is then rotated into gse. [NB qrtn routines at
// present (July 2001) assume vectors are _xyz]

pos_in_seq       = QdRMatrixSeq_var::narrow(pos_in);
pos_in_const     = QdRMatrix_var::narrow(pos_in);
if( (!pos_in_seq.is_nil()) && (!pos_in_const.is_nil()) ){
  QplugAppendTextDisplay("BNCoords>> pos_in both const and sequence!?\n");
  return QPLUG_FAILURE;
}

if( !pos_in_seq.is_nil() ) {
  QplugAppendTextDisplay ("BNCoords>> Vector Sequence. Interpolating ...\n");
  ret_status = QBNCentrePosition(vec_gsep,pos_in_seq,time_c);
  if (ret_status != QPLUG_SUCCESS){
   QplugAppendTextDisplay("BNCoords>> Failed to calculate crossing position\n");
   return QPLUG_FAILURE;
  }
} else if (!pos_in.is_nil()){
  ret_status = QBNConstPosition(vec_gsep, pos_in_const, time_c);
  if (ret_status != QPLUG_SUCCESS){
    QplugAppendTextDisplay("BNCoords>> Failed to calc const. position\n");
    return QPLUG_FAILURE;
  }
} else {
  QplugAppendTextDisplay("BNCoords>> pos_in neither const nor sequence!?\n");
  return QPLUG_FAILURE;
}

//QrVector_diag_print(vec_gsep);
vec_gsep->out_frame = GSE;
check_rtn = vec_gsep->QrDoRotation();  // make sure vec_gse in gse
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay( "BNCoords>> Failed to put pos into gse\n");
   return QPLUG_FAILURE;
}
//QrVector_diag_print(vec_gsep);
vec_gsep->in_frame = GSE;        // advertise in GSE so don't double rotate

return QPLUG_SUCCESS;
}

/**********************************************************************/
QplugReturnStatus  QBNConstPosition(QrVector *vec_gsep, 
                                    QdRMatrix_var pos_in, 
				    QdTime_var time_c){
/**********************************************************************/

QplugReturnStatus ret_status;
const char * xyz_comp[] = {"x", "y", "z"};
int i;

// grab constant vector and fill vec_gsep.

if (!QuIsThreeVector(pos_in)){
   QplugAppendTextDisplay("BNCoords>> Const pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

QdTimeSeq_var tt = new QdTimeSeq(1, *time_c);
vec_gsep->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(vec_gsep->ptr)->copy_xrefs_from(pos_in);
QdRMatrixSeq_var pos_seq = new QdRMatrixSeq(1, (*pos_in) );
pos_seq->copy_xrefs_from(pos_in);
for (i=0; i<3; i++){
  QdRScalarSeq_var comp = 
     QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_seq,xyz_comp[i]));
  if (comp.is_nil()){
     QplugAppendTextDisplay("BNCoords>> Const pos comp narrow failed\n");
     return QPLUG_FAILURE;
  }
  (vec_gsep->ptr)[0][i] = comp[0];
}
vec_gsep->n_recs = 1;
vec_gsep->t = tt;

vec_gsep->in_frame = vec_gsep->QrGetFrameFromAttr(pos_in);

if (vec_gsep->in_frame == UNKNOWN){
  QplugAppendTextDisplay("BNCoords>> Const pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

ret_status = QBNMakeGSERe(vec_gsep);
if (ret_status != QPLUG_SUCCESS){
  QplugAppendTextDisplay("BNCoords>> QBNMakeGSERe failed\n");
  return QPLUG_FAILURE;
}

return QPLUG_SUCCESS;
}


/**********************************************************************/
QplugReturnStatus  QBNCentrePosition(QrVector *vec_gsep, 
                                    QdRMatrixSeq_var pos_in_seq, 
				    QdTime_var time_c){
/**********************************************************************/

int i,nttin,ntrue,il,ih;
QplugReturnStatus ret_status;
double cl,ch, frac;
const char * xyz_comp[] = {"x", "y", "z"};

if (!QuIsThreeVector(pos_in_seq)){
   QplugAppendTextDisplay("BNCoords>> Sequence pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

// prepare vec_gsep 
QmjdTime tc = *time_c;
QdTimeSeq_var tt = new QdTimeSeq(1, tc);
vec_gsep->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(vec_gsep->ptr)->copy_xrefs_from(pos_in_seq);
vec_gsep->t = tt;
vec_gsep->n_recs = 1;
vec_gsep->in_frame = vec_gsep->QrGetFrameFromAttr(pos_in_seq);
if (vec_gsep->in_frame == UNKNOWN){
  QplugAppendTextDisplay("BNCoords>> Const pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

// linearly interpolate crossing position from bracketting time tags
// If any problems, default to 1st vector in pos_in_seq
// There are probably much simpler ways to do this. Note, though, that
// Qrtn routines only give correct answers if given x,y,z representation

QdTimeSeq_var tt_in = get_timetags(pos_in_seq);
if (tt_in.is_nil()){
  QplugAppendTextDisplay(
     "BNCoords>> No timetags for input pos seq. Taking 1st pt\n");
     il = ih = 0;
} else {
  nttin = tt_in->size();
  if (nttin <= 0){
     QplugAppendTextDisplay("BNCoords>> Input seq 0 length\n");
     return QPLUG_FAILURE;
  } else if (nttin == 1){
     il = ih = 0;
  } else {
     SequenceMask msk(tt_in->get_mask( QdTimeSeq::LE( tc)));
     ntrue = msk.num_true();
     if (ntrue <= 0) il = ih = 0;
     else {
       il = ntrue -1;
       ih =  (ntrue < nttin) ?   il + 1 : il; 
     }  
  }
}

if(ih == il) frac = 0.0;   // safe
else frac = (tc-tt_in[il])/(tt_in[ih] - tt_in[il]); // calc lin interpol factor

/* printf("tl: %s\ntc: %s\nth: %s\nfrac = %g\n",
                tt_in[il].iso_srep().c_str(),
		tc.iso_srep().c_str(),
		tt_in[ih].iso_srep().c_str(),
		frac);
*/

for (i=0; i<3; i++){                          // fill crossing position
  QdRScalarSeq_var comp = 
     QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_in_seq,xyz_comp[i]));
  if (comp.is_nil()){
     QplugAppendTextDisplay("BNCoords>> Centre Pos comp narrow failed\n");
     return QPLUG_FAILURE;
  }
  if (ih==il) {
    vec_gsep->ptr[0][i] = (double)comp[il];
  } else {
    cl = (double)comp[il];
    ch = (double)comp[ih];
    vec_gsep->ptr[0][i] = cl + frac*(ch-cl);
  }
}

ret_status = QBNMakeGSERe(vec_gsep);
if (ret_status != QPLUG_SUCCESS){
  QplugAppendTextDisplay("BNCoords>> QBNMakeGSERe failed\n");
  return QPLUG_FAILURE;
}

return QPLUG_SUCCESS;
}

/**********************************************************************/
QplugReturnStatus QBNMakeGSERe (QrVector *vecp){
/**********************************************************************/

// rotate vecp to GSE and scale to Re

int check_rtn;
double factor, SIfactor;

// Rotate

vecp->out_frame = GSE;
//printf("QBNMakeGSEre received:\n");QrVector_diag_print(vecp);
check_rtn = vecp->QrDoRotation();
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay("BNCoords>> QBNMakeGSERe rotate to gse failed\n");
   printf("QBNMakeGSERe QrDoRotation returned %d\n",check_rtn);
   fflush(stdout);
   return QPLUG_FAILURE;
}
vecp->in_frame = GSE;    // if don't do this, another call to QrDoRot will
                         // rotate again!
// Scale
// Algorithm:
//    Take SI_conversion, verify base unit = m, and use SI_conv/Re as factor
//    If SI_conversion exists but isn't -> m or not positive, assume factor=1.
//    If no SI_conversion, try to make sense of UNITS; default to factor = 1

factor = 1.0;  // assume in Re if all else fails; actually this is done
               // explicitly below, but this is here for safety.

//check whether SI_conversion exists
string attr = QuGetAttrText("SI_conversion", vecp->ptr);
if(attr.empty()){                // no SI_conversion
   QplugAppendTextDisplay (
    "BNCoords>>WARNING:Input Vector SI_conversion not defined, trying UNITS.\n");
   attr = QuGetAttrText("UNITS", vecp->ptr);
   if(attr.empty()) {
     QplugAppendTextDisplay (
     "BNCoords>>WARNING: UNITS not defined, assuming [Re].\n");
     factor = 1.0;
   } else {                           // yes, we have UNITS
     if(QuToUpper(attr).find("Re") != string::npos) factor=1.0;
     else if(attr.find("km") !=string::npos) factor=1000.0/Re;
     else {
       QplugAppendTextDisplay (
         "BNCoords>>WARNING: UNITS not recognized, assuming [Re].\n");
       factor = 1.0;
     }
   }   
} else {      // Found SI_conversion.
  // check whether it is position vector, but don't change vecp->IsPos...
  // so can rotate it
  if(vecp->QrIsPositionVector(vecp->ptr)!= 1){  // checks base SI == m
      QplugAppendTextDisplay("BNCoords>> Input Vector not position vector. ");
      QplugAppendTextDisplay("BNCoords>> WARNING: Re will be assumed.\n");
      factor = 1.0;
  } else {
      SIfactor = SIConv_factor(vecp->ptr);  // get SI_conversion
      if(SIfactor <= 0.0) {
	QplugAppendTextDisplay("BNCoords>> Input Vec SI_conversion invalid.\n");
	QplugAppendTextDisplay ("BNCoords>> WARNING: Re will be assumed.\n");
	factor = 1.0;
      } else {
        factor = SIfactor/Re;
      }
  }
}

*(vecp->ptr) *= factor;

//printf("QBNMakeGSEre returned:\n");QrVector_diag_print(vecp);

return QPLUG_SUCCESS;
}

/***********************************************************************/
double SIConv_factor(QdObject_var in1_ptr){ 
/***********************************************************************/

 string siconv1;
 double value1;

  siconv1 = QuGetAttrText("SI_conversion", in1_ptr);
  
  if(siconv1.empty()) {
    siconv1 = "1 > ()";
  }
  
  if( siconv1.find(">") == string::npos) {
     return 0.0;
  }
  value1 = (double) atof(siconv1.c_str());
   
  return value1;
  
} // end SIConv_factor

/**************************************************************************/
void QrVector_diag_print(QrVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
     (double)vecp->ptr[0][0], (double)vecp->ptr[0][1], (double)vecp->ptr[0][2],
     vecp->t[0].iso_srep().c_str());
	 fflush(stdout);
  return;
}



