/* shockDistance.cc Jeremy Mitchell 28/08/2009
Uses qbowshock (sjs) routines 
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "qbowshock.h"

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include "Xrefs.h"

extern "C" QplugReturnStatus shockDistance(QplugArgList *);

using namespace QSAS;

QplugReturnStatus shockDistance(QplugArgList *call_list) {
  QdObject_var pos_in = (* call_list)[0];
  QdObject_var sw_vel_in = (* call_list)[1];
  QdObject_var model_in = (* call_list)[2];
  
  QdObject_var pos_SI=QarConvertUnitsTo(pos_in, "1>m");  
  
  if (pos_SI.is_nil()) {
    QplugAppendTextDisplay("shockDistance Error: Can't convert units to m");
    return QPLUG_FAILURE;
  }
  
  if (!QuIsThreeVector(pos_SI)) {
    QplugAppendTextDisplay("shockDistance Error: Distance is not a vector");
    return QPLUG_FAILURE;
  }
  
  bool is_series = true;
  
  QdRMatrix_var pos_ind;
  QdRMatrixSeq_var pos_seq = QdRMatrixSeq_var::narrow(pos_SI);
  if (pos_seq.is_nil()) {
    pos_ind = QdRMatrix_var::narrow(pos_SI);
    if (pos_ind.is_nil()) {
      QplugAppendTextDisplay("shockDistance Error: Unable to unpack distance\n");
      return QPLUG_FAILURE;
    } else
    is_series = false;
  }
  
  QdRScalar_var sw_vel_qd = QdRScalar_var::narrow(sw_vel_in);
  if (sw_vel_qd.is_nil()) {
    QplugAppendTextDisplay("shockDistance Error: Unable to unpack SW vel\n");
    return QPLUG_FAILURE;
  }
  float vel_sw = (float)(* sw_vel_qd);
  if (vel_sw < 0) {
    QplugAppendTextDisplay("shockDistance Error: SW vel < 0\n");
    return QPLUG_FAILURE;
  }
  
  QdRScalar_var model_qd = QdRScalar_var::narrow(model_in);
  if (model_qd.is_nil()) {
    QplugAppendTextDisplay("shockDistance Error: Unable to unpack model\n");
    return QPLUG_FAILURE;
  }
  int model = (int)(* model_qd);
  if ((model < 0) || (model > 13)) {
    QplugAppendTextDisplay("shockDistance Error: Must have 0 <= model <= 13\n");
    return QPLUG_FAILURE;
  }
  
  /************************ Units ****************************/
  
  // convert sw_vel from m to km if necessary
  
  QdRScalar_var vel_sw_var = new QdRScalar(vel_sw);
  QdString_var sic = new QdString(QuGetAttrText(SI_CONVERSION, sw_vel_in).c_str());
  vel_sw_var->change_xref(SI_CONVERSION, sic);

  QarConvertUnitsTo(vel_sw_var, "1.0e3>m s^-1");
  
  vel_sw = *vel_sw_var;
  
  /****************************************************/
  
  QBSMP_models model_arg = (QBSMP_models) model;
  
  QdObject_var shock_out;
  QdObject_var normal_out;
  
  if (is_series) {
    int size = pos_seq->sequence_size();
    
    QdRMatrixSeq_var normal = new QdRMatrixSeq(size, Dim(3));
    QdRMatrixSeq_var shock  = new QdRMatrixSeq(size, Dim(3));
    
    for (int i=0; i < size; i++) {
      
      float pos[3], dist, r_out[3], n_out[3];
      pos[0] = pos_seq[i][0]; pos[1] = pos_seq[i][1]; pos[2] = pos_seq[i][2];
      
      QBSDistanceAlongNormal(pos, vel_sw, model_arg, 1., &dist, r_out);
      QBSNormalAtSurfaceGSE (pos, vel_sw, model_arg, n_out);
      
      cout << r_out[0] << " " << r_out[1] << " " << r_out[2] << " " << endl;
      
      shock[i][0] = r_out[0] - (float)(pos_seq[i][0]);
      shock[i][1] = r_out[1] - (float)(pos_seq[i][1]);
      shock[i][2] = r_out[2] - (float)(pos_seq[i][2]);
      
      normal[i][0] = n_out[0];
      normal[i][1] = n_out[1];
      normal[i][2] = n_out[2];   
      
    }
    
    normal_out = (QdObject_var) normal;
    shock_out = (QdObject_var) shock;
 
  } else {
    
    QdRMatrix_var shock = new QdRMatrix(Dim(3));
    QdRMatrix_var normal = new QdRMatrix(Dim(3));
    
    float pos[3], dist, r_out[3], n_out[3];
    pos[0] = (*pos_seq)[0]; pos[1] = (*pos_seq)[1]; pos[2] = (*pos_seq)[2];
    
    QBSDistanceAlongNormal(pos, vel_sw, model_arg, 1., &dist, r_out);
    QBSNormalAtSurfaceGSE (pos, vel_sw, model_arg, n_out);
    
    (*shock)[0] = r_out[0] * 6.3172e6 - (*pos_ind)[0];
    (*shock)[1] = r_out[1] * 6.3172e6 - (*pos_ind)[1];
    (*shock)[2] = r_out[2] * 6.3172e6 - (*pos_ind)[2];
    
    (*normal)[0] = n_out[0];
    (*normal)[1] = n_out[1];
    (*normal)[2] = n_out[2];
    
    normal_out = (QdObject_var) normal;
    shock_out = (QdObject_var) shock;
  }
  
  /********************** Set Attributes *************************/
  
  char attrib[20];
  
  QdRScalar_var doubleFill = new QdRScalar(1.e30);
  
  QuSetTxtAttr(FRAME, "vector>gse_xyz", shock_out);
  QuSetTxtAttr(FRAME, "vector>gse_xyz", normal_out);
  QuSetTxtAttr(SI_CONVERSION, "1>m", shock_out);
  QuSetTxtAttr(SI_CONVERSION, "1>", normal_out);
  QuSetTxtAttr(UNITS, "m", shock_out);
  QuSetTxtAttr(UNITS, "", normal_out);
  QuSetTxtAttr(LABLAXIS, "r_GSE", shock_out);
  QuSetTxtAttr(VALIDMIN, "-1.e30", shock_out);
  QuSetTxtAttr(VALIDMAX,  "1.e30", shock_out);
  shock_out->change_xref(FILL_VALUE, doubleFill );
  QuSetTxtAttr(SCALETYP,  "lin", shock_out);
  
 	// Add timeline (or DEPEND_0) : Either timeline is safe if it exists
	QdObject_var qd0 = get_Depend_0( pos_in );
	if ( !qd0.is_nil() ){
    	shock_out->change_xref(DEPEND_0 , qd0 );
		normal_out->change_xref(DEPEND_0 , qd0 );
	}
  
  call_list->push_back( shock_out );
  call_list->push_back( normal_out );
  
  return QPLUG_SUCCESS;
  
}
