/****************************************************************************
							LICENSE

  sf_ChampToGeo - plug-in for QSAS software.
  Copyright (C) 2010 - 2012 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file sf_ChampToGeo.cc
 *
 * \brief Converts variable from Champ's CDF into a vector in GSE frame.
 *
 *
 */

#include <fstream>
#include <iostream>
//the next line is needed for visual studio to find M_PI and others.
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "qar.h"
#include "Qdos.h"

#include "qrtn.h"  //coordinates rotation

using namespace QSAS;
using namespace std;


// ==========================================================================

const short kScalarRank = 1; ///compare nr_coordinates with this to decided if it is a scalar
const short kVectorRank = 3; ///compare nr_coordinates with this to decided if it is a vector

///Same as strcmp, but transforms strings to lowercase first.
int strcmp_ignorecase(const char* str1, const char* str2)
{
	size_t n = strlen(str1);
	char *str1_lower = (char*) malloc(n);
	for (size_t i = 0; i < n; ++i)
		str1_lower[i] = static_cast<char> (tolower(str1[i]));
	n = strlen(str2);
	char *str2_lower = (char*) malloc(n);
	for (size_t i = 0; i < n; ++i)
		str2_lower[i] = static_cast<char> (tolower(str2[i]));
	int result = strcmp(str1_lower, str2_lower);
	free(str1_lower);
	free(str2_lower);
	return result;
}

// =========  class Status  =================================================

/**
 * The class Status has the purpose to print messages either to user or to the Console.
 */
class Status
{
public:
	static const bool kPrintToUserAndConsole = true;
	static const bool kPrintToConsoleOnly = false;
	static void PrintMessage(bool to_user, const char fmt[], ...);
public:
	enum ReturnStatus
	{
		kOk = 0, kUnknownError, kWrongInput, kNoInput, kVectorFieldRequired
	} code;
	std::string message() const;
	inline void set_message(const std::string message)
	{
		message_ = message;
	}
	;
	Status();
	virtual ~Status();
private:
	std::string message_;
};
// ==========================================================================

QplugReturnStatus  ConvertToDesiredFrame (QdRMatrixSeq_var &ts_in, int frame, QdRMatrixSeq_var &ts_out);

QplugReturnStatus UnderstandInputsFromCallList (QplugArgList *call_list, QdRScalarSeq_var &geo_alt, double &ref_radius
		, QdRScalarSeq_var &geo_lat, QdRScalarSeq_var &geo_lon, QdRMatrixSeq_var &vec_nec_frame
		, int &output_ref_frame);

QplugReturnStatus ConvertFrameNEC2GEO (QdRMatrixSeq_var &vec_nec_frame
		, QdRScalarSeq_var &geo_lat, QdRScalarSeq_var &geo_lon, QdRMatrixSeq_var &vec_geoc_frame);

extern "C" QplugReturnStatus sf_ChampToGeo(QplugArgList *call_list)
{
	//INPUT 0	SCALAR_TIME_SERIES geo_alt,
	//INPUT 1	SCALAR_TIME_SERIES geo_lat,
	//INPUT 2	SCALAR_TIME_SERIES geo_lon,
	//OUTPUT	VECTOR_TIME_SERIES geo_pos)

	QplugReturnStatus status = QPLUG_SUCCESS;
	QdRScalarSeq_var geo_alt = NULL, geo_lat = NULL, geo_lon = NULL;
	QdRMatrixSeq_var vec_nec_frame = NULL;
	int output_ref_frame = 2;  // GEOC frame
	double ref_radius = -1;
	status = UnderstandInputsFromCallList(call_list, geo_alt, ref_radius, geo_lat, geo_lon
			, vec_nec_frame, output_ref_frame);

	// Compute the position vector  -----------------------------------------
	QdRMatrixSeq_var champ_pos = NULL;
	if ((status == QPLUG_SUCCESS) && (!geo_alt.is_nil()) && (!geo_lat.is_nil()) && (!geo_lon.is_nil()) ) {
		Status::PrintMessage(Status::kPrintToUserAndConsole, "Processing the data...\n");
		size_t nr_measurements = geo_alt->sequence_size();
		champ_pos = new QdRMatrixSeq(nr_measurements, RMatrix(Dim(3), 0.0));
		for (size_t t = 0; t < nr_measurements; ++t) {
			double radius = (double) (*geo_alt)[t] + ref_radius;
			double rcos = radius * cos(((double) (*geo_lat)[t]) / 180 * M_PI);
			(*champ_pos)[t][0] = rcos * cos(((double) (*geo_lon)[t]) / 180 * M_PI);
			(*champ_pos)[t][1] = rcos * sin(((double) (*geo_lon)[t]) / 180 * M_PI);
			(*champ_pos)[t][2] = radius * sin(((double) (*geo_lat)[t]) / 180 * M_PI);
		}
	}

	// Convert the vector frome NEC to GEOC frame  --------------------------
	QdRMatrixSeq_var vec_geoc_frame = NULL;
	if ((status == QPLUG_SUCCESS) && (!vec_nec_frame.is_nil()) ) {
		status = ConvertFrameNEC2GEO (vec_nec_frame, geo_lat, geo_lon, vec_geoc_frame);
	}

	// Preparing output champ pos ---------------------------------------------------
	if ((status == QPLUG_SUCCESS) && (!champ_pos.is_nil())) {
		QuSetFrame((QdObject_var) champ_pos, "vector>geoc_xyz");
		QuSetTxtAttr(SI_CONVERSION, "1.0E3>m", (QdObject_var) champ_pos);
		QuSetTxtAttr(UNITS, "km", (QdObject_var) champ_pos);
		string label_axis = "Champ position";
		QuSetTxtAttr("LABLAXIS", label_axis.c_str(), (QdObject_var) champ_pos);
		QuSetTxtAttr("FIELDNAM", label_axis.c_str(), (QdObject_var) champ_pos);
		set_timetags((QdObject_var) champ_pos, get_timetags(geo_alt));
	}

	if ((status == QPLUG_SUCCESS) && (!vec_geoc_frame.is_nil()) && (!vec_nec_frame.is_nil())) {
		vec_geoc_frame->copy_xrefs_from(vec_nec_frame);
		QuSetFrame((QdObject_var) vec_geoc_frame, "vector>geoc_xyz");
		set_timetags((QdObject_var) vec_geoc_frame, get_timetags(vec_nec_frame));
	}

	// Convert from GEOC to the desired frame -------------------------------
	QplugReturnStatus sub_status = QPLUG_SUCCESS;
	QdRMatrixSeq_var champ_pos_new_frame = NULL, vec_new_frame = NULL;;
	if (status == QPLUG_SUCCESS) {
		if (output_ref_frame == 2) {
			//no need for conversion
			sub_status = QPLUG_FAILURE;
		}
		else {
			sub_status = ConvertToDesiredFrame (champ_pos, output_ref_frame, champ_pos_new_frame);
			if ((sub_status != QPLUG_FAILURE) && (!vec_geoc_frame.is_nil()))
				sub_status = ConvertToDesiredFrame (vec_geoc_frame, output_ref_frame, vec_new_frame);
			if (sub_status == QPLUG_FAILURE) {
				Status::PrintMessage(Status::kPrintToUserAndConsole
						, "Warning: Could not convert to the desired frame. Output will be in GEOC frame.\n");
			}
		}
	}

	//5. Output.  -------------------------------------------------------------
	if ((status == QPLUG_SUCCESS) && (!champ_pos.is_nil() || !champ_pos_new_frame.is_nil())) {
		if (sub_status == QPLUG_SUCCESS) {
			call_list->push_back((QdObject_var) champ_pos_new_frame);
			call_list->push_back((QdObject_var) vec_new_frame);
		}
		else {
			call_list->push_back((QdObject_var) champ_pos);
			call_list->push_back((QdObject_var) vec_geoc_frame);
		}
	}
	Status::PrintMessage(Status::kPrintToUserAndConsole, "Finished.\n");
	return status;
}

QplugReturnStatus UnderstandInputsFromCallList (QplugArgList *call_list, QdRScalarSeq_var &geo_alt, double &ref_radius
		, QdRScalarSeq_var &geo_lat, QdRScalarSeq_var &geo_lon, QdRMatrixSeq_var &vec_nec_frame
		, int &output_ref_frame)
{
	const int kCallListIndexOfAltitude  = 0;
	const int kCallListIndexOfLatitude  = 1;
	const int kCallListIndexOfLongitude  = 2;
	const int kCallListIndexOfNecVector  = 3;
	const int kCallListIndexOfOutputFrame  = 4;

	QplugReturnStatus status = QPLUG_SUCCESS;
	// ------------------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToUserAndConsole, "Understanding the input data...\n");
	size_t nr_measurements = 0;
	//get altitude
	ref_radius = 6371.2;
	geo_alt = NULL;
	if (status == QPLUG_SUCCESS) {
		geo_alt = QdRScalarSeq_var::narrow((*call_list)[kCallListIndexOfAltitude]);
		if (geo_alt.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand altitude time series.\n");
		}
		else {
			const char kRefRadiusAttribute[] = "REFRADIUS";
			if (geo_alt->xref_exists(kRefRadiusAttribute)) {
				ref_radius = atof(QuGetAttrText(kRefRadiusAttribute, geo_alt).c_str());
				if (ref_radius < 1.0) {
					Status::PrintMessage(Status::kPrintToUserAndConsole,
							"warning: reference radius is either negative or almost zero.");
				}
			}
			else {
				Status::PrintMessage(Status::kPrintToUserAndConsole,
						"warning: reference radius was not found in meta-data. Using r = %.1f km.\n", ref_radius);
			}
			nr_measurements = geo_alt->sequence_size();
		}
	}
	//get latitude
	geo_lat = NULL;
	if (status == QPLUG_SUCCESS) {
		geo_lat = QdRScalarSeq_var::narrow((*call_list)[kCallListIndexOfLatitude]);
		if (geo_lat.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand latitude time series.\n");
		}
		else {
			if (nr_measurements != geo_lat->sequence_size()) {
				status = QPLUG_FAILURE;
				Status::PrintMessage(Status::kPrintToUserAndConsole,
						"error: Latitude time tags disagree with those of altitude.\n");
			};
		}
	};
	//get longitude
	geo_lon = NULL;
	if (status == QPLUG_SUCCESS) {
		geo_lon = QdRScalarSeq_var::narrow((*call_list)[kCallListIndexOfLongitude]);
		if (geo_lon.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand longitude time series.\n");
		}
		else {
			if (nr_measurements != geo_lon->sequence_size()) {
				status = QPLUG_FAILURE;
				Status::PrintMessage(Status::kPrintToUserAndConsole,
						"error: longitude time tags disagree with those of altitude.\n");
			};
		}
	};
	//get NEC vector
	vec_nec_frame = NULL;
	if (status == QPLUG_SUCCESS) {
		vec_nec_frame = QdRMatrixSeq_var::narrow((*call_list)[kCallListIndexOfNecVector]);
		if (vec_nec_frame.is_nil())
			Status::PrintMessage(Status::kPrintToUserAndConsole, "No optional vector sequence in the NEC frame detected.\n");
	}
	//Get the reference frame of the output vector.
	output_ref_frame = 2;  // GEOC frame
	if (status == QPLUG_SUCCESS) {
		QdObject_var fr_ptr = (*call_list)[kCallListIndexOfOutputFrame];
		QdRScalar_var new_frame_ptr = QdRScalar_var::narrow(fr_ptr);

		if ( new_frame_ptr.is_nil() )
			Status::PrintMessage(Status::kPrintToUserAndConsole, "Warning: frame argument not an integer flag!\n");
		else
			output_ref_frame = (int)(*new_frame_ptr);
	}
	return status;
}

QplugReturnStatus ConvertFrameNEC2GEO (QdRMatrixSeq_var &vec_nec_frame
		, QdRScalarSeq_var &geo_lat, QdRScalarSeq_var &geo_lon, QdRMatrixSeq_var &vec_geoc_frame)
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	vec_geoc_frame = NULL;
	if (!vec_nec_frame.is_nil() && !geo_lat.is_nil() && !geo_lon.is_nil()) {
		size_t nr_measurements = vec_nec_frame->sequence_size();
		vec_geoc_frame = new QdRMatrixSeq(nr_measurements, RMatrix(Dim(3), 0.0));
		double Bsph[3];  //frame defined as: [in colatitude direction; radially; eastward]
		double colat = -1, lon = -1, sin_col = -1, sin_lon = -1, cos_col = -1, cos_lon = -1;
		for (size_t t = 0; t < nr_measurements; ++t) {
			Bsph[0] = -((double) (*vec_nec_frame)[t][2]);
			Bsph[1] = -((double) (*vec_nec_frame)[t][0]);
			Bsph[2] =  ((double) (*vec_nec_frame)[t][1]);
			colat = M_PI/2 - ((double) (*geo_lat)[t]) / 180 * M_PI;
			sin_col = sin(colat);
			cos_col = cos(colat);
			lon = ((double) (*geo_lon)[t]) / 180 * M_PI;
			sin_lon = sin(lon);
			cos_lon = cos(lon);
			(*vec_geoc_frame)[t][0] = Bsph[0] * sin_col * cos_lon  +  Bsph[1] * cos_col * cos_lon  +  Bsph[2] * (-sin_lon);
			(*vec_geoc_frame)[t][1] = Bsph[0] * sin_col * sin_lon  +  Bsph[1] * cos_col * sin_lon  +  Bsph[2] * ( cos_lon);
			(*vec_geoc_frame)[t][2] = Bsph[0] * cos_col            +  Bsph[1] * (-sin_col);
		}
	}
	return status;
}

/**
 * Converts a vector object to the desired new frame.
 * \param[in] ts_in is a vector TS object, which has a well-defined frame
 * \param[in] frame is an integer. Acronyms are set to integer values by #define in qrtn.h
 *        0  GSE,  Geocentric Solar Ecliptic
 *        1  GSM,  Geocentric Solar Magnetic
 *        2  GEOC, Geographic (geocentric)
 *        3  SM,   Solar Magnetic
 *        4  GEI,  Geocentric Inertial
 *        5  MAGD,  Geomagnetic (dipole)
 *        6  HAE,  Heliospheric Aries Ecliptic
 *        7  HEE,  Heliospheric Earth Ecliptic
 *        8  HEEQ, Heliospheric Earth Equatorial
 *        9  GSEQ, Geocentric Solar Equatorial
 * \param[out] ts_out is the converted (rotated) TS object.
 */
QplugReturnStatus ConvertToDesiredFrame (QdRMatrixSeq_var &ts_in, int frame, QdRMatrixSeq_var &ts_out)
{
	QplugReturnStatus	status = QPLUG_SUCCESS;
	QrVector vec;
	vec.n_recs = ts_in->sequence_size();
	vec.in_frame = QrGetFrameFromAttr(ts_in);
	vec.out_frame = frame;
	//       next item in list is int identifying target frame
	//       acronyms are set to integer values by #define in qrtn.h
	//       0  GSE,  Geocentric Solar Ecliptic
	//       1  GSM,  Geocentric Solar Magnetic
	//       2  GEOC, Geographic (geocentric)
	//       3  SM,   Solar Magnetic
	//       4  GEI,  Geocentric Inertial
	//       5  MAGD,  Geomagnetic (dipole)
	//       6  HAE,  Heliospheric Aries Ecliptic
	//       7  HEE,  Heliospheric Earth Ecliptic
	//       8  HEEQ, Heliospheric Earth Equatorial
	//       9  GSEQ, Geocentric Solar Equatorial


	vec.IsPositionVector = QrIsPositionVector(ts_in);
	if(vec.IsPositionVector == 1) {
		vec.IsPositionVector = 0; /* issue warning & continue, delta r rotates */
		if(QrOriginsDiffer(vec.in_frame, vec.out_frame)) {
			QplugAppendTextDisplay
			("WARNING: CoordRotation plug-in provides frame rotation only.\n");
			QplugAppendTextDisplay
			("Transformation is between geocentric and heliocentric frames.\n");
			QplugAppendTextDisplay
			("Position vector coordinate transformations will be WRONG.\n");
			QplugAppendTextDisplay
			("Note: Spacecraft separations will rotate OK.\n");
		}
	}

	ts_out = (QdRMatrixSeq_var) Factory::clone(ts_in);
	ts_out->copy_xrefs_from(ts_in);
	vec.ptr = ts_out;

	vec.t = get_timetags(ts_in);
	if( (vec.t).is_nil() ) {
		QplugAppendTextDisplay ("Cannot find time tags, some frames are time dependent\n");
		return QPLUG_FAILURE;
	}

	// call C function to do rotation
	long check = QrDoRotation(&vec);
	//Check that the COORDINATE_SYSTEM metadata is correctly written after the coordinate transformation.
	if (check == QRTN_OK || check == SAME_FRAME) {
		QdTimeSeq_var new_tt = Factory::clone(vec.t);
		if(new_tt.is_nil()){
			QplugAppendTextDisplay ("Cannot create new time tags\n");
			return QPLUG_FAILURE;
		}
		set_timetags((QdObject_var)vec.ptr, new_tt);

		ts_out = vec.ptr;
	}
	else{
		ts_out = NULL;
		QplugAppendTextDisplay (QrErrorMessage(check));
		status = QPLUG_FAILURE;
	} // end if on check
	return status;
}

// =========  class Status  =================================================

Status::Status()
{
	message_ = "";
}

Status::~Status()
{
}

string Status::message() const
{
	string result = "ChampToGeo: ";
	result += message_;
	return result;
}

void Status::PrintMessage(bool to_user, const char fmt[], ...)
{
	try {
		va_list args;
		char buffer[2048];
		va_start(args, fmt);
		vsprintf(buffer, fmt, args);
		va_end(args);
		if (to_user)
			QplugAppendTextDisplay(buffer);
		printf("%s", buffer);
		fflush(stdout);
	}
	catch (...) {
	}
}

