/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	CorrelFFT.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS CorrelFFT wrapper
*
* PURPOSE:      Wrapper for Fast Fourier Transform.
* 
* DESCRIPTION:  Plug In module checking input parameters and
*               constructing CorrelFFT arrays.
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*              see CorrelFFT.qtpl file
*   OUT:
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	four1.c cdft.c
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*
* HISTORY: created  20/03/2001 Markus Fraenz 
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

extern "C" void rdft(int n, int isgn, double *a, int *ip, double *w);
extern "C" QplugReturnStatus CorrelFFT(QplugArgList *);

using namespace QSAS;


QplugReturnStatus	CorrelFFT( QplugArgList *call_list )
{

// define some C variables

int	        attribute_ok,list_length;
unsigned long	Ndata,Ndata2=2,Ndata4,i,np,nm,Ndatah,icorr=0,iacorr=0;
int             method,*ip_tak,j,j1;

int qar_status;

char str_out[150];
double  *data,*dataG,*dataP,*dataPF,*dataPG,deltat,T,*w_tak,dnm,dum;
double  sumf=0.0,sumg=0.0,dcorr=-1.0,dacorr=1.0;

// define QSAS data objects


QdRScalarSeq_var  out_tag_series,out_correl_series;
QdRScalarSeq_var  out_freq_series,out_coh_series,out_phas_series;



//
// ...first, unpack the input data objects
//
QplugAppendTextDisplay("CorrelFFT 1.0 11 Apr 2001\n");

// ...first item is input Series

QdObject_var F_in = (* call_list)[0];

QdRScalarSeq_var ff = QdRScalarSeq_var::narrow(F_in);
if(ff.is_nil() ){
      QplugAppendTextDisplay ("ComplexProduct>> ERROR - F sequence type not handled\n");
      return QPLUG_FAILURE;  
}

Ndata = ff->sequence_size();


QdObject_var G_in = (* call_list)[1];

QdRScalarSeq_var gg = QdRScalarSeq_var::narrow(G_in);
if(gg.is_nil() ){
      QplugAppendTextDisplay ("ComplexProduct>> ERROR - G sequence type not handled\n");
      return QPLUG_FAILURE;  
}


// Test validity of inputs...

qar_status =  QarTestConformal(F_in, G_in);
  if(qar_status == QAR_ERR)   
  {
    QplugAppendTextDisplay ("Arithmetic>> input sequences not joined properly\n");
     return QPLUG_FAILURE; 
  } 
  
//	..4th item is option 'Method'
method = (int) *QdRScalar_var::narrow( (* call_list)[2]);



//	..extract the time tags from the data series object ###SUBSETTING END

QdTimeSeq_var tt;
QdRScalarSeq_var ss;
QdObject_var d0;

if(F_in->xref_exists(DEPEND_0) ){
	d0 = F_in->get_xref(DEPEND_0);
	tt = QdTimeSeq_var::narrow(d0);
	ss = QdRScalarSeq_var::narrow(d0);
}
else if(G_in->xref_exists(DEPEND_0) ){
	d0 = G_in->get_xref(DEPEND_0);
	tt = QdTimeSeq_var::narrow(d0);
	ss = QdRScalarSeq_var::narrow(d0);
}

if(!tt.is_nil()) QuIsTSRegular(tt, &deltat);
else if(!ss.is_nil()) QuIsSSRegular(ss, &deltat);
else deltat=1.0;
 


if (Ndata <= 1) {
	QplugAppendTextDisplay ("CorrelFFT>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; }

//      (notice that NR routines start indices with 1!!) 

QplugAppendTextDisplay ("CorrelFFT>> using Ooura CDFT8G.\n");


// zero - padding :
while (Ndata2 < Ndata) Ndata2 *= 2;

Ndatah=Ndata2/2;

data    = new double[Ndata2]; 
dataG    = new double[Ndata2];
dataP    = new double[Ndata2];
dataPF    = new double[Ndata2];
dataPG    = new double[Ndata2];

// initialize Tak data arrays and transform input
ip_tak = new int[(long) sqrt((float) Ndatah) +3];
ip_tak[0]=0;
w_tak = new double[Ndatah];



//first find mean values of original series
for (i=0; i<Ndata; i++) {
 sumf +=  ff[i];
 sumg +=  gg[i];
}
sumf/=(double)Ndata;
sumg/=(double)Ndata;


for (i=0; i<Ndata; i++) data[i] =  ff[i]-sumf;
for (i=Ndata; i<Ndata2; i++) data[i] = 0.0; 
rdft((int) Ndata2, 1, data, ip_tak, w_tak);

for (i=0; i<Ndata; i++) dataG[i] =  gg[i]-sumg;
for (i=Ndata; i<Ndata2; i++) dataG[i] = 0.0; 
rdft((int) Ndata2, 1, dataG, ip_tak, w_tak);

// now multiply transforms:
for (i =0; i<Ndata2; i+=2) {
 dataP[i]=(data[i]*dataG[i]+data[i+1]*dataG[i+1]);
 dataP[i+1]=(data[i+1]*dataG[i]-data[i]*dataG[i+1]);
}
dataP[1]=data[1]*dataG[1];

//now calculate autocorr for each
for (i =0; i<Ndata2; i+=2) {
 dataPF[i]=(data[i]*data[i]+data[i+1]*data[i+1]);
 dataPF[i+1]=0.0;
}
dataPF[1]=data[1]*data[1];

for (i =0; i<Ndata2; i+=2) {
 dataPG[i]=(dataG[i]*dataG[i]+dataG[i+1]*dataG[i+1]);
 dataPG[i+1]=0.0;
}
dataPG[1]=dataG[1]*dataG[1];

//now calculate freq and coherence radius and phase
if (method > 0) {
 out_freq_series  = new QdRScalarSeq(Ndatah);
 out_coh_series = new QdRScalarSeq(Ndatah);
 out_phas_series = new QdRScalarSeq(Ndatah);
 T  = (double)Ndata2 * deltat;
 for (i =0; i<Ndatah; i++) {
  j=2*i;
  out_freq_series[i] =  (double)i / T;
  dum = dataP[j]*dataP[j];
  dum /= dataPF[j]*dataPG[j];
  out_coh_series[i] = sqrt(dum);
  out_phas_series[i] = atan2(dataP[j+1],dataP[j]);
 }
}

if (method !=1) {
 out_tag_series  = new QdRScalarSeq(Ndata2);
 out_correl_series = new QdRScalarSeq(Ndata2);

 //and inverse transform
 rdft((int) Ndata2,-1, dataPF, ip_tak, w_tak);
 rdft((int) Ndata2,-1, dataPG, ip_tak, w_tak);
 rdft((int) Ndata2,-1, dataP, ip_tak, w_tak);

 // to make F_in leading sign(lag) is changed:
 dnm = (double)Ndatah/sqrt(dataPF[0]*dataPG[0]);

 for (i =0; i<Ndatah; i++) {
  np=Ndatah+i;
  out_tag_series[np] = -(double)i * deltat;
  out_tag_series[i] = (double)(Ndatah-i) * deltat; 
  out_correl_series[np] = dataP[i]*dnm/(double)(Ndatah-i);
  out_correl_series[i] = dataP[i+Ndatah]*dnm/(double)(i+1);
 }

 //determine extrema in interval -N/8,+N/8:
 for (i =3*Ndatah/4; i<5*Ndatah/4; i++) {
  if ((double)out_correl_series[i]>dcorr) {
    dcorr=(double) out_correl_series[i];
    icorr=i;
  }
  else if ((double)out_correl_series[i]<dacorr) {
    dacorr=(double)out_correl_series[i];
    iacorr=i;
  }
 }
}

//	...set some attributes

if (method !=1) {
  QuSetTxtAttr("SI_conversion", " ", out_correl_series);
  QuSetTxtAttr("LABLAXIS","CorrelFFT", (QdObject_var) out_correl_series);
  QuSetTxtAttr("FIELDNAM","CorrelFFT", (QdObject_var) out_correl_series);
  QuSetFrame((QdObject_var) out_correl_series, "scalar");
  QuSetTxtAttr("UNITS", " ", (QdObject_var) out_correl_series);


  QuSetTxtAttr("SI_conversion", "1", (QdObject_var) out_tag_series);
  QuSetTxtAttr("LABLAXIS", "TimeLag", (QdObject_var) out_tag_series);
  QuSetTxtAttr("FIELDNAM", "TimeLag", (QdObject_var) out_tag_series);
  QuSetFrame((QdObject_var) out_tag_series, "scalar");
  QuSetTxtAttr("UNITS", "s", (QdObject_var) out_tag_series);
}
if (method >0) {
  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_freq_series);
  QuSetTxtAttr("LABLAXIS", "frequency", (QdObject_var) out_freq_series);
  QuSetTxtAttr("FIELDNAM", "frequency", (QdObject_var) out_freq_series);
  QuSetFrame((QdObject_var) out_freq_series, "scalar");
  QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_freq_series);

  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_coh_series);
  QuSetTxtAttr("LABLAXIS", "Coherence", (QdObject_var) out_coh_series);
  QuSetTxtAttr("FIELDNAM", "Coherence", (QdObject_var) out_coh_series);
  QuSetFrame((QdObject_var) out_coh_series, "scalar");
  sprintf(str_out,"%s", QuGetAttrText("UNITS", F_in).c_str());
  QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_coh_series);

  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_phas_series);
  QuSetTxtAttr("LABLAXIS", "Phase", (QdObject_var) out_phas_series);
  QuSetTxtAttr("FIELDNAM", "Phase", (QdObject_var) out_phas_series);
  QuSetFrame((QdObject_var) out_phas_series, "scalar" );
  QuSetTxtAttr("UNITS", "rad", (QdObject_var) out_phas_series);
}

//	..some messages to the analysis interface window..

sprintf(str_out, "CorrelFFT>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out, "CorrelFFT>> %ld zero padded data points\n", Ndata2);
QplugAppendTextDisplay (str_out);
if (method !=1) {
 sprintf(str_out, "CorrelFFT>> Max. Correlation %e, at lag %e\n", dcorr,(double)out_tag_series[icorr]);
 QplugAppendTextDisplay (str_out);
 sprintf(str_out, "CorrelFFT>> Max. AntiCorrelation %e, at lag %e\n", dacorr,(double)out_tag_series[iacorr]);
 QplugAppendTextDisplay (str_out);
}
QplugAppendTextDisplay ("CorrelFFT>> finished...\n");

//	..add them to the list

  call_list->push_back( (QdObject_var) out_tag_series);    
  call_list->push_back( (QdObject_var) out_correl_series);     
  call_list->push_back( (QdObject_var) out_freq_series);    
  call_list->push_back( (QdObject_var) out_coh_series);     
  call_list->push_back( (QdObject_var) out_phas_series);     




//	..clean up and exit

delete[] data; 
delete[] dataG; 
delete[] dataP;
delete[] dataPF;
delete[] dataPG;
delete ip_tak;
delete w_tak;


return QPLUG_SUCCESS;


}



