/**********************************************************************
* SCCS id: @(#)CorCoef.c	1.1 14:32:01 12/17/98 
***********************************************************************
* PROJECT:   Cluster UK CDHF
* COMPONENT: CorCoef Correlation Coefficient
* MODULE:    CorCoef.c
* LANGUAGE:  C  
* PURPOSE:    Compute the correlation coefficient between two data series
* DESCRIPTION: 
*
* These routines are based on SJS's fortran routines dated 1991
* with some up-dating to treat end conditions better
*
c  This routine computes the correlation coefficient
c
c  rho_xy(r)   = Cxy(r)/sqrt ( Cxx(0) Cyy(0) )
c
c  between arrays x(i) and y(i) at lags   -nlags <= r <= +nlags
c  Here Cxy is the Covariance function, which is just Rxy(r) = <xy>(r), the 
c  cross-correlation function found by removing the means from
c  the data points x(i) and y(i). That is:
*
*  Cxy(r) = < (x-xbar)(y-ybar) >(r)
*
c  Cxx(0) is equal to the standard deviation squared of the x(i)'s
c  Some trivial algebra yields
c
c  Cxy(r) = Rxy(r) - <x><y>
*
*  In the above <xy>(r) = [ sum (i=0, N-r) x(i) y(i+r) ] / (N-r) for r>0
c
* <x>, <y> and their STD's (for Cxx(0) and Cyy(0)) must be calculated
* over the (N-r) points used in the summations above, that is, the
* non-overlapping regions are removed and the effective length of the
* data series reduced. The alternative, employed in the FFT approach,
* is to zero pad the ends. This replaces (N-r) divisors by N everywhere
* in the algorithm. If these two approaches are mixed, rho can take on 
* values larger than unity. Zero padding would clearly (?) diminish the
* values of rho for larger |r|. Alternatively, the discrete, non-padded
* algorithm here over-emphasises rho at large |r|.
*
c      NOTE:   rho must have dimension >= 2*nlags +1
*
* Arguments:
*   IN
*      CorCoefpacket *ccpacketp   pointer to corcoefpacket xith x,y,npts,nlags
*   OUT
*      rho_xy(r) as shown above in corcoef_results[r+nlags] r=-nlags ... +nlags
*
*   Sign convention: r>0 pairs x[i] with y[i+r], i.e., moves x forward
*   relative to y. 
*	
* Modules Used:  Xcor.c and stats.c
*
* HISTORY: SJ Schwartz Dec 98
*/

/* *************************************************************************** */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "CorCoef.h"


CorCoefReturnStatus CorCoef(CorCoefpacket *ccpacketp){

StatsSuccessorFailure   stats_ok;
CorCoefReturnStatus xcor_ok;

int i,r;

int  npts, nlags;                    /* local copies */
float xbar, ybar, xsigma, ysigma, tol;   

float xparsum1, xparsum2, yparsum1, yparsum2;

float *rzero;             /* pointer to r=0 pos in results array */

/**** check input; need at least 2 pts for all r (to calculate sigmas) */

nlags = ccpacketp->nlags;
npts  = ccpacketp->npts ;          /* make local copies for convenience */

if (nlags >= npts-1) return CORCOEF_TOO_MANY_LAGS_ERROR;


/* find whole data series means and sigmas for tol */
stats_ok=stats(ccpacketp->x, npts, &xbar, &xsigma);
if (stats_ok != STATSSUCCESS) return CORCOEF_STATS_ERROR;
stats_ok=stats(ccpacketp->y, npts, &ybar, &ysigma);
if (stats_ok != STATSSUCCESS) return CORCOEF_STATS_ERROR;

tol = ( xsigma < ysigma ) ? xsigma : ysigma ;   /* min of sigmas */
tol /= 1000.0 ;         /* tolerance for test if any sigma = 0 */


/****************************************** compute cross correl. fcn */
xcor_ok = Xcor(ccpacketp);

if (xcor_ok != CORCOEF_OK) return CORCOEF_STATS_ERROR;
#ifdef VERBOSE
printf("Xcor OK");
#endif
/******************** fill partial sums and calculate final vals  */

rzero = ccpacketp->corcoef_results + nlags;   /* point to r=0 pos in array */

for( r=0; r < nlags +1 ; r++){

	/* compute mean and std over pts in overlap region for r>=0 */
	stats_ok=stats( ccpacketp->x, npts-r,&xbar, &xsigma);
	if (stats_ok != STATSSUCCESS) return CORCOEF_STATS_ERROR;
	stats_ok=stats( ccpacketp->y+r, npts-r,&ybar, &ysigma);
	if  (stats_ok != STATSSUCCESS) return CORCOEF_STATS_ERROR;

	/* rzero array already contains Xcor results for Rxy*/
	rzero[r]  -= xbar*ybar;
	if ((xsigma > tol) && (ysigma > tol )) rzero[r]  /= (xsigma*ysigma);
	else rzero[r] = 0.0;
	
	if(r!=0){      /* only calculate r=0 once as Rxy overwritten */

	/* compute mean and std over pts in overlap region for r<0 */
	stats_ok=stats( ccpacketp->x+r, npts-r,&xbar, &xsigma);
	if  (stats_ok != STATSSUCCESS)  return CORCOEF_STATS_ERROR;
	stats_ok=stats( ccpacketp->y, npts-r,&ybar, &ysigma);
	if  (stats_ok != STATSSUCCESS) return CORCOEF_STATS_ERROR;
	
	rzero[-r]  -= xbar*ybar;
	if ((xsigma > tol) && (ysigma > tol )) rzero[-r]  /= (xsigma*ysigma);
	else  rzero[-r] =0.0;
	}
}

return CORCOEF_OK;

}   /* end CorCoef  */



/***********************************************************************
************************************************************* Xcor
***********************************************************************/

CorCoefReturnStatus Xcor(CorCoefpacket *ccpacketp){

/*
c Compute the Cross-correlation function Rxy(r) between the sets of data
c points xi and yi for a lag of r increments (-nlags .le. r .le. +nlags)
c according to
c
c  Rxy(r) = (1/(N-r)) sum x(i) y(i+r)  where sum is over (N-r) pts
c
c Note Rxy(-r) = Ryx(r) so can compute using positive lags only, i.e. 
c for  0        < r   < nlags   use        x(i)     y(i+r)
c for -nlags < r   < 0         use        x(i+r)  y(i)
c 
c NOTE: rxy must have dimension at least 2*nlags +1
c
c Result gets passed back to CorCoef in ccpacket->corcoef_results
*/

int	i,r;
float	rpos, rneg;
float	*rzerop;            /* pointer to center of results array */
int	npts, nlags;

npts = ccpacketp->npts;
nlags = ccpacketp->nlags;

if ( nlags >= npts-1) return CORCOEF_TOO_MANY_LAGS_ERROR;

rzerop = ccpacketp->corcoef_results + nlags;  /* set pointer to r=0 position */


for (r=0; r<nlags+1 ; r++){
	for (	i=0, rpos = 0.0, rneg = 0.0 ;
		i < npts - r        ;
		i++){
		rpos += ccpacketp->x[i]*ccpacketp->y[i+r];
		rneg += ccpacketp->x[i+r]*ccpacketp->y[i];
/*		printf("%g, %d, %g\n",rpos, i, ccpacketp->x[i]); */
	}
	rzerop[r]  = rpos/(npts-r);
	rzerop[-r] = rneg/(npts-r);

#ifdef VERBOSE
  printf("r = %d, rpos = %g, rneg = %g, rzerop[r] = %g, rzerop[-r] = %g\n",
         r,rpos, rneg, rzerop[r],rzerop[-r]);
  printf("Check: ccpacketp->corcoef_results[nlags+r] = %g,  ccpacketp->corcoef_results[nlags - r] = %g\n",
            ccpacketp->corcoef_results[nlags+r],
            ccpacketp->corcoef_results[nlags-r]);

#endif

}

return CORCOEF_OK;
}  


/******************************************************************
******************************************************* stats *****
*******************************************************************/

StatsSuccessorFailure stats (float *x, int n, float *xbarp, float *sigmap){

/* compute mean and standard deviation */

/*  Arguments:

float *x      pointer to array of values
int   n       number of values in array x
float *xbarp  pointer to location where mean will be returned
float *sigmap pointer to location where standard deviation will be returned

Algorithm:
  xbar = [ sum xi ] / n
  sigma = sqrt{ [ sum (x-xbar)^2 ] / n }    NB not /(n-1)
        = sqrt{ [ sum x^2]/n  - xbar^2 }
*/

int i;

if (n<1) return STATSFAILURE;   /* must have at least one value */

for (i=0, *xbarp = 0, *sigmap=0;
     i < n ;
     i++){
     *xbarp += x[i];
     *sigmap += x[i]*x[i];
}

*xbarp /= n;
*sigmap /= n;
*sigmap -= (*xbarp)*(*xbarp);

if(*sigmap < 0.0)   return STATSFAILURE;              /* be dead safe! */
else if             (*sigmap == 0.0) *sigmap=0.0;
else                *sigmap = sqrt(*sigmap);



return STATSSUCCESS;
}
