// ------------------------------------------------------------------
// ISSI QSAS PlugIn : deHoffman-Teller analysis
// 
// This plugin takes a magnetic field vector and a plasma velocity
// vector as inputs, and calculates the deHoffman-Teller velocity
//
// Required inputs :
// 	bfield		- vector time series with magnetic field
//	velocity	- vector time series with plasma velocity
//
// Optional inputs (leave empty or '0' if not used):
//      density         - scalar time series with plasma density
//      t_par           - scalar time series with paralell plasma temperature
//      t_perp          - scalar time series with perpendicular plasma temperature
//      normal          - vector unit vector with boundary normal to dot inth V_HT
//
// Outputs
//      E_ht            - vector time series with Vht x B
//      E_C             - vector time series with V x B
//
//      Also, the following outputs are valid if a density is given :
//      V_rest          - vector time series with Vr = V - Vht 
//      V_alfven        - vector time series with the Alfven velocity 
//
//
// SEH, 19 Oct, 2001
//      29 May, 2002 - added a normal as input and V_ht * n
//      12 Aug, 2002 - added density as optional input
//      05 Sep, 2002 - temporary bug fix : return all WS outputs
//      15 Jan, 2003 - V_rest as output also without density + ccoeff for Va/Vrest
//      12 Dec, 2003 - modified for QSAS_2.0
//      21 Jan, 2004 - time interval output + QString for output window
//      30 Mar, 2004 - pressure correction introduced
//      05 May, 2004 - use qdos classes and methods -> remove 'minvert.c' dependence
//      06 May, 2004 - incorporated AJA's input checking  + Doxygen info
//      18 Jun, 2004 - logfile possible
//      30 Jan, 2005 - corrected bug in pressure anisotropy
//      13 May, 2005 - catch sigularity in alpha calculation; if B^2==0, or alpha >> 1, set alpha = 0
//      21 Jan, 2008 - hack to allow Themis temperatur tensor in addition to T_par, T_perp
//      22 Jan, 2008 - removed (comment out) calls to minvar
//	27 Jan, 2008 - temporary removed QT dependencies
//      30 Jan, 2008 - reinserted QT dependencies + added hack for Themis SI conversion n = [10^6>m^-3]
//
//
// Doxygen info :
//
/* -*- mode: C++ ; c-file-style: "stroustrup" */
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include <qstring.h>
#include <qdatetime.h>
#ifndef WINDOWS
#include "GenericMinVar.h"
#endif

#ifndef WINDOWS
QdRMatrix_var Q_MVAB_calc(QdRMatrixSeq_var B,   QdRMatrixSeq_var V, QdRMatrix_var U, 
			  QdRMatrixSeq_var Valf,  QdRScalarSeq_var N, 
			  double *wl, double *cc, int nestsize, int full);
double slope(QdRMatrixSeq_var X_in,QdRMatrixSeq_var Y_in,int n);
double correlationCoeff(QdRMatrixSeq_var X_in,QdRMatrixSeq_var Y_in,int n);
#else
double vecsum2(QdRMatrix_var Vector)
  //
  // returns the norm^2 of a vector
  //
{
  double sum=0.0;

  for (int i=0; i<3; i++) sum += (*Vector)[i] * (*Vector)[i];

  return(sum);
}
#include "../GenericResidue/Q_MVAB_calc.cc"
#endif
double QSAS_vecsum2(QdRMatrix_var Vector);
double QSAS_vecsum2(QdRMatrixSeq_var VectorSeq, long index);
double minimizeD(int M, QdRMatrix_var V, QdRMatrixSeq_var vm, QdRMatrixSeq_var Bm);
QdRMatrix_var V_HTcalc(long nestsize, double deltat, QdRMatrixSeq_var B,QdRMatrixSeq_var V, QdRMatrix_var K0c,     
	     QdRMatrix_var K, QdRMatrix_var K1, QdRMatrix_var K2, QdRMatrix_var KV,  QdRMatrix_var KVT);

QdRMatrix_var str2Vec(QString s);
QdRMatrix_var aveVectorOver(QdRMatrixSeq_var VectorSeq, long sta, long sto);
double aveScalarOver(QdRScalarSeq_var ScalarSeq, long sta, long sto);


extern "C" QplugReturnStatus  HofTel  ( QplugArgList *);

using namespace QSAS;




QplugReturnStatus  HofTel  ( QplugArgList *call_list )
// ---------------------------------------------------------------------------
/*!
  \brief Perform a deHoffmann-Teller analysis and return the results as QSAS_objects and outputs 
  \brief performs a deHoffmann-Teller analysis
  \param call_list - pointer to a list of input objects
*/
//
// SEH, 15 Aug, 2002
//      06 May, 2004
//
// ---------------------------------------------------------------------------
{ 
 QTime t;
 t.start();                     // start clock

 long    nestsize;
 int     i,j,k,l,n=3,qar_status;
 int     nseq = 2;
 double  vpar_scalar,avg_width;
 double  deltat;
 double  bb = 0.00;             // to hold B^2
 char    txtbuf[255];
 bool    LogFile = FALSE;

 // Logfiles - i.e. output to files applies to MPE/ISSI ONLY :
 LogFile = false;//qwuiWindow::UseLogFile->isChecked();
 if (LogFile) {
 }

 // ---- get density if supplied -> 3 instead of 2 series to join ----------

 QdRScalarSeq_var density = QdRScalarSeq_var::narrow((*call_list)[2]);
 if (!density.is_nil()) {
   if (is_empty(density) ){
     QplugAppendTextDisplay("density object is empty\n");
   }
   else{
     QplugAppendTextDisplay("HofTel> Density supplied.. ");
     nseq = 3;
   }
 }
 else {
   QplugAppendTextDisplay("HofTel> Density not supplied.. ");
 }



 // ----- added by AJA :  test if frames are same for B and V
 
 int frame_status = QarTestConformalForVecProds((*call_list)[0], (*call_list)[1]);
  if ( frame_status == QAR_ERR ){
   QplugAppendTextDisplay("B field and velocity in different frames\n");
   return QPLUG_FAILURE;
 }
 else if(frame_status == QAR_WARN){
   QplugAppendTextDisplay("Caution: Frame unknown for either B or V\n");
 }
 
 // ---- get Pressure if supplied -> 5 instead of 2 series to join ----------

 bool alpha = false; 
 QdRScalarSeq_var temp_par = QdRScalarSeq_var::narrow((*call_list)[4]);
 QdRScalarSeq_var temp_perp = QdRScalarSeq_var::narrow((*call_list)[5]);
 QdRMatrixSeq_var temp_tensor =  QdRMatrixSeq_var::narrow((*call_list)[4]);
 if (!density.is_nil() && !temp_par.is_nil() && !temp_perp.is_nil()) {
    QplugAppendTextDisplay("HofTel> Temp supplied as t_par and t_perp.. ");
   nseq += 2;
   alpha = true;
 } 
 else if (!density.is_nil() && !temp_tensor.is_nil()) {
   QplugAppendTextDisplay("HofTel> Temp supplied as tensor.. ");
   nseq += 1;
   alpha = true;
 }  
 else {
    QplugAppendTextDisplay("HofTel> Temp not supplied.. ");
 }

 

 
 

 //  --- first we join B measurements to V measurements -----

 printf("HT : join B, V and N, T (if supplied) \n");
 fflush(stdout);
 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 vector <double> gapValue(nseq);
 QdStringSeq_var gapHandler = new QdStringSeq(nseq);
 QdStringSeq_var joinMethod = new QdStringSeq(nseq);
 QdRMatrixSeq_var  bfield_nest;
 QdRMatrixSeq_var  velocity_nest;
 QdRScalarSeq_var  density_nest;
 QdRScalarSeq_var  Tpar_nest;
 QdRScalarSeq_var  Tperp_nest;
 QdTimeSeq_var tt_out;                                      // tt of masked V and B
 QdRMatrix_var     normal;                                  // the normal to be dotted into Vht
 int slots_use[5] = {0,1,2,4,5};                            // input slots to be joined

//  ---- get Time interval and mask ----------------
 QdTimeSeq_var ttJoined = get_timetags((*call_list)[1]);    // tt of joined V and B

 printf("HT : get time interval \n");
 fflush(stdout);
 QdTimeInterval_var start_end = get_timeinterval((* call_list)[3]);
 SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 tt_out = new QdTimeSeq();
 ttJoined->get_subsequence(*tt_out, msk);

 if (tt_out.is_nil()) {
   QplugAppendTextDisplay("Bad or missing time tags in input data\n");
   return QPLUG_FAILURE;
 }

 QuIsTSRegular(tt_out,&deltat);
 avg_width = 1.5 * deltat;
 printf("delta t = %f  ->  gap_fill if t > %f\n",deltat,avg_width);
 fflush(stdout);

 for (i=0; i<nseq; i++) {
   toJoin[i] = ((*call_list)[slots_use[i]]);

   joinMethod[i] =  QdString(LINEAR);
   gapValue[i] = avg_width;
   gapHandler[i] =  QdString(LINEAR);   
 }

 try {
   printf("HT : avg width = %f\n",avg_width);
   QuMultiJoin(toJoin, tt_out, joined, gapValue, gapHandler, joinMethod);
   printf("%f\n",avg_width);
   fflush(stdout);
 }
 catch(Exception &e) {
   e.print_msg();
   printf("HT : Exception in join : avg width %f\n",avg_width);
   fflush(stdout);
   return QPLUG_FAILURE;
 }


 

 //  ---- get B-field -------------------------------

 printf("HT : get joined B-field \n");
 fflush(stdout);
 QdRMatrixSeq_var masked_B = QdRMatrixSeq_var::narrow(joined[0]);

 if(masked_B.is_nil())
 {
	 printf("B is nil");
	 fflush(stdout);
 }
  // added by AJA to ensure B field units are correct (nT)
   string SIcB = QuGetAttrText("SI_conversion", ((* call_list)[0]));
   if ( SIcB.empty() )  {
     QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for B field");
     bfield_nest = (QdRMatrixSeq_var)  masked_B;
   }
   else if (QarSameBaseSI(SIcB.c_str(), "1.0e-9>T")){
     // base units the same, ensure conversion is too
     // take care as masked_B does not have orig xref
     masked_B->change_xref("SI_conversion", new QdString(SIcB));
     bfield_nest = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_B, "1.0e-9>T"));
   }
   else {
     QplugAppendTextDisplay("ERROR: cannot correct units of B field");
     return QPLUG_FAILURE;
   }

 if( is_empty(bfield_nest) ){
     QplugAppendTextDisplay("ERROR: no B field data in selected interval");
     return QPLUG_FAILURE;
 }

 nestsize = bfield_nest->sequence_size();
 printf("HT : Nestsize B = %ld\n",nestsize);
 fflush(stdout);



 //  ---- get Velocity (or E-field) ----------------------

 printf("HT : get velocity or E-field \n");
 fflush(stdout);
 QdRMatrixSeq_var masked_V = QdRMatrixSeq_var::narrow(joined[1]);
 QdRMatrixSeq_var nested_E;

  // added by AJA to ensure velocity units are correct (km s^-1)
   string SIcVel = QuGetAttrText("SI_conversion", ((* call_list)[1]));
   if ( SIcVel.empty() )  {
     QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for velocity");
     velocity_nest = (QdRMatrixSeq_var)  masked_V;
   }
   else if (QarSameBaseSI(SIcVel.c_str(), "1.0e3>m s^-1")){
     // base units the same, ensure conversion is too
     QplugAppendTextDisplay("HofTel> Input is plasma velocity...");
     masked_V->change_xref("SI_conversion", new QdString(SIcVel));
     velocity_nest = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_V, "1.0e3>m s^-1"));
     nested_E = Factory::clone( QdRMatrixSeq_var::narrow(velocity_nest));
   }
   else if (QarSameBaseSI(SIcVel.c_str(), "1.0e-3>V m^-1")){
     // --- SEH : input is E-field -> generate V from E x B ----
     QplugAppendTextDisplay("HofTel> Input is electric field...");
     masked_V->change_xref("SI_conversion", new QdString(SIcVel));
     velocity_nest = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_V, "1.0e-3>V m^-1"));
     nested_E = Factory::clone( QdRMatrixSeq_var::narrow(velocity_nest));
     for (i=0; i<nestsize; i++) {
       //       for (j=0; j<3; j++) (*B_k)[j] = (*bfield_nest)[i][j];
       bb = 1000.0/ QSAS_vecsum2(bfield_nest,i);
       velocity_nest[i][0] =  bb * (nested_E[i][1] * bfield_nest[i][2]  -  nested_E[i][2] * bfield_nest[i][1]);
       velocity_nest[i][1] =  bb * (nested_E[i][2] * bfield_nest[i][0]  -  nested_E[i][0] * bfield_nest[i][2]);
       velocity_nest[i][2] =  bb * (nested_E[i][0] * bfield_nest[i][1]  -  nested_E[i][1] * bfield_nest[i][0]);
     }
   }
   else {
     QplugAppendTextDisplay("ERROR: cannot correct units of velocity");
     return QPLUG_FAILURE;
   }
 
 if( is_empty(velocity_nest) ){
     QplugAppendTextDisplay("ERROR: no V or E data in selected interval");
     return QPLUG_FAILURE;
 }

 nestsize = velocity_nest->sequence_size();
 printf("Nestsize V = %ld\n",nestsize);
 fflush(stdout);

//  for (int i=0; i<nestsize; i++) {
//    printf("B = %10.3f  %10.3f  %10.3f\n",(double)(*bfield_nest)[i][0],(double)(*bfield_nest)[i][1],(double)(*bfield_nest)[i][2]);
//    printf("E = %10.3f  %10.3f  %10.3f\n",(double)(*nested_E)[i][0],(double)(*nested_E)[i][1],(double)(*nested_E)[i][2]);
//    printf("V = %10.3f  %10.3f  %10.3f\n",(double)(*velocity_nest)[i][0],(double)(*velocity_nest)[i][1],(double)(*velocity_nest)[i][2]);
//  }

 // ---- get density if supplied -------


 printf("HT : get density if supplied \n");
 fflush(stdout);
 if (!density.is_nil()) {
   QdRScalarSeq_var masked_D = QdRScalarSeq_var::narrow(joined[2]);

   // -- added by AJA to ensure density units are correct (1/cc) --
   string SIcDen = QuGetAttrText("SI_conversion", ((* call_list)[2]));
   if (SIcDen.empty())  {
     QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for density");
     density_nest = (QdRScalarSeq_var) masked_D;
   }
   // -- SEH 30 Jan 2008 > hack for erronous Themis SI conversion --
   else if (QarSameBaseSI(SIcDen.c_str(), "1.0e6>m^-3") ||  QarSameBaseSI(SIcDen.c_str(), "10^6>m^-3")){
     // base units the same, ensure conversion is too
     masked_D->change_xref("SI_conversion", new QdString(SIcDen.c_str()));
     density_nest = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_D, "1.0e6>m^-3"));
   }
   else {
     QplugAppendTextDisplay("ERROR: cannot correct units of density");
     return QPLUG_FAILURE;
   }
   if( is_empty(density_nest) ){
     QplugAppendTextDisplay("ERROR: no density data in selected interval");
     return QPLUG_FAILURE;
   }
   nestsize = density_nest->sequence_size();
 }
 printf("Nestsize rho = %ld\n",nestsize);
 fflush(stdout);



 // ---- get pressure if supplied -------

 printf("HT : get temperature if supplied \n");
 fflush(stdout);
 if (alpha) {
   if (!temp_tensor.is_nil()) {
     // -- SEH 20 Jan, 2008 : extract tpar and t_perp from Themis magt3
     double inv_k_boltzmann = 1.602e-19/1.38054e-17; // [MK/ev]
     QdRMatrixSeq_var masked_Ttens = QdRMatrixSeq_var::narrow(joined[3]);

     string SIcTpar = QuGetAttrText("SI_conversion", ((* call_list)[4]));
     Tpar_nest = new QdRScalarSeq(masked_Ttens->sequence_size());
     Tperp_nest = new QdRScalarSeq(masked_Ttens->sequence_size());
     if (SIcTpar.empty() || SIcTpar.find("1.602e-19>J") == string::npos) {
       QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) in Temp tensor");       
     } 

     for (i=0; i<masked_Ttens->sequence_size(); i++) {
       (*Tpar_nest)[i] = inv_k_boltzmann * (double)(*masked_Ttens)[i][2];
       (*Tperp_nest)[i] = inv_k_boltzmann * 0.5 * ((double)(*masked_Ttens)[i][0]+(double)(*masked_Ttens)[i][1]);
     }    
   } else {    
     QdRScalarSeq_var masked_Tpar = QdRScalarSeq_var::narrow(joined[3]);

     string SIcTpar = QuGetAttrText("SI_conversion", ((* call_list)[4]));
     if (SIcTpar.empty())  {
       QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for T_parallel");
       Tpar_nest = (QdRScalarSeq_var) masked_Tpar;
     }
     else if (QarSameBaseSI(SIcTpar.c_str(), "1.0e6>K")){
       // base units the same, ensure conversion is too
       masked_Tpar->change_xref("SI_conversion", new QdString(SIcTpar));
       Tpar_nest = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_Tpar, "1.0e6>K"));
     }
     else {
       QplugAppendTextDisplay("ERROR: cannot correct units of T_parallel");
       return QPLUG_FAILURE;
     }
   

     QdRScalarSeq_var masked_Tperp = QdRScalarSeq_var::narrow(joined[4]);

     string SIcTperp = QuGetAttrText("SI_conversion", ((* call_list)[5]));
     if (SIcTperp.empty())  {
       QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for T_parallel");
       Tperp_nest = (QdRScalarSeq_var) masked_Tperp;
     }
     else if (QarSameBaseSI(SIcTperp.c_str(), "1.0e6>K")){
       // base units the same, ensure conversion is too
       masked_Tperp->change_xref("SI_conversion", new QdString(SIcTperp.c_str()));
       Tperp_nest = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_Tperp, "1.0e6>K"));
     }
     else {
       QplugAppendTextDisplay("ERROR: cannot correct units of T_parallel");
       return QPLUG_FAILURE;
     }
   }

   if (is_empty(Tperp_nest)) alpha= FALSE;
   if (is_empty(Tpar_nest)) alpha = FALSE;
     
   nestsize = Tperp_nest->sequence_size();
 }
 printf("HT : Nestsize temp = %ld\n",nestsize);
 fflush(stdout);




 // ---- get normal if supplied ----------

 printf("HT : get normal if supplied \n");
 fflush(stdout);
 normal = QdRMatrix_var::narrow((*call_list)[6]);
 QdString_var n_string = QdString_var::narrow((*call_list)[6]);
 if (!normal.is_nil()) {
   printf("Normal supplied as QSAS object...\n");
   fflush(stdout);
 }
 else {
   if (!n_string.is_nil()) {
     normal = new QdRMatrix(Dim(3),0.00);
     //     normal = InputParseString(n_string,3);
      QString s = n_string->c_str();
      normal = str2Vec(s);

     printf("Normal supplied as string...[%g %g %g]\n",(double)(*normal)[0],(double)(*normal)[1],(double)(*normal)[2]);
   } else {
     normal = new QdRMatrix(Dim(3),0.00);
     printf("Normal not supplied...\n");
   }
   fflush(stdout);
}
 
 // --- declare various objects and call Vht function ----

 printf("HT : declare HT objects \n");
 fflush(stdout);
 QdRMatrixSeq_var Vht_x_B = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 QdRMatrixSeq_var V_x_B   = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 QdRMatrixSeq_var V_rest   = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 QdRMatrixSeq_var V_alf = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));  
 
 QdRMatrix_var K0c  = new QdRMatrix(RMatrix(Dim(3,3), 0.0));
 QdRMatrix_var K = new QdRMatrix(Dim(3,3), 0.00);
 QdRMatrix_var K1 = new QdRMatrix(Dim(3,3), 0.00);
 QdRMatrix_var K2 = new QdRMatrix(Dim(3,3), 0.00);
 QdRMatrix_var KV = new QdRMatrix(Dim(3), 0.00);
 QdRMatrix_var KVT = new QdRMatrix(Dim(3), 0.00);
 QdRMatrix_var V_ht0 = new QdRMatrix(Dim(3), 0.00);
 QdRMatrix_var a_ht  = new QdRMatrix(Dim(3), 0.00);

 QdRMatrix_var V_ht = V_HTcalc(nestsize,deltat,bfield_nest,velocity_nest,K0c,K,K1,K2,KV,KVT);



 // ----- Accelerating HT frame - see blue book sec 9.5 ----------
 
 printf("HT : accelerating HT frame \n");
 fflush(stdout);
 QdRMatrix_var K66c = new QdRMatrix(RMatrix(Dim(6,6), 0.00));


 for (i=0; i<3; i++) for (j=0; j<3; j++) {
   (*K66c)[i][j] = (*K0c)[i][j];
   (*K66c)[i][j+3] = (*K1)[i][j];
   (*K66c)[i+3][j] = (*K1)[i][j];
   (*K66c)[i+3][j+3] = (*K2)[i][j]; }

 //QdRMatrix_var Kinv66 = QdRMatrix_var::narrow(invert(K66c));
 QdRMatrix_var Kinv66 = new QdRMatrix(RMatrix(Dim(6,6), 0.00));

 
 for (i=0; i<3; i++) {
   (*V_ht0)[i] = (*Kinv66)[i][0]*(*KV)[0] + (*Kinv66)[i][1]*(*KV)[1] + (*Kinv66)[i][2]*(*KV)[2] + (*Kinv66)[i][3]*(*KVT)[0] + (*Kinv66)[i][4]*(*KVT)[1] + (*Kinv66)[i][5]*(*KVT)[2];
   (*a_ht)[i] = (*Kinv66)[i+3][0]*(*KV)[0] + (*Kinv66)[i+3][1]*(*KV)[1] + (*Kinv66)[i+3][2]*(*KV)[2] + (*Kinv66)[i+3][3]*(*KVT)[0] + (*Kinv66)[i+3][4]*(*KVT)[1] + (*Kinv66)[i+3][5]*(*KVT)[2];
 }


 // ----- calc the quantities dotted into the normal --------------

 printf("HT : calc n dot values \n");
 fflush(stdout);
 double V_ht_n=0.0, V_ht0_n=0.0, a_ht_n=0.0;

 V_ht_n   = (double)(*V_ht)[0] * (double)(*normal)[0]  + (double)(*V_ht)[1] * (double)(*normal)[1]  + (double)(*V_ht)[2] * (double)(*normal)[2];
 V_ht0_n  = (double)(*V_ht0)[0] * (double)(*normal)[0] + (double)(*V_ht0)[1] * (double)(*normal)[1] + (double)(*V_ht0)[2] * (double)(*normal)[2];
 a_ht_n   = (double)(*a_ht)[0] * (double)(*normal)[0]  + (double)(*a_ht)[1] * (double)(*normal)[1]  + (double)(*a_ht)[2] * (double)(*normal)[2];


 // ---- error analysis - see 9.4.1 in blue book -------------

 printf("HT : error analysis \n");
 fflush(stdout);
 QdRMatrix_var Suv = new QdRMatrix(Dim(3,3), 0.00);
 double sigma_n=0.00;
 double  dvht = 1.0E12 * minimizeD(nestsize, V_ht, velocity_nest, bfield_nest);  // 10^12 doe to nT^2 and km^/s^2 units 

 QdRMatrix_var K0invc = QdRMatrix_var::narrow(invert(K0c));

 for (i=0; i<3; i++) for (j=0; j<3; j++) (*Suv)[i][j] = (*K0invc)[i][j] *  dvht / (2.0*(double)nestsize - 3.0);
 for (i=0; i<3; i++) for (j=0; j<3; j++) sigma_n += (double)(*normal)[i] * (double)(*Suv)[i][j] * (double)(*normal)[j];


 // ---- send some output to QSAS window -------

 printf("HT : send outputs \n");
 fflush(stdout);
 QdTime t_start = (*tt_out)[0];
 QdTime t_stop = (*tt_out)[nestsize-1];
 QdTime t_center = t_start + (t_stop - t_start) / 2;

 QString Status;
 sprintf(txtbuf,"HofTel > t = %s to %s (%ld samples)\n",t_start.iso_srep().c_str()+11,t_stop.iso_srep().c_str()+11,nestsize);
 Status += txtbuf;
 sprintf(txtbuf,"HofTel > Param                 Value                   dot normal:\n");
 Status += txtbuf;
 sprintf(txtbuf,"HofTel > V_HT  = [%10.4f %10.4f %10.4f]  (%9.4f +/- %9.4f) km/s\n",(double)(*V_ht)[0], (double)(*V_ht)[1], (double)(*V_ht)[2], V_ht_n, sqrt(sigma_n));
 Status += txtbuf;
 sprintf(txtbuf,"HofTel > V_HT0 = [%10.4f %10.4f %10.4f]  (%9.4f) km/s\n",(double)(*V_ht0)[0], (double)(*V_ht0)[1], (double)(*V_ht0)[2], V_ht0_n);
 Status += txtbuf;
 sprintf(txtbuf,"HofTel > a_HT  = [%10.4f %10.4f %10.4f]  (%9.4f) km/s^2\n",(double)(*a_ht)[0], (double)(*a_ht)[1], (double)(*a_ht)[2], a_ht_n);
 Status += txtbuf;

 // ----- SEH 12 Jul 2004 : Want to compare V-Vht and V_magnetospheath ----------------

 if (!density.is_nil()) {
   long upper = 1+(long) nestsize/10;
   printf("HT : upper = %ld\n",upper);

   upper = (upper > 5) ? 5 : upper;
   printf("HT: upper = %ld\n",upper);

   double n0 = aveScalarOver(density_nest,0,upper);                          // ave density on one side
   double n1 = aveScalarOver(density_nest,nestsize-upper-1,nestsize-1);      // ave density on other side
   long start = (n0 > n1) ? 0 : nestsize-upper-1;
   long stop = start+upper;
   printf("n0, ni, upper = %f  %f  %ld - %ld  (%ld)\n",n0, n1, start,stop,nestsize);

   fflush(stdout);
   
   QdRMatrix_var Vsh = new QdRMatrix(Dim(3), 0.00);
   Vsh = aveVectorOver(velocity_nest,start,stop);
   for (i=0; i<3; i++) (*Vsh)[i] -= (*V_ht)[i];
   sprintf(txtbuf,"\nHofTel > Vsh  = [%10.4f %10.4f %10.4f]  km/s\n",(double)(*Vsh)[0], (double)(*Vsh)[1], (double)(*Vsh)[2]);
   Status += txtbuf;
   sprintf(txtbuf,"(Vsh = average of samples %ld - %ld - Vht)\n",start,stop);
   Status += txtbuf;
 }

 //  --- generate the two outputs Vht x B and V x B ---------------------


 double cc = 1.0e-03;          // conversion due to B in nT and V in km/s
 double mu_0 = 1.25664e-06;    // modified mu_0 (due to B in nT and dens in 1/ccm)
 double rho,m_p= 1.67e-27;     // rho = mass dens with proton_mass m_p
 double alfa = 0.00;           // for pressure anisotropy
 double kb = 1.38054e-23;      // boltzmann const
 FILE *f2;  
 //  ------ output also to the file /tmp/qsas_dht.txt ---

 QdTime timetag=(*tt_out)[0];

 if (LogFile) {
   QplugAppendTextDisplay("Log file written to /tmp/qsas_dht.txt");
   f2=fopen("/tmp/qsas_dht.txt","w"); 
   QDateTime today = QDateTime::currentDateTime();
   QString logTxt;
   logTxt += "#QSAS HofTel run : "+today.toString() + "\n#\n";
   fprintf(f2,"%s", logTxt.toLatin1().data());
   fprintf(f2,"#Time = %s\n",timetag.iso_srep().c_str());
   fprintf(f2,"#Vht = %10.4f  %10.4f  %10.4f \n", (double)(*V_ht)[0], (double)(*V_ht)[1], (double)(*V_ht)[2]);
   fprintf(f2,"#  Time                 V_Alfven [km/s]                     V_plasma [km/s]                     B_field [nT]             Dens [1/ccm]\n");
   fprintf(f2,"# \n");
 }
 
 for (i=0; i<nestsize; i++) {
   Vht_x_B[i][0] = cc * (bfield_nest[i][1] * (double)(*V_ht)[2]  - bfield_nest[i][2] * (double)(*V_ht)[1]);
   Vht_x_B[i][1] = cc * (bfield_nest[i][2] * (double)(*V_ht)[0]  - bfield_nest[i][0] * (double)(*V_ht)[2]);
   Vht_x_B[i][2] = cc * (bfield_nest[i][0] * (double)(*V_ht)[1]  - bfield_nest[i][1] * (double)(*V_ht)[0]);

   V_x_B[i][0] = cc * (bfield_nest[i][1] * velocity_nest[i][2]  -  bfield_nest[i][2] * velocity_nest[i][1]);
   V_x_B[i][1] = cc * (bfield_nest[i][2] * velocity_nest[i][0]  -  bfield_nest[i][0] * velocity_nest[i][2]);
   V_x_B[i][2] = cc * (bfield_nest[i][0] * velocity_nest[i][1]  -  bfield_nest[i][1] * velocity_nest[i][0]);

   for (j=0; j<3; j++) V_rest[i][j] =  velocity_nest[i][j] - (double)(*V_ht)[j];

   if (!density.is_nil()) {
     timetag = (*tt_out)[i];
     rho = (double)(density_nest[i]) * 1e06 * m_p;
     if (alpha) {
       bb = 1e-18*QSAS_vecsum2(bfield_nest,i);                                              // B^2 in [T]
       alfa = mu_0 * 1e12 * density_nest[i] * kb * (fabs(Tpar_nest[i]) - fabs(Tperp_nest[i])) / bb;     // 1e12 = cm^3 -> m^3 and MK -> K
       if (bb < 1e-31 || alfa > 1.0) alfa = 0.00;
       //       printf("bb = %g  %g   %g  %g..\n",bb, alfa, (double)(*Tpar_nest)[i],(double)(*Tperp_nest)[i]);
     }

     for (j=0; j<3; j++)  V_alf[i][j]  = sqrt(1.0-alfa) * 1e-03 * (double)(bfield_nest[i][j]) * 1e-09 / sqrt(mu_0 * rho);
     if (LogFile) fprintf(f2,"%s  %10.4f  %10.4f  %10.4f  %10.4f  %10.4f  %10.4f   %10.4f  %10.4f  %10.4f  %10.4f\n", 
			  timetag.iso_srep().c_str()+11,
                          (double)(*V_alf)[i][0],   (double)(*V_alf)[i][1],  (double)(*V_alf)[i][2],
			  (double)(velocity_nest)[i][0], (double)(velocity_nest)[i][1], (double)(velocity_nest)[i][2], 
			  (double)(bfield_nest)[i][0],   (double)(bfield_nest)[i][1],   (double)(bfield_nest)[i][2],
			  (double)(density_nest[i]) );
   }
 }   
 
 printf("HT : checking frame \n");
 fflush(stdout);
 string frame_in = QuGetAttrText("Frame", ((* call_list)[0]));
 if( frame_in.empty()) 
       frame_in = QuGetAttrText("Frame", ((* call_list)[1]));
 if( frame_in.empty()) 
       frame_in = "vector>???_xyz";


 QuSetTxtAttr("Frame",frame_in.c_str(),(QdObject_var) V_x_B);
 QuSetTxtAttr("UNITS","mV/m",(QdObject_var) V_x_B);
 QuSetTxtAttr("SI_conversion","0.001>V m^-1",(QdObject_var) V_x_B);
 QuSetTxtAttr("LABLAXIS","Ec",(QdObject_var) V_x_B);
 set_timetags((QdObject_var) V_x_B, tt_out);

 Vht_x_B->copy_xrefs_from(V_x_B);
 QuSetTxtAttr("LABLAXIS","E_Ht",(QdObject_var) Vht_x_B);



 // ------ Calculate correlation - se first eq on p.228 in blue book ------


 printf("HT : correlation coeff & slopw \n");
 fflush(stdout);
 QdRMatrix_var nullvec=new QdRMatrix(RMatrix(Dim(3), 0.00));
 double minD0 =  minimizeD(nestsize,nullvec,velocity_nest,bfield_nest);
 double minDVht = minimizeD(nestsize,V_ht,velocity_nest,bfield_nest);
 double Ccoeff = sqrt(1.0 - minDVht/minD0);
 sprintf(txtbuf,"\nCorrelation coeffecient (E_HT/E_c) =%10.6f  (after Blue Book, first Eq, p228)\n",Ccoeff);
 Status += txtbuf;

 Ccoeff = correlationCoeff(V_x_B,Vht_x_B,nestsize);
 sprintf(txtbuf,"Correlation coeffecient (E_HT/E_c) =%10.6f  (after classic correlation coeff)\n",Ccoeff);
 Status += txtbuf;


 double Cslope = slope(Vht_x_B,V_x_B,nestsize);
 sprintf(txtbuf,"Regression line slope (E_HT/E_C)   =%10.6f\n",Cslope);
 Status += txtbuf;
 if (!density.is_nil()) {
   Ccoeff = correlationCoeff(V_alf,V_rest,nestsize);
   sprintf(txtbuf,"\nCorrelation coeffecient (Va/Vrest) =%10.6f  (after classic correlation coeff)\n",Ccoeff);
   Status += txtbuf;
   Cslope = slope(V_alf,V_rest,nestsize);
   sprintf(txtbuf,"Regression line slope (Va/Vrest)   =%10.6f\n",Cslope);
   Status += txtbuf;
   if (LogFile) fprintf(f2,"# Slope = %f\n",Cslope);
   if (LogFile) fprintf(f2,"# Ccoeff = %f\n",Ccoeff);
 }
 QplugAppendTextDisplay(Status.toLatin1().data());




// --- SEH 13 mar 2007 --- add MVAB for log file
 if (LogFile) {
   QdRMatrix_var ProjMx  = new QdRMatrix(Dim(3,3), 0.0);
   QdRMatrix_var Evec = new QdRMatrix(Dim(3,3), 0.0);  
   QdRMatrix_var Evalues = new QdRMatrix(Dim(3), 0.0);
   QdRMatrix_var dummy=new QdRMatrix(RMatrix(Dim(3)));
   QdRMatrix_var maxBabs  = new QdRMatrix(Dim(3), 0.0);
   QdRMatrix_var maxBmax  = new QdRMatrix(Dim(3), 0.0);
   double cc=0.0, wl=0.0, Babs,BabsMax=-99.0,BabsMin=999.0, Bn=0.0,  Bmax,BmaxMax=0.0, normfac=(double)nestsize ;
   int constrained=0;
   
   QdRMatrix_var Q_MVAB = Q_MVAB_calc(bfield_nest, velocity_nest, dummy, V_alf, density_nest, &wl, &cc, nestsize, 0);
   try{
     GenericMinVar minvar(constrained,*Q_MVAB,*ProjMx);  
     *Evec = minvar.eigenvectors().duplicate(SHALLOW);
     *Evalues = minvar.eigenvalues().duplicate(SHALLOW); 
   }
   catch(Exception& e) 
     {return QPLUG_FAILURE; }

   for (int i=0; i<3; i++){     
     fprintf(f2,"# LMN_%d %12.6f  [%10.6f, %10.6f, %10.6f]\n", i, 
	    (double)(*Evalues)[i],
	    (double)(*Evec)[i][0],
	    (double)(*Evec)[i][1], 
	    (double)(*Evec)[i][2]);
   }

   QdRMatrix_var aveB = aveVectorOver(bfield_nest, 0,nestsize);
   for (k=0; k<nestsize; k++) {              // find <B>
     Babs = sqrt(QSAS_vecsum2(bfield_nest,k));
     if (Babs > BabsMax) BabsMax = Babs;                              // find largest |B|
     if (Babs < BabsMin) BabsMin = Babs;                              // find smalles |B|
   }

   for (j=0; j<3; j++) {
     Bn += (*aveB)[j] * (double)(*Evec)[0][j];
   }
   fprintf(f2,"# <B>   = [%8.3f  %8.3f  %8.3f]\n",(float)(*aveB)[0],(float)(*aveB)[1],(float)(*aveB)[2]);
   fprintf(f2,"# AveBn = %f\n",Bn);   
   fprintf(f2,"# Bjump = %f\n",BabsMax-BabsMin);
   fprintf(f2,"# Bmax  = %f\n",BabsMax); 

 }
 // --- eo SEH 13 Mar 2007 ---


 if (LogFile) fclose(f2);


 // ----- Push V_conv x B and V_ht x B back on WL --------------

 printf("HT : push back \n");
 call_list->push_back((QdObject_var) Vht_x_B);
 call_list->push_back((QdObject_var) V_x_B);
 if (!density.is_nil()) {
    QuSetTxtAttr("Frame",frame_in.c_str(),(QdObject_var) V_rest);
    QuSetTxtAttr("UNITS","km/s",(QdObject_var) V_rest);
    QuSetTxtAttr("SI_conversion","1.0e3 > m s^-1",(QdObject_var) V_rest);
    set_timetags((QdObject_var) V_rest, tt_out);
    call_list->push_back((QdObject_var) V_rest);

    V_alf->copy_xrefs_from(V_rest);
    QuSetTxtAttr("LABLAXIS","V_Alfven",(QdObject_var) V_alf);
    set_timetags((QdObject_var) V_alf, tt_out);
    call_list->push_back((QdObject_var) V_alf);
 }
 else{
   // be safe and put null object on argument list
   // this would allow these possibly unused outputs 
   // to not be at the end of output list
   
    QdObject_var empty1;
    QdObject_var empty2;
    call_list->push_back(empty1);
    call_list->push_back(empty2);
 } 

 
  qDebug( "Time elapsed : %d\n", t.elapsed() ); // prints the number of msecs elapsed

 return QPLUG_SUCCESS;
}






double minimizeD(int M, QdRMatrix_var V, QdRMatrixSeq_var vm, QdRMatrixSeq_var Bm)
// -----------------------------------------------------------------------------------------
/*!
  \brief Calculate the least square residue between two vectors
  \return the least square residue
  \param M - number of elements in the input vectors
  \param V - vector time series 
  \param vm - vector time series 
  \param Bm - vector time series 

  Calculates the residual from Eq 9.8 in blue ISSI book

*/
//
// SEH, Nov, 2001
//      May, 2002
//      May, 2004,  using qdos classes rather than own VEC structures
//
// V,vm [km/s], B [nT]
//
// -----------------------------------------------------------------------------------------
{
  int i;
  double E_abs, Dsum=0.0;
  QdRMatrix_var Em = new QdRMatrix(RMatrix(Dim(3),0));

  for (i=0; i<M; i++) {
    (*Em)[0] = (((double)vm[i][1] - (double)(*V)[1])*(double)Bm[i][2] - ((double)vm[i][2] - (double)(*V)[2])*(double)Bm[i][1]);
    (*Em)[1] = (((double)vm[i][0] - (double)(*V)[0])*(double)Bm[i][2] - ((double)vm[i][2] - (double)(*V)[2])*(double)Bm[i][0]);
    (*Em)[2] = (((double)vm[i][1] - (double)(*V)[1])*(double)Bm[i][0] - ((double)vm[i][0] - (double)(*V)[0])*(double)Bm[i][1]);
    E_abs = 1.0E-12 * QSAS_vecsum2(Em);
    Dsum = Dsum + E_abs;
  }
  return (Dsum / (double) M);   // in units of (V/m)^2
}





QdRMatrix_var V_HTcalc(long nestsize, double deltat, QdRMatrixSeq_var B, QdRMatrixSeq_var V,
	     QdRMatrix_var K0, QdRMatrix_var K, QdRMatrix_var K1, QdRMatrix_var K2, QdRMatrix_var KV, QdRMatrix_var KVT)
// ---------------------------------------------------------------------------
/*!
  \brief calculates the deHoffmann-Teller velcoity
  \return a vector with the deHoffmann-Teller velocity in the same frame as the input data
  \param nestsize - number of elements in the B-field and Velcity vectors
  \param deltat - sampling time - used for the accellerated HT frame
  \param B - vector time series with magnetic field
  \param V - vector time series with plasma velocity
  \param K0 - 3x3 projection matrix - see Blue ISSI book, p 226
  \param K - 3x3 projection matrix - see Blue ISSI book, p 226
  \param K1 - 3x3 projection matrix - see Blue ISSI book, p 240
  \param K2 - 3x3 projection matrix - see Blue ISSI book, p 240
  \param KV - 6x6 projection matrix - see Blue ISSI book, p 240
  \param KVT - 6x6 projection matrix - see Blue ISSI book, p 240

  This routine calculates a deHoffmann-Teller velocity and fills in the various
  projection matrices as given by the ISSI book on Multi-spacecraft analysis
  (ISSI SR-001)

*/
//
// SEH, Nov, 2001
//      May, 2002
//      Dec, 2003 Modified for Qsas 2. More to be done (use more qdos classes)
//      May, 2004 use more qdos classes and methods, e.g. invert
//      May, 2004 Doxygen docs
//
// -----------------------------------------------------------------------------------------
{
  int i,j,k,l;
  double Q, bb;
  double time=0.0;

  QdRMatrix_var V_ht = new QdRMatrix(RMatrix(Dim(3), 0.00));
  QdRMatrix_var B_k = new QdRMatrix(RMatrix(Dim(3)));


  for(k=0; k<nestsize; k++)  {         
    for (j=0; j<3; j++) {
      (*B_k)[j] = (*B)[k][j];
    }
    bb = QSAS_vecsum2(B_k);

    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	(*K)[i][j] = (double)(i==j) * bb - ((double)(*B_k)[i]*(double)(*B_k)[j]);
      }
    }
     
     for (i=0; i<3; i++) {
       for (j=0; j<3; j++) {
	 (*K0)[i][j] += (*K)[i][j];
         (*K1)[i][j] += (*K)[i][j] * time;
	 (*K2)[i][j] += (*K)[i][j] * time * time;
	 (*KV)[i]    += (*K)[i][j] * (*V)[k][j];
         (*KVT)[i]   += (*K)[i][j] * (*V)[k][j] *time;           // see 9.26 in blue book
       }
     }
     time +=deltat;  
  }

  (*KV)  /= (double) nestsize;
  (*KVT) /= (double) nestsize;
  (*K0)  /= (double) nestsize;                                   // K0 == <K>
  (*K1)  /= (double) nestsize;
  (*K2)  /= (double) nestsize;
    
  QdRMatrix_var Kinv = QdRMatrix_var::narrow(invert(K0));

  for (i=0; i<3; i++) {
    (*V_ht)[i] = (*Kinv)[i][0]*(*KV)[0] + (*Kinv)[i][1]*(*KV)[1] + (*Kinv)[i][2]*(*KV)[2]; 
  }

  return (V_ht);
}



#ifndef WINDOWS
double slope(QdRMatrixSeq_var X_in, QdRMatrixSeq_var Y_in,int n)
// ---------------------------------------------------------------------------
/*!
  \brief calculates the regression line 
  \return a double with the slope, i.e., the coeff b in y(x) = a + bx
  \param X_in - 3 component vector with time series 1
  \param Y_in - 3 component vector with time series 2
  \param n - number of elements in the two vectors
*/
//
// SEH, 15 Aug, 2002
//
// ---------------------------------------------------------------------------

{
  int numpts = 3*n;

  ScalarSeq<double> Xc;   Xc.resize(numpts);
  ScalarSeq<double> Yc;   Yc.resize(numpts);
  QdRScalarSeq_var Tc = new QdRScalarSeq(numpts);

  for (int i=0; i<n; i++) {
    Xc[i] = (double) (*X_in)[i][0];
    Xc[i+n] = (double) (*X_in)[i][1];
    Xc[i+2*n] = (double) (*X_in)[i][2];
    Yc[i] = (double) (*Y_in)[i][0];
    Yc[i+n] = (double) (*Y_in)[i][1];
    Yc[i+2*n] = (double) (*Y_in)[i][2];
  }
  
  double sum_X = Xc.sum();
  double sum_Y = Yc.sum();
  double ave_X = sum_X / (double) (numpts);
  double ave_Y = sum_Y / (double) (numpts);
  double b=0.0, tt=0.0;

  for (int i=0; i<numpts; i++) {
    Tc[i] = (double) Xc[i] - ave_X;
    b +=(double) Tc[i] * (double)Yc[i];
    tt += (double) (*Tc)[i] * (double) (*Tc)[i];
  }

  b /= tt;
  return(b);
}


double correlationCoeff(QdRMatrixSeq_var X_in, QdRMatrixSeq_var Y_in, int n)
// ---------------------------------------------------------------------------
/*!
  \brief calculates the correlation coefficient between two 3 element vectors
  \return a double with correlation coefficient
  \param X_in - 3 component vector with time series 1
  \param Y_in - 3 component vector with time series 2
  \param n - number of elements in the two vectors
*/
//
// SEH, 15 Jan 2003 
//
// ---------------------------------------------------------------------------
{

  double sum_upper=0.0, sum_xx=0.0, sum_yy=0.0;
  int numpts = 3*n;

  ScalarSeq<double> Xc;   Xc.resize(numpts);
  ScalarSeq<double> Yc;   Yc.resize(numpts);
  QdRScalarSeq_var Tc = new QdRScalarSeq(numpts);

  for (int i=0; i<n; i++) {
    Xc[i] = (double) (*X_in)[i][0];
    Xc[i+n] = (double) (*X_in)[i][1];
    Xc[i+2*n] = (double) (*X_in)[i][2];
    Yc[i] = (double) (*Y_in)[i][0];
    Yc[i+n] = (double) (*Y_in)[i][1];
    Yc[i+2*n] = (double) (*Y_in)[i][2];
  }

  double sum_x = Xc.sum() / (double) (numpts);
  double sum_y = Yc.sum() / (double) (numpts);

  printf("In correlationcoeff : sum_x, sum_y = %f %f\n",sum_x, sum_y);
  fflush(stdout);

  for (int i=0; i<numpts; i++) {
    sum_upper += (Xc[i] - sum_x) * (Yc[i] - sum_y);
    sum_xx += (Xc[i] - sum_x)*(Xc[i] - sum_x);
    sum_yy += (Yc[i] - sum_y)*(Yc[i] - sum_y);
  }

  return (double) (sum_upper / (sqrt(sum_xx) * sqrt(sum_yy)));
}
#endif




double QSAS_vecsum2(QdRMatrix_var Vector)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the length of a vector
  \return the norm^2 of a vector
  \param Vector - vector 

*/
//
// SEH, 05 May 2004
//
// ---------------------------------------------------------------------------
{
  double sum=0.0;

  for (int i=0; i<3; i++) sum += (*Vector)[i] * (*Vector)[i];

  return(sum);
}



double QSAS_vecsum2(QdRMatrixSeq_var VectorSeq, long index)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the length of a vector
  \return the norm^2 of a vector TS
  \param VectorSeq - vector series 
  \param index - select element in vector series to calculate norm^2
  
  Overloaded of below

*/
//
// SEH, 05 May 2004
//
// ---------------------------------------------------------------------------
{
  double sum=0.0;

  for (int i=0; i<3; i++) sum += (*VectorSeq)[index][i] * (*VectorSeq)[index][i];

  return(sum);
}


QdRMatrix_var aveVectorOver(QdRMatrixSeq_var VectorSeq, long start, long stop)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the average of a vector between two indices
  \return a Vector with the average
  \param VectorSeq - vector series 
  \param start, stop - indices to average over
*/
//
// SEH, 05 May 2004
//
// ---------------------------------------------------------------------------
{

  QdRMatrix_var v = new QdRMatrix(RMatrix(Dim(3), 0.00));;

  for (int i=start; i<stop; i++) 
    for (int j=0; j<3; j++) 
      (*v)[j] += (*VectorSeq)[i][j];

  (*v) /= (double) abs(stop - start);

  return(v);
}



double aveScalarOver(QdRScalarSeq_var ScalarSeq, long start, long stop)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the average of a scalar between two indices
  \return a Scalar with the average
  \param ScalarSeq - vector series 
  \param start, stop - indices to average over

  use overloading ?
*/
//
// SEH, 12 Jul 2004
//
// ---------------------------------------------------------------------------
{

  double s = 0;
  for (int i=start; i<stop; i++) 
    s += (*ScalarSeq)[i];

  s /= (double) abs(stop - start);

  return(s);
}



// -------------------------------------------------------------------
/*!
  \brief Parses the a text and converts text to a Vector
  \return  QSAS matrix with the parsed values of the string
  \param s - a string on the form '[2.00 -23.0  9.0]'
*/
//
// SEH, Mar 27, 2003
//
// -------------------------------------------------------------------
QdRMatrix_var str2Vec(QString s)
{
    QdRMatrix_var v = new QdRMatrix(RMatrix(Dim(3)));;
    QString str=s.simplified();
    str=str.replace(QChar('['),"").replace(QChar(']'),"").replace(QChar(',')," ");
    str=str.simplified();
    (*v)[0] = str.section(' ',0,0).toDouble();
    (*v)[1] = str.section(' ',1,1).toDouble();
    (*v)[2] = str.section(' ',2,2).toDouble();
    return(v);
}


int main()
{
  return(0);
}
