/****************************************************************************
							LICENSE

  Copyright (C) 2010 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file NablaEstimation.cpp
 * \author: eugenatjacobs
 *
 * \brief Implementation of ClusterAnalysis class (the gradient estimate part)
 *
 *
 */

#include "qplug_if.h"
#include "qdutil.h"
#include "qar.h"
#include "Qdos.h"
#include "ClusterCommons.h"
#include "ConstrainedMinVariance.h"

using namespace std;

namespace cluster {

//------- class CEstimateNormal -----------------------------
ClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::CEstimateNormalComponentFor3SpacecraftMethod(const unsigned short option)
{
	CEstimateNormalComponentFor3SpacecraftMethod();
	set_option(option);
}

void ClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::set_option(const unsigned short option)
{
	switch (option) {
		case 1: option_ = kGradientParallel;break;
		case 2: option_ = kGradientPerpendicular;break;
		case 3: option_ = kStationarity;break;
		default: option_ = kNone;
	}
}

int ClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::computeSlopeOfTimeSeries(QdRMatrixSeq_var &g)
{
	//If it is not stationarity assumption, we still don't have any error.
	if ((option_ == kStationarity) && (slope_.is_nil())) {
		int nestsize = g->sequence_size();
		unsigned short nr_spacecrafts = (unsigned short)((g[0]).dims())[0];
		unsigned short nr_dimensions = (unsigned short)((g[0]).dims())[1];

		//Compute g_mc (the field in the spacecraft constellation mass center):
		QdRMatrixSeq_var V_mc = new QdRMatrixSeq(nestsize, RMatrix(Dim(nr_dimensions), 0.0));
		slope_ = new QdRMatrixSeq(nestsize, RMatrix(Dim(nr_dimensions), 0.0));
		for (unsigned short coord = 1; coord < nr_dimensions; coord++) {
			for (unsigned int t = 0; t < nestsize; t++) {
				(*V_mc)[t][coord] = 0;
				for (unsigned int satellite = 0; satellite < 3; satellite++)
					(*V_mc)[t][coord] += (double) (*g)[t][satellite][coord];
				(*V_mc)[t][coord] /= 3;
			}

			//Linear fit:
			const unsigned int linear_fit_window = 15;
			unsigned int half_fit_window = linear_fit_window / 2;
			double time_window = linear_fit_window * 4, half_time_window = time_window / 2; //4 seconds

			double s_t = 0, s_t2 = 0, s_g = 0, s_tg = 0, current_time;
			QdTimeSeq_var time_tags = get_timetags(g);
			int t_back = 0, t_current = 0, t_front = 0; //intems from t_back to t_front will be used for slope computation.
			for (t_current = 0; t_current < nestsize; t_current++) {
				while ((time_tags[t_front].getDiffSecs(time_tags[t_current]) < half_time_window) && (t_front+1 < nestsize)) {
					current_time = time_tags[t_front].getDiffSecs(time_tags[0]); //everything is done relatively to the first
					s_t += current_time;
					s_t2 += current_time * current_time;
					s_g += (double) (*V_mc)[t_front][coord];
					s_tg += current_time * (double) (*V_mc)[t_front][coord];
					t_front++;
				}
				while (time_tags[t_front].getDiffSecs(time_tags[t_back]) > time_window) {
					current_time = time_tags[t_back].getDiffSecs(time_tags[0]); //everything is done relatively to the first
					s_t -= current_time;
					s_t2 -= current_time * current_time;
					s_g -= (double) (*V_mc)[t_back][coord];
					s_tg -= current_time * (double) (*V_mc)[t_back][coord];
					t_back++;
				}
				(*slope_)[t_current][coord] = ((t_front - t_back + 1) * s_tg - s_g * s_t) /
											((t_front - t_back + 1) * s_t2 - s_t * s_t);
			}
		}
		slope_->copy_xrefs_from(g);
	}
	return 0;
}

double ClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::get_slope(const int t, const unsigned long coord) const
{
	if (slope_.is_nil()) return 0.0;
	return (double)((*slope_)[t][coord]);
}

// -------- ClusterAnalysis ----------------------------

Status::ReturnStatus ClusterAnalysis::Compute4PointsEstimate()
{
	Status::ReturnStatus result = Status::kOk;
	if (nr_satellites_ != 4) {
		status.set_message("This function is for 4 satellites only.");
		result = Status::kWrongInput;
		return result;
	}
	else if (normal_TS_.is_nil() || vol_TS_.is_nil()) {
		status.set_message("The Normal and the Volume are required but at least one of them is null.");
		result = Status::kWrongInput;
		return result;
	}

	QdRMatrix_var K = new QdRMatrix(Dim(nr_satellites_, 3), 0.0); // Reciprocalvector - see Blue book, sec 12.4

	QdRMatrix_var Vs = new QdRMatrix(Dim(nr_satellites_, nr_coordinates_), 0.0);
	QdRMatrix_var GradMagn = NULL;
	QdRMatrix_var Grad = NULL;

	grad_magn_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3), 0.0)); //if the value is scalar => it is equal to its magnitude.
	if (nr_coordinates_ == 1) {
		//For a scalar field, we don't have any curls, divergents,
		grad_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3), 0.0));
		Grad = new QdRMatrix(Dim(3), 0.0);
	}
	else {
		grad_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3, 3), 0.0));
		GradMagn = new QdRMatrix(Dim(3), 0.0);
		Grad = new QdRMatrix(Dim(3, 3), 0.0);
	}

	if (result == Status::kOk)
		Status::PrintMessage(false, "NablaEstimator: DEBUG: break 2");


	int t;
//#pragma omp parallel for private(t)
	for (t = 0; t < nr_measurements_; t++) {
		//1. Make sure the variables are initialized:

		//2. Compute reciprocal vectors (a function of spacecraft positions only).

		// The following lines compute the so-called volumetric tensor. I leave it here for future applications.
		//		QdRMatrix_var R = new QdRMatrix(Dim(3, 3), 0.0); // Volumetric tensor - see Blue book sec 12.2.1 + 14.2.1
		//		for (short i = 0; i < 3; i++)
		//			for (short j = 0; j < 3; j++) {
		//				for (unsigned short sc = 0; sc < nr_satellites_; sc++)
		//					(*R)[i][j] += (double) (*position_TS_)[t][sc][i] * (double) (*position_TS_)[t][sc][j];
		//			}

		(*vol_TS_)[t] = ComputeReciprocal4Spacecrafts(t, K); //the output is the total volume of the tetrahedron.
		for (unsigned short sc = 0; sc < nr_satellites_; sc++) {
			for (unsigned short coord = 0; coord < nr_coordinates_; coord++) {
				(*Vs)[sc][coord] = (*field_TS_)[t][sc][coord];
			}
		}

		if (nr_coordinates_ == 1) {
			computeGradScalar(Vs, K, Grad);
			for (short i = 0; i < 3; i++) {
				(*grad_magn_TS_)[t][i] = (double) (*Grad)[i];
				(*grad_TS_)[t][i] = (double) (*Grad)[i];
			}
		}
		else {
			computeGradMagnitude(Vs, K, GradMagn);
			computeGradVec(Vs, K, Grad);
			for (short i = 0; i < 3; i++) {
				(*grad_magn_TS_)[t][i] = (double) (*GradMagn)[i];
				for (short j = 0; j < 3; j++)
					(*grad_TS_)[t][i][j] = (double) (*Grad)[i][j];
			}
		}
	}
	if (result == Status::kOk)
		Status::PrintMessage(false, "NablaEstimator: DEBUG: break 3");

	return result;
}

Status::ReturnStatus ClusterAnalysis::Compute3PointsEstimate()
{
	Status::ReturnStatus result = Status::kOk;
	if (nr_satellites_ != 3) {
		status.set_message("This function is for 3 satellites only.");
		result = Status::kWrongInput;
		return result;
	}
	else if (normal_TS_.is_nil() || vol_TS_.is_nil()) {
		status.set_message("The Normal and the Volume are required but at least one of them is null.");
		result = Status::kWrongInput;
		return result;
	}

	QdRMatrix_var K = new QdRMatrix(Dim(nr_satellites_,3), 0.0);  // Reciprocals - see Vogt et al. 2009, formula (12)
	QdRMatrix_var Normal = new QdRMatrix(Dim(kVectorRank), 0.0);  //Normal to the plane of spacecrafts.

	QdRMatrix_var Vs = new QdRMatrix(Dim(nr_satellites_, nr_coordinates_), 0.0);
	QdRMatrix_var Grad = NULL;

	if (nr_coordinates_ == 1){
		grad_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3), 0.0));
		Grad = new QdRMatrix(Dim(3), 0.0);
	}
	else{
		grad_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3,3), 0.0));
		Grad = new QdRMatrix(Dim(3,3), 0.0);
	}

	//temporary variables:
	QdRMatrix_var qtemp = new QdRMatrix(Dim(3), 0.0);
	QdRScalar dgdn = 0.0;

	estimator.computeSlopeOfTimeSeries(field_TS_);

	for (int t=0; t<nr_measurements_; t++) {
		//1. Make sure the variables are initialized:

		//2. Compute reciprocal vectors (a function of spacecraft positions only).
		for (short i=0; i<3; i++) (*Normal)[i] = (*normal_TS_)[t][i];
		//the output is the total volume of the tetrahedron:
		computeReciprocal3Spacecrafts(t, Normal, static_cast<double>((*vol_TS_)[t]), K);

		for (unsigned short sc=0; sc<nr_satellites_; sc++)
			for (unsigned short i=0; i<nr_coordinates_; i++)
				(*Vs)[sc][i] = static_cast<double>((*field_TS_)[t][sc][i]);

		dgdn = 0.0;
		if (nr_coordinates_ == 1){
			computeGradScalar(Vs, K, Grad);
			computeEstimateDgDn(Grad, Normal, estimator, t, 0, dgdn);
			for (short i=0; i<3; i++)
			{
				(*Grad)[i] += (*Normal)[i] * dgdn;
				(*grad_TS_)[t][i] = static_cast<double>((*Grad)[i]);
			}
		}
		else{
			computeGradVec(Vs,K, Grad);
			for (unsigned short coord = 0; coord<3; coord++)
			{
				for (unsigned short jj = 0; jj<3; jj++) (*qtemp)[jj] = (double)(*Grad)[coord][jj];
				computeEstimateDgDn(qtemp, Normal, estimator, t, coord, dgdn);
				for (unsigned short jj = 0; jj<3; jj++) (*Grad)[coord][jj] += (double)(*Normal)[jj] * dgdn;
			}

			for (short i=0; i<3; i++)
				for (short j=0; j<3; j++)
					(*grad_TS_)[t][i][j] = static_cast<double>((*Grad)[i][j]);
		}
	}
	return result;
}

Status::ReturnStatus ClusterAnalysis::ComputeGradientEstimate(signed short nr_spacecrafts /*= -1*/)
{
	Status::ReturnStatus result = Status::kOk;
	if (field_TS_.is_nil() || position_TS_.is_nil()) {
		status.set_message("No input for the gradient estimate.");
		result = Status::kNoInput;
	}
	else {
		nr_measurements_ = field_TS_->sequence_size();
		if (nr_measurements_ < 1) {
			status.set_message("No input for the gradient estimate.");
			result = Status::kNoInput;
		}
		else if (nr_measurements_ != position_TS_->sequence_size()) {
			status.set_message("Position array and measurements array sizes do not match.");
			result = Status::kWrongInput;
		}
		else {
			nr_satellites_ = static_cast<unsigned short>(((field_TS_[0]).dims())[0]);
			nr_coordinates_ = static_cast<unsigned short>(((field_TS_[0]).dims())[1]);
			if ((nr_coordinates_ != 1) && (nr_coordinates_ != 3)) {
				status.set_message("Only scalars or 3 dimensional vectors are accepted.");
				result = Status::kWrongInput;
			}
			else if (nr_satellites_ < 3) {
				status.set_message("Data from at least 3 satellites are required.");
				result = Status::kWrongInput;
			}
			else if (nr_spacecrafts == -1) {
				nr_satellites_ = std::min((unsigned short)4, nr_satellites_);
			}
			else if ((nr_spacecrafts != 3) && (nr_spacecrafts != 4)) {
				status.set_message("Currently only 3 or 4 satellites methods are supported.");
				result = Status::kWrongInput;
			}
			else {
				nr_satellites_ = nr_spacecrafts;
			}
		}
	}
	if (result == Status::kOk) {
		grad_magn_TS_ = NULL;
		grad_TS_ = NULL;
		curl_TS_ = NULL;
		div_TS_ = NULL;
		vol_TS_ = NULL;
		bary_vector_TS_ = NULL;
		bary_scalar_TS_ = NULL;
		normal_TS_ = NULL;
		curl_TS_MVA_ = NULL;
		bary_TS_MVA_ = NULL;
		velo_TS_MVA_ = NULL;
		rotation_matrix_ = NULL;
		computeNormalToSurfaceOfFirst3Sat(nr_measurements_, vol_TS_, normal_TS_); //valid for all methods with at least 3 satellites.
		result = (nr_satellites_ == 3)? Compute3PointsEstimate() : Compute4PointsEstimate();
	}
	return result;
}

Status::ReturnStatus ClusterAnalysis::ComputeCurlAndDiv(double curl_scale_factor /*= 1*/)
{
	Status::ReturnStatus result = Status::kOk;
	if (grad_TS_.is_nil()) {
		result = Status::kNoInput;
	}
	else if (nr_coordinates_ != 3) {
		result = Status::kVectorFieldRequired;
	}
	else {
		div_TS_ = new QdRScalarSeq(nr_measurements_);
		curl_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(kVectorRank), 0.0));
		int c[] = { 0, 1, 2, 0, 1 };
		for (int t = 0; t < nr_measurements_; t++) {
			for (unsigned short coord = 0; coord < kVectorRank; coord++) {
				//taken from the IDL code of atmuso:
				(*div_TS_)[t] += static_cast<double>((*grad_TS_)[t][coord][coord]);
				(*curl_TS_)[t][c[coord]] = (static_cast<double>((*grad_TS_)[t][c[coord + 2]][c[coord + 1]])
						- static_cast<double>((*grad_TS_)[t][c[coord + 1]][c[coord + 2]])) / curl_scale_factor;
			}
		}
	}
	return result;
}

Status::ReturnStatus ClusterAnalysis::ComputeBarycenter()
{
	Status::ReturnStatus result = Status::kOk;

	//cout << "Hello" << endl;
	//Compute Barycenter
	if (!field_TS_.is_nil()) {

		bary_coordinates_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(kVectorRank), 0.0));
		for (int t = 0; t < nr_measurements_; t++)
			for (unsigned short satellite = 0; satellite < nr_satellites_; satellite++)
			{
				for (unsigned short coord = 0; coord < kVectorRank; coord++)
					(*bary_coordinates_TS_)[t][coord] += static_cast<double>((*position_TS_)[t][satellite][coord]) / nr_satellites_;
			}
		if (nr_coordinates_ == kVectorRank) {
			bary_vector_TS_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(nr_coordinates_), 0.0));
			for (int t = 0; t < nr_measurements_; t++)
				for (unsigned short satellite = 0; satellite < nr_satellites_; satellite++)
				{
					for (unsigned short coord = 0; coord < nr_coordinates_; coord++)
						(*bary_vector_TS_)[t][coord] += static_cast<double>((*field_TS_)[t][satellite][coord]) / nr_satellites_;
				}
		}
		else if (nr_coordinates_ == kScalarRank) {
			bary_scalar_TS_ = new QdRScalarSeq(nr_measurements_, 0.0);
			for (int t = 0; t < nr_measurements_; t++)
				for (unsigned short satellite = 0; satellite < nr_satellites_; satellite++)
				{
					for (unsigned short coord = 0; coord < nr_coordinates_; coord++)
						(*bary_scalar_TS_)[t] += static_cast<double>((*field_TS_)[t][satellite][coord]) / nr_satellites_;
				}
		}

	}
	else {
		result = Status::kNoInput;
	}
	return result;
}

double NormalizeToTen(QdRMatrixSeq_var &matr_TS, QdRMatrixSeq_var &destination)
{
	double result = 1.0;
	if (!matr_TS.is_nil()) {
		int nr_measurements = matr_TS->sequence_size();
		if (nr_measurements > 0) {
			unsigned short nr_coordinates = static_cast<unsigned short>(((matr_TS[0]).dims())[0]);
			destination = new QdRMatrixSeq(nr_measurements, RMatrix(Dim(3), 0.0));
			//1. Determine the maximum
			double maxim = static_cast<double>((*matr_TS)[0][0]);
			for (int t = 0; t< nr_measurements; ++t)
				for (unsigned short coord = 0; coord < nr_coordinates; ++coord)
					maxim = max(static_cast<double>((*matr_TS)[t][coord]), maxim);

			//2.Normalize
			result = maxim / 10;
			for (int t = 0; t< nr_measurements; ++t)
				for (unsigned short coord = 0; coord < nr_coordinates; ++coord)
					(*destination)[t][coord] = static_cast<double>((*matr_TS)[t][coord]) / result;
		}
	}
	return result;
}

Status::ReturnStatus ClusterAnalysis::Postprocess()
{
	Status::ReturnStatus result = Status::kOk;
	if (grad_TS_.is_nil() || curl_TS_.is_nil()) {
		result = Status::kNoInput;
	}
	else if (nr_coordinates_ == kVectorRank) {
		unsigned int i;
		Status::PrintMessage(true, "Computing MinVar of the curl\n");
		//  ----- do a MinVar of Curl Vector vector -------
		rotation_matrix_ = new QdRMatrix(Dim(3, 3), 0.0);
		QdRMatrix_var Eigenvalues = new QdRMatrix(Dim(3), 0.0);
		char evalsevecs[80];
		QString str("Eigenvalues, eigenvectors of MVAJ are :\n");
		const int constrained = 0;

//		printf("curl V_0 = %f  %f  %f\n", (double)(*curl_TS_)[0][0], (double)(*curl_TS_)[0][1], (double)(*curl_TS_)[0][2]);
//		printf("curl V_1 = %f  %f  %f\n", (double)(*curl_TS_)[1][0], (double)(*curl_TS_)[1][1], (double)(*curl_TS_)[1][2]);
//		fflush(stdout);

		QdRMatrixSeq_var temp_curl = NULL;
		double normal_const = NormalizeToTen(curl_TS_, temp_curl);
		normal_const = normal_const * normal_const;
		try {
			ConstrainedMinVariance minvar(*temp_curl, *temp_curl, constrained); // qdos minvariance call
			*rotation_matrix_ = minvar.eigenvectors().duplicate(SHALLOW);
			*Eigenvalues = minvar.eigenvalues().duplicate(SHALLOW);
		}
		catch (Exception& e) {
			e.print_msg();
			//return QPLUG_FAILURE;
			return Status::kUnknownError;
		}
		if ((double) (*rotation_matrix_)[0][0] < 0)
			(*rotation_matrix_) *= -1.0;

		for (i = 0; i < 3; i++) {
			printf("%10.5g  [%10.6g, %10.6g, %10.6g ]\n", (double) (*Eigenvalues)[i] * normal_const,
					(double) (*rotation_matrix_)[i][0], (double) (*rotation_matrix_)[i][1],
					(double) (*rotation_matrix_)[i][2]);
			sprintf(evalsevecs, "%10.6g  [%10.6g, %10.6g, %10.6g ]\n", (double) (*Eigenvalues)[i] * normal_const,
					(double) (*rotation_matrix_)[i][0], (double) (*rotation_matrix_)[i][1],
					(double) (*rotation_matrix_)[i][2]);
			str += evalsevecs;
		} // end for
		fflush(stdout);
		// ----- rotate J into the MinVar system -------


		double mu0 = 1.25664e-6;
		curl_TS_MVA_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3), 0.0));
		bary_TS_MVA_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3), 0.0));
		velo_TS_MVA_ = new QdRMatrixSeq(nr_measurements_, RMatrix(Dim(3), 0.0));
		if (!bary_vector_TS_.is_nil()) {
			for (int k = 0; k < nr_measurements_; k++) {
				curl_TS_MVA_[k] = (*rotation_matrix_) * curl_TS_[k];
				bary_TS_MVA_[k] = (*rotation_matrix_) * bary_vector_TS_[k];
			}
		}

		QdRMatrix_var Jint = new QdRMatrix(Dim(3), 0.0);
		QdRMatrix_var Jave = new QdRMatrix(Dim(3), 0.0);
		QdRMatrix_var Jint_GSE = new QdRMatrix(Dim(3), 0.0);
		QdRMatrix_var Jave_GSE = new QdRMatrix(Dim(3), 0.0);

		QdRMatrix_var Bint = new QdRMatrix(Dim(3), 0.0);
		QdRMatrix_var v_n = new QdRMatrix(Dim(3), 0.0);
		QdRMatrix_var deltaB = new QdRMatrix(Dim(3), 0.0);
		double nf = (double) nr_measurements_;

		char l_txtbuf[150] = "";
		for (i = 0; i < 3; i++) {
			(*deltaB)[i] = (double) (*bary_TS_MVA_)[nr_measurements_ - 1][i] - (double) (*bary_TS_MVA_)[0][i];
			sprintf(l_txtbuf, "\nDelta B_%d = %f", i, (double) (*deltaB)[i]);
			str += l_txtbuf;
		}

		for (int k = 0; k < nr_measurements_; k++) {
			for (i = 0; i < 3; i++) {
				(*Bint)[i] += (double) (*bary_TS_MVA_)[k][i];
				(*Jint)[i] += (double) (*curl_TS_MVA_)[k][i] * deltat_;
				(*Jave)[i] += (double) (*curl_TS_MVA_)[k][i];
				(*Jint_GSE)[i] += (double) (*curl_TS_)[k][i] * deltat_;
				(*Jave_GSE)[i] += (double) (*curl_TS_)[k][i];
			}
		}

		double bave = (double) (*rotation_matrix_)[constrained][0] * (double) (*Bint)[0] / nf
				+ (double) (*rotation_matrix_)[constrained][1] * (double) (*Bint)[1] / nf
				+ (double) (*rotation_matrix_)[constrained][2] * (double) (*Bint)[2] / nf;

		double df = (double) (*rotation_matrix_)[constrained][0] * (double) (*Jave_GSE)[0] / nf
				+ (double) (*rotation_matrix_)[constrained][1] * (double) (*Jave_GSE)[1] / nf
				+ (double) (*rotation_matrix_)[constrained][2] * (double) (*Jave_GSE)[2] / nf;

		printf("<J> = %d %g  %g  %g  (%g)\n", i, (double) (*Jave)[0] / nf, (double) (*Jave)[1] / nf, (double) (*Jave)[2]
				/ nf, df);
		sprintf(l_txtbuf, "\n\n<J> = [ %g  %g  %g] (%g along n_MVAJ)\n", (double) (*Jave)[0] / nf, (double) (*Jave)[1]
				/ nf, (double) (*Jave)[2] / nf, df);
		str += l_txtbuf;

		sprintf(l_txtbuf, "\n\n<J> = [ %g  %g  %g] (%g along n_MVAJ)\n", (double) (*Jave_GSE)[0] / nf,
				(double) (*Jave_GSE)[1] / nf, (double) (*Jave_GSE)[2] / nf, df);
		str += l_txtbuf;

		double maxJ = maxVal(curl_TS_);
		sprintf(l_txtbuf, "J_max = %10.3g\n", maxJ);
		str += l_txtbuf;

		printf("<B> = %d %g  %g  %g  (%g)\n", i, (double) (*Bint)[0] / nf, (double) (*Bint)[1] / nf, (double) (*Bint)[2]
				/ nf, bave);
		sprintf(l_txtbuf, "<B> = [ %g  %g  %g] (%g along n_MVAJ)\n\n", (double) (*Bint)[0] / nf, (double) (*Bint)[1]
				/ nf, (double) (*Bint)[2] / nf, bave);
		str += l_txtbuf;
		fflush(stdout);

		if (!bary_vector_TS_.is_nil()) {
			for (int k = 1; k < nr_measurements_; k++) {
				(*velo_TS_MVA_)[k][0] = ((double) (*bary_vector_TS_)[k - 1][1] - (double) (*bary_vector_TS_)[k][1]) * 1e-12 / (deltat_
						* mu0 * (double) (*curl_TS_MVA_)[k][2]);
				(*velo_TS_MVA_)[k][1] = ((double) (*bary_vector_TS_)[k - 1][2] - (double) (*bary_vector_TS_)[k][2]) * 1e-12 / (deltat_
						* mu0 * (double) (*curl_TS_MVA_)[k][1]);
				(*velo_TS_MVA_)[k][2] = ((double) (*velo_TS_MVA_)[k][0] + (double) (*velo_TS_MVA_)[k][1]) / 2.0; //avg of above
				//   printf("k = %d, %f  %f  -> %f  \n",k, (double)(*velo_TS_MVA_)[k][0],(double)(*velo_TS_MVA_)[k][1],(double)(*velo_TS_MVA_)[k][2] );
			}
		}
		if (constrained == 1) {
			(*v_n)[0] = (double) (*deltaB)[0] * 1e-9 / (mu0 * (double) (*Jint)[2]);
			(*v_n)[1] = -(double) (*deltaB)[2] * 1e-9 / (mu0 * (double) (*Jint)[0]);
			(*v_n)[2] = ((double) (*v_n)[0] + (double) (*v_n)[1]) / 2.0; //avg of above
		}
		else {
			(*v_n)[0] = (double) (*deltaB)[1] * 1e-9 / (mu0 * (double) (*Jint)[2]);
			(*v_n)[1] = -(double) (*deltaB)[2] * 1e-9 / (mu0 * (double) (*Jint)[1]);
			(*v_n)[2] = ((double) (*v_n)[0] + (double) (*v_n)[1]) / 2.0;
		}

		Status::PrintMessage(true, "--\n");
		for (i = 0; i < 2; i++) {
			//   printf("J_%d =  %g\n",i,(double)(*Jint)[i]);
			sprintf(l_txtbuf, "V_%d    = %g\n", i, (double) (*v_n)[i]);
			printf("%s\n", l_txtbuf);
			str += l_txtbuf;
		}
		sprintf(l_txtbuf, "V_ave = %g\n", (double) (*v_n)[2]);
		str += l_txtbuf;

		// ---- do MVA of barycentric B-field -----------
		QdRMatrix_var RmatxB = new QdRMatrix(Dim(3, 3), 0.0);
		QdRMatrix_var Rmatx_tmp = new QdRMatrix(Dim(3, 3), 0.0);
		str += "\nEigenvals and Eigenvecs for MVAB_bary :\n";
		normal_const = NormalizeToTen(bary_vector_TS_, temp_curl);
		normal_const = normal_const * normal_const;
		if (!bary_vector_TS_.is_nil()) {
			try {
				ConstrainedMinVariance minvar(*temp_curl, *temp_curl, constrained); // qdos minvariance call
				*RmatxB = minvar.eigenvectors().duplicate(SHALLOW);
				*Eigenvalues = minvar.eigenvalues().duplicate(SHALLOW);
			}
			catch (Exception& e) {
				e.print_msg();
				return Status::kUnknownError;
			}
		}
		if ((double) (*RmatxB)[0][0] < 0)
			(*RmatxB) *= -1.0;

		for (i = 0; i < 3; i++) {
			printf("%10.5g  [%10.6g, %10.6g, %10.6g ]\n", (double) (*Eigenvalues)[i] * normal_const, (double) (*RmatxB)[i][0],
					(double) (*RmatxB)[i][1], (double) (*RmatxB)[i][2]);
			sprintf(evalsevecs, "%10.6g  [%10.6g, %10.6g, %10.6g ]\n", (double) (*Eigenvalues)[i] * normal_const,
					(double) (*RmatxB)[i][0], (double) (*RmatxB)[i][1], (double) (*RmatxB)[i][2]);
			str += evalsevecs;
		} // end for
		fflush(stdout);
		bave = (double) (*RmatxB)[constrained][0] * (double) (*Bint)[0] / nf + (double) (*RmatxB)[constrained][1]
				* (double) (*Bint)[1] / nf + (double) (*RmatxB)[constrained][2] * (double) (*Bint)[2] / nf;
		df = (double) (*RmatxB)[constrained][0] * (double) (*Jave)[0] / nf + (double) (*RmatxB)[constrained][1]
				* (double) (*Jave)[1] / nf + (double) (*RmatxB)[constrained][2] * (double) (*Jave)[2] / nf;

		sprintf(l_txtbuf, "\n\n<J> = [ %g  %g  %g] (%g along n_MVAB_bary)\n", (double) (*Jave)[0] / nf,
				(double) (*Jave)[1] / nf, (double) (*Jave)[2] / nf, df);
		str += l_txtbuf;

		printf("<B> = %d %g  %g  %g  (%g)\n", i, (double) (*Bint)[0] / nf, (double) (*Bint)[1] / nf, (double) (*Bint)[2]
				/ nf, bave);
		fflush(stdout);
		sprintf(l_txtbuf, "<B> = [ %g  %g  %g] (%g along n_MVAB_bary)\n\n", (double) (*Bint)[0] / nf,
				(double) (*Bint)[1] / nf, (double) (*Bint)[2] / nf, bave);
		str += l_txtbuf;
		Status::PrintMessage(true, str.toLatin1().data());
	}
	return result;
}
// --------------------------------------------

double ClusterAnalysis::ComputeReciprocal4Spacecrafts(int t, QdRMatrix_var &K)
{
	int c[] = { 0, 1, 2, 0, 1 };
	int i, j, sc, c1 = 0, c2 = 1, c3 = 2, c4 = 3;
	double totvol = 0.0, vol[] = { 0.0, 0.0, 0.0, 0.0 };

	// Reciprocal vectors (a function of positions only) according to Eq. (14.7) in Blue Book
	for (i = 0; i < 3; i++) {
		(*K)[c1][i] = ((double) (*position_TS_)[t][c3][c[i + 1]] - (double) (*position_TS_)[t][c2][c[i + 1]])
				* ((double) (*position_TS_)[t][c4][c[i + 2]] - (double) (*position_TS_)[t][c2][c[i + 2]])
				- ((double) (*position_TS_)[t][c3][c[i + 2]] - (double) (*position_TS_)[t][c2][c[i + 2]])
						* ((double) (*position_TS_)[t][c4][c[i + 1]] - (double) (*position_TS_)[t][c2][c[i + 1]]);
		(*K)[c2][i] = -(((double) (*position_TS_)[t][c4][c[i + 1]] - (double) (*position_TS_)[t][c3][c[i + 1]])
				* ((double) (*position_TS_)[t][c3][c[i + 2]] - (double) (*position_TS_)[t][c1][c[i + 2]])
				- ((double) (*position_TS_)[t][c4][c[i + 2]] - (double) (*position_TS_)[t][c3][c[i + 2]])
						* ((double) (*position_TS_)[t][c3][c[i + 1]] - (double) (*position_TS_)[t][c1][c[i + 1]]));
		(*K)[c3][i] = ((double) (*position_TS_)[t][c4][c[i + 1]] - (double) (*position_TS_)[t][c1][c[i + 1]])
				* ((double) (*position_TS_)[t][c4][c[i + 2]] - (double) (*position_TS_)[t][c2][c[i + 2]])
				- ((double) (*position_TS_)[t][c4][c[i + 2]] - (double) (*position_TS_)[t][c1][c[i + 2]])
						* ((double) (*position_TS_)[t][c4][c[i + 1]] - (double) (*position_TS_)[t][c2][c[i + 1]]);
		(*K)[c4][i] = ((double) (*position_TS_)[t][c2][c[i + 1]] - (double) (*position_TS_)[t][c1][c[i + 1]])
				* ((double) (*position_TS_)[t][c3][c[i + 2]] - (double) (*position_TS_)[t][c1][c[i + 2]])
				- ((double) (*position_TS_)[t][c2][c[i + 2]] - (double) (*position_TS_)[t][c1][c[i + 2]])
						* ((double) (*position_TS_)[t][c3][c[i + 1]] - (double) (*position_TS_)[t][c1][c[i + 1]]);
	}

	//Normally, the volume must be same. But there are always some round-off errors.
	vol[c1] = -((double) (*position_TS_)[t][c2][0] - (double) (*position_TS_)[t][c1][0]) * (double) (*K)[c1][0]
			- ((double) (*position_TS_)[t][c2][1] - (double) (*position_TS_)[t][c1][1]) * (double) (*K)[c1][1]
			- ((double) (*position_TS_)[t][c2][2] - (double) (*position_TS_)[t][c1][2]) * (double) (*K)[c1][2];
	vol[c2] = -((double) (*position_TS_)[t][c3][0] - (double) (*position_TS_)[t][c2][0]) * (double) (*K)[c2][0]
			- ((double) (*position_TS_)[t][c3][1] - (double) (*position_TS_)[t][c2][1]) * (double) (*K)[c2][1]
			- ((double) (*position_TS_)[t][c3][2] - (double) (*position_TS_)[t][c2][2]) * (double) (*K)[c2][2];
	vol[c3] = -((double) (*position_TS_)[t][c4][0] - (double) (*position_TS_)[t][c3][0]) * (double) (*K)[c3][0]
			- ((double) (*position_TS_)[t][c4][1] - (double) (*position_TS_)[t][c3][1]) * (double) (*K)[c3][1]
			- ((double) (*position_TS_)[t][c4][2] - (double) (*position_TS_)[t][c3][2]) * (double) (*K)[c3][2];
	vol[c4] = +((double) (*position_TS_)[t][c4][0] - (double) (*position_TS_)[t][c1][0]) * (double) (*K)[c4][0]
			+ ((double) (*position_TS_)[t][c4][1] - (double) (*position_TS_)[t][c1][1]) * (double) (*K)[c4][1]
			+ ((double) (*position_TS_)[t][c4][2] - (double) (*position_TS_)[t][c1][2]) * (double) (*K)[c4][2];

	for (sc = 0; sc < 4; sc++) {
		for (i = 0; i < 3; i++) {
			(*K)[sc][i] /= vol[sc];
			totvol += 0.25 * fabs(vol[sc]);
		}
	}
	return totvol;
}

void ClusterAnalysis::computeReciprocal3Spacecrafts(const int t, QdRMatrix_var &normal, const double surface, QdRMatrix_var &K)
{
	//Assume that QdRMatrix_var P = new QdRMatrixSeq(nestsize, RMatrix(Dim(nr_spacecrafts,3)));
	//Assume that: QdRMatrix_var normal = new QdRMatrix(Dim(3), 0.0);
	//Assume that: QdRMatrix_var K = new QdRMatrix(Dim(nr_spacecrafts,3), 0.0);

	int c[]={0,1,2,0,1};  //to compute the cross product: cyclic permutations

	//Formula (12), but normal is normalized here.
	for (unsigned short sat = 0; sat < 3; sat++)
		for (unsigned short i = 0; i < 3; i++) //coordinate
			(*K)[sat][i] = ((double) (*normal)[c[i + 1]] *
					(static_cast<double>((*position_TS_)[t][c[sat + 2]][c[i + 2]]) - static_cast<double>((*position_TS_)[t][c[sat + 1]][c[i + 2]]))
					- (double) (*normal)[c[i + 2]] * (static_cast<double>((*position_TS_)[t][c[sat + 2]][c[i + 1]])
					- static_cast<double>((*position_TS_)[t][c[sat + 1]][c[i + 1]]))) / surface;
}

void ClusterAnalysis::computeNormalToSurfaceOfFirst3Sat(int nr_position_measurements,
		QdRScalarSeq_var &surface_TS, QdRMatrixSeq_var &normal_TS)
{
	int nr_meas = min(position_TS_->sequence_size(), nr_position_measurements);  //for safety reasons.
	//Assume that normal_TS_ was not pre-allocated:
	normal_TS = new QdRMatrixSeq(nr_meas, RMatrix(Dim(kVectorRank), 0.0));
	QdRMatrix_var Ns = new QdRMatrix(Dim(kVectorRank,kVectorRank),0.0);  //normal coordinate as viewed from each satellite (temp. variable).
	surface_TS = new QdRScalarSeq(nr_meas, 0.0);
	QdRMatrix_var normal = new QdRMatrix(Dim(kVectorRank), 0.0);

	int c[]={0,1,2,0,1};  //to compute the cross product: cyclic permutations

	for (int t=0; t<nr_meas; t++) {

		for (short sat=0; sat<3; sat++)  //we always have at least 3 satellites.
			for (short i=0; i<3; i++) {  //coordinate (position is always a 3 vector)
				(*Ns)[sat][i] =	((double)(*position_TS_)[t][c[sat+1]][c[i+1]] - (double)(*position_TS_)[t][c[sat+0]][c[i+1]]) *
								((double)(*position_TS_)[t][c[sat+2]][c[i+2]] - (double)(*position_TS_)[t][c[sat+0]][c[i+2]]) -
								((double)(*position_TS_)[t][c[sat+1]][c[i+2]] - (double)(*position_TS_)[t][c[sat+0]][c[i+2]]) *
								((double)(*position_TS_)[t][c[sat+2]][c[i+1]] - (double)(*position_TS_)[t][c[sat+0]][c[i+1]]);
			}
		//	Average over normals viewed from different points:
		for (short i=0; i<3; i++)  //coordinate
			(*normal)[i] = ((*Ns)[0][i] + (*Ns)[1][i] + (*Ns)[2][i])/3;
		(*surface_TS)[t] = sqrt(computeSquareVectorNorm(normal));
		for (short i=0; i<3; i++)  //coordinate
			(*normal_TS)[t][i] = static_cast<double>((*normal)[i]) / (double)(*surface_TS)[t];
	}
}

double ClusterAnalysis::maxVal(QdRMatrixSeq_var &V)
{
	double sumvec, maxval = 0.0;
	int n = V->sequence_size();
	for (unsigned int t = 0; t < n; t++) {
		sumvec = (*V)[t][0] * (*V)[t][0] + (*V)[t][1] * (*V)[t][1] + (*V)[t][2] * (*V)[t][2];
		if (sumvec > maxval)
			maxval = sumvec;
	}
	return sqrt(maxval);
}

double ClusterAnalysis::sum_SCs(QdRMatrix_var &V1, QdRMatrix_var &V2, int coordV1, int coordV2, unsigned short nr_spacecrafts)
{
	double sum = 0.0;

	for (int i = 0; i < nr_spacecrafts; i++) {
		sum += static_cast<double>((*V1)[i][coordV1]) * static_cast<double>((*V2)[i][coordV2]);
	}
	return (sum);
}

double ClusterAnalysis::DivergenceVec(QdRMatrix_var &V, QdRMatrix_var &R)
{
	// see Eq. 14.16 in Blue Book
	return (sum_SCs(V, R, 0, 0, nr_satellites_) + sum_SCs(V, R, 1, 1, nr_satellites_) + sum_SCs(V, R, 2, 2, nr_satellites_));
}

void ClusterAnalysis::CurlVec(QdRMatrix_var &V, QdRMatrix_var &R, const unsigned short nr_spacecrafts, QdRMatrix_var &Curl)
{
	// see Eq. 14.17 in Blue Book
	// SEH, 2 May 2005; flipped

	//Assume that the last variable is: QdRMatrix_var Curl = new QdRMatrix(Dim(3), 0.0);
	(*Curl)[0] = (-sum_SCs(V, R, 1, 2, nr_spacecrafts) + sum_SCs(V, R, 2, 1, nr_spacecrafts));
	(*Curl)[1] = (-sum_SCs(V, R, 2, 0, nr_spacecrafts) + sum_SCs(V, R, 0, 2, nr_spacecrafts));
	(*Curl)[2] = (-sum_SCs(V, R, 0, 1, nr_spacecrafts) + sum_SCs(V, R, 1, 0, nr_spacecrafts));
}

void ClusterAnalysis::computeGradVec(QdRMatrix_var &V, QdRMatrix_var &R, QdRMatrix_var &Grad)
{
	//Gradient of a vector is a 3x3 matrix.
	//Assume that: QdRMatrix_var Grad = new QdRMatrix(Dim(3,3), 0.0);

	for (unsigned int coord = 0; coord < 3; coord++) {
		for (unsigned int j = 0; j < 3; j++) {
			(*Grad)[coord][j] = sum_SCs(V, R, coord, j, nr_satellites_);
		}
	}
}

void ClusterAnalysis::computeGradScalar(QdRMatrix_var &Vs, QdRMatrix_var &R, QdRMatrix_var &Grad)
{
	//Assume that: QdRMatrix_var Grad = new QdRMatrix(Dim(3), 0.0);

	//Multiply the magnitude with reciprocal vectors:
	for (unsigned int i=0; i<3; i++) (*Grad)[i] = sum_SCs(Vs,R,0,i, nr_satellites_);
}

void ClusterAnalysis::computeGradMagnitude(QdRMatrix_var &V, QdRMatrix_var &R, QdRMatrix_var &Grad)
{
	//First parameter is the field and second parameter contains reciprocal vectors.
	//Assume that: QdRMatrix_var Grad = new QdRMatrix(Dim(3), 0.0);
  QdRMatrix_var Magn = new QdRMatrix(Dim(nr_satellites_,1), 0.0);      // magnitude of vector from each SC
  QdRMatrix_var  Vec = new QdRMatrix(Dim(3), 0.0);
  int i,j;

	//1. Compute the magnitude
  for (i=0; i<nr_satellites_; i++) {
    for (j=0; j<3; j++) (*Vec)[j] = (*V)[i][j];
    (*Magn)[i][0] = sqrt(computeSquareVectorNorm(Vec));
  }

	//2. Compute gradient of scalars:
	computeGradScalar(Magn, R, Grad);
}

double ClusterAnalysis::computeSquareVectorNorm(const QdRMatrix_var &V)
{
  double sum=0.0;
  for (int i=0; i<3; i++) sum += (*V)[i] * (*V)[i];
  return sum;
}

void ClusterAnalysis::normalizeVector(const QdRMatrix_var &vector, QdRMatrix_var &unit_vector)
{
	//Assume that: QdRMatrix_var vector = new QdRMatrix(Dim(3), 0.0); [in]
	//Assume that: QdRMatrix_var unit_vector = new QdRMatrix(Dim(3), 0.0); [out]

	double norm = computeSquareVectorNorm(vector);
	for (unsigned short i = 0; i<3; i++)
		(*unit_vector)[i] = (double)((*vector)[i]) / norm;
}

double ClusterAnalysis::computeEstimateDgDn(const QdRMatrix_var &grad_p, const QdRMatrix_var &normal,
													 const CEstimateNormalComponentFor3SpacecraftMethod &estimator,
													 const unsigned int t,
													 const unsigned short coord,
													 QdRScalar &dgdn)
{
	//Assume that: QdRMatrix_var grad_p = new QdRMatrix(Dim(3), 0.0); [in]
	//Assume that: QdRMatrix_var normal = new QdRMatrix(Dim(3), 0.0); [in]
	//Assume that: QdRMatrix_var vec_e = new QdRMatrix(Dim(3), 0.0); [in]

	double quality = 0.0;  //how reliable is the result. Quality should not be too small.
	dgdn = 0;
	if (estimator.get_option() == CEstimateNormalComponentFor3SpacecraftMethod::kNone) {
		if ((t==0) && (coord==0)) Status::PrintMessage(true, "No estimation for the normal component\n");
		return quality;
	}

	const double epsilon = 1e-50;
	QdRMatrix_var estvector;
	if (!estimator.qvector.is_nil()) {
		estvector = estimator.qvector;
	}
	else {
		estvector = new QdRMatrix(Dim(3), 0.0);
		for (unsigned short j=0; j<3; j++)
			(*estvector)[j] = (double) ((*estimator.qvector_seq)[t][j]);
	}

	switch (estimator.get_option()){
		case CEstimateNormalComponentFor3SpacecraftMethod::kGradientParallel: {
			//Gradient parallel to a given vector (formula 32)
			if ((t==0) && (coord==0)) Status::PrintMessage(true, "Estimating normal component using a vector parallel to gradient.\n");
			int c[]={0,1,2,0,1};  //to compute the cross product: cyclic permutations
			double prod1=0, prod2=0;
			QdRMatrix_var qnormalized_estvector = new QdRMatrix(Dim(3), 0.0);
			normalizeVector(estvector, qnormalized_estvector);
			QdRMatrix_var prodvect = new QdRMatrix(Dim(3), 0.0);
			for (unsigned int i=0; i<3; i++){  //vector components
				prod1 += (*(qnormalized_estvector))[i] * (*grad_p)[i];
				prod2 += (*(qnormalized_estvector))[i] * (*normal)[i];
				//vector product:
				(*prodvect)[c[i]] = (*(qnormalized_estvector))[c[i+1]] * (*normal)[c[i+2]] -
														(*(qnormalized_estvector))[c[i+2]] * (*normal)[c[i+1]];
			}
			quality = computeSquareVectorNorm(prodvect);
			if (quality>epsilon)
				dgdn = ((double)prod1 * (double)prod2) / quality;
		}
			break;
		case CEstimateNormalComponentFor3SpacecraftMethod::kGradientPerpendicular: {
			//Gradient perpendicular to a given vector (formula 34)
			if ((t==0) && (coord==0)) Status::PrintMessage(true, "Estimating normal component using a vector perpendicular to gradient.\n");
			QdRMatrix_var qnormalized_estvector = new QdRMatrix(Dim(3), 0.0);
			normalizeVector(estvector, qnormalized_estvector);
			double prod1=0, prod2=0;
			for (unsigned int i=0; i<3; i++){  //vector components
				prod1 += (*(qnormalized_estvector))[i] * (*grad_p)[i];
				prod2 += (*(qnormalized_estvector))[i] * (*normal)[i];
			}
			quality = prod2;
			if (quality>epsilon)
				dgdn = - prod1 / prod2;
		}
			break;
		case CEstimateNormalComponentFor3SpacecraftMethod::kStationarity: {
			if ((t==0) && (coord==0)) Status::PrintMessage(true, "Estimating normal component using stationarity assumption.\n");
			//Here the estimator vector is the velocity of the plasma frame relative to the spacecraft array: U = U_p +U_n n
			double U_n = 0;
			for (unsigned short i=0; i<3; i++) U_n += (*(estvector))[i] * (*normal)[i];
			quality = abs(U_n);
			QdRMatrix_var U_p = new QdRMatrix(Dim(3), 0.0);
			double UpDpg = 0;
			for (unsigned short i=0; i<3; i++) {
				(*U_p)[i] = (double)(*(estvector))[i] - U_n * (*normal)[i];
				UpDpg += (*U_p)[i] * (*grad_p)[i];
			}
			dgdn = -(estimator.get_slope(t, coord) + UpDpg)/U_n;
		}
			break;
        case CEstimateNormalComponentFor3SpacecraftMethod::kNone: {
            break;
        }
            
	}
	return quality;
}

}  //namespace cluster
