#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

using namespace QSAS;

extern "C"
{
    DECLDIR QplugReturnStatus Sort(QplugArgList *);
    DECLDIR QplugReturnStatus Union(QplugArgList *);
    DECLDIR QplugReturnStatus Intersection(QplugArgList *);
    DECLDIR QplugReturnStatus Exclusion(QplugArgList *);
}



struct sortCriterionIncStart : public std::binary_function< QdTimeInterval&, QdTimeInterval&, bool > 
{
    inline bool operator()(const QdTimeInterval& i1, const QdTimeInterval& i2)
    {
        return (i1.start() < i2.start() || (i1.start()==i2.start() && i1.end() < i2.end()));
    }
};

struct sortCriterionDecStart : public std::binary_function< QdTimeInterval&, QdTimeInterval&, bool > 
{
    inline bool operator()(const QdTimeInterval& i1, const QdTimeInterval& i2)
    {
        return (i2.start() < i1.start() || (i2.start()==i1.start() && i2.end() < i1.end()));
    }
};


struct sortCriterionIncEnd : public std::binary_function< QdTimeInterval&, QdTimeInterval&, bool > 
{
    inline bool operator()(const QdTimeInterval& i1, const QdTimeInterval& i2)
    {
        return (i1.end() < i2.end() || (i1.end()==i2.end() && i1.start() < i2.start()));
    }
};

struct sortCriterionDecEnd : public std::binary_function< QdTimeInterval&, QdTimeInterval&, bool > 
{
    inline bool operator()(const QdTimeInterval& i1, const QdTimeInterval& i2)
    {
        return (i2.end() < i1.end() || (i2.end()==i1.end() && i2.start() < i1.start()));
    }
};

struct sortCriterionIncCentre : public std::binary_function< QdTimeInterval&, QdTimeInterval&, bool > 
{
    inline bool operator()(const QdTimeInterval& i1, const QdTimeInterval& i2)
    {
        QmjdTime centre1 = i1.centre();
        QmjdTime centre2 = i2.centre();
        return (centre1 > centre2 || (centre1==centre2 && i1.start() < i2.start()));
    }
};

struct sortCriterionDecCentre : public std::binary_function< QdTimeInterval&, QdTimeInterval&, bool > 
{
    inline bool operator()(const QdTimeInterval& i1, const QdTimeInterval& i2)
    {
        QmjdTime centre1 = i1.centre();
        QmjdTime centre2 = i2.centre();
        return (centre2 < centre1 || (centre2==centre1 && i2.start() < i1.start()));
    }
};

QplugReturnStatus Sort(QplugArgList *call_list)
{
    QdObject_var temp=(*call_list)[0];
    QdTimeIntervalSeq_var table=QdTimeIntervalSeq_var::narrow(temp);
    
    if(table.is_nil())
    {
        QplugAppendTextDisplay("Failed: Input is not an event table.");
        return QPLUG_FAILURE;
    }
    
    QdObject_var temp_sort=(*call_list)[1];
    QdRScalar_var sortType=QdRScalar_var::narrow(temp_sort);
    
    if(sortType.is_nil())
    {
        QplugAppendTextDisplay("Failed: sort criterion.");
        return QPLUG_FAILURE;
    }
    
    QdTimeIntervalSeq seq=*table;
    
    switch((int)*sortType)
    {
        case 1: std::sort(seq.begin(), seq.end(), sortCriterionIncCentre());
                break;
       
        case 2: std::sort(seq.begin(), seq.end(), sortCriterionIncEnd());
                break;
                
        case 3: std::sort(seq.begin(), seq.end(), sortCriterionDecStart());
                break;
       
        case 4: std::sort(seq.begin(), seq.end(), sortCriterionDecCentre());
                break;
                
        case 5: std::sort(seq.begin(), seq.end(), sortCriterionDecEnd());
                break;
       
        default: std::sort(seq.begin(), seq.end(), sortCriterionIncStart());
                break;
    }
    
    QdTimeIntervalSeq_var res=new QdTimeIntervalSeq(seq);
    call_list->push_back((QdObject_var) res);
    return QPLUG_SUCCESS;
}

typedef struct
{
    QdTimeIntervalSeq_var elements;
    std::vector<bool> isInSet1;
    std::vector<bool> isInSet2;
} set_elements;

void getSetElements(QdTimeIntervalSeq_var& l1, QdTimeIntervalSeq_var& l2, set_elements* res)
{
    res->elements->clear();
    res->isInSet1.clear();
    res->isInSet2.clear();
    
    std::list<QmjdTime> points;
    
    for(int i=0; i<l1->size(); i++)
    {
        points.push_back(l1->elt(i).start());
        points.push_back(l1->elt(i).end());
    }
    
    for(int i=0; i<l2->size(); i++)
    {
        points.push_back(l2->elt(i).start());
        points.push_back(l2->elt(i).end());
    }
    
    points.sort();
    
    QmjdTime t1, t2;
    t1=points.front();
    points.pop_front();
    while(!points.empty())
    {
        t2=points.front();
        points.pop_front();
        
        res->elements->push_back(QmjdTimeInterval(t1, t2));
        t1=t2;
    }
    
    int nb_elements=res->elements->size();
    
    res->isInSet1.resize(nb_elements);
    res->isInSet2.resize(nb_elements);
//     res->isInSet1=false;
//     res->isInSet2=false;
    
    for(int i=0; i<l1->size(); i++)
    {
        t1=l1->elt(i).start();
        t2=l1->elt(i).end();
        bool end=false;
        for(int j=0; j<nb_elements && !end; j++)
        {
            if(res->elements->elt(j).start()>=t1 && res->elements->elt(j).end()<=t2)
            {
                res->isInSet1[j]=true;
            }
            if(res->elements->elt(j).start()>t2) end=true;
        }
    }
    
    for(int i=0; i<l2->size(); i++)
    {
        t1=l2->elt(i).start();
        t2=l2->elt(i).end();
        bool end=false;
        for(int j=0; j<nb_elements && !end; j++)
        {
            if(res->elements->elt(j).start()>=t1 && res->elements->elt(j).end()<=t2)
            {
                res->isInSet2[j]=true;
            }
            if(res->elements->elt(j).start()>t2) end=true;
        }
    }
}

QplugReturnStatus Union(QplugArgList *call_list)
{
    QdObject_var temp1=(*call_list)[0];
    QdObject_var temp2=(*call_list)[1];
    
    QdTimeIntervalSeq_var times1=QdTimeIntervalSeq_var::narrow(temp1);
    QdTimeIntervalSeq_var times2=QdTimeIntervalSeq_var::narrow(temp2);
    
    if(times1.is_nil())
    {
        QplugAppendTextDisplay("Failed: First object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    if(times2.is_nil())
    {
        QplugAppendTextDisplay("Failed: Second object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    set_elements elts;
    elts.elements=new QdTimeIntervalSeq();
    getSetElements(times1, times2, &elts);
    
    QdTimeIntervalSeq_var res=new QdTimeIntervalSeq();
    
    QmjdTime t1, t2;
    bool isInInterval=false;
    for(int i=0; i<elts.elements->size(); i++)
    {
        if(elts.isInSet1[i] || elts.isInSet2[i]) // If in union
        {
            if(!isInInterval) // We have just entered the union interval
            {
                t1=elts.elements->elt(i).start();
                t2=elts.elements->elt(i).end();
                isInInterval=true;
            }
            else
            {
                t2=elts.elements->elt(i).end();
            }
        }
        else
        {
            if(isInInterval) // We have just left the union interval
            {
                res->push_back(QmjdTimeInterval(t1, t2));
                isInInterval=false;
            }
        }
    }
    call_list->push_back((QdObject_var) res);
    return QPLUG_SUCCESS;
}

QplugReturnStatus Intersection(QplugArgList *call_list)
{
    QdObject_var temp1=(*call_list)[0];
    QdObject_var temp2=(*call_list)[1];
    
    QdTimeIntervalSeq_var times1=QdTimeIntervalSeq_var::narrow(temp1);
    QdTimeIntervalSeq_var times2=QdTimeIntervalSeq_var::narrow(temp2);
    
    if(times1.is_nil())
    {
        QplugAppendTextDisplay("Failed: First object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    if(times2.is_nil())
    {
        QplugAppendTextDisplay("Failed: Second object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    set_elements elts;
    elts.elements=new QdTimeIntervalSeq();
    getSetElements(times1, times2, &elts);
    
    QdTimeIntervalSeq_var res=new QdTimeIntervalSeq();
    
    QmjdTime t1, t2;
    bool isInInterval=false;
    for(int i=0; i<elts.elements->size(); i++)
    {
        if(elts.isInSet1[i] && elts.isInSet2[i]) // If in union
        {
            if(!isInInterval) // We have just entered the union interval
            {
                t1=elts.elements->elt(i).start();
                t2=elts.elements->elt(i).end();
                isInInterval=true;
            }
            else
            {
                t2=elts.elements->elt(i).end();
            }
        }
        else
        {
            if(isInInterval) // We have just left the union interval
            {
                res->push_back(QmjdTimeInterval(t1, t2));
                isInInterval=false;
            }
        }
    }
    call_list->push_back((QdObject_var) res);
    return QPLUG_SUCCESS;
}

QplugReturnStatus Exclusion(QplugArgList *call_list)
{
    QdObject_var temp1=(*call_list)[0];
    QdObject_var temp2=(*call_list)[1];
    
    QdTimeIntervalSeq_var times1=QdTimeIntervalSeq_var::narrow(temp1);
    QdTimeIntervalSeq_var times2=QdTimeIntervalSeq_var::narrow(temp2);
    
    if(times1.is_nil())
    {
        QplugAppendTextDisplay("Failed: First object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    if(times2.is_nil())
    {
        QplugAppendTextDisplay("Failed: Second object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    set_elements elts;
    elts.elements=new QdTimeIntervalSeq();
    getSetElements(times1, times2, &elts);
    
    QdTimeIntervalSeq_var res=new QdTimeIntervalSeq();
    
    QmjdTime t1, t2;
    bool isInInterval=false;
    for(int i=0; i<elts.elements->size(); i++)
    {
        if(elts.isInSet1[i] && !elts.isInSet2[i]) // If in union
        {
            if(!isInInterval) // We have just entered the union interval
            {
                t1=elts.elements->elt(i).start();
                t2=elts.elements->elt(i).end();
                isInInterval=true;
            }
            else
            {
                t2=elts.elements->elt(i).end();
            }
        }
        else
        {
            if(isInInterval) // We have just left the union interval
            {
                res->push_back(QmjdTimeInterval(t1, t2));
                isInInterval=false;
            }
        }
    }
    call_list->push_back((QdObject_var) res);
    return QPLUG_SUCCESS;
}
