// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : V_perp and V_par
//
// This plugin takes a magnetic field vector and a plasma velocity
// vector as inputs, and calculates the plasma velocity components
// along and perpendicular to the magnetic field
//
// Inputs :
//      bfield          - vector time series with magnetic field
//      velocity        - vector time series with plasma velocity
//
// Outputs
//      v_perp          - plasma velocity perpendicular to B-field
//      v_par           - plasma velocity along B-field
//
//
// SEH, 19 Sep, 2001
//      11 Mar, 2003    - join V and B in PlugIn
//      26 Dec, 2003    - QSAS 2.0 version; Changed headers + 'external'
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

// needed for QSAS 2.0
extern "C" QplugReturnStatus  ISSI_Vperp  ( QplugArgList *);

typedef struct VEC {float x,y,z;} vec;
float vecsum(float x, float y, float z);

using namespace QSAS;

QplugReturnStatus  ISSI_Vperp  ( QplugArgList *call_list )
{
QdRMatrixSeq_var  v_perp;
QdRScalarSeq_var  v_par;

long    n_vecs, n_plas;
vec     b_vec,vpar_vec,v_vec;
float   bmag,vpar_scalar;
double  deltat,avg_width;
int     nseq=2,triage_cnt;

//AR Added validity checks for inputs
QdRMatrixSeq_var test_B=QdRMatrixSeq_var::narrow((*call_list)[0]);
QdRMatrixSeq_var test_V=QdRMatrixSeq_var::narrow((*call_list)[1]);
if(test_B.is_nil())
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - Wrong type for bfield\n");
	return QPLUG_FAILURE;  
}
QdObject_var Units_test=QarConvertUnitsTo(test_B, "1 > T");
if(Units_test.is_nil())
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - Wrong units for bfield\n");
	return QPLUG_FAILURE;  
}

if(test_V.is_nil())
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - Wrong type for velocity\n");
	return QPLUG_FAILURE;  
}
Units_test=QarConvertUnitsTo(test_V, "1 > m s^-1");
if(Units_test.is_nil())
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - Wrong units for velocity\n");
	return QPLUG_FAILURE;  
}
if(QarTestConformalForVecProds(test_B, test_V)==QAR_ERR)
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - bfield and velocity are not conformal \n");
	return QPLUG_FAILURE;  
}

 //  --- first we join B measurements to plasma measurements -----

 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 QdObjectSeq_var restSig = new QdObjectSeq(nseq);
 QdStringSeq_var joinMethod = new QdStringSeq(nseq);
 vector <double> gapValue(nseq);
 QdStringSeq_var gapHandler = new QdStringSeq(nseq);
 QdRMatrixSeq_var  bfield_nest;
 QdRScalarSeq_var  density_nest;
 QdRScalarSeq_var  Tpar_nest;
 QdRScalarSeq_var  Tperp_nest;
 QdTimeSeq_var ttJoined = get_timetags((*call_list)[1]);     // tt of joined = from plasma
 QdTimeSeq_var tt_out;                                      // tt of masked

 if (ttJoined.is_nil()) {
   QplugAppendTextDisplay("Bad or missing time tags in input data\n");
   return QPLUG_FAILURE;
 }

 QuIsTSRegular(ttJoined,&deltat);
 avg_width = 1.5 * deltat;
 printf("delta t = %f  ->  gap_fill if t > %f\n",deltat,avg_width);
 fflush(stdout);

 for (int i=0; i<nseq; i++) {
   toJoin[i] = ((*call_list)[i]);

   joinMethod[i] = QdString(LINEAR);
   gapValue[i] = avg_width;
   gapHandler[i] = QdString(LINEAR);   
 }

 try {
   printf("%f\n",avg_width);
   QuMultiJoin(toJoin, ttJoined, joined, gapValue, gapHandler, joinMethod);
   printf("%f\n",avg_width);
   fflush(stdout);
 }
 catch(Exception &e) {
   e.print_msg();
   printf("%f\n",avg_width);
   fflush(stdout);
   return QPLUG_FAILURE;
 }

 QuSetTxtAttr("SI_conversion", QuGetAttrText("SI_conversion", test_B).c_str(), joined[0]);
 QuSetTxtAttr("UNITS", QuGetAttrText("UNITS", test_B).c_str(), joined[0]);
 QuSetFrame((QdObject_var) joined[0], QuGetFrameAttr(test_B).c_str());
 QuSetTxtAttr("SI_conversion", QuGetAttrText("SI_conversion", test_V).c_str(), joined[1]);
 QuSetTxtAttr("UNITS", QuGetAttrText("UNITS", test_V).c_str(), joined[1]);
 QuSetFrame((QdObject_var) joined[1], QuGetFrameAttr(test_V).c_str());
 std::cout << QuGetAttrText("SI_conversion", joined[0]) <<  " " << QuGetAttrText("UNITS", joined[0]) << " " << QuGetAttrText("FRAME", joined[0]) << std::endl;
 joined[0]=QarConvertUnitsTo(joined[0], "1 > T");
 joined[1]=QarConvertUnitsTo(joined[1], "1 > m s^-1");
 if(joined[0].is_nil()) std::cout << "0 is nil" << std::endl;
 if(joined[1].is_nil()) std::cout << "1 is nil" << std::endl;
 QdRMatrixSeq_var joined_B = QdRMatrixSeq_var::narrow(joined[0]);
 QdRMatrixSeq_var joined_V = QdRMatrixSeq_var::narrow(joined[1]);
 if(joined_B.is_nil()) std::cout << "B is nil" << std::endl;
 if(joined_V.is_nil()) std::cout << "V is nil" << std::endl;


// --- get Plasma velocity vector -----

n_vecs = joined_V->sequence_size();
printf("No of positions %ld\n",n_vecs);
fflush(stdout);



// declare and calculate the three output objects

v_perp = new QdRMatrixSeq(n_vecs,RMatrix(Dim(3)));
v_par  = new QdRScalarSeq(n_vecs);

  for(int i=0; i<n_vecs; i++)  {

     b_vec.x = (float) (joined_B[i][0]);
     b_vec.y = (float) (joined_B[i][1]);
     b_vec.z = (float) (joined_B[i][2]);
     bmag = vecsum(b_vec.x,b_vec.y,b_vec.z);                                            //   calculate |B|

     b_vec.x = b_vec.x / bmag;                                                          //   calculate B/|B|
     b_vec.y = b_vec.y / bmag;
     b_vec.z = b_vec.z / bmag;

     v_vec.x = (float) (joined_V[i][0]);
     v_vec.y = (float) (joined_V[i][1]);
     v_vec.z = (float) (joined_V[i][2]);

     vpar_scalar = v_vec.x*b_vec.x  +  v_vec.y*b_vec.y  +  v_vec.z*b_vec.z;             //   scalar(v_||) = V dot B
     v_par[i] = vpar_scalar/1000.;

     vpar_vec.x = (float) (vpar_scalar * b_vec.x);                                      //   V_|| = scalar(v_||) * B
     vpar_vec.y = (float) (vpar_scalar * b_vec.y);
     vpar_vec.z = (float) (vpar_scalar * b_vec.z);

     v_perp[i][0] = (v_vec.x - (float) vpar_vec.x)/1000.;                                     //   V_perp = V - V_||
     v_perp[i][1] = (v_vec.y - (float) vpar_vec.y)/1000.;
     v_perp[i][2] = (v_vec.z - (float) vpar_vec.z)/1000.;
     
  }



set_timetags(v_perp,ttJoined);
QuSetFrame((QdObject_var) v_perp, QuGetFrameAttr(test_B).c_str());

QuSetTxtAttr("SI_conversion","1.0e3>m s^-1",(QdObject_var) v_perp);
QuSetTxtAttr("UNITS","km/s",(QdObject_var) v_perp);
QuSetTxtAttr("LABLAXIS","V_perp",(QdObject_var) v_perp);

set_timetags(v_par,ttJoined);
QuSetTxtAttr("SI_conversion","1.0e3>m s^-1",(QdObject_var) v_par);
QuSetTxtAttr("UNITS","km/s",(QdObject_var) v_par);
QuSetTxtAttr("LABLAXIS","V_par",(QdObject_var) v_par);


call_list->push_back((QdObject_var) v_perp);
call_list->push_back((QdObject_var) v_par);

QplugAppendTextDisplay ("ISSI_Vperp>> data object created - exiting successfully\n");

return QPLUG_SUCCESS;

}



float vecsum(float x, float y, float z)
{
  return sqrt(x*x +y*y +z*z);
}
