/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	MaxEntSpectrum.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Maximum Entropy Power spectrum wrapper
*
* PURPOSE:      Wrapper for NR 'evlmem' power spectrum.
* 
* DESCRIPTION:  Plug In module checking input parameters.
*
* ARGUMENTS:
*    IN:       QdCDOList  list of plug in objects containing
*              QdCTimeSeries*   vector or scalar regular time series object
*              QdCTimeIntervall inter vall object for subsetting
*              int  npoles  number of poles
*              int  nfreq   number of frequencies
*              int  linlog  linear or logarithmic frequency spacing
*              float minfreq minimal frequency
*              float maxfreq maximal frequency
*
*   OUT:
*         float *frequency   frequency tags		
*         float *power_mag       power spectral density of magnitude
*         vector series *power_vec       power spectral density of components
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	Qcoef Qevlp.c
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*
* HISTORY: created  24/11/97 Markus Fraenz 
* Migrated to QDOS 2 and C arrays 09/2000 A.J.Allen 
* indexing repaired 16/1/2001 Markus Fraenz
* vector input and subsetting 18/01/2001 Markus Fraenz
*
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"

extern "C" QplugReturnStatus  MaxEntSpectrum( QplugArgList *);

using namespace QSAS;

static float sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

void Qcoef(float data[],unsigned long n,unsigned long m,float *xms,float d[]);
float Qevlp(float fdt, float d[], unsigned long m, float xms);


QplugReturnStatus	MaxEntSpectrum( QplugArgList *call_list )
{

// define some C variables

long	i, j;

int	    linlog,vectorset=0;
unsigned long	Ndata,Ndata2, nfreq, npoles;
float           minfreq,maxfreq;

char *attr_value, str_out[150];

float	xms,fdt=0.0,freqstep=0.0,freqwidth=0.0;
float  *data,*d,*freq;
double *tags,deltat=0.0,deltat2=0.0;


// define QSAS data objects


QdTimeInterval_var start_end;
QdObject_var ts_out;
QdTimeSeq_var tt_out;

QdRScalarSeq_var  ts_inr;
QdRScalarSeq_var  out_freq_series;
QdRScalarSeq_var  out_spmag_series;
QdObject_var  out_spvec_series;
QdRMatrixSeq_var ts_inm;


// ...first item is input Time Series

QdObject_var ts_in = (* call_list)[0];

//  check QSAS datatype

QdTimeSeq_var tt_in = get_timetags(ts_in);
if( tt_in.is_nil() ) {
  QplugAppendTextDisplay ("MaxEntSpectrumFFT>> Failed to get time tags\n");
  return QPLUG_FAILURE;
}
if ( !QuIsTSRegular(tt_in, &deltat) ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - time series not evenly spaced\n");
	return QPLUG_FAILURE; 
}

//	..2nd item is option 'time_interval'

 QdObject_var dobj1 = (* call_list)[1];
 start_end = get_timeinterval(dobj1);
 
//	..3rd item is option 'npoles'


QdObject_var opt_ptr = (* call_list)[2];
QdRScalar_var option_np = QdRScalar_var::narrow(opt_ptr);

if ( option_np.is_nil() ) { 
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - npoles must be integer.\n");
	return QPLUG_FAILURE; 
}

npoles = (int)(*option_np);
 
if (npoles <= 0) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - npoles must be greater than 0.\n");
	return QPLUG_FAILURE; 
}


//	..3rd item is option 'nfreq'

opt_ptr = (* call_list)[3];
QdRScalar_var option_nf = QdRScalar_var::narrow(opt_ptr);

if ( option_nf.is_nil() ) { 
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - nfreq must be integer.\n");
	return QPLUG_FAILURE; 
}

nfreq = (int)(*option_nf);
if (nfreq <= 0) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - nfreq must be greater than 0.\n");
	return QPLUG_FAILURE; }

//	..4th item is option 'linlog'


opt_ptr = (* call_list)[4];
QdRScalar_var option_ll = QdRScalar_var::narrow(opt_ptr);

if ( option_ll.is_nil()  ) { 
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - linlog must be integer.\n");
	return QPLUG_FAILURE; 
}

linlog = (int)(*option_ll);
if (linlog != 0 && linlog != 1) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - linlog must be 0 or 1.\n");
	return QPLUG_FAILURE; 
}


//	..5th item is option 'minfreq'

opt_ptr = (* call_list)[5];
QdRScalar_var option_mf = QdRScalar_var::narrow(opt_ptr);

if ( option_mf.is_nil()  ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - minfreq must be float.\n");
	return QPLUG_FAILURE; 
}
minfreq = (float)(*option_mf);

if (linlog > 0 && minfreq <= 0) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - for log spacing minfreq must be > 0.\n");
	return QPLUG_FAILURE; 
}



//	..6th item is option 'maxfreq'

opt_ptr = (* call_list)[6];
QdRScalar_var option_Mf = QdRScalar_var::narrow(opt_ptr);

if ( option_Mf.is_nil() ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - maxfreq must be float.\n");
	return QPLUG_FAILURE; 
}
maxfreq = (float)(*option_Mf);

// now take the subset for the time interval chosen  ###SUBSETTING START

QplugAppendTextDisplay ("MaxEntSpectrum>> Subsetting Time Series...\n");
SequenceMask msk(tt_in->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 
tt_out = new QdTimeSeq();
  
tt_in->get_subsequence(*tt_out, msk);

  
QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(ts_in);
if(!ss.is_nil() ){
    QdRScalarSeq_var  ss_out = new QdRScalarSeq();
    ss->get_subsequence(*ss_out, msk);
    ts_out = (QdObject_var) ss_out;
}
else{
    QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(ts_in);
    if(ms.is_nil()  ){
      QplugAppendTextDisplay ("MaxEntSpectrumFFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }
    if(QuIsThreeVector(ms) ){
      QdRMatrixSeq_var  ms_out = new QdRMatrixSeq();
      ms->get_subsequence(*ms_out, msk);
      ts_out = (QdObject_var) ms_out;
    }
    else {
      QplugAppendTextDisplay ("MaxEntSpectrumFFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }

}
 

if ( ts_out.is_nil() ){
    QplugAppendTextDisplay ("MaxEntSpectrumFFT>> ERROR - subset failed:\n");
    return QPLUG_FAILURE;
} 
if ( ts_out->sequence_size() <1 ){
    QplugAppendTextDisplay ("MaxEntSpectrumFFT>> Empty subset:\n");
    return QPLUG_FAILURE;
}

ts_in->copy_xrefs_to(ts_out);
set_timetags(ts_out, tt_out);

//	..extract the time tags from the data series object ###SUBSETTING END

Ndata = tt_out->sequence_size();

if (Ndata <= 1) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; }

if (npoles > Ndata) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - npoles must be smaller than number of data points.\n");
	return QPLUG_FAILURE; 
}
if (nfreq > Ndata) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - nfreq must be smaller than number of data points.\n");
	return QPLUG_FAILURE; }



//	..extract the time tags and data from the data series object

tags  = new double[Ndata];
data  = new float[Ndata];

if ( QuIsThreeVector(ts_out) ) {
  vectorset=1;	
  QplugAppendTextDisplay ("MaxEntSpectrum>> Decomposing Vector Timeseries...\n");
  ts_inm = QdRMatrixSeq_var::narrow(ts_out);
  if ( ts_inm.is_nil() ) { return QPLUG_FAILURE; }
  ts_inr = QdRScalarSeq_var::narrow(QuGetVectorComponent(ts_out, "r(mag)")); 
}
else
 ts_inr = QdRScalarSeq_var::narrow(ts_out);



for (int i=0; i<Ndata; i++){
  tags[i] = (double) tt_out[i].getCDFepoch();
  data[i] = (float) (ts_inr[i]);
}
  
// get time interval in secs and compare with timeinterval of half the set :

Ndata2=Ndata/2;

deltat=(tags[Ndata-1]-tags[0])/(Ndata-1)/1000.0;
deltat2=(tags[Ndata2-1]-tags[0])/(Ndata2-1)/1000.0;
if (fabs(deltat-deltat2) > 0.001 ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - time series not evenly spaced, or data gaps\n");
        delete[] tags;
	return QPLUG_FAILURE; 
}

// check Nyquist interval

if (fabs(minfreq*deltat) > 0.5) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - |minfreq*deltat| not <0.5 \n");
        delete[] tags;
	return QPLUG_FAILURE; 
}
if (fabs(maxfreq*deltat) > 0.5) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - |maxfreq*deltat| not <0.5 \n");
        delete[] tags;
	return QPLUG_FAILURE; 
}
if (maxfreq <= minfreq) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - maxfreq must be greater than minfreq.\n");
        delete[] tags;
	return QPLUG_FAILURE; 
}

QplugAppendTextDisplay ("MaxEntSpectrum>> Calculating Spectrum.\n");
//	call the NR memcof power spectrum calculation
//      (notice that NR routines start indices with 1!!) 

d = new float[npoles];
freq = new float[nfreq];

if (linlog > 0) 
  freqstep=(log(maxfreq)-log(minfreq))/nfreq;
else 
  freqstep=(maxfreq-minfreq)/nfreq;

for(i=0;i<nfreq;i++) { 
 if (linlog > 0) 
  freq[i]=exp(freqstep*i+log(minfreq));
 else
  freq[i]=freqstep*i+minfreq;
}

Qcoef(data, Ndata, npoles, &xms, d);

//	..create the output data series objects


out_freq_series  = new QdRScalarSeq( nfreq );
out_spmag_series  = new QdRScalarSeq( nfreq );
for(int i=0; i<nfreq; i++) {
  out_freq_series[i] =  (double) freq[i];
  out_spmag_series[i] =  (double) Qevlp(freq[i]*deltat, d, npoles, xms);
}



if(vectorset) { 
 QdRMatrixSeq_var spvec_series = new QdRMatrixSeq(nfreq,RMatrix(Dim(3)));
 
 for(long i=0; i<Ndata; i++)  data[i] = (float) (ts_inm[i][0]);
 Qcoef(data, Ndata, npoles, &xms, d);
 for (long i =0; i<nfreq; i++) spvec_series[i][0] = (double) Qevlp(freq[i]*deltat, d, npoles, xms);

 for(long i=0; i<Ndata; i++)  data[i] = (float) (ts_inm[i][1]);
 Qcoef(data, Ndata, npoles, &xms, d);
 for (long i =0; i<nfreq; i++) spvec_series[i][0] = (double) Qevlp(freq[i]*deltat, d, npoles, xms);

 for(long i=0; i<Ndata; i++)  data[i] = (float) (ts_inm[i][2]);
 Qcoef(data, Ndata, npoles, &xms, d);
 for (long i =0; i<nfreq; i++) spvec_series[i][0] = (double) Qevlp(freq[i]*deltat, d, npoles, xms);
 out_spvec_series = (QdObject_var) spvec_series;
}


//	...set some attributes

QuSetTxtAttr("LABLAXIS", "Maximum Entropy Power", (QdObject_var) out_spmag_series);
QuSetTxtAttr("FIELDNAM", "Maximum Entropy Spectrogram", (QdObject_var) out_spmag_series);
QuSetFrame((QdObject_var) out_spmag_series, "scalar");
QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_spmag_series);
sprintf(str_out,"(%s)^2\n", QuGetAttrText("UNITS", ts_in).c_str());
QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_spmag_series);


 
if(vectorset) {
    QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_spvec_series);
    QuSetTxtAttr("LABLAXIS", "MaxEnt Spectrogram", (QdObject_var) out_spvec_series);
    QuSetTxtAttr("FIELDNAM", "MaxEnt Spectrogram", (QdObject_var) out_spvec_series);
    QuSetFrame((QdObject_var) out_spvec_series,  "vector", "na", "na");
    QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_spvec_series);
    QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_spvec_series);
}

QuSetTxtAttr("LABLAXIS", "frequency", (QdObject_var) out_freq_series);
QuSetTxtAttr("FIELDNAM", "frequency", (QdObject_var) out_freq_series);
QuSetFrame((QdObject_var) out_freq_series, "scalar");
QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_freq_series);
QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_freq_series);


//	..add them to the list

  call_list->push_back((QdObject_var) out_freq_series);     
  call_list->push_back((QdObject_var) out_spmag_series);     
  call_list->push_back( (QdObject_var) out_spvec_series); 
  
//	..some messages to the analysis interface window..

sprintf(str_out, "MaxEntSpectrum>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"MaxEntSpectrum>> %ld frequencies\n", nfreq);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"MaxEntSpectrum>> mean square discrepancy : %f \n", xms);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("MaxEntSpectrum>> finished...\n");

    


//	..clean up and exit

delete[] data;
delete[] tags;
delete[] d;
delete[] freq;

return QPLUG_SUCCESS;


}


float Qevlp(float fdt, float d[], unsigned long m, float xms)
{
	int i;
	float sumr=1.0,sumi=0.0;
	double wr=1.0,wi=0.0,wpr,wpi,wtemp,theta;

	theta=6.28318530717959*fdt;
	wpr=cos(theta);
	wpi=sin(theta);
	for (i=0;i<m;i++) {
		wr=(wtemp=wr)*wpr-wi*wpi;
		wi=wi*wpr+wtemp*wpi;
		sumr -= d[i]*wr;
		sumi -= d[i]*wi;
	}
	return xms/(sumr*sumr+sumi*sumi);
}

void Qcoef(float data[], unsigned long n, unsigned long m, float *xms, float d[])
{
	int k,j,i;
	float p=0.0,*wk1,*wk2,*wkm;

	wk1= (float *) malloc( sizeof(float) * n );
	wk2= (float *) malloc( sizeof(float) * n );
	wkm= (float *) malloc( sizeof(float) * m );
	for (j=0;j<n;j++) p += SQR(data[j]);
	*xms=p/n;
	wk1[0]=data[0];
	wk2[n-2]=data[n-1];
	for (j=1;j<n-1;j++) {
		wk1[j]=data[j];
		wk2[j-1]=data[j];
	}
	for (k=0;k<m;k++) {
		float num=0.0,denom=0.0;
		for (j=0;j<(n-k-1);j++) {
			num += wk1[j]*wk2[j];
            denom += SQR(wk1[j]);
            denom += SQR(wk2[j]);
		}
		d[k]=2.0*num/denom;
		*xms *= (1.0-SQR(d[k]));
		for (i=0;i<k;i++)
			d[i]=wkm[i]-d[k]*wkm[k-i-1];
		if (k == m-1) {
			free(wk1);
			free(wk2);
			free(wkm);
			return;
		}
		for (i=0;i<=k;i++) wkm[i]=d[i];
		for (j=0;j<(n-k-2);j++) {
			wk1[j] -= wkm[k]*wk2[j];
			wk2[j]=wk2[j+1]-wkm[k]*wk1[j+1];
		}
	}
	printf("never get here in Qcoef.\n");
}



