// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Plasma calculator
// 
//
// Inputs [optional]:
// 	bfield		- vector time series with magnetic field
//      density         - plasma density
//      velocity        - plasma velocity
// 	t_perp   	- scalar time series with perpendicular temp
// 	t_par	        - scalar time series with parallelle plasma temp
//
// Outputs
//      P_b             - magnetic pressure (B^2/2my0)
//      P_pperp         - plasma pressure (nkT_perp)
//      P_p             - full plasma pressure (nkT_par +2nkT_perp)/3
//      P_tot           - magnetic + plasma pressure
//      Bal             - pressure balance = P_b / P_pperp
//      A_p             - plasma anisotropy (T_par/T_perp - 1).
//      gyro            - ion gyro radius
//      inert           - ion inertial length
//      valf            - local Alfven speed
//      cs              - sound speed
//      fpe             - plasma frequency
//
// SEH, 08 Jan, 2003
//      14 Feb, 2003 - added rho as output
//      20 Nov, 2003 - added ion gyro radii and inertial length as output
//      26 Dec, 2003 - QSAS 2.0 version
//      07 Oct, 2004 - added Mach number + Alfven speed 
//      28 Jun, 2006 - added fce + cs (electron freq + sound speed
//      01 Feb, 2008 - reorganized + added temp tensor (for i.e. Themis), removed rho, mack, (1-alpha)
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

typedef struct VEC {double x,y,z;} vec;
double vecsum(double x, double y, double z);
float sqr(float x);

extern "C" QplugReturnStatus  PlasmaParams  ( QplugArgList *);

using namespace QSAS;

QplugReturnStatus  PlasmaParams  ( QplugArgList *call_list )

{

 QdRMatrixSeq_var  bfield,babs,rho;
 QdRScalarSeq_var  P_b, P_p, P_pfull, P_tot, Beta, A_p, gyro, inert, Valf, Mach, Alpha, OneAlpha,Entropy;
 bool    Bsup=false, Nsup=false, Vsup=false, Tparsup=false, Tperpsup=false, Ttenssup=false;
 long    n_vecs,nestsize;
 double  L0, btot2,theta,deltat,avg_width,Vsw,TotalEnergy=0.0,Bt;
 double  ionMass= 1.67E-6;                                         
 int     nseq =0, triage_cnt=0,j, joinSeq[10];
 vec     B,V;
 char    txtbuf[200];


 // --- establish what inputs are -------

 QdRMatrixSeq_var B_in = QdRMatrixSeq_var::narrow((*call_list)[0]);
 if (!B_in.is_nil()) {
   if (!is_empty(B_in)) {
     Bsup=true;
     nseq++;
     QplugAppendTextDisplay("B-field supplied\n");
   } 
 }

 QdRMatrixSeq_var V_in = QdRMatrixSeq_var::narrow((*call_list)[1]);
 if (!V_in.is_nil()) {
   if (!is_empty(V_in)) {
     Vsup=true;
     nseq++;
     QplugAppendTextDisplay("Velocity supplied\n");
   } 
 }

 
 QdRScalarSeq_var N_in = QdRScalarSeq_var::narrow((*call_list)[2]);
 if (!N_in.is_nil()) {
   if (!is_empty(N_in) ){
     QplugAppendTextDisplay("Density supplied\n");
     Nsup=true;
     nseq++;
   }  
 }


 QdRScalarSeq_var Tpar_in = QdRScalarSeq_var::narrow((*call_list)[3]);
 QdRScalarSeq_var Tperp_in = QdRScalarSeq_var::narrow((*call_list)[4]);
 QdRMatrixSeq_var Ttens_in =  QdRMatrixSeq_var::narrow((*call_list)[3]);
 if (!Tperp_in.is_nil()) {
   QplugAppendTextDisplay("T_perp supplied\n");
   nseq ++;
   Tperpsup=true;
 } 
 if (!Tpar_in.is_nil()) {
   QplugAppendTextDisplay("T_par supplied\n");
   nseq ++;
   Tparsup=true;
 } 
 if (!Ttens_in.is_nil()) {
   QplugAppendTextDisplay("Temp supplied as tensor..\n");
   nseq ++;
   Ttenssup=true;
 }  
 
 
 // --- do we have sufficient input ---
 if (nseq == 0 || (nseq == 1 && Vsup)) {
    QplugAppendTextDisplay("To few input params\n");
    return(QPLUG_FAILURE);
 }


 // ---- check if B and V in same frame ----
 if (Vsup && Bsup){
   int frame_status = QarTestConformalForVecProds((*call_list)[0], (*call_list)[1]);
   if ( frame_status == QAR_ERR ){
     QplugAppendTextDisplay("B field and velocity in different frames\n");
     return QPLUG_FAILURE;
   }
   else if(frame_status == QAR_WARN){
     QplugAppendTextDisplay("Caution: Frame unknown for either B or V\n");
   }
 }




 //  --- first we join B measurements to plasma measurements -----

 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 QdObjectSeq_var restSig = new QdObjectSeq(nseq);
 QdStringSeq_var joinMethod = new QdStringSeq(nseq);
 vector <double> gapValue(nseq);
 QdStringSeq_var gapHandler = new QdStringSeq(nseq);
 QdRMatrixSeq_var  bfield_nest;
 QdRMatrixSeq_var  velocity_nest;
 QdRScalarSeq_var  density_nest;                            
 QdRScalarSeq_var  Tpar_nest;                            
 QdRScalarSeq_var  Tperp_nest;                            
 QdTimeSeq_var ttJoined = get_timetags((*call_list)[1]);     // tt of joined = from plasma
 QdTimeSeq_var tt_out;                                      // tt of masked

 int seq=0;
 if (Bsup) {
   toJoin[seq] = B_in;
   ttJoined = get_timetags(B_in);
   joinSeq[0]=seq;
   seq++;
 }
 if (Vsup) {
   toJoin[seq] = V_in;
   ttJoined = get_timetags(V_in);
   joinSeq[1]=seq;
   seq++;
 }
 if (Nsup) {
   toJoin[seq] = N_in;
   ttJoined = get_timetags(N_in);
   joinSeq[2]=seq;
   seq++;
 }
 if (Tparsup) {
   toJoin[seq] = Tpar_in;
   ttJoined = get_timetags(Tpar_in);
   joinSeq[3]=seq;
   seq++;
 }
 if (Tperpsup) {
   toJoin[seq] = Tperp_in;
   ttJoined = get_timetags(Tperp_in);
   joinSeq[4]=seq;
   seq++;
 }
 if (Ttenssup) {
   toJoin[seq] = Ttens_in;
   ttJoined = get_timetags(Ttens_in);
   joinSeq[3]=seq;
   seq++;
 }
 
 for (int i=0; i<nseq; i++) {
   joined[i]  = toJoin[i]->create();
   joinMethod[i] = QdString(LINEAR);
   gapValue[i] = avg_width;
   gapHandler[i] = QdString(LINEAR);
 }

 //  ---- get Time interval and mask ----------------

 QdTimeInterval_var start_end = get_timeinterval((* call_list)[5]);
 if(start_end.is_nil() ) {  
   QplugAppendTextDisplay("Time interval missing\n");
   return(QPLUG_FAILURE);
 }
 SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 tt_out = new QdTimeSeq();
 ttJoined->get_subsequence(*tt_out, msk);
 if(tt_out.is_nil() || tt_out->sequence_size() <= 0) {
   QplugAppendTextDisplay("No data in interval\n");
   return(QPLUG_FAILURE);
 }
 printf("Mask OK..\n");
 fflush(stdout);
 QplugAppendTextDisplay("Mask OK\n");

 try {
   QuMultiJoin(toJoin, tt_out, joined, gapValue, gapHandler, joinMethod);
 }
 catch(Exception &e) {
   e.print_msg();
   return QPLUG_FAILURE;
 }

 printf("Join OK..\n");
 fflush(stdout);
 QplugAppendTextDisplay("Join OK\n");


 // -- B field ---
 if (Bsup) {
   QdRMatrixSeq_var masked_B = QdRMatrixSeq_var::narrow(joined[joinSeq[0]]);
   if(masked_B.is_nil() || masked_B->sequence_size() <= 0) {
     QplugAppendTextDisplay("No B data after joining\n");
     return(QPLUG_FAILURE);
   }
   bfield_nest = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_B, "1.0e-9>T")); // ensure units are in those assumed

   if(bfield_nest.is_nil() || bfield_nest->sequence_size() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm B units are 1.0e-9>T\n");
     bfield_nest = (QdRMatrixSeq_var) masked_B;
   }                         
   nestsize = bfield_nest->sequence_size();
   printf("Nestsize B = %ld\n",nestsize);
   fflush(stdout);
 }


 // -- Velocity field ---
 if (Vsup) {
   QdRMatrixSeq_var masked_V = QdRMatrixSeq_var::narrow(joined[joinSeq[1]]);
   if(masked_V.is_nil() || masked_V->sequence_size() <= 0) {
     QplugAppendTextDisplay("No Vel data after joining\n");
     return(QPLUG_FAILURE);
   }
  
   velocity_nest = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_V, "1.0e3>m s^-1")); // ensure units are in those assumed
   if(velocity_nest.is_nil() || velocity_nest->sequence_size() == 0) {
     QplugAppendTextDisplay("WARNING: Cannot confirm V units are 1.0e3>m s^-1\n");
     velocity_nest = (QdRMatrixSeq_var) masked_V;
   }
   nestsize = velocity_nest->sequence_size();
   printf("Nestsize V = %ld\n",nestsize);
   fflush(stdout);
 }


 // -- Density ---
 if (Nsup) {
   QdRScalarSeq_var masked_N = QdRScalarSeq_var::narrow(joined[joinSeq[2]]);
   if(masked_N.is_nil() || masked_N->sequence_size() <= 1) {
     QplugAppendTextDisplay("No density data after joining\n");
     return(QPLUG_FAILURE);
   }
   
   nestsize = masked_N->sequence_size();
   printf("Nestsize N = %ld\n",nestsize);
   fflush(stdout);
   density_nest = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_N, "1.0e6>m^-3")); // ensure units are in those assumed
   if(density_nest.is_nil() || density_nest->sequence_size() == 0) {
     QplugAppendTextDisplay("WARNING: Cannot confirm N units are 1.0e6>m^-3\n");
     density_nest = (QdRScalarSeq_var) masked_N;
   }
 }
 

 // -- T_par ---
 if (Tparsup) {
   QdRScalarSeq_var masked_Tpar = QdRScalarSeq_var::narrow(joined[joinSeq[3]]);
   if(masked_Tpar.is_nil() || masked_Tpar->sequence_size() <= 0) {
     QplugAppendTextDisplay("No T par data after joining\n");
     return(QPLUG_FAILURE);
   }

   Tpar_nest = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_Tpar, "1.0e6>K")); // ensure units are in Cluster units assumed
   if(Tpar_nest.is_nil() ) {
     QplugAppendTextDisplay("WARNING: Cannot confirm Tpar units are 1.0e6>K\n");
     Tpar_nest = (QdRScalarSeq_var) masked_Tpar;
   }

   nestsize = Tpar_nest->sequence_size();
   printf("Nestsize Tpar = %ld\n",nestsize);
   fflush(stdout);
 }
 

 // -- T_perp ---
 if (Tperpsup) {
   QdRScalarSeq_var masked_Tperp = QdRScalarSeq_var::narrow(joined[joinSeq[4]]);
   if(masked_Tperp.is_nil() || masked_Tperp->sequence_size() <= 0) {
     QplugAppendTextDisplay("No T perp data after joining\n");
     return(QPLUG_FAILURE);
   }
   
   Tperp_nest = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_Tperp, "1.0e6>K")); // ensure units are in Cluster units assumed
   if(Tperp_nest.is_nil() ) {
     QplugAppendTextDisplay("WARNING: Cannot confirm Tperpunits are 1.0e6>K\n");
     Tperp_nest = (QdRScalarSeq_var) masked_Tperp;
   }
   
   nestsize = Tperp_nest->sequence_size();
   printf("Nestsize Tperp = %ld\n",nestsize);
   fflush(stdout);
 }


 // --- Temp tensor, eventually split this into Tpar, Tperp --------
 if (Ttenssup) {
   QdRMatrixSeq_var masked_Ttens = QdRMatrixSeq_var::narrow(joined[joinSeq[3]]);
   double inv_k_boltzmann = 1.602e-19/1.38054e-17; // [MK/ev]
   string SIcTpar = QuGetAttrText("SI_conversion", ((* call_list)[4]));
   Tpar_nest = new QdRScalarSeq(masked_Ttens->sequence_size());
   Tperp_nest = new QdRScalarSeq(masked_Ttens->sequence_size());
   if ( SIcTpar.empty() || SIcTpar.find("1.602e-19>J") == string::npos ) {
       QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) in Temp tensor");       
   };     
   
   for (int i=0; i<masked_Ttens->sequence_size(); i++) {
     (*Tpar_nest)[i] = inv_k_boltzmann * (double)(*masked_Ttens)[i][2];
     (*Tperp_nest)[i] = inv_k_boltzmann * 0.5 * ((double)(*masked_Ttens)[i][0]+(double)(*masked_Ttens)[i][1]);
   }    

   Tperpsup=true;
   Tparsup=true;
   nestsize = Tperp_nest->sequence_size();
   printf("Nestsize Tperp = %ld\n",nestsize);
   fflush(stdout);
 }


 QplugAppendTextDisplay("Calculating output..\n");

// ---- declare and calculate the output objects -----

 P_b    = new QdRScalarSeq(nestsize);
 P_p    = new QdRScalarSeq(nestsize);
 P_pfull    = new QdRScalarSeq(nestsize);
 P_tot  = new QdRScalarSeq(nestsize);
 Beta   = new QdRScalarSeq(nestsize);
 Alpha  = new QdRScalarSeq(nestsize);
 Entropy = new QdRScalarSeq(nestsize);
 Valf   = new QdRScalarSeq(nestsize);
 A_p    = new QdRScalarSeq(nestsize);
 gyro   = new QdRScalarSeq(nestsize);
 inert  = new QdRScalarSeq(nestsize);
 QdRScalarSeq_var cs  = new QdRScalarSeq(nestsize);
 QdRScalarSeq_var fpe  = new QdRScalarSeq(nestsize);
 babs   = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 rho   = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));

 float pi = 3.14159;        // good enough here....
 float Mu_0 = 1.25664E-06;
 float k = 1.38054e-23;    // Boltzmann constant
 float e = 1.609e-19;      //  elementary charge C (Vs)
 float sfac = 1.0E21;      //  1E6 * 1E6 * 1E9 due to n in 1/ccm and T in units of 1E6K -> nPa
 float mp= 1.67e-27;       //  proton mass
 float me= 9.1091e-31;     //  electron mass
 float vtot2=0.00;

 for(int i=0; i<nestsize; i++)  {
    vtot2 = 0.00;
    if (Bsup) {
      B.x = (double) bfield_nest[i][0] * 1.0E-9;
      B.y = (double) bfield_nest[i][1] * 1.0E-9;
      B.z = (double) bfield_nest[i][2] * 1.0E-9;
      btot2 = (B.x*B.x + B.y*B.y + B.z*B.z);
      P_b[i] = btot2 / (2.0 * Mu_0) * 1.0E9;                                    // Pb in nPa;
     }
    if (Nsup) {
      inert[i] = 227.0/sqrt((double) density_nest[i]);
      fpe[i] =8968.0*sqrt((double) density_nest[i]);
      if (Bsup) Valf[i] = 21.83e9 * sqrt(btot2)/sqrt((double) density_nest[i]); // assuming ions==protons
    }
    if (Nsup && Tperpsup) {
      P_p[i] = (double) density_nest[i] * k * (double) Tperp_nest[i] * sfac;    // Pp in nPa;
    }
    if (Nsup && Tperpsup && Bsup) {
      	Beta[i] = (double) P_p[i] / (double) P_b[i];                             // Plasma beta
    }
    if (Tperpsup && Bsup) {
      gyro[i] = 1300.0 * sqrt((double)Tperp_nest[i]) / (sqrt(btot2) * 1e09) ;
    }
    if (Nsup && Tperpsup && Tparsup) {
      P_pfull[i] = (double) density_nest[i] * k * (2.0 * (double) Tperp_nest[i] + (double) Tpar_nest[i])/3.0 * sfac;    // Ppfull in nPa;
      A_p[i] = (double) Tpar_nest[i]/ (double) Tperp_nest[i] - 1.00;            // temp anisotropy
      if (Bsup) {
	P_tot[i] = P_p[i] + P_b[i];
      }
    }
    if (Vsup) {
      for (j=0; j<3; j++) {
	vtot2 += (double) (*velocity_nest)[i][j] * (double)  (*velocity_nest)[i][j];
      }  
    }
    if (Nsup && Tparsup && Tperpsup && Bsup) {
      Alpha[i] = (double) density_nest[i] * k * sfac * ((double) Tpar_nest[i] - (double)Tperp_nest[i]) /  (btot2 / (Mu_0) * 1.0E9);       
      Entropy[i] = (double) P_pfull[i] / pow(((double) density_nest[i]),(5.0/3.0));   
      cs[i] = 989.0 * sqrt((double) (*P_pfull)[i]/(double) density_nest[i]);
    }    
 }    

 // printf("Completed calculations..\n");

 // --- send some output to QSAS window ---
 sprintf(txtbuf,"Nest size (# of data points) = %ld",nestsize);
 QplugAppendTextDisplay(txtbuf);


 // --- insert timetags + attributes and push to WL ---


 QdObject_var empty;//,empty2,empty3,empty4,empty5,empty6,empty7,empty8,empty9,empty10,empty11,empty12,empty13;
 seq=0;
 if (Bsup) {
   set_timetags(P_b,tt_out);
   QuSetTxtAttr("SI_conversion","1E-9 Pa",(QdObject_var) P_b);
   QuSetTxtAttr("UNITS","nPa",(QdObject_var) P_b);
   QuSetTxtAttr("LABLAXIS","Mag_press",(QdObject_var) P_b);
   QuSetTxtAttr("CATDESC","Magnetic pressure",(QdObject_var) P_b);
   call_list->push_back((QdObject_var) P_b);
 } else  call_list->push_back(empty);

 if (Nsup && (Tperpsup)) {
   set_timetags(P_p,tt_out);
   QuSetTxtAttr("SI_conversion","1E-9 Pa",(QdObject_var) P_p);
   QuSetTxtAttr("UNITS","nPa",(QdObject_var) P_p);
   QuSetTxtAttr("LABLAXIS","Plasma_press",(QdObject_var) P_p);
   QuSetTxtAttr("CATDESC","Plasma pressure P = nkT_perp",(QdObject_var) P_p);
   call_list->push_back((QdObject_var) P_p);
 }  else  call_list->push_back(empty);

 if (Nsup && Tperpsup && Tparsup) {
   set_timetags(P_pfull,tt_out);
   QuSetTxtAttr("SI_conversion","1E-9 Pa",(QdObject_var) P_pfull);
   QuSetTxtAttr("UNITS","nPa",(QdObject_var) P_pfull);
   QuSetTxtAttr("LABLAXIS","Plasma_press",(QdObject_var) P_pfull);
   QuSetTxtAttr("CATDESC","Plasma pressure P = nkT",(QdObject_var) P_pfull);
   call_list->push_back((QdObject_var) P_pfull);
 } else  call_list->push_back(empty);

 if (Bsup && Nsup && Tparsup && Tperpsup) {
   set_timetags(P_tot,tt_out);
   QuSetTxtAttr("SI_conversion","1E-9 Pa",(QdObject_var) P_tot);
   QuSetTxtAttr("UNITS","nPa",(QdObject_var) P_tot);
   QuSetTxtAttr("LABLAXIS","Pp_tot",(QdObject_var) P_tot);
   QuSetTxtAttr("CATDESC","Total plasma pressure (Pp + Pp_perp",(QdObject_var) P_tot);
   call_list->push_back((QdObject_var) P_tot);
 } else  call_list->push_back(empty);

 if (Bsup && Nsup && Tperpsup) {
   set_timetags(Beta,tt_out);
   QuSetTxtAttr("LABLAXIS","Beta",(QdObject_var) Beta);
   QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) Beta);
   call_list->push_back((QdObject_var) Beta);
   QplugAppendTextDisplay("4 - Bsup Nsup & Tperp-> Beta");
 } else  call_list->push_back(empty);

 if (Tparsup && Tperpsup) {
   set_timetags(A_p,tt_out);
   QuSetTxtAttr("LABLAXIS","A_p",(QdObject_var) A_p);
   QuSetTxtAttr("CATDESC","Temperature anisotropy",(QdObject_var) A_p);
   call_list->push_back((QdObject_var) A_p);
 } else  call_list->push_back(empty);

 if (Nsup && Tparsup && Tperpsup && Bsup) {
   set_timetags(Alpha,tt_out);
   QuSetTxtAttr("LABLAXIS","Alpha",(QdObject_var) Alpha);
   QuSetTxtAttr("CATDESC","Alpha",(QdObject_var) Alpha);
   QuSetTxtAttr("UNITS","NA",(QdObject_var) Alpha);
   call_list->push_back((QdObject_var) Alpha);

   set_timetags(Entropy,tt_out);
   QuSetTxtAttr("LABLAXIS","Entropy",(QdObject_var) Entropy);
   QuSetTxtAttr("CATDESC","Entropy",(QdObject_var) Entropy);
   call_list->push_back((QdObject_var) Entropy);
   
   set_timetags(cs,tt_out);
   QuSetTxtAttr("LABLAXIS","CS",(QdObject_var) cs);
   QuSetTxtAttr("Frame","scalar>na",(QdObject_var) cs);
   QuSetTxtAttr("UNITS","m s^-1",(QdObject_var) cs);
   QuSetTxtAttr("CATDESC","Sound speed",(QdObject_var) cs);
   call_list->push_back((QdObject_var) cs);
 } else {
   call_list->push_back(empty);
   call_list->push_back(empty);
   call_list->push_back(empty);
 } 
  
 if (Nsup) {
   set_timetags(fpe,tt_out);
   QuSetTxtAttr("LABLAXIS","fpe",(QdObject_var) fpe);
   QuSetTxtAttr("Frame","scalar>na",(QdObject_var) fpe);
   QuSetTxtAttr("UNITS","s^-1",(QdObject_var) fpe);
   QuSetTxtAttr("CATDESC","Plasma frequency",(QdObject_var) fpe);
   call_list->push_back((QdObject_var) fpe);

   set_timetags(inert,tt_out);
   QuSetTxtAttr("LABLAXIS","Ion inertial length",(QdObject_var) inert);
   QuSetTxtAttr("Frame","scalar>na",(QdObject_var) inert);
   QuSetTxtAttr("UNITS","km",(QdObject_var) inert);
   QuSetTxtAttr("CATDESC","Ion inertial length",(QdObject_var) inert);
   call_list->push_back((QdObject_var) inert);
 } else {
   call_list->push_back(empty);
   call_list->push_back(empty);
 }   

 if (Tperpsup && Bsup) {
   set_timetags(gyro,tt_out);
   QuSetTxtAttr("LABLAXIS","GyroRadius",(QdObject_var) gyro);
   QuSetTxtAttr("Frame","scalar>na",(QdObject_var) gyro);
   QuSetTxtAttr("UNITS","km",(QdObject_var) gyro);
   QuSetTxtAttr("CATDESC","Ion gyro radius",(QdObject_var) gyro);
   call_list->push_back((QdObject_var) gyro);
 }else {
   call_list->push_back(empty);
 }
 
 if (Nsup && Bsup) {
   set_timetags(Valf,tt_out);
   QuSetTxtAttr("LABLAXIS","Alfven Velocity",(QdObject_var) Valf);
   QuSetTxtAttr("Frame","scalar>na",(QdObject_var) Valf);
   QuSetTxtAttr("UNITS","km/s",(QdObject_var) Valf);
   QuSetTxtAttr("CATDESC","Alfven velocity",(QdObject_var) Valf);
   call_list->push_back((QdObject_var) Valf);
 }else {
   call_list->push_back(empty);
 }
 

 return QPLUG_SUCCESS;

}



// ---------------------------------------------------------------------------
/*!
  \brief calculates the length of a vector
  \param x,y,z - vector time series with magnetic field
*/
// -----------------------------------------------------------------------------------------
double vecsum(double x, double y, double z)
{
    return sqrt(x*x +y*y +z*z);
}


// ---------------------------------------------------------------------------
/*!
  \brief calculates x*x
  \param x - float to be squared
*/
// -----------------------------------------------------------------------------------------
float sqr(float x)
{
    return(x*x);
}


// ---------------------------------------------------------------------------
/*!
  \brief dummy
*/
// -----------------------------------------------------------------------------------------
int main()
{
  // do nothing
}


