/****************************************************************************
							LICENSE

  sf_PommeForChamp -  plug-in for the QSAS software.
  Copyright (C) 2010 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file sf_pomme_champ.cc
 * 
 * \brief
 * 
 * 
 */

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "qar.h"
#include "Qdos.h"

#include "./pomme.h"
#include "./POMME7_ext.h"

using namespace QSAS;
using namespace std;

/**
 * The class Status has the purpose to print messages either to user or to the Console.
 */
class Status
{
public:
	static const bool kPrintToUserAndConsole = true;
	static const bool kPrintToConsoleOnly = false;
	static void PrintMessage(bool to_user, const char fmt[], ...);
public:
	enum ReturnStatus
	{
		kOk = 0, kUnknownError, kWrongInput, kNoInput, kVectorFieldRequired
	} code;
	std::string message() const;
	inline void set_message(const std::string message)
	{
		message_ = message;
	}
	;
	Status();
	virtual ~Status();
private:
	std::string message_;
};

std::string QplugNewSTR_2string(const char *in_str)
{
	char * str = QplugNewSTR(in_str);
	std::string out_string = str;
	free(str);
	return out_string;
}

// ==========================================================================

const int kCallListIndexOfAlt  		= 0;
const int kCallListIndexOfLat 		= 1;
const int kCallListIndexOfLon 		= 2;
const int kCallListIndexOfTimeRange = 3;
const int kCallListIndexOfIMFBy		= 4;
const int kCallListIndexOfEm		= 5;
const int kCallListIndexOfEst 		= 6;
const int kCallListIndexOfIst 		= 7;
const int kCallListIndexOff107 		= 8;
const int kCallListIndexOfMaxOrder	= 9;

struct TExternalParameters
{
	//printf("IMF-By 35 minutes earlier (enter '0' if unknown)? ");
	double IMF_By;
	//printf("Merging electric field, 60 minutes earlier (enter '0.5' if unknown)? ");
	double Em;
	//printf("E_st (enter '0' if unknown)? ");
	double E_st;
	//printf("I_st (enter '0' if unknown)? ");
	double I_st;
	//printf("F10.7, 81-day average, 20 months earlier (enter '120' if unknown)? ");
	double f107;
	int oder_internal_field;
};

QplugReturnStatus getPositions(QplugArgList *call_list, SequenceMask &mask
		, QdRScalarSeq_var &geo_alt, QdRScalarSeq_var &geo_lat, QdRScalarSeq_var &geo_lon);

QplugReturnStatus compute_pomme6model_in_nec_frame(QdRScalarSeq_var &geo_alt, QdRScalarSeq_var &geo_lat
		, QdRScalarSeq_var &geo_lon, TExternalParameters &parameters, QdRMatrixSeq_var &B_field);

QplugReturnStatus get_external_parameters(QplugArgList *call_list, TExternalParameters &parameters);

// ==========================================================================
extern "C" QplugReturnStatus sf_pomme_champ(QplugArgList *call_list)
{
//	   INPUT   SCALAR_TIME_SERIES geo_alt,
//	   INPUT   SCALAR_TIME_SERIES geo_lat,
//	   INPUT   SCALAR_TIME_SERIES geo_lon,
//	   INPUT  TIME_INTERVAL      start_end,
//	   INPUT  FLOAT              IMF_By,
//	   INPUT  FLOAT              Em,
//	   INPUT  FLOAT              E_st,
//	   INPUT  FLOAT              I_st,
//	   INPUT  FLOAT              f107,
//	   INPUT  INTEGER            model_no,
//	   OUTPUT VECTOR_TIME_SERIES B_field
	QplugReturnStatus status = QPLUG_SUCCESS;

	// 1. Importing data into memory --------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "1. Extracting data from the working list\n");

	QdTimeInterval_var start_end = get_timeinterval((*call_list)[kCallListIndexOfTimeRange]);
	if (start_end.is_nil()) {
		status = QPLUG_FAILURE;
		Status::PrintMessage(Status::kPrintToUserAndConsole, "error: Could not get the time interval.\n");
	}
	QdRScalarSeq_var geo_alt = NULL, geo_lat = NULL, geo_lon = NULL;
	QdTimeSeq_var tt_out = NULL;
	if (status == QPLUG_SUCCESS) {
		QdTimeSeq_var time_tags = get_timetags((*call_list)[kCallListIndexOfAlt]);
		SequenceMask mask(time_tags->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
		tt_out = new QdTimeSeq(); // tt of masked V and Pos
		time_tags->get_subsequence(*tt_out, mask);

		status = getPositions(call_list, mask, geo_alt, geo_lat, geo_lon);
		if ((status == QPLUG_SUCCESS) && (geo_alt.is_nil() || geo_lat.is_nil() || geo_lon.is_nil())) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: could not read data. Contact the developer.\n");
		}
	}
	// Extract external parameters
	TExternalParameters external_params;
	if (status == QPLUG_SUCCESS)
		status = get_external_parameters(call_list, external_params);

	// 2. Computing model -------------------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "2. Computing the POMME model.\n");
	QdRMatrixSeq_var B_field = NULL;
	if (status == QPLUG_SUCCESS) {
		try {
			status = compute_pomme6model_in_nec_frame(geo_alt, geo_lat, geo_lon, external_params, B_field);
		}
		catch (...) {
			status = QPLUG_FAILURE;
		}
	}

	// 3. Placing output into the working list ----------------------------------------------------
	if ((status == QPLUG_SUCCESS) && (!B_field.is_nil()))
	{
		Status::PrintMessage(Status::kPrintToConsoleOnly, "3. Placing output into the working list\n");
		try{
			//Units and SI conversion:
			QuSetFrame((QdObject_var) B_field, "vector>nec_xyz");
			QuSetTxtAttr(UNITS, "nT", (QdObject_var) B_field);
			QuSetTxtAttr(SI_CONVERSION, "1.0E-9>T",	(QdObject_var) B_field);
			QuSetTxtAttr(LABLAXIS, "POMME6 B field", (QdObject_var) B_field);
			QuSetTxtAttr(FIELDNAM, "POMME6 B field", (QdObject_var) B_field);
			set_timetags((QdObject_var) B_field, tt_out);
		}
		catch (...) {
			status = QPLUG_FAILURE;
		}
	}

	if (status == QPLUG_SUCCESS) {
		try {
			Status::PrintMessage(Status::kPrintToConsoleOnly, "4. Push back\n");
			call_list->push_back((QdObject_var) B_field);
		}
		catch (...) {
			status = QPLUG_FAILURE;
		}
	}
	return status;
}

QplugReturnStatus getOneScalaSeq(QplugArgList *call_list, int call_list_index
		, SequenceMask &mask, QdRScalarSeq_var &scalar_ts)
{
	QplugReturnStatus status = QPLUG_SUCCESS; //success
	QdRScalarSeq_var scalar_full = QdRScalarSeq_var::narrow((*call_list)[call_list_index]);
	if (scalar_full.is_nil()) {
		status = QPLUG_FAILURE;
		Status::PrintMessage(Status::kPrintToUserAndConsole, "error: Input is not a sequence of scalars.\n");
	}
	else {
		scalar_ts = new QdRScalarSeq();
		scalar_full->get_subsequence(*scalar_ts, mask);
		scalar_ts->copy_xrefs_from((*call_list)[call_list_index]);
	}
	return status;
}

QplugReturnStatus getPositions(QplugArgList *call_list, SequenceMask &mask
		, QdRScalarSeq_var &geo_alt, QdRScalarSeq_var &geo_lat, QdRScalarSeq_var &geo_lon)
{
	QplugReturnStatus status = QPLUG_SUCCESS; //success
	size_t nr_measurements = 0;
	//get altitude
	if (status == QPLUG_SUCCESS) {
		status = getOneScalaSeq(call_list, kCallListIndexOfAlt, mask, geo_alt);
		if ((status == QPLUG_FAILURE) || geo_alt.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand altitude time series.\n");
		}
		else {
			nr_measurements = geo_alt->sequence_size();
		}
	}
	//get latitude
	if (status == QPLUG_SUCCESS) {
		status = getOneScalaSeq(call_list, kCallListIndexOfLat, mask, geo_lat);
		if ((status == QPLUG_FAILURE) || geo_lat.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand latitude time series.\n");
		}
		else {
			if (nr_measurements != geo_lat->sequence_size()) {
				status = QPLUG_FAILURE;
				Status::PrintMessage(Status::kPrintToUserAndConsole,
						"error: Latitude time tags disagree with those of altitude.\n");
			};
		}
	};
	//get longitude
	if (status == QPLUG_SUCCESS) {
		status = getOneScalaSeq(call_list, kCallListIndexOfLon, mask, geo_lon);
		if ((status == QPLUG_FAILURE) || geo_lon.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand longitude time series.\n");
		}
		else {
			if (nr_measurements != geo_lon->sequence_size()) {
				status = QPLUG_FAILURE;
				Status::PrintMessage(Status::kPrintToUserAndConsole,
						"error: longitude time tags disagree with those of altitude.\n");
			};
		}
	};
	return status;
}

QplugReturnStatus get_external_parameters(QplugArgList *call_list, TExternalParameters &parameters)
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	//IMF By
	QdRScalar_var temp_scalar  = QdRScalar_var::narrow((*call_list)[kCallListIndexOfIMFBy]);
	if (temp_scalar.is_nil()) {
		status = QPLUG_FAILURE;
		Status::PrintMessage(Status::kPrintToUserAndConsole,
				"error: Could not get IMF By. Check that it is a scalar.\n");
	}
	else {
		parameters.IMF_By = *temp_scalar;
	};
	//Em
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOfEm]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Em. Check that it is a scalar.\n");
		} else {
			parameters.Em = *temp_scalar;
		}
	};
	//Est
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOfEst]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Est. Check that it is a scalar.\n");
		} else {
			parameters.E_st = *temp_scalar;
		}
	};
	//Ist
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOfIst]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Ist. Check that it is a scalar.\n");
		} else {
			parameters.I_st = *temp_scalar;
		}
	};
	//F107
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOff107]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get F107. Check that it is a scalar.\n");
		} else {
			parameters.f107 = *temp_scalar;
		}
	};
	//Main field order.
	if (status != QPLUG_FAILURE) {
		temp_scalar = QdRScalar_var::narrow((*call_list)[kCallListIndexOfMaxOrder]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get the max order of the main field. Check that it is an integer.\n");
		} else {
			parameters.oder_internal_field = static_cast<unsigned int>(*temp_scalar);
		}
	};
	return status;
}

// =========  Model functions  =================================================

extern void pomme_geocentric (int control[MAXCONTROL],
		double fday, double r, double lat, double lon,
		double est, double ist, double imf_by, double f107, double Em,  /* time lagged but not peaked or subtracted mean */
		double *bx, double *by, double *bz);

void convert_geoxyz2long_lat_r(double x, double y, double z, double &lon, double &lat, double &r)
{
	double temp = x*x + y*y;
	r = sqrt(temp + z*z);
	lon = atan2(y, x);
	lat = atan2(z, sqrt(temp));
}

QplugReturnStatus compute_pomme6model_in_nec_frame(QdRScalarSeq_var &geo_alt, QdRScalarSeq_var &geo_lat
		, QdRScalarSeq_var &geo_lon, TExternalParameters &parameters, QdRMatrixSeq_var &B_field)
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	B_field = NULL;
	if (geo_alt.is_nil() || geo_lat.is_nil() || geo_lon.is_nil())
	{
		status = QPLUG_FAILURE;
	}
	else
	{
		size_t nr_measurements = geo_lat->sequence_size();
		B_field = new QdRMatrixSeq(nr_measurements, RMatrix(Dim(3), 0.0));

		//Initialize model's variables:
		tcontrol control;        /* array of user-control parameters */

		/* mainfield_open_file_mp initializes the control parameters. You can now modify these: */
		control[POS_0] =  parameters.oder_internal_field;  //133;   /* default: model.deg_int, 0=off */
		control[POS_1] =         15;    /* default: model.deg_1st, 0=off */

		control[POS_SM] =        DEG_SM;       /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_SM_IND] =    DEG_SM;       /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107] =      DEG_F107;     /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107_IND] =  DEG_F107;     /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_EST] =       1;                  /* Est dependent field: 1=on, 0=off */
		control[POS_IST] =       1;                  /* Ist dependent field: 1=on, 0=off */
		control[POS_EST_FAC] =   1;                  /* 1 = use model Est factor, 0 = use factor 1.0 */

		control[POS_GSM] =       DEG_GSM;     /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_GSM_IND] =   DEG_GSM;     /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_IMFBY] =     1;                 /* IMF-By dependent field: 1=on, 0=off */
		control[POS_IMFBY_IND] = 1;                 /* Induction by IMF-By dependent field: 1=on, 0=off */
		control[POS_EM] =        DEG_EM;      /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_EM_IND] =    DEG_EM;      /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */


		// Computation of the model for each position
		QmjdTime day_Jan2000(2000, 1, 1, 0, 0, 0);
		QdTimeSeq_var times = get_timetags(geo_lat);
		double r = 0, fday = 0.0;
		double B_nec[3];   /* Resulting field */
		for (size_t t = 0; (t < nr_measurements) && (status != QPLUG_FAILURE); ++t) {
			// Universal time in floating point days since 00:00:00 on 1/1/2000
			fday = times[t].getDiffDays(day_Jan2000);
			r = (double)(*geo_alt)[t] + 6371.2;

			pomme_geocentric (control, fday, r, (double)(*geo_lat)[t], (double)(*geo_lon)[t], parameters.E_st
					, parameters.I_st, parameters.IMF_By, parameters.f107
					, parameters.Em  /* time lagged but not peaked or subtracted mean */
					, &(B_nec[0]), &(B_nec[1]), &(B_nec[2]) );
			(*B_field)[t][0] = B_nec[0];
			(*B_field)[t][1] = B_nec[1];
			(*B_field)[t][2] = B_nec[2];
		}
	}
	return status;
}

// =========  class Status  =================================================

Status::Status()
{
	message_ = "";
}

Status::~Status()
{
}

string Status::message() const
{
	string result = "sf_PommeForChamp: ";
	result += message_;
	return result;
}

void Status::PrintMessage(bool to_user, const char fmt[], ...)
{
	try {
		va_list args;
		char buffer[2048];
		va_start(args, fmt);
		vsprintf(buffer, fmt, args);
		va_end(args);
		if (to_user)
			QplugAppendTextDisplay(buffer);
		printf("%s", buffer);
		fflush(stdout);
	}
	catch (...) {
	}
}
