
#include "qdutil.h"
#include "qie_caa.h"
#include "QdTime.h"

// global block size


// ------------------------------------------------------------------------------------
//    QieVariable CLASS 
//    -------------------

QieVariable::QieVariable()
{	
	name = " ";
	col_start = -1; // unused for cdf
	col_width = -1; // unused for cdf
//	  ftpackets.timeformat = NOT_A_TIME; // SJS re FT - safe initialisation
	recVary = false;
	datatype = string("CHAR");
	n_posn = -1; // unused for cdf
	cachedFrom = -1; // used to keep rec number for data range read already
	cachedTo = -1;   // used to keep rec number for data range read already
	
	// create dummy xref object to hold attributes until put on WL
	
	Vxref = new QdRScalar(0.0);
}

QieVariable::~QieVariable()
{

}


// -----------------------------------------------------------------------------------------
//    QieFile CLASS  
//    -------------------


/// \fn QieFile::QieFile()
/// \brief Constructor of the QieFile class
/// \param     none
/// \exception None
/// \return    None
///  

QieFile::QieFile()

{

  fp = NULL;
  name = NULL;
  is_gz = false;
  includedFile = NULL;
  fileType = 0;
  data_until = std::string("");
  start_after = std::string("");
  
   // set defaults
  end_of_line = '\n';
  commentMark = '!';
  recNumberingOn = false;
  attr_delim = ',';
  data_delim = ',';
  blockSize = 100;

}

// ------------------------

QieFile::~QieFile()

{
	if(includedFile != NULL) delete includedFile;
	this->closeFile();
	if( name != NULL) free (name);  
  	
}

// ---------------------------

void QieFile::openIncludedFile(const char * fullName, const char *mode){

	std::string incFileName;
	if(includedFile == NULL) {
		// create QieFile object with constructor before attaching file to it
		includedFile = new QieFile();
		includedFile->end_of_line = end_of_line;
		includedFile->commentMark = commentMark;
		includedFile->fileType = fileType;
		
		// use full path in file or same path as parent
		fp=fopen(fullName, "r"); // try name as absolute path
     	if( fp == NULL ) {
			// try with relative path
			incFileName = this->name;
			int ptr = incFileName.rfind("/");
			incFileName.replace(ptr+1, incFileName.length()-ptr-1, fullName);
     	}
     	else{
			// it exists, close and let openFile() do it properly
			fclose(fp);
			incFileName = fullName;
     	}

		includedFile->openFile(incFileName.c_str(), mode);
		if(includedFile->fp == NULL){
			// open failed
			
			QieMsg ("ERROR", "Cannot open data file... \n", incFileName.c_str(), "\n");
			delete includedFile;
			includedFile = NULL;
		}
	}
	else {
		// descend recursively until we find empty pointer 
		// (which will be in current open file)
		includedFile->openIncludedFile(fullName, mode);
	}
 
}
 
// --------------------------------

void QieFile::openFile(const char * fullName, const char *mode, bool overwrite){
	
	// get safe name expanding ~ etc. This uses a static char array, so copy it
	std::string safeName = QieSafePath(&(fullName[0]));
	this->name = (char *) malloc(safeName.length()*sizeof(char) +1);
	strcpy(this->name, safeName.c_str());
	
	if(strcmp(mode,"r") == 0 || strcmp(mode,"rb") == 0) {
		// open file for reading, allow for gz format
 	    if(strstr(this->name, ".cef.gz") != 0 || strstr(this->name, ".CEF.GZ") != 0){
	    	  this->is_gz = true;
	       char * unzipname = (char *) malloc(strlen(this->name) + 13);
	       strcpy(unzipname, "gzip -cd ");
	       strcat(unzipname, this->name);
            fp=popen(unzipname, mode);
	       free(unzipname);
         }
         else{
	       // open regular file
       	  fp=fopen(this->name, mode);
         }
	}
  	else if( !overwrite){ // replaces '-' option in open call

     	/* try reading to see if file of that name exists */
     	fp=fopen(this->name, "r");
     	if( fp == NULL ) {
       		fp=fopen(this->name, mode);
     	}
     	else{
       		/* it exists, overwrite not allowed, bale out */
       		fclose(fp);
       		fp = NULL;
     	}
   	}
   	else{
     	/* open file with mode supplied */
     	fp=fopen(this->name, mode);
   }
     //  pointer this->fp is NULL if open fails
}

// -----------------------------------------------------
// 
// returns true if something was read OK otherwise false
// Leaves fp ready to read next record
 
bool QieFile::skipNextRecord(){
	
	char c;
     // descend include hierarchy
	if (includedFile != NULL) {
	   if( includedFile->skipNextRecord() ) return true;
	   else {
	       // end of file reached
	       includedFile->closeFile();
		   delete includedFile;
		   includedFile = NULL;
	   }
	}
	
	// read from this File ptr
	if(fp == NULL) return false;
	c=fgetc(fp);
	
	// ignore commented line
	while(c == '!') {
		c=skipRestOfLine();	
		return skipNextRecord();
	}
	
    // skip empty lines
	while (c == ' '  || c == '\t' || c == '\n' || c == '\r') {
		if(c == EOF) return false;  // ran out of data without finding end of line
		c=fgetc(fp);
	}
	
    // start of non-commented, non-empty line
	while (c != end_of_line ){
		if(c == EOF) return false;  // ran out of data without finding end of line
		c=fgetc(fp);	
    }
	return true;
}


// ------------------------------------------------------

 bool QieFile::readRecord(string &record){
 	
	// concatenates lines of input up to end of line marker
	// to get full cef record or header line
	// returns false if end of file reached or no file to read

     // descend include hierarchy
	if (includedFile != NULL) {
		
		if( includedFile->readRecord(record) ) {
			return true;	
		}
		else {
			// end of file reached
	       includedFile->closeFile();
		   delete includedFile;
		   includedFile = NULL;
		}
	}
	
	// read from this File ptr
	if(fp == NULL) return false; // will be empty
	record.assign(blockSize, '\0');
	int len=0;
	char endChar;
	while( (endChar = assembleLine(record, len, end_of_line)) != EOF){
		if(data_until != std::string("")){
			if( record.find(data_until) != -1 ) return false; // end of input
		}

		if(endChar == end_of_line) return true;
    }
	// ran out of data without finding end of line
	return false;
}


// ------------------------------------------------------

char QieFile::readLine(string &line_ptr, int &len){
	
	len=0;	
		
	line_ptr.assign(blockSize, '\0');
	
	char c = assembleLine(line_ptr, len, '\n'); // assembleLine can be called recursively
		
	return c; 
 }
 

// ------------------------------------------------------

char QieFile::assembleLine(string &line_ptr, int &posn, char endMark){
	
	// Reads input lines up to '\n' and concatenates on continuation marker
	// Used internally by readRecord() and header reads
    
	// it strips white space from start and adds one space at the end
	// to make concatenation safe.

	// endMark may be '\n' or end of record
	
	int available = line_ptr.size();
	
	
	char c = fgetc(fp);
	// remove leading white space and empty lines
	while(c == ' ' || c == '\t' || c == '\n'|| c == '\r') c = fgetc(fp);
	
	// skip comment/blank lines 
	if(  c == commentMark) {
			c = skipRestOfLine();
			return assembleLine(line_ptr, posn, endMark); // try next line
	}
	
	while(c != endMark &&  c != EOF){
	
		if(  c == '\n' || c == '\r') {
			// skip new line 
			c=fgetc(fp);
			continue;
		}
		if(  c == commentMark) {
			// skip comment after valid line 
			c = skipRestOfLine();
			if(posn == 0) assembleLine(line_ptr, posn, endMark); // try next line
			else return c; // there is something in this line
		}
		if(  c == '\\' ) {
			// skip continuation marker, skip rest of line and add next
			c =skipRestOfLine();
			return assembleLine(line_ptr, posn, endMark); // try next line
		}
		else if(c == '\"') {
			// take quoted text verbatim
			addChar(posn, available, line_ptr, c);
			c=fgetc(fp);
			while( c != '\"'){
				if( c == EOF ) {
					// unclosed quote
					// string may be rest of file, so truncate
					char cutText[301];
					strncpy(&(cutText[0]), line_ptr.c_str(), 100);
					QieMsg("Error",  "Unclosed quotes after... \n", &(cutText[0]), "\n");
					return EOF; 
				}
				addChar(posn, available, line_ptr, c);
				c=fgetc(fp);
			}
			addChar(posn, available, line_ptr, c);
			c=fgetc(fp);
		}	
		else if(c == end_of_line) {
			return end_of_line;
		}
		else{
			// add this character and test next
			addChar(posn, available, line_ptr, c);
			c=fgetc(fp);
		}					
	}
	
	return c; // normally 'endMark'
 }
 

// -------------------------------------------------------

void QieFile::addChar(int & posn, int &available, string &line_ptr, char c){

	if (posn >= available) {
		line_ptr.append(blockSize, '\0');
		available += blockSize;
	}
	
	line_ptr[posn] = c;
	posn++;
	
	return;
}

// -------------------------------------------------------

int QieFile::skipRestOfLine(){
	// ignore input up to the end of line (or file if sooner)
	
	// look for version stamp and file type identifier first
	// as these flags are escaped with !@
	int c = fgetc(fp);
	if(c == '@'){
		string localLine="";
		while(c != '\n'  ){
			if( c != '\r') localLine += c;
			c = fgetc(fp);
		}
		const char *linePtr = localLine.c_str();
		if ( strncmp(linePtr, ".qft", 4) == 0){
         fileType = _QFT;
         QieMsg("WARNING", "version identifier found, qft ");
       }
       else if ( strncmp(linePtr, ".qfd", 4) == 0){
         fileType = _QFD;
         QieMsg("WARNING", "version identifier found, qfd ");
       }
       else if ( strncmp(linePtr, ".cef2", 4) == 0 ){
         fileType = _CAA;
       }
       else if ( strncmp(linePtr, ".cef", 4) == 0 ){
         fileType = _CEF;
         QieMsg("WARNING", "version identifier found,  cef 1 ");
       }
	   
	}
	while(c != '\n' &&  c != EOF) {
		c = fgetc(fp);
	}
	return c;
}

// -------------------------------------------------------

void QieFile::closeFile(){

	if(fileType == _CDF) 
	{
		CDFlib(CLOSE_, CDF_, NULL_);  // safe against multiple closes
		fp = NULL;
	}
	
	if (fp == NULL) return; // trap multiple close on flat files

	if(is_gz) {
		// close compressed file
		pclose(fp);
	}
	else {
		// close regular file
		fclose(fp);
	}

	fp = NULL;
}

// -------------------------------------------------------

void QieFile::rewindFile(){

	if (fp == NULL) return; // trap closed file
	
	if(is_gz) {
		// close compressed file and reopen
 		pclose(fp);
		this->openFile(name,"r");
	}
	else {
		// rewind regular file
 		rewind(fp);
	}
}

// -------------------------------------------------------

void QieFile::setEOL(char eol){
	end_of_line = eol;
	if (includedFile != NULL) includedFile->setEOL(eol);
}

// -------------------------------------------------------

void QieFile::setCMark(char cmark){
	commentMark = cmark;
	if (includedFile != NULL) includedFile->setCMark(cmark);
}

// -------------------------------------------------------

void QieFile::setFileType(int ftype){
	fileType = ftype;
	if (includedFile != NULL) includedFile->setFileType(ftype);
}

// -------------------------------------------------------

void QieFile::setDataUntil(std::string txt){
	data_until = txt.c_str();
	if (includedFile != NULL) includedFile->setDataUntil(txt);
}

// -------------------------------------------------------

void QieFile::setStartAfter(std::string txt){
	start_after = txt.c_str();
	if (includedFile != NULL) includedFile->setStartAfter(txt);
}

// -------------------------------------------------------

void QieFile::setAttrDelim(char c){
	attr_delim = c;
	if (includedFile != NULL) includedFile->setAttrDelim(c);
}

// -------------------------------------------------------

void QieFile::setDataDelim(char c){
	data_delim = c;
	if (includedFile != NULL) includedFile->setDataDelim(c);
}

// -------------------------------------------------------

void QieFile::setRecNumbering(bool recNumOn){
	recNumberingOn = recNumOn;
	if (includedFile != NULL) includedFile->setRecNumbering(recNumOn);
}


// ------------------------------------------------------------------------------------
//    QieData CLASS 
//    -------------------


/// \fn QieData::QieData()
/// \brief Constructor of the QieData class
/// \param     none
/// \exception None
/// \return    None
///  

QieData::QieData()
{
	// start with empty variable and global lists
	QieVars.clear();
	GlobalList = new ObjectNamesTable;
	recordNumEntries = 0;
	d0_var_num = -1;
	inFile.fileType = 0;
}

QieData::~QieData()
{
  inFile.closeFile(); 
  outFile.closeFile(); 
  attrFile.closeFile(); 
  QieVars.clear();
  dataRecords.clear();
  tokenStart.clear();
  tokenLen.clear();
}

bool QieData::readFlatFile(const char * fullName, const char * headerName)
{
	int maxNumRecs;
	// utility to query file (open and read header) then get data for whole file
	if( !queryFile( fullName,  headerName)) return false;

	// reads in flat records and counts max rec number
        maxNumRecs = readDataRecords();

	return true;
	
}

// ----------------------------------------------------------

int QieData::countDataRecords()
{
  // this method is to be called after the header has been read  
  // without rewinding the file (fp is positioned at first record

	long pos = ftell(inFile.fp);
	
	int nrecs = 0;
	while(inFile.skipNextRecord() ) nrecs++;
	
	// return to start of block for reading
	fseek(inFile.fp, pos, SEEK_SET);  
  
	return nrecs;
  
}

// ----------------------------------------------------------

bool QieData::queryFile(const char * fullName, const char * headerName)
{
	if(fullName == NULL) return false;
	if(strcmp(fullName, "") == 0) return false;
	
	string readMode;
#ifdef WINDOWS
	// ftell() and fseek() fail on Windows if file has Unix end of line
	// open in binary mode to force verbatim read of file
	readMode="rb";
#else
	readMode="r";
#endif

	// initialise guess of file type if not set by calling routine
	if(inFile.fileType == 0) {
		const char *fullNameInUpper = QieToUpper(fullName);
		int ext = strlen(fullNameInUpper) - 4;
		if( strcmp(&(fullNameInUpper[ext]), ".CEF") == 0 ) inFile.fileType = _CAA; // CEF1 must be set by calling fn or in file
		else if( strcmp(&(fullNameInUpper[ext-3]), ".CEF.GZ") == 0 ) inFile.fileType = _CAA;
		else if( strcmp(&(fullNameInUpper[ext]), ".CDF") == 0) inFile.fileType = _CDF;
	//	else if( strcmp(&(fullNameInUpper[ext]), ".QFT") == 0) inFile.fileType = _QFT;
	//	else if( strcmp(&(fullNameInUpper[ext]), ".QFD") == 0) inFile.fileType = _QFD;
		else return false; // default
	}
	bool header_OK;

	// branch to cdf reading options

	if(inFile.fileType == _CDF) return queryCDFfile(fullName);
	
	// default is CAA syntax cef file
	// first read detached header if provided
	if(headerName != NULL){
		inFile.openFile(headerName, readMode.c_str());
		if(inFile.fp != NULL){
			header_OK = readHeader();
			inFile.closeFile();	
			inFile.openFile(fullName, readMode.c_str());
		}
		else{
			inFile.openFile(fullName, readMode.c_str());
			header_OK = readHeader();
		}
	}
	else{	
		inFile.openFile(fullName, readMode.c_str());
		if(inFile.fp==NULL) return false;
		header_OK = readHeader();
	}

	// if header was not read, look for header in same directory as file
	if( header_OK ) {
		// Add attributes from helper file
		getAttributeHelper(fullName);
		return true;
	}
	else {
		inFile.closeFile();	
		inFile.openFile(fullName, readMode.c_str()); //reset ready to read data
		
		// read detached header, so attribute helper file is not loaded
		vector <std::string> hdrList = QieFindHelperFiles(fullName, ".qfh");
		const char *hdrName;
		// cannot tell which headers are applicable, so only use first found
		if(hdrList.size() > 0) hdrName = (hdrList[0]).c_str(); 
		else hdrName = "";
		
		if ( !QieStrIsNULL(hdrName) ){
			inFile.closeFile(); // close data file and open detached header
			inFile.openFile(hdrName, readMode.c_str());
			if (inFile.fp != NULL) {
				header_OK = readHeader();
				inFile.closeFile();	
				if( !header_OK ){
					QieMsg ("ERROR", "Cannot open header\n", hdrName);
					return false;
				}
				inFile.openFile(fullName, readMode.c_str());
				return true;
			}
		}
		else{
			QieMsg ("ERROR", "Cannot locate a valid header\n");
			inFile.closeFile(); // close data file
			return false;
		}	
	}
	// This leaves file open and positioned ready to read data	
    return true;

}

// ----------------------------------------------------------

void QieData::getAttributeHelper(const char * fullName)
{
	vector <std::string> helperList = QieFindHelperFiles(fullName, ".qat");

	for (int i=0; i<helperList.size(); i++){
		const char *helperName = (helperList[i]).c_str();

		if ( !QieStrIsNULL(helperName) ){
			attrFile.openFile(helperName, "r");
			if (attrFile.fp != NULL) {
				QieMsg ("INFO:", helperName, "being read for additional attribute information\n"); 
				readAttrHelper();
				attrFile.closeFile();
			}
		}
	}
	return;
} 

// ----------------------------------------------------------

void QieData::readAttrHelper()
{
	char endOfRecord = '\n';
    
 	// CAA defaults may not be written in the file, so set in advance
	attrFile.attr_delim = ',';
	attrFile.data_delim = ',';
	
	// note: readLine skips comment lines 
	std::string nextLine;
	int len;
	while(attrFile.readLine(nextLine, len) != EOF){
		
		if( !parseLine(nextLine, len) ){
			// detect (usually) data lines with no header or Start_data
			if( isdigit(parameter[0]) ) break; // non-fatal error
			else QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
			continue;
		}
		
		
		if(strncmp(QieToUpper(parameter.c_str()), "START_VARIABLE", 10) == 0) {

			int vnum = findVarNumByName(value.c_str());
			// keep reading even if var is not in file being read to skip var block
			
			// read attribute info for this variable
			while(attrFile.readLine(nextLine, len) != EOF){
				if( !parseLine(nextLine, len) ){
					QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
					continue;
				}
				if(strncmp(QieToUpper(parameter.c_str()), "END_VARIABLE", 10) == 0) break;
				if(vnum >= 0){
					 /* read attribute */
					QdObject_var nextAttr;
					QieVariable *V = &(QieVars[vnum]);
					try{
						nextAttr = this->QieText_to_Attr(*V, value); // V needed for type and dimensions
					}
					catch(Exception &e){
						QieMsg( "WARNING", "failed to make attribute from: ", value.c_str());
					}
		 			if( !nextAttr.is_nil() ){
						// REPLACE that in the file if present
						V->Vxref->change_xref(parameter.c_str(), nextAttr);
					}
					
				}
			}
		}
		
		
		if(strncmp(QieToUpper(parameter.c_str()), "COMMENT_MARKER", 14) == 0) {
			attrFile.setCMark(value[0]);
			continue;
		}
		
		if(strncmp(QieToUpper(parameter.c_str()), "ATTRIBUTE_DELIMITER", 19) == 0) {
			attrFile.setAttrDelim(QieGetDelim(value));
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "DATA_DELIMITER", 14) == 0) {
			attrFile.setDataDelim(QieGetDelim(value));
			continue;
		}	

	} // end while
 
}

// ----------------------------------------------------------

bool QieData::readHeader()
{
	int j;
	int i;
	long n;
	char endOfRecord = '\n';
    // initialise some traps
	year_offset = -1;  /* acts as trap if YY format used */
	strictMonotonic = false;
	yp_last=-2000000000;

 		// CAA defaults may not be written in the file, so set in advance
	inFile.attr_delim = ',';
	inFile.data_delim = ',';

	// note: readLine skips comment lines 
	std::string nextLine;
	int len;
	while(inFile.readLine(nextLine, len) != EOF){
		if(inFile.start_after != std::string("") && nextLine.find(inFile.start_after)) {
			QieMsg( "INFO", "Start reading data after line =", nextLine.c_str());
			break; // located start of data, stop reading header
		}
		
		if( !parseLine(nextLine, len) ){
			// detect (usually) data lines with no header or Start_data
			//if( isdigit(parameter[0]) ) break; // non-fatal error
		   // else QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
			continue;
		}
		if(strcmp(QieToUpper(parameter.c_str()), "FILE_FORMAT_VERSION") == 0 && 
		   strstr(QieToUpper(value.c_str()), "CEF-2") != NULL) {
				inFile.setFileType(_CAA);
				continue;
		}
	  
		if(strncmp(QieToUpper(parameter.c_str()), "START_DATA", 10) == 0) {
			QieMsg("INFO", value.c_str(), "records expected");
			break;
		}
		
		if(strncmp(QieToUpper(parameter.c_str()), "DATA_UNTIL", 10) == 0) {
			if(inFile.fileType == 0) inFile.setFileType(_CAA);
			inFile.setDataUntil(QieStripQuotes(value));
			break;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "START_VARIABLE", 10) == 0) {
			if( this->readVariable(value.c_str()) ) continue;
			else return false;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "START_META", 10) == 0) {
			if( this->readGlobal(value.c_str()) ) continue;
			else return false;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "YY_OFFSET", 9) == 0) {
			year_offset = atol(value.c_str()); // global year_offset 
			if(year_offset == 0 && value.find('0') == -1 ) {
				QieMsg("ERROR", "Bad year offset = ", value.c_str());
				return false;
			}
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "YEAR_INCREASING", 17) == 0) {
			if(value[0] == 'y' || value[0] == 'Y')
			strictMonotonic = true;
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "FILE_TYPE", 9) == 0) {
			if(value[0] == 't' || value[0] == 'T' ){
				if(inFile.fileType != _CAA) inFile.setFileType(_QFT);
			}
			else if(value[0] == 'd' || value[0] == 'D' ){
				if(inFile.fileType != _CAA) inFile.setFileType(_QFD);
			}
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "COMMENT_MARKER", 14) == 0) {
			inFile.setCMark(value[0]);
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "START_AFTER", 11) == 0) {
			inFile.setStartAfter(QieStripQuotes(value));
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "ATTRIBUTE_DELIMITER", 19) == 0) {
			inFile.setAttrDelim(QieGetDelim(value));
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "DATA_DELIMITER", 14) == 0) {
			inFile.setDataDelim(QieGetDelim(value));
			continue;
		}	
		if(strncmp(QieToUpper(parameter.c_str()), "END_OF_RECORD_MARKER", 20) == 0) {
			if (value[0] != ' ') {
				std::string delimValue = QieStripQuotes(value);
				if( delimValue[0] == '\n' || delimValue[0] == '\0') 
					endOfRecord = '\n';
				else 
					endOfRecord = QieGetDelim(delimValue);
			}
			continue;

		}
		if(strncmp(QieToUpper(parameter.c_str()), "RECORD_NUMBERING", 16) == 0) {
			if(value.find("on") != -1 && inFile.fileType != _CAA ) inFile.setRecNumbering(true);
			else inFile.setRecNumbering(false);
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "FILE_NAME", 9) == 0) {
			if(QuSlotDebug()) QieMsg("INFO", "file: ", value.c_str());
			continue;
		}

	} // end while
 
	// validate cef 2 inputs found 
	if(inFile.fileType == _CAA){
		if( inFile.data_until == std::string("") )
			QieMsg( "WARNING", "No DATA_UNTIL parameter found in header. ", " Continuing to attempt read" );
		if( inFile.start_after != std::string("") )
			QieMsg( "WARNING", "START_AFTER parameter should not be used in CAA file. ", " Continuing to read");
		if(inFile.commentMark != '!')
			QieMsg( "WARNING", "CEF 2 format: Comment marker not exclamation mark '!'. ", " Continuing to read" );
		if(inFile.attr_delim != ',')
			QieMsg( "WARNING", "CEF 2 format: Attribute delimiter must be a comma. ", " Continuing to read" );
		if( inFile.data_delim != ',')
			QieMsg( "WARNING", "CEF 2 format: Data delimiter must be a comma. ", " Continuing to read" );
		if (inFile.recNumberingOn)
			QieMsg("WARNING", "CEF 2 format: Record numbering is not supported in CAA", " Continuing to read" );
   	}

	inFile.setEOL( endOfRecord ); // only set this AFTER the header has been read as only records use end_of_line
	//   QieFTinit_parsers();  // initalise parsers for SJS free time routines 

	// ensure there is a DEPEND_0 set if time line exists
	ensureDependZero();

	return true;
} 

// ------------------------------------------------

void QieData::countEntries(std::string &gattr_datatype_0, int &num_elems, vector <string> &valueList)
{
	int n_entries=0;
	gattr_datatype_0 = "CHAR";
	string gattr_datatype = "CHAR";

	std::string nextLine;
	int len;
   	// read the global block 
   	while(inFile.readLine(nextLine, len) != EOF)
	{
		if( !parseLine(nextLine, len) )  {
		      QieMsg( "WARNING", "Failed to parse line in Global block =", nextLine.c_str());
			 // failed to parse line
			continue;
		}
				
       	if(strncmp(QieToUpper(parameter.c_str()), "END_META", 8) == 0) return;
       	
		if(strncmp(QieToUpper(parameter.c_str()), "G_ATTR_DATA_TYPE", 16) == 0 ||
		   strncmp(QieToUpper(parameter.c_str()), "VALUE_TYPE", 10) == 0){
			
				// current data type
				gattr_datatype = QieToUpper(value.c_str());  // deep copy
				
				// once first entry is found, don't change datatype (QSAS can't cope yet)
				if(n_entries == 0) gattr_datatype_0 = QieToUpper(value.c_str());  // deep copy	
							
        	  	continue;
		}

		if(strncmp(QieToUpper(parameter.c_str()), "ENTRY", 5) == 0)
		{
			if(n_entries == 0) 
			{
				num_elems = countEntryElements(value, inFile.attr_delim);
			}
			if(gattr_datatype == gattr_datatype_0)
			{
				valueList.push_back(value.c_str());
				n_entries++;
			}
		}
	}
	
	return;
}

// ------------------------------------------------

bool QieData::readGlobal(const char *gAttrName)
{
	// make local copy as parameter/value re-used
	string globalAttrName(gAttrName);
	
   	// create new QdObjectSeq_var for this global

	std::string gattr_datatype;
	std::string gattr_datatype_0;
	int num_elems_0;
	
	vector <string> valueList;
	// count number of entries (involves re-reading of file)
	countEntries(gattr_datatype_0, num_elems_0, valueList);
	
	if( globalAttrName != value)  {
		// last line read should be end meta, so value contains meta name
		QieMsg( "WARNING", "Global block start/end names differ", globalAttrName.c_str(), value.c_str());
		return false;
	}

	int nEntries =  valueList.size();	
	gattr_datatype = gattr_datatype_0.c_str();
	
	if(nEntries == 0) {
		QieMsg( "ERROR", "no G entries counted");
		return false;
	}	
	if(num_elems_0 == 0) {
		QieMsg( "ERROR", "G entry has no elements");
		return false;
	}
	
    QdObject_var Gobj;
	
	if(nEntries == 1){
		// single entry
		if(gattr_datatype == "CHAR"  ){
			if(num_elems_0 == 1)
			{
				QdString_var Sentry = new QdString();
				getAttrString(gattr_datatype_0, num_elems_0, *Sentry, valueList[0] );
				Gobj = Sentry;
			}
			else
			{
				QdSMatrix_var SAentry = new QdSMatrix(num_elems_0);
				getAttrStringMatrix(gattr_datatype_0, num_elems_0, *SAentry, valueList[0] );
				Gobj = SAentry;
			}
		}
		else if(gattr_datatype == "DOUBLE" || gattr_datatype == "FLOAT" || gattr_datatype == "INT" ||  gattr_datatype == "BYTE"){
			if(num_elems_0 == 1)
			{
				QdRScalar_var Dentry = new QdRScalar();
				getAttrScalar(gattr_datatype_0, num_elems_0, *Dentry, valueList[0] );
				Gobj = Dentry;
			}
			else
			{
				QdRMatrix_var Mentry = new QdRMatrix(Dim(num_elems_0));
				getAttrMatrix(gattr_datatype_0, num_elems_0, *Mentry, valueList[0] );
				Gobj = Mentry;
			}
		}
		else if(gattr_datatype == "ISO_TIME"){
			if(num_elems_0 == 1)
			{
				QdTime_var Tentry = new QdTime();
				getAttrTime(gattr_datatype_0, num_elems_0, *Tentry, valueList[0] );
				Gobj = Tentry;
			}
			else if(num_elems_0 == 2)
			{
				QdTimeInterval_var TIentry = new QdTimeInterval();
				getAttrTimeIvl(gattr_datatype_0, num_elems_0, *TIentry, valueList[0] );
				Gobj = TIentry;
			}
			else
			{
				QieMsg( "ERROR", "QSAS cannot handle array of times");
				return false;
			}
		}
		else if(gattr_datatype == "ISO_TIME_RANGE"){
			if(num_elems_0 == 1)
			{
				QdTimeInterval_var TIentry = new QdTimeInterval();
				getAttrTimeRange(gattr_datatype_0, num_elems_0, *TIentry, valueList[0] );
				Gobj = TIentry;
			}
			else
			{
				QieMsg( "ERROR", "QSAS cannot handle array of times");
				return false;
			}
		}
		else {
			// use char, but should never get here as default is set
			QieMsg( "WARNING", "Global block value type not recognised", gattr_datatype.c_str());
			QdString_var Sentry = new QdString();
			getAttrString(gattr_datatype_0, num_elems_0, *Sentry, valueList[0] );
			Gobj = Sentry;
		}
	}
	else{
		// sequence of entries
		if(gattr_datatype == "CHAR"  )
		{
			if(num_elems_0 == 1)
			{
				QdStringSeq_var Sentry = new QdStringSeq(nEntries);
				for(int i=0; i < nEntries; i++)
				{
					getAttrString(gattr_datatype_0, num_elems_0, Sentry[i], valueList[i] );
				}
				Gobj = Sentry;
			}
			else
			{
				QdSMatrixSeq_var SAentry = new QdSMatrixSeq(nEntries, Dim(num_elems_0));
				for(int i=0; i < nEntries; i++)
				{
					getAttrStringMatrix(gattr_datatype_0, num_elems_0, SAentry[i], valueList[i] );
				}
				Gobj = SAentry;
			}
		}
		else if(gattr_datatype == "DOUBLE" || gattr_datatype == "FLOAT" || gattr_datatype == "INT" || gattr_datatype == "BYTE")
		{
			if(num_elems_0 == 1)
			{
				QdRScalarSeq_var Sentry = new QdRScalarSeq(nEntries);
				for(int i=0; i < nEntries; i++)
				{
					getAttrScalar(gattr_datatype_0, num_elems_0, Sentry[i], valueList[i] );
				}
				Gobj = Sentry;
			}
			else
			{
				QdRMatrixSeq_var MSentry = new QdRMatrixSeq(nEntries);
				for(int i=0; i < nEntries; i++)
				{
					getAttrMatrix(gattr_datatype_0, num_elems_0, MSentry[i], valueList[i] );
				}
				Gobj = MSentry;
			}
		}
		else if(gattr_datatype == "ISO_TIME"){
			if(num_elems_0 == 1)
			{
				QdTimeSeq_var Tentry = new QdTimeSeq(nEntries);
				for(int i=0; i < nEntries; i++)
				{
					getAttrTime(gattr_datatype_0, num_elems_0, Tentry[i], valueList[i] );
				}
				Gobj = Tentry;
			}
			else if(num_elems_0 == 2)
			{
				QdTimeIntervalSeq_var TIentry = new QdTimeIntervalSeq(nEntries);
				for(int i=0; i < nEntries; i++)
				{
					getAttrTimeIvl(gattr_datatype_0, num_elems_0, TIentry[i], valueList[i] );
				}
				Gobj = TIentry;
			}
			else
			{
				QieMsg( "ERROR", "QSAS cannot handle array of times");
				return false;
			}
		}
		else if(gattr_datatype == "ISO_TIME_RANGE"){
			if(num_elems_0 == 1)
			{
				QdTimeIntervalSeq_var TIentry = new QdTimeIntervalSeq(nEntries);
				for(int i=0; i < nEntries; i++)
				{
					getAttrTimeRange(gattr_datatype_0, num_elems_0, TIentry[i], valueList[i] );
				}
				Gobj = TIentry;
			}
			else
			{
				QieMsg( "ERROR", "QSAS cannot handle array of time intervals");
				return false;
			}
		}
	} // end if on sequence
	
	if( Gobj.is_nil() ) {
		return false;
	}
	
	Name GobjName(globalAttrName);
	  
	QdString_var typeXref = new QdString(gattr_datatype_0.c_str());
	Gobj->change_xref("value_type", typeXref);
	GlobalList->bindx(GobjName, Gobj);

	return true;
}


// ------------------------------------------------

void QieData::getAttrString(string gattr_datatype_0, int num_elems_0, string &Sentry, string &entry )
{
	int numEntryElements = countEntryElements(entry, inFile.attr_delim);
	if(1 != numEntryElements)
	{	
		QieMsg( "WARNING", "Global entries number of elements differ");
		return;
	}
	Sentry = QieStripQuotes(entry);
}

// ------------------------------------------------

void QieData::getAttrScalar(string gattr_datatype_0, int num_elems_0, RScalar &Dentry, string &entry )
{
	int numEntryElements = countEntryElements(entry, inFile.attr_delim);
	if(1 != numEntryElements)
	{	
		QieMsg( "WARNING", "Global entries number of elements differ");
		return;
	}
	Dentry = atof((QieStripQuotes(entry)).c_str());
}

// ------------------------------------------------

void QieData::getAttrTime(string gattr_datatype_0, int num_elems_0, QmjdTime &Tentry, string &entry )
{
	int numEntryElements = countEntryElements(entry, inFile.attr_delim);
	if(1 != numEntryElements)
	{	
		QieMsg( "WARNING", "Global entries number of elements differ");
		return;
	}
	Tentry.setFromISOstring((QieStripQuotes(entry)).c_str());
}

// ------------------------------------------------

void QieData::getAttrStringMatrix(string gattr_datatype_0, int num_elems_0, SMatrix &SAentry, string &entry )
{
	int numEntryElements = countEntryElements(entry, inFile.attr_delim);
	
	if(num_elems_0 != numEntryElements)
	{	
		QieMsg( "WARNING", "Global entries number of elements differ");
		return;
	}
	
	int startAt = 0;
	string token;
	for(int i=0; i<num_elems_0; i++){
		getNextEntry(entry, token, startAt, inFile.attr_delim);
		SAentry[i] = token.c_str();
	}
}

// ------------------------------------------------

void QieData::getAttrMatrix(string gattr_datatype_0, int num_elems_0, RMatrix &Mentry, string &entry )
{
	int numEntryElements = countEntryElements(entry, inFile.attr_delim);
	
	if(num_elems_0 != numEntryElements)
	{	
		QieMsg( "WARNING", "Global entries number of elements differ");
		return;
	}
	
	int startAt = 0;
	string token;
	for(int i=0; i<num_elems_0; i++){
		getNextEntry(entry, token, startAt, inFile.attr_delim);
		Mentry[i] = strtod(token.c_str(), NULL);
	}
}

// ------------------------------------------------

void QieData::getAttrTimeIvl(string gattr_datatype_0, int num_elems_0, QmjdTimeInterval &TIentry, string &entry )
{
	int numEntryElements = countEntryElements(entry, inFile.attr_delim);
	
	if(num_elems_0 != numEntryElements)
	{	
		QieMsg( "WARNING", "Global entry for interval does not have two date/times");
		return;
	}
	
	int startAt = 0;
	string token;
	getNextEntry(entry, token, startAt, inFile.attr_delim);
	QmjdTime ts;
	ts.setFromISOstring(token.c_str());
	
	getNextEntry(entry, token, startAt, inFile.attr_delim);
	QmjdTime te;
	te.setFromISOstring(token.c_str());
		
	TIentry.set(ts, te);
}

// ------------------------------------------------

void QieData::getAttrTimeRange(string gattr_datatype_0, int num_elems_0, QmjdTimeInterval &TIentry, string &entry )
{	
	string rangeStr = QieStripQuotes(entry).c_str();
	int pos = rangeStr.find("/");
	string startStr = rangeStr.substr(0, pos);
	string endStr = rangeStr.substr(pos+1);
	QmjdTime ts;
	ts.setFromISOstring(startStr.c_str());
	QmjdTime te;
	te.setFromISOstring(endStr.c_str());
		
	TIentry.set(ts, te);
}

// ------------------------------------------------

bool QieData::readVariable(const char *variableName)
{
// initialise variable and format first (set defaults)
	QieVariable V;
	V.name = variableName;
	V.col_start = -1;
	V.col_width = -1;
	V.objDim.clear();
//	V.ftpackets.timeformat = NOT_A_TIME; // SJS re FT - safe initialisation
	V.recVary = true;
	V.datatype = "CHAR";
	V.n_posn = recordNumEntries;
	V.cachedFrom = -1;
	V.cachedTo  = -1;
	V.nRecs = -1;

	std::string dataValues=" ";
	std::string nextLine;
	int nEntries = 1;
	int len;

	while(inFile.readLine(nextLine, len) != EOF){
		if( !parseLine(nextLine, len) ) {
			QieMsg( "WARNING", "Failed to parse line in Variable block =", nextLine.c_str());
			continue;
		}

		if(strncmp(QieToUpper(parameter.c_str()), "END_VARIABLE", 12) == 0) {
        		if(V.name != value) {
		     	 	QieMsg( "WARNING", "Variable block start/end names differ", V.name.c_str(), value.c_str());
         		}
          		break; // end variable block, now make object
		}

		if(strncmp(QieToUpper(parameter.c_str()), "DATA_TYPE", 9) == 0 ||
		   strncmp(QieToUpper(parameter.c_str()), "VALUE_TYPE", 10) == 0){
				V.datatype = QieToUpper(value.c_str());  // deep copy
				// Note: first time var is default time tags
				if(d0_var_num == -1 && value == "ISO_TIME") d0_var_num = QieVars.size(); // not pushed back yet
        	  		continue;
		}
		
		if(strcmp(QieToUpper(parameter.c_str()), "DATA") == 0) {
			/* NOVARY variable */
			V.recVary = false;
			dataValues = value.c_str(); // deep copy for creating object after while loop completes
			continue;
		}
		
		if(strncmp(QieToUpper(parameter.c_str()), "COLUMN_START", 12) == 0) {
			V.col_start = atol(value.c_str());
			if(inFile.fileType == _CAA) QieMsg( "WARNING", "COLUNM_START detected for '", V.name.c_str(), "' being ignored");
			continue;
		}

		if(strncmp(QieToUpper(parameter.c_str()), "COLUMN_WIDTH", 12) == 0) {
			V.col_width = atol(value.c_str());
			if(inFile.fileType == _CAA) QieMsg( "WARNING", "COLUNM_WIDTH detected for '", V.name.c_str(), "' being ignored");
			continue;
		}
		 
		if(strncmp(QieToUpper(parameter.c_str()), "IGNORE_ON_READ", 14) == 0) {
           	if (value[0] == 'y' || value[0] == 'Y') V.ignore = true;
           	if(inFile.fileType == _CAA) QieMsg( "WARNING", "non-CAA option IGNORE_ON_READ detected for '",  V.name.c_str(), "', flag being ignored");
          	continue;
		}

		if(strcmp(QieToUpper(parameter.c_str()), "TIME_FORMAT") == 0) {
			QieMsg( "WARNING", "TIME_FORMAT descriptor detected for  ", V.name.c_str(), " not in CEF 2 syntax");
			if(strncmp(QieToUpper(value.c_str()),"FREE_TIME_FORMAT", 16 ) == 0) {
   //>>>>           V.ftpackets.timeformat = FREE_TIME_FORMAT;
			}
			else if(strncmp(QieToUpper(value.c_str()),"ISO", 3 ) == 0){
     //>>>          V.ftpackets.timeformat = ISO;
			}             // Non-time vars don't have this, and carry initialised NOT_A_TIME value, SJS re FT */
			else{
				QieMsg( "WARNING", "Unknown time format designation '", value.c_str(), "' being ignored");
			}
			continue;
		}
			 		 
		if(strncmp(QieToUpper(parameter.c_str()), "MAXSTRLEN", 9) == 0) {
			// no longer needed
			if(inFile.fileType == _CAA) QieMsg( "WARNING", "MAXSTRLEN detected for '", V.name.c_str(), "' not in CEF 2 syntax");
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "SIZES", 5) == 0) {

			int n_elems = countEntryElements(value, inFile.attr_delim);
			// set up Dim object 
			
			if(n_elems == 0)  continue;  
		
			V.objDim.clear();
			V.objDim.assign(n_elems, 0);	
			int startAt = 0;
			string token;
			for(int i=0; i<n_elems; i++){
					
				getNextEntry(value, token, startAt, inFile.attr_delim);
				int Ndim = strtol(token.c_str() , NULL, 10);
				if(Ndim == 0) {
					QieMsg( "ERROR", "Variable dimension size zero from '", value.c_str());
					return false;
				}
				nEntries = nEntries * Ndim;
				V.objDim[i] = Ndim;
			}
			continue;
		}

        	 /* else read attribute */
		QdObject_var nextAttr;
		try{
			 nextAttr = this->QieText_to_Attr(V, value); // V needed for type and dimensions
		}
		catch(Exception &e){
			QieMsg( "WARNING", "failed to make attribute from: ", value.c_str());
			e.print_msg();
		}
		if( !nextAttr.is_nil() ){
			V.Vxref->change_xref(parameter.c_str(), nextAttr);
		}
	} // end while loop

	// If NRV object, fill with data now
	if(dataValues != " "){
		V.Vobj = QieText_to_Obj(V, dataValues);
	}
	
	QieVars.push_back(V);
	if (V.recVary) recordNumEntries += nEntries;
	
	return true;  
}

// ----------------------------------------------------------

QdObject_var QieData::QieText_to_Attr(QieVariable &V, std::string valueStr)
{
	std::string datatype;
	QdObject_var dobj;

	// create data object from valueStr
	int n_elems = countEntryElements(valueStr, inFile.attr_delim);
	if( n_elems == 0) return dobj;
   
	int startAt = 0;
	string token;
	bool isChar = getNextEntry(valueStr, token, startAt, inFile.attr_delim);


	if(isChar) datatype = "CHAR";
	
	else if(QieIsISOtimeRange(token)) datatype = "ISO_TIME_RANGE";

	else if(QieIsISOtime(token)) datatype = "ISO_TIME";
	 
	else{
       // test for numeric data type 
		char * endPtr;
		strtod(token.c_str(), &endPtr);
		if( token == string(endPtr) ) datatype = "CHAR"; // no numeric value found, use char as safe
		else datatype = "DOUBLE"; 
	}

	
	if(datatype == "CHAR")
	{
		if(n_elems == 1)
		{
			QdString_var sobj = new QdString(token.c_str());
			dobj = sobj;
		}
		else
		{
			QdSMatrix_var smobj = new QdSMatrix(Dim(n_elems));
			(*smobj)[0] = token.c_str();
			for(int i=1; i<n_elems; i++){
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				(*smobj)[i] = token.c_str();
			}
			dobj = smobj;
		}
	}

	else if(datatype == "ISO_TIME")
	{		
		if(n_elems == 1){
	
			try{
				if( QieIsISOtime(token) ){
					QdTime_var tob = new QdTime(token); 
					dobj = tob;
				}
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ", token.c_str());
				e.print_msg();
			}
		}
		else if(n_elems == 2){
			// assume to be a time interval (NRV) 
			try{
				QmjdTime iv_start(token.c_str());
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				QmjdTime iv_end(token.c_str());
				QdTimeInterval_var ti = new QdTimeInterval(iv_start, iv_end); 
	 
				dobj = ti;
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ",token.c_str());
				e.print_msg();
			}
		}
		else{ 
			// treat as time sequence
 			try{
				QdTimeSeq_var ts = new QdTimeSeq(n_elems);
				ts[0] = QmjdTime(token.c_str());
				for (int i=i; i < n_elems; i++){
					getNextEntry(valueStr, token, startAt, inFile.attr_delim);
					ts[i] = QmjdTime(token.c_str());
				}
				dobj = ts;
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ", token.c_str());
				e.print_msg();
			}
			
		}
	}

	else if(datatype == "ISO_TIME_RANGE")
	{		
		if(n_elems == 1){
	
			try{
				if( QieIsISOtimeRange(token) ){
					QdTimeInterval_var tiob = new QdTimeInterval();
					tiob->setFromISOstring(token.c_str());
					dobj = tiob;
				}
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ", token.c_str());
				e.print_msg();
			}
		}
	}

   	else if(datatype == "DOUBLE" )
	{
		
		if(n_elems == 1)
		{
			QdRScalar_var sobj = new QdRScalar(strtod(token.c_str(), NULL) );
			dobj = sobj;
		}
		else
		{
			QdRMatrix_var mobj = new QdRMatrix(Dim(n_elems));
			(*mobj)[0] =  strtod(token.c_str(), NULL) ;
			for(int i=1; i<n_elems; i++){
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				(*mobj)[i] =  strtod(token.c_str(), NULL);
			}
			dobj = mobj;
		}
	}
	
    return dobj;
}


// ----------------------------------------------------------

QdObject_var QieData::QieText_to_Obj(QieVariable &V, std::string valueStr)
{
	// used for NRV data, uses attr_delim
	QdObject_var dobj;
	
	// create data object from valueStr
	int num_elems = countEntryElements(valueStr, inFile.attr_delim);
	if(num_elems == 0) return dobj;
	
	if(V.datatype.find("CHAR") != string::npos)
	{
		if(num_elems == 1)
		{
			int startAt=0;
			string token;
			getNextEntry(valueStr, token, startAt, inFile.attr_delim);
			QdString_var sobj = new QdString(token.c_str());
			dobj = sobj;
		}
		else{
			int startAt=0;
			string token;
			QdSMatrix_var sobj = new QdSMatrix(num_elems);
			for(int i=0; i<num_elems; i++){
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				(*sobj)[i] = token.c_str();
			}
			dobj = sobj;
		}
	}

	else if(V.datatype.find("ISO_TIME_RANGE") != string::npos)
	{		
		// do ISO_TIME_RANGE first as ISO_TIME is substring
		
		if(num_elems == 1){
			int startAt=0;
			string token;
			getNextEntry(valueStr, token, startAt, inFile.attr_delim);
		    try{
				QdTimeInterval_var tio = new QdTimeInterval(); 
				tio->setFromISOstring(token.c_str());
				dobj = tio;
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
	}

	else if(V.datatype.find("ISO_TIME") != string::npos)
	{		
		if(num_elems == 1){
			int startAt=0;
			string token;
			getNextEntry(valueStr, token, startAt, inFile.attr_delim);
		    try{
				QdTimeSeq_var to = new QdTimeSeq(1, token.c_str()); 
				dobj = to;
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
		else if(num_elems == 2){
			// assume to be a time interval (NRV) 
			try{
				int startAt=0;
				string token;
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				QmjdTime iv_start(token.c_str());
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				QmjdTime iv_end(token.c_str());
				QdTimeInterval_var ti = new QdTimeInterval(iv_start, iv_end); 
	 
				dobj = ti;
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
		else{ 
			// treat as time sequence
 			try{
				QdTimeSeq_var ts = new QdTimeSeq(num_elems);
				int startAt=0;
				string token;
				for (int i=0; i < num_elems; i++){
					getNextEntry(valueStr, token, startAt, inFile.attr_delim);
					ts[i] = QmjdTime(token.c_str());
				}
   
				dobj = ts;
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
	}

   	else if(V.datatype.find("DOUBLE") != string::npos || 
			V.datatype.find("FLOAT")  != string::npos|| 
			V.datatype.find("INT")  != string::npos|| 
			V.datatype.find("BYTE")  != string::npos )
	{
		if( num_elems == 1)
		{
			int startAt=0;
			string token;
			getNextEntry(valueStr, token, startAt, inFile.attr_delim);
			QdRScalar_var sobj = new QdRScalar( strtod(token.c_str(), NULL) );
			dobj = sobj;
		}
		else
		{
			int startAt=0;
			string token;
			QdRMatrix_var mobj = new QdRMatrix(Dim(V.objDim));
			
			valarray<double> v_data(num_elems);

			// peel off entries using shallow copies
			for(register int i=0; i<num_elems; i++){
				getNextEntry(valueStr, token, startAt, inFile.attr_delim);
				v_data[i] = strtod(token.c_str(), NULL);
			}
		
			mobj->data()[mobj->getgslice()]=v_data;
	
			dobj = mobj;
		}
	}
	
	// attach xrefs now as NRV vars are not touched later
//	if(!dobj.is_nil()) dobj->copy_xrefs_from(V.Vxref);
	
	return dobj;

}


// ----------------------------------------------------------

void QieData::clearGetThis()
{
	//  sets the getThis flag to false for all vars 	
	for (int i=0; i< QieVars.size(); i++){
		QieVars[i].getThis = false;
	}
}

// ----------------------------------------------------------

void QieData::setGetThis(int vnum)
{
	// recursively sets the getThis flag to true for var and xrefs
	QieVariable *V = &(QieVars[vnum]);

	if(V->getThis) return; // already set
	
	V->getThis = true;
	 
	// go through xrefs and find vars pointed to
	vector <string> aNames;
	V->Vxref->list_xref_names(aNames);

	for(int i=0; i< aNames.size(); i++){
		QdObject_var aObj = V->Vxref->get_xref( aNames[i]);
		if(aObj.is_nil()) continue;
		
		QdString_var depVarName = QdString_var::narrow(aObj);
		if( !depVarName.is_nil() ){
			int attr_num = findVarNumByName( depVarName->c_str());
			if(attr_num != -1) {
				// found dependent variable
				setGetThis(attr_num);
			}
		}
	}
}

// ----------------------------------------------------------

bool QieData::initSequence(QieVariable * V, int nRecs)
{
	// determine dimensionality
	V->nEntries = 1;
	V->nRecs = -1; // set flag in case we fail to read data (was set to record number by qdi for sliders)
	for(int i=0; i<V->objDim.size(); i++){
		V->nEntries = V->nEntries * V->objDim[i];
	}
		
	// initialise sequence 
	if(V->datatype == "CHAR"){
		if( V->nEntries > 1 ) {
			QdSMatrixSeq_var stms = new QdSMatrixSeq(nRecs, Dim(V->objDim));	
				
			V->Vobj = stms;
		}
		else{
			QdStringSeq_var sts = new QdStringSeq(nRecs);
			V->Vobj = sts;
		}
			
	}
	else if (V->datatype == "ISO_TIME"){

		if(V->nEntries == 1) {
			// time sequence
			QdTimeSeq_var ts = new QdTimeSeq(nRecs);
			V->Vobj = ts;
		}
		else{
			// cannot handle time arrays yet
			return false;
		}
	}
	else if (V->datatype == "ISO_TIME_RANGE"){

		if(V->nEntries == 1) {
			// time interval sequence
			QdTimeIntervalSeq_var tis = new QdTimeIntervalSeq(nRecs);
			V->Vobj = tis;
		}
		else{
			// cannot handle time interval arrays yet
			return false;
		}
	}
	else{
		// everything else stored as double 
		if( V->nEntries > 1 ) {
			QdRMatrixSeq_var  ms = new QdRMatrixSeq(nRecs, Dim(V->objDim));		
			V->Vobj = ms;
		}
		else{
			QdRScalarSeq_var ss = new QdRScalarSeq(nRecs);
			V->Vobj = ss;
		}
	}
	V->nRecs = nRecs;
	return true;
}

// ----------------------------------------------------------

QdTimeSeq_var QieData::readTT(int vnum, int nRecs)
{
	// Used only to get time vars, does not attach xrefs
	
	QieVariable *var = &(QieVars[vnum]);
						
	QdTimeSeq_var tt;
	// find time tags variable
	if(var->Vxref->xref_exists("DEPEND_0"))
	{
		QdObject_var tt_obj = var->Vxref->get_xref("DEPEND_0");
		tt =  QdTimeSeq_var::narrow(tt_obj);
		if(!tt.is_nil()) return tt; // case that tt object has been determined
		
		QdString_var tt_str = QdString_var::narrow(tt_obj);
		if(!tt_str.is_nil()) {
			int ttVarNum = findVarNumByName( tt_str->c_str());
			if(ttVarNum == -1) return tt;

			// check if this has been read before
			tt = QdTimeSeq_var::narrow(QieVars[ttVarNum].Vobj);
			if(tt.is_nil() || tt->sequence_size() != nRecs)
			{
				if(inFile.fileType == _CDF){
					clearGetThis();
					setGetThis(ttVarNum); 

					// create list of vars to fetch and initialise
					vector <int> getList;
					for(int i=0; i<QieVars.size(); i++){
						if(QieVars[i].getThis)  getList.push_back(i);
					}
					
					// Now we have Themis components too, get everything
					getCDFvarDataList(getList, 0, -1);
					tt = QdTimeSeq_var::narrow(QieVars[ttVarNum].Vobj); // cdf changes object
				}
				else
				{
					// first call for this variable
					if( !initSequence(&(QieVars[ttVarNum]), nRecs) ) return tt;
					tt = QdTimeSeq_var::narrow(QieVars[ttVarNum].Vobj);
					
					for( int recNow=0; recNow<nRecs; recNow++ ) 
					{
						// locate start and length of each token in record
						int ptr = findEntryTT(QieVars[ttVarNum].n_posn, recNow);
						if( ptr < 0 ) {
							QieMsg ("ERROR", "failed to find tt entries in record\n");
							continue;
						}
		
						try{
							tt[recNow].setFromISOstring(&(dataRecords[recNow].c_str()[ptr]));
						}
						catch(Exception &e){
							e.print_msg();
						}
					}
					
					QieVars[ttVarNum].cachedFrom = 0;
					QieVars[ttVarNum].cachedTo = nRecs-1;
				}
			}
			// else tt already read, will use them
		}
	}
	else{
		// variable has no DEPEND_0, can we make it into time sequence?
		QdTimeIntervalSeq_var tis = QdTimeIntervalSeq_var::narrow(var->Vobj);
		if( !tis.is_nil() ){
			int len = tis->sequence_size();
			tt = new QdTimeSeq(len);
			for(int i=0; i<len; i++){
				tt[i] = tis[i].start();
			}
		}
		
	}
	return tt;
}
// ----------------------------------------------------------

bool QieData::readVarData(int vnum, int recStart, int recEndFlag)
{

	clearGetThis(); // ensure no vars are set to be read
	int recEnd;
	if(recEndFlag == -1) recEnd = dataRecords.size() -1 ; // set to  -1 fetches whole range
	else recEnd = recEndFlag;
	
	QieVariable *V = &(QieVars[vnum]);
	// NOTE: NRV variables are read already in queryFile() but not queryCDFfile()
	if ( !(V->recVary) && inFile.fileType != _CDF) return true; 

	
	// recursively set the getThis flag to true for var and  variable type xrefs (incl time tags)
	setGetThis(vnum); 

	// create list of vars to fetch and initialise
	vector <int> getList;
	int nRecs = recEnd - recStart + 1;
	for(int i=0; i<QieVars.size(); i++){
		if(QieVars[i].getThis){

			if(inFile.fileType == _CDF){
			 	getList.push_back(i);
			}
			else if( QieVars[i].recVary){
				// excludes NRV vars for flat files as read in header
				
				// Use cached object if possible works for all file types
				if(QieVars[i].cachedFrom == recStart && QieVars[i].cachedTo == recEnd) {
					// variable already read for this exact range
					continue; 
				}
				else if( QieVars[i].cachedFrom != -1 && QieVars[i].cachedFrom <= recStart && QieVars[i].cachedTo >= recEnd){
					// Take G slice
					QieVars[i].Vobj = QuGetSubset(QieVars[i].Vobj, recStart - QieVars[i].cachedFrom, recEnd - QieVars[i].cachedFrom);
					QieVars[i].cachedFrom = recStart;
					QieVars[i].cachedTo = recEnd;

					if( !(QieVars[i].Vobj.is_nil()) ) continue; // else fall through and add to list to get
				}	
				else {
					// fall through and add to list of vars to get
					QieVars[i].cachedFrom = recStart;
					QieVars[i].cachedTo = recEnd;
				}
				// not cached if we get here
				if( initSequence(&(QieVars[i]), nRecs) ) getList.push_back(i);


			}
		}
	}

	if(inFile.fileType == _CDF) {
	   // fetch data for variables from CDF file
	   getCDFvarDataList(getList, recStart, recEndFlag);
	}
	else{
		// get records requested for appropriate data type
		int recNum=0;
		for( int recNow=recStart; recNow<dataRecords.size(); recNow++ ){
			// locate start and length of each token in record
			if( !findEntries(recNow) ) {
				QieMsg ("ERROR", "failed to find entries in record\n");
				continue;
			}
			for(int nv=0; nv<getList.size(); nv++){

				if ( !(getVarRecordEntry(getList[nv], recNum, recNow)) ) {
					// don't continue if data itself fails, but do continue for metadata
					if(QieVars[getList[nv]].name == QieVars[vnum].name) {
						QieMsg ("ERROR", "failed to read entries in record\n");
						return false;
					}
				}
			}

			recNum++;
			if(recNow == recEnd) break;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
	}
	

	return true;
}		


// ------------------------------------------------

void QieData::attachXrefs(int varN, string varName){

	QdObject_var dobj = QieVars[varN].Vobj; 

	vector <string> aNames;
	dobj->list_xref_names(aNames);
	
	if(aNames.size() > 0) return; // object has them already 
	else aNames.clear();
	
	// Attach xrefs, changing attributes that point to vars. 
	QieVars[varN].Vxref->list_xref_names(aNames);
	for(int i=0; i<aNames.size(); i++){
		// skip Themis epoch xrefs as tt now cleaned up
		if(aNames[i] == "VIRTUAL") continue;
		if(aNames[i] == "COMPONENT_0") continue;
		if(aNames[i] == "COMPONENT_1") continue;
		
		// Find attributes that point to vars
		try
		{
			QdObject_var aObj = QieVars[varN].Vxref->get_xref( aNames[i]);
			if(aObj.is_nil()) continue;

			QdString_var depVarName = QdString_var::narrow(aObj);
			if( !depVarName.is_nil() ){
				// attach var pointed to as xref
				int attr_num = findVarNumByName( depVarName->c_str());
				if(attr_num != -1	&& *depVarName != QieVars[varN].name // skip self reference, e.g. Ptr_var_name
									&& *depVarName  != varName ) // skip reference to parent var name (inf loop)
				{ 
						// use var as attribute
						// recursively attach xrefs
						attachXrefs(attr_num, QieVars[varN].name);

						QdString_var ptrVarName = new QdString(QieVars[attr_num].name.c_str());
						QieVars[attr_num].Vobj->change_xref(string("Ptr_var_name"), ptrVarName);
						// change this xref on target var (dobj may not be same as var in QieDataCont)
						dobj->change_xref(aNames[i].c_str(), QieVars[attr_num].Vobj);
						continue;
				}
			}

			// otherwise attach attribute as it is
			dobj->change_xref(aNames[i].c_str(), aObj);

		}
		catch(Exception &e){
			cout << "in attach xrefs, exception for: " << QieVars[varN].name.c_str() << " \n attribute " << aNames[i].c_str() << "\n     " << e.msg().c_str() << endl;
		}
	}

}


// ------------------------------------------------

bool QieData::getVarRecordEntry(int nToGet, int recNum, int recNow){

	// recNum is position in output sequence
	// recNow is position in input vector of records
	
	QieVariable *V = &(QieVars[nToGet]);

	if(V->datatype == "CHAR")
	{

  
		if( V->nEntries > 1 ) {
			QdSMatrixSeq_var stms = QdSMatrixSeq_var::narrow(V->Vobj);
			if(stms.is_nil()) return false;

			if( !getEntriesChar(V->n_posn, V->nEntries, stms[recNum], recNow ) ){
				QieMsg ("ERROR", "unable to extract variable char array data for\n", V->name.c_str());
				return false;
			}
		}
		else{				
			QdStringSeq_var   sts = QdStringSeq_var::narrow(V->Vobj);
			if(sts.is_nil()) return false;
			
			if( !getEntriesChar(V->n_posn,  sts[recNum], recNow) ){
				QieMsg ("ERROR", "unable to extract variable char data for\n", V->name.c_str());
				return false;
			}
		}

	}
	else if(V->datatype == "ISO_TIME"){
	
		QdTimeSeq_var ts = QdTimeSeq_var::narrow(V->Vobj);
		if(ts.is_nil()) return false;
		
		if( !getEntriesTime(V->n_posn, ts[recNum], recNow) ){
			QieMsg ("ERROR", "unable to extract variable time data for\n", V->name.c_str());
			return false;
		}
		
	}
	else if(V->datatype == "ISO_TIME_RANGE"){
	

		QdTimeIntervalSeq_var tis = QdTimeIntervalSeq_var::narrow(V->Vobj);
		if(tis.is_nil()) return false;

		if( !getEntriesTimeIvl(V->n_posn, tis[recNum], recNow) ){
			QieMsg ("ERROR", "unable to extract variable time interval data for\n", V->name.c_str());
			return false;
		}
		
	}
	else{		
		// treat as double
		if( V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] == 1) ) {				
		   // scalar
			QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(V->Vobj);
			if(ss.is_nil()) return false;

			if( !getEntriesNumeric(V->n_posn, ss[recNum], recNow) ){
				QieMsg ("ERROR", "unable to extract variable numeric data for\n",  V->name.c_str());
				return false;
			}
		}
		else{
			QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(V->Vobj);
			if(ms.is_nil()) return false;

		    // matrix
			if( !getEntriesNumeric(V->n_posn, V->nEntries, ms[recNum], recNow) ){
				 QieMsg ("ERROR", "unable to extract variable numeric array data for ", V->name.c_str());
				return false;
			}
		}
	}
	return true;
}		

// ------------------------------------------------

bool QieData::parseLine(string &nextLine, int len)
{
  int ptr;
  int startAt=0;
  ptr  = nextLine.find("="); // note we remove white space on param later 
  if(ptr == -1) {
	QieMsg ("ERROR", "No '=', cannot parse line\n", nextLine.c_str());
    return false;
  }
  if(ptr == 0) {
	QieMsg ("ERROR", "Line starts '=', cannot parse line\n", nextLine.c_str());
    return false;
  }

  /* remove leading white space in param */
  while( (nextLine[startAt] == ' ' || nextLine[startAt] == '\t') ) startAt++;
 
   /* remove possible trailing white space in param */
  int tokLen = ptr-1; 
  while( nextLine[tokLen] == ' ' || nextLine[tokLen] == '\t' ) tokLen--;
  tokLen++;
  tokLen -= startAt;
  if(tokLen <= 0) {
	QieMsg ("ERROR", "No parameter, cannot parse line\n", nextLine.c_str());
    return false;
  }
  parameter = nextLine.substr(startAt, tokLen);
 
  ptr ++;
  // remove leading white space in value 
  while( nextLine[ptr] == ' ' || nextLine[ptr] == '\t')  ptr++;
 
  // remove newline and possible trailing white space in value 
  tokLen = len-1;  
  
  while( nextLine[tokLen] == ' ' || nextLine[tokLen] == '\t' || nextLine[tokLen] == '\n'  || nextLine[tokLen] == '\r'  ) tokLen--;
  tokLen++;
  tokLen -= ptr;
  if(tokLen <= 0) {
	// replace empty RHS with single space
	value = " ";
  }
  else{
	value = nextLine.substr( ptr, tokLen);
  }

  return true;

} 

// -----------------------------------------------
bool QieData::findEntries(int recNow){

	tokenStart.assign(recordNumEntries, 0);
	
	// locates start and length of each token
	
	if(dataRecords[recNow] == "") return false;
	int ptr = 0;
	int startAt=0;
	int NextLineLen = dataRecords[recNow].length();
	int i=0;

	while(ptr < NextLineLen && i < recordNumEntries) {
		// find next delimiter or end of line 
		while(ptr < NextLineLen && dataRecords[recNow][ptr] != inFile.data_delim  && dataRecords[recNow][ptr] != '\0') ptr++;
	
		tokenStart[i] = startAt;
		dataRecords[recNow][ptr] = '\0'; // replace delimiters with nulls
		ptr++;
		startAt = ptr;
		i++;
	}
	return true;
	
}

// -----------------------------------------------
int QieData::findEntryTT(int n_pos, int recNow){

	// locates start and length of each token, keep TT only
	
	if(dataRecords[recNow] == "") return -1;
	int ptr = 0;
	int startAt=0;
	int NextLineLen = dataRecords[recNow].length();
	int i=0;
	
	while(ptr < NextLineLen) {
		// find next delimiter or end of line (multiple time lines may have NULLS from earlier tt search)
		while(ptr < NextLineLen && dataRecords[recNow][ptr] != inFile.data_delim && dataRecords[recNow][ptr] != '\0') ptr++;
		dataRecords[recNow][ptr] = '\0'; // replace delimiters with nulls
		if( i == n_pos) return startAt;
		ptr++;
		startAt = ptr;
		i++;
	}
	
	return -1;
	
}


// -----------------------------------------------
/*
bool QieData::getEntriesChar(int nStart, int nEntries, int recNum, QdSMatrixSeq_var stms){
	
	if(nStart + nEntries > tokenStart.size() ) return false;
	
	try{
		valarray <string> v_data(nEntries);
		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			std::string token = currentRecord.substr(tokenStart[nStart+i],  tokenLen[nStart+i]);
			v_data[i] = token.c_str();
		}
		stms[recNum].data()[stms[recNum].getgslice()]=v_data;
		return true;	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}
*/
// -----------------------------------------------

bool QieData::getEntriesChar(int nStart, int nEntries, SMatrix &stm, int recNow){
	
	if(nStart + nEntries > tokenStart.size() ) return false;
	
	try{
		valarray <string> v_data(nEntries);
		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			v_data[i] = string(&(dataRecords[recNow].c_str()[tokenStart[nStart+i]]));
		}
		stm.data()[stm.getgslice()]=v_data;
		return true;	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}

// -----------------------------------------------
/*
bool QieData::getEntriesChar(int nStart, int recNum, QdStringSeq_var sts){
	
	if(nStart  > tokenStart.size() ) return false;
	
	try{
		std::string token = currentRecord.substr(tokenStart[nStart],  tokenLen[nStart]);
		sts[recNum] = token.c_str();
		return true;
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}
*/
// -----------------------------------------------

bool QieData::getEntriesChar(int nStart, string &st, int recNow){
	
	if(nStart  > tokenStart.size() ) return false;
	
	try{
		st.assign(&(dataRecords[recNow].c_str()[tokenStart[nStart]]));
		return true;
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}


// -----------------------------------------------

bool QieData::getEntriesNumeric(int nStart, int nEntries, RMatrix &m, int recNow){

	if(nStart + nEntries > tokenStart.size() ) return false;

	try{	
		valarray<double> v_data(nEntries);

		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			v_data[i] = strtod(&(dataRecords[recNow].c_str()[tokenStart[nStart+i]]), NULL);
		}
		
		m.data()[m.getgslice()]=v_data;

		return true;
	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}

}

// -----------------------------------------------

bool QieData::getEntriesNumeric(int nStart, RScalar &s, int recNow){
	if(nStart > tokenStart.size() ) return false;

	try{	
		s = strtod(&(dataRecords[recNow].c_str()[tokenStart[nStart]]), NULL);
		
		return true;		
	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}

}
// -----------------------------------------------

bool QieData::getEntriesTime(int nStart, QmjdTime &t, int recNow){

	if(nStart  >= tokenStart.size() ) return false;
	
	int tok_start = tokenStart[nStart];
	try{
		t.setFromISOstring(&(dataRecords[recNow].c_str()[tok_start]));
	}
	catch(Exception &e){
		e.print_msg();
	}
	
	return true;
		
}

// -----------------------------------------------

bool QieData::getEntriesTimeIvl(int nStart, QmjdTimeInterval &t, int recNow){

	if(nStart  >= tokenStart.size() ) return false;
	
	int tok_start = tokenStart[nStart];
	try{
		t.setFromISOstring(&(dataRecords[recNow].c_str()[tok_start]));
	}
	catch(Exception &e){
		e.print_msg();
	}

	return true;
		
}

// -----------------------------------------------

int QieData::countEntryElements(std::string &inputStr, const char delim)
{
	int ptr = 0;
	int LineLen = inputStr.length();
	const char *cStr = inputStr.c_str();
	if(LineLen == 0) return 0;
	
	int numElems = 1;

	while(ptr < LineLen) {

		if(cStr[ptr] == delim)
		{
			numElems++;
		}
		else if(cStr[ptr] == '\"') {
			// ignore anything INSIDE quotes
			ptr++;
			while(cStr[ptr] != '\"' && ptr < LineLen) ptr++;
		}
		ptr++;
	}
	return numElems;
}
	
// -----------------------------------------------

bool QieData::getNextEntry(string &inputStr, string &token, int &startAt, const char delim){
	
	int ptr = startAt;
	int LineLen = inputStr.length();
	const char *cStr = inputStr.c_str();
	int truncateAt;
	bool isChar = false;
	
	// peel off entries using shallow copies

	while(ptr < LineLen )
	{
		// find next delim if there is one
		if( cStr[ptr] == delim) break;

		// skip to end of quoted text, ignore delim INSIDE quotes
		if(cStr[ptr] == '\"') {
			isChar = true;
			ptr++;

			while(cStr[ptr] != '\"' ) 
			{
				ptr++;
				if(ptr == LineLen) 
				{
					// back up as it gets incremented again
					ptr--;
					break;
				}
			}
		}
		ptr++;
	}
	// ptr is now just passed end of line or on next delimiter

	// remove leading white space and quotes
	while((cStr[startAt] == ' ' || cStr[startAt] == '\"') && startAt < ptr) startAt++;
		
	// remove trailing white space and quotes
	truncateAt = ptr-1;
	while ((cStr[truncateAt] == ' ' || cStr[truncateAt] == '\n' || cStr[truncateAt] == '\"') && startAt < truncateAt) truncateAt--;

	if( truncateAt - startAt >= 0 ) 	
		token = inputStr.substr(startAt, truncateAt - startAt +1);
	else
		token = " ";
		
	ptr++;
	startAt = ptr;

	return isChar;
}

// -----------------------------------------------

int QieData::findVarNumByName(const char * name){

	for(int i=0; i<QieVars.size(); i++){
		if(QieVars[i].name == std::string(name) ) return i;
	}
	
	return -1; // not found
}


// ------------------------------------------------------

int QieData::readDataRecords(){
	dataRecords.clear();
	bool OK=true;
	inFile.blockSize = 100; 		

	if(inFile.fileType == _CDF) return maxCDFnumRecs;
		
	while(OK){
		string record;
		if( !inFile.readRecord(record) ) break;

		dataRecords.push_back(record);
		if (record.size() > inFile.blockSize) inFile.blockSize = record.size()+1; // try to optimise block size
	}
	return dataRecords.size();
}

// --------------------------------------------------------

QdObject_var QieData::getD0_obj(int vnum){

	QdObject_var Nil_obj;
	int tt_num = d0_var_num; // pick up file default if set

	if(QieVars[vnum].Vxref->xref_exists("DEPEND_0")){
		QdObject_var D0 = QieVars[vnum].Vxref->get_xref("DEPEND_0");
				
		QdString_var strD0 = QdString_var::narrow(D0);
		if( !strD0.is_nil() ){
			tt_num = findVarNumByName(strD0->c_str());
		}
		else return D0; // object itself not pointer to obj
		
		if (tt_num == -1) tt_num = d0_var_num; // failed, so reset to default
	}
	if (tt_num != -1) {
		return QieVars[tt_num].Vobj;
	}

	return Nil_obj;
}

// --------------------------------------------------------

void QieData::ensureDependZero(){

	// ensure an attribute called DEPEND_0 exists if time tags found
	for(int i=0; i< QieVars.size(); i++){
		// NRV variables should not need DEPEND_0 adding, and they will
		if( !(QieVars[i].recVary) ) continue;

		// if any variable has DEPEND_0 set leave it alone
		if(QieVars[i].Vxref->xref_exists("DEPEND_0")) continue;

		if (d0_var_num != -1){
			// set default tt. Note variable may not be read yet so create ptr to var attribute
			QdString_var tName =  new QdString(QieVars[d0_var_num].name.c_str() );
			QieVars[i].Vxref->change_xref("DEPEND_0", tName);
		}
	}
	return;

}


