#include "PlotWindow.h"

#include <cmath>

#include <QMessageBox>

QList<PlotWindow*> GlobalPlotWindows;
	 
PlotWindow::PlotWindow(int id, int width, int height, double saveWidth, double saveHeight, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation, QWidget* parent)
	:QMainWindow(parent)
{
	QMenu * plotMenu = menuBar()->addMenu("Plot");

	plotMenu->addAction("Save Plot As PNG", this, SLOT(savePlotPNG()));
	plotMenu->addAction("Save Plot As JPG", this, SLOT(savePlotJPG()));
	plotMenu->addAction("Save Plot As SVG", this, SLOT(savePlotSVG()));
	plotMenu->addAction("Save Plot As PDF", this, SLOT(savePlotPDF()));

	plotMenu->addAction("&Print", this, SLOT(printPlot()), Qt::CTRL+Qt::Key_P);

	printWeightMenu = menuBar()->addMenu("Print Weight (M)");
	printWeightMenu->addAction("Light", this, SLOT(setWeightL()));
	printWeightMenu->addAction("Medium", this, SLOT(setWeightM()));
	printWeightMenu->addAction("Heavy", this, SLOT(setWeightH()));
	
	setWeightM();
	
	this->saveWidth=saveWidth;
	this->saveHeight=saveHeight;
	this->pageSize=pageSize;
	this->orientation=orientation;
	this->resolution=resolution;
	
	plot=new QtPLWidget(width, height, saveWidth, saveHeight, resolution, this);
	setCentralWidget(plot);
	connect(plot, SIGNAL(resized(double)), this, SLOT(changeDisplayedRatio(double)));
        
		resize(width, height);
        setId(id);
        plmkstrm(&strm);
        plsetqtdev((void*)plot);
        plsdev ("qt");
		plwid(0);
        plinit();
	
	connect(plot, SIGNAL(clicked(double, double, int, int)), this, SIGNAL(clicked(double, double, int, int))); // redirection of widget "clicked" events
	connect(plot, SIGNAL(doubleClicked(double, double, int, int)), this, SIGNAL(doubleClicked(double, double, int, int))); // redirection of widget "doubleClicked" events
	
	intervalDialog=new IntervalDialog(this);
	intervalDialog->hide();

	intervalDialog->setVisible(false);
	
	connect(intervalDialog, SIGNAL(rescaleScalar(double, double)), this, SIGNAL(rescaleScalar(double, double)));
	connect(intervalDialog, SIGNAL(rescaleTime(const QmjdTimeInterval&)), this, SIGNAL(rescaleTime(const QmjdTimeInterval&)));
	
	qApp->processEvents(); // ensure initialisation is finished
	
}

void PlotWindow::setPrintRes(int res)
{
    plot->setPrintRes(res);
}

void PlotWindow::setWeightL()
{
    weight = 1.75;
	printWeightMenu->setTitle("Print Weight (L)");
}

void PlotWindow::setWeightM()
{
    weight = 2.5;
	printWeightMenu->setTitle("Print Weight (M)");
}

void PlotWindow::setWeightH()
{
    weight = 3.5;
	printWeightMenu->setTitle("Print Weight (H)");
}

void PlotWindow::setPLFocus()
{
    plsstrm(strm);
}

void PlotWindow::setTitle(const QString& title)
{
	this->title=title;
	
	setWindowTitle(title);
}

void PlotWindow::changeDisplayedRatio(double r)
{
	setWindowTitle(QString("%1 (%2%)").arg(title).arg((int)(r*100.)));
}

                        
void PlotWindow::setSmooth(bool b)
{
    if(plot) plot->setSmooth(b);
}

void PlotWindow::setId(int id){this->id=id;}
                        
int PlotWindow::getId(){return id;}

void PlotWindow::setSize(double width, double height)
{
	bool mustResize=(width!=saveWidth||height!=saveHeight);
    saveWidth=width;
    saveHeight=height;
    PLINT old_strm;
    plgstrm(&old_strm);
    plsstrm(strm);
    plot->setNewSaveSize(width, height);
    plsstrm(old_strm);
    if(mustResize)
    {
		resize((width<1000?(int)width: 1000), (height<700?(int)height:700));
    	
    }
}

void PlotWindow::savePlotPNG()
{
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "PNG file (*.png)");
	
	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);

	if(!saveName.endsWith(".png", Qt::CaseInsensitive))
	{
		saveName.append(".png");
	}
	plot->savePixmap(saveName, "PNG", (int)saveWidth, (int)saveHeight, 85, weight);
}

void PlotWindow::savePlotJPG()
{
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "JPG file (*.jpg *.jpeg)");

	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);

	if(!saveName.endsWith(".jpg", Qt::CaseInsensitive) && !saveName.endsWith(".jpeg", Qt::CaseInsensitive))
	{
		saveName.append(".jpg");
	}
	plot->savePixmap(saveName, "JPEG", (int)saveWidth, (int)saveHeight, 85, weight);
}


void PlotWindow::savePlotPDF()
{
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "PDF document (*.pdf)");

	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);
	
	if(!saveName.endsWith(".pdf", Qt::CaseInsensitive))
	{
		saveName.append(".pdf");
	}
	
	QPrinter p;
	p.setResolution(resolution);
	p.setOutputFileName(saveName);
	p.setOrientation(orientation);
	p.setColorMode(QPrinter::Color);

	p.setOutputFormat(QPrinter::PdfFormat);
	
#if QT_VERSION < 0x040400
	p.setPageSize(pageSize);
#else

	fixAspectRatio(saveWidth, saveHeight);
	
	if(orientation == QPrinter::Landscape) p.setPaperSize(QSizeF(saveHeight, saveWidth), QPrinter::DevicePixel);
	else p.setPaperSize(QSizeF(saveWidth, saveHeight), QPrinter::DevicePixel);

#endif

	double plotFactor = (saveWidth + saveHeight) / 2000.;
	plot->print(p, weight*plotFactor); 
	
}

void PlotWindow::fixAspectRatio(double &saveWidth, double &saveHeight)
{
	// ensure aspect ratio is not too large for pdf,ps and svg drivers
	
	double aspect = saveWidth/saveHeight;
	
	// reset height if ratio too great
	if(aspect > 2.5 ) {
		saveHeight = saveWidth * 0.4;
		return;
	}
	else if( aspect < 0.4) {
		saveHeight = saveWidth * 2.5;
	}
	
}

void PlotWindow::savePlotSVG()
{
#if QT_VERSION >= 0x040300
			
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "SVG file (*.svg)");

	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);

	if(!saveName.endsWith(".svg", Qt::CaseInsensitive))
	{
		saveName.append(".svg");
	}
	
	fixAspectRatio(saveWidth, saveHeight);

	plot->saveSVG(saveName, (int)saveWidth, (int)saveHeight, resolution, weight);
	
	this->hide();
	this->show();		// force repaint on Mac
#endif


	

}

void PlotWindow::printPlot()
{
	QPrinter printer;
	printer.setResolution(resolution);
	printer.setOrientation(orientation);
	printer.setOutputFileName("QSAS_plot");
	printer.setColorMode(QPrinter::Color);
	QPrintDialog printDialog(&printer, this);
	
#if QT_VERSION < 0x040400
	printer.setPageSize(pageSize);        
#else
	// vector graphics so rescalable.

	fixAspectRatio(saveWidth, saveHeight);
	
	if(orientation == QPrinter::Landscape) printer.setPaperSize(QSizeF(saveHeight/resolution, saveWidth/resolution), QPrinter::Inch);
	else printer.setPaperSize(QSizeF(saveWidth/resolution, saveHeight/resolution), QPrinter::Inch);
#endif

#if !defined(WINDOWS) && !defined(Q_OS_MACX)
	if(QSAS_profiles->current()->UseKPrinter)
	{
		printer.setPrintProgram(QString("kprinter"));
		printer.setPrinterName("foo"); // Requires a name. Ideally a printer name, but anything is fine in practice.
		plot->print(printer, weight*0.75);
	}
	else
	{
		if (printDialog.exec() == QDialog::Accepted)
		{
			printer.setPrintProgram(QString("lpr"));
			plot->print(printer, weight*0.75);
		}
	}
#else
	if (printDialog.exec() == QDialog::Accepted)
	{
		plot->print(printer, weight*0.75);
	}
#endif

	
}

PlotWindow* newPlotWindow(double width, double height, int* id, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation)
{
	// find free id
	
	bool found_id=false;
	bool this_one_ok;
	*id=1;
	while(!found_id)
	{
		this_one_ok=true;

		for(QList<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end() && this_one_ok; ++i)
		{
			this_one_ok=this_one_ok && ((*i)->getId()!=*id);
		}
		if(this_one_ok)
		{
			found_id=true;
		}
		else
		{
			(*id)++;
		}
	}

	
	PlotWindow* res=new PlotWindow(*id, (width<1000?(int)width: 1000), (height<700?(int)height:700), width, height, resolution, pageSize, orientation, NULL);

	GlobalPlotWindows.append(res);

	res->setTitle(QString("QSAS Plot Window #%1").arg(*id));
	res->show();
	return res;
}

PlotWindow* focusPlotWindow(int id, double width, double height, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation)
{
	PlotWindow * res;

	for(QList<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end(); ++i)
	{
		if((*i)->getId()==id)
		{
			(*i)->show();
			if(width>0. && height > 0.)
			{
				(*i)->setSize(width, height);
				(*i)->resolution=resolution;
				(*i)->pageSize=pageSize;
				(*i)->orientation=orientation;
			}
			(*i)->setPLFocus();
			return (*i);
		}
	}
	
	// If reached this point, the window does not exist.
	// Force creation of Window with this id
	
	res=new PlotWindow(id, (width<1000?(int)width: 1000), (height<700?(int)height:700), width, height, resolution, pageSize, orientation, NULL);

	GlobalPlotWindows.append(res);

	res->setTitle(QString("QSAS Plot Window #%1").arg(id));
	res->show();
	return res;
}

void clearPlotWindows()
{
	for(QList<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end(); ++i)

    {
        PlotWindow * temp=*i;
	temp->close();
        delete temp;
    }
    GlobalPlotWindows.clear();
}

