#include "qpui.h"

extern int qsas_restore_release;
int importedPaletteOffset;

////// QpuiItem

QpuiItem::QpuiItem(const QString& name, QpuiItem* parent):
	_name(name)
{
	_parent=NULL;
	if(parent!=NULL)
	{
		parent->addChild(this);
	}
	else
	{
		_path=_name;
	}
}

bool QpuiItem::setName(const QString& newName)
{
	if(_parent==NULL)
	{
		_name=newName;
		refreshPath();
		return true;
	}
	
	QString tempName(newName);
	QpuiItem* item;
	item=_parent->getChildByName(tempName);
	while(item!=NULL && item!=this)
	{
		tempName.append("A");
		item=_parent->getChildByName(tempName);
	}
	if(tempName!=_name)
	{
		_name=tempName;
		refreshPath();
	}

    if(root()!=NULL)
    {
        if(root()->I_PlotView!=NULL)
        {
            root()->I_PlotView->resizeColumnToContents(0); //required, to have the columns resize properly
        }
    }
	
	return true;
}

bool QpuiItem::isParentOf(QpuiItem* other)
{
	if(this==other) return true;
	
	bool res=false;
	for(QList<QpuiItem*>::iterator iter=_children.begin(); iter!=_children.end() && !res; ++iter)
	{
		res=(*iter)->isParentOf(other);
	}
	return res;
}

bool QpuiItem::isDescendentOf(QpuiItem* other)
{
	if(this==other) return true;
	
	if(_parent==NULL) return false;
	
	return _parent->isDescendentOf(other);
}

QpuiItem* QpuiItem::getChildByName(const QString& name) const
{
	if(_children.isEmpty()) return NULL;
	for(QList<QpuiItem*>::const_iterator iter=_children.begin(); iter!=_children.end(); ++iter)
	{
		if((*iter)->_name==name) return *iter;
	}
	return NULL;
}

QpuiItem* QpuiItem::getChildByPath(const QString& path) const
{
	QString tempPath(path);
	while(tempPath.startsWith(PATH_SEPARATOR)) tempPath.remove(0,QString(PATH_SEPARATOR).length());
	if(path.contains(PATH_SEPARATOR))
	{
		QStringList splitList=path.split(PATH_SEPARATOR, QString::SkipEmptyParts);
		
		QpuiItem* child=getChildByName(splitList.at(0));
		if(child)
		{
			splitList.removeFirst();
			return child->getChildByPath(splitList.join(PATH_SEPARATOR));
		}
	}
	else
	{
		return getChildByName(path);
	}
	
	return NULL;
}

QpuiItem* QpuiItem::getChildByNumber(int i) const
{
	if(i<0 || i>=_children.size()) return NULL;

	return _children.at(i);
}

int QpuiItem::childNumber()
{

	if(_parent!=NULL)
	{
		return _parent->_children.indexOf(this);
	}
	return -1;
}


bool QpuiItem::removeChild(QpuiItem* child, bool alsoDelete)
{
	int i=_children.indexOf(child);
	
	if(i<0) return false;
	
	_children.removeAt(i);
	
	if(alsoDelete)
	{
		delete child;
	}
	return true;
}

bool QpuiItem::removeChild(int i, bool alsoDelete)
{
	if(i<0 || i>=_children.size())
		return false;
	
	if(alsoDelete)
	{
		QpuiItem* child=_children.takeAt(i);
		delete child;
	}
	else
	{
		_children.removeAt(i);
	}
	return true;
}


void QpuiItem::refreshPath()
{
	if(_parent==NULL)
	{
		if(_path!=_name)
		{
			QString oldPath=_path;
			_path=_name;
// 			emit pathChanged(oldPath, _path);
		}
	}
	else
	{
		QString oldPath=_path;
		_path=_parent->_path+PATH_SEPARATOR+_name;
		if(_path.startsWith(PATH_SEPARATOR)) _path=_path.remove(0,QString(PATH_SEPARATOR).length());
		if(_path!=oldPath)
		{
// 			emit pathChanged(oldPath, _path);
		}
		foreach(QpuiItem* child, _children)
		{
			child->refreshPath();
		}
	}
}

void QpuiItem::addCustomColourTableTo(ColourTablePalette& palette)
{
    foreach(QpuiItem* child, _children)
    {
        child->addCustomColourTableTo(palette);
    }
}

void QpuiItem::clear()
{
	qDeleteAll(_children);
	_children.clear();
}

bool QpuiItem::ancestorInList(QList <QpuiItem*> &list)
{
	for(int i=0; i<list.size(); i++)
	{
		if(this != list[i] && this->isDescendentOf(list[i]) ) return true;	
	}
	return false;
}


Vistream& operator>>(Vistream& v_is, QpuiItem& item)
{
	int with;
	v_is >> with;
	string text;
	v_is >> text;

	if( qsas_restore_release < 20400) cleanWLtext(text);
	item.setName(text.c_str());
	bool lock;
	v_is >>lock;

	int i;
	v_is >> i;

	int type;
	v_is >> type;

	return v_is;
}

Vostream& save(Vostream& v_os, QpuiItem& item)
{
	v_os << (int)0; //item.with in older version
	v_os << std::string(item.getName().toLatin1().data());
	v_os << (bool)false;//item.Lock in older version

	return v_os;
}

QDomElement QpuiItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Plot_Item");
    return xml;
}

void QpuiItem::fromXMLElement(QDomElement& el)
{
}

////// RootItem

RootItem::RootItem(PlotView* plotView)
	:QpuiItem("", NULL)
{
	I_PlotView=plotView;
}

RootItem::~RootItem(){;}

bool RootItem::addChild(QpuiItem* newChild, int position)
{
	if(newChild==NULL || _children.indexOf(newChild)>=0) return false;
	
	PageItem* pageChild=dynamic_cast<PageItem*>(newChild);
	FrameItem* frameChild=dynamic_cast<FrameItem*>(newChild);
	PanelItem* panelChild=dynamic_cast<PanelItem*>(newChild);
	PlotItem* plotChild=dynamic_cast<PlotItem*>(newChild);
	
	if(pageChild!=NULL)
	{
		QString childName=newChild->_name;
		while(getChildByName(childName))
		{
			childName.append("A");
		}
		
		newChild->_parent=this;
		if(position<0 || position>=_children.size())
		{
			_children.push_back(newChild);
		}
		else
		{
			_children.insert(position, newChild);
		}
		newChild->_name=childName;
		newChild->refreshPath();
	}
	else if(frameChild!=NULL || panelChild!=NULL || plotChild!=NULL)
	{
		// create new page, insert child
		PageItem* newPage=new PageItem("", this, position);
		newPage->addChild(newChild);
	}
	else return false;
	
	return true;
}

// Icons for list view
 
 static const char * page_xpm[] = {
  "12 16 4 1",
  " 	c None",
  "X	c #A0A0A0",
  ".	c #FFFFFF",
  "#	c #E0E0E0",
  "            ",
  " XXXXXX     ",
  " X....XX    ",
  " X....X.X   ",
  " X....XXXX# ",
  " X.......X# ",
  " X.......X# ",
  " X.......X# ",
  " X.......X# ",
  " X.......X# ",
  " X.......X# ",
  " X.......X# ",
  " X.......X# ",
  " XXXXXXXXX# ",
  "  ######### ",
  "            "};
  
static const char * frame_xpm[] = {
   "12 16 4 1",
   " 	c None",
   "X	c #A0A0A0",
   ".	c #FFFFFF",
   "#	c #E0E0E0",
   "            ",
   "   X X  X   ",
   "  XXXX XXX  ",
   "            ",
   " XXXXXXXXX  ",
   " X.......X# ",
   " X.......X# ",
   " XXXXXXXXX# ",
   " X.......X# ",
   " X.......X# ",
   " XXXXXXXXX# ",
   " X.......X# ",
   " X.......X# ",
   " XXXXXXXXX# ",
   "  ######### ",
   "            "};

static const char * panel_xpm[] = {
    "12 16 6 1",
    " 	c None",
    "X	c #A0A0A0",
    ".	c #FFFFFF",
    "1	c #FF4040",
    "2	c #40B040",
    "#	c #E0E0E0",
    "            ",
    " XXXXXXXXX  ",
    " X.......X# ",
    " X.......X# ",
    " XX.2...XX# ",
    " X.222...X# ",
    " X22.2221X# ",
    " XX...21XX# ",
    " X....11.X# ",
    " X..111..X# ",
    " XX111..XX# ",
    " X11.....X# ",
    " X1......X# ",
    " XXXXXXXXX# ",
    "  ######### ",
    "            "};
    
static const char * constxt_xpm[] = {
     "12 16 4 1",
     " 	c None",
     "X	c #FF7070",
     ".	c #FFFFFF",
     "#	c #DF5050",
     "            ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     " ....X....  ",
     "            "};
     
static const char * consty_xpm[] = {
	     "12 16 4 1",
      " 	c None",
      "X	c #FF7070",
      ".	c #FFFFFF",
      "#	c #DF5050",
      "            ",
      " .......... ",
      " .......... ",
      " .......... ",
      " .......... ",
      " .......... ",
      " .......... ",
      " XXXXXXXXXX ",
      " .......... ",
      " .......... ",
      " .......... ",
      " .......... ",
      " .......... ",
      " .......... ",
      "            ",
      "            "};
      
static const char * trace_xpm[] = {
	      "12 16 4 1",
       " 	c None",
       "X	c #FF7070",
       ".	c #FFFFFF",
       "#	c #DF5050",
       "            ",
       " .......... ",
       " .......... ",
       " .......... ",
       " .......... ",
       " ..XX...... ",
       " .X..X..... ",
       " X....X.... ",
       " X....X.... ",
       " ......X..X ",
       " .......XX. ",
       " .......... ",
       " .......... ",
       " .......... ",
       " .......... ",
       "            "};

static const char * linear_xpm[] = {
	      "12 16 4 1",
       " 	c None",
       "X	c #FF7070",
       ".	c #FFFFFF",
       "#	c #DF5050",
       "            ",
       " .......... ",
       " X......... ",
       " .X........ ",
       " ..X....... ",
       " ...X...... ",
       " ....X..... ",
       " .....X.... ",
       " ......X... ",
       " .......X.. ",
       " ........X. ",
       " .........X ",
       " .......... ",
       " .......... ",
       " .......... ",
       "            "};

  
  static const char * spectrum_xpm[] = {
   "12 16 6 1",
   " 	c None",
   "D	c #0178FE",
   "B	c #17DEDE",
   ".	c #40FE35",
   "Y	c #E9F500",
   "R	c #FF5218",
   "            ",
   " DDDDBDDDBD ",
   " DBDBDBDDDB ",
   " BBBDBBBDBB ",
   " .BBB.BBB.B ",
   " BBB.BBB.BB ",
   " .B...B...B ",
   " ..Y....Y.. ",
   " Y...Y...Y. ",
   " YY.Y.Y.YYY ",
   " YYYYYYYYRY ",
   " RRYRYRYYYR ",
   " RRRYRYRYRR ",
   " RYRRRRYRRR ",
   " RRRRRRRRRR ",
   "            "};
 
 static const char * event_xpm[] = {
	 "12 16 4 1",
  " 	c None",
  "X	c #A0A0A0",
  ".	c #FFFFFF",
  "#	c #E0E0E0",
  "            ",
  " XXXXXXXXX  ",
  " X.......X# ",
  " X..XXX..X# ",
  " X..X....X# ",
  " X..XX...X# ",
  " X..X....X# ",
  " X..XXX..X# ",
  " X.......X# ",
  " XXXXXXXXX# ",
  "  ###X##### ",
  "     X#     ",
  "     X#     ",
  "     X#     ",
  "      #     ",
  "            "};
       
PlotModel::PlotModel(PlotView* parent):
	QAbstractItemModel((QObject*)parent)
{
	_root=new RootItem(parent);
	_root->tabWidget=parent->tabWidget;
	
	_view=parent;
	connect(this,SIGNAL(pageLayoutChanged(PageItem*)), this,SLOT(setPageLayout(PageItem*)));
	
	LinkNumber=0;
}

PlotModel::~PlotModel()
{
	delete _root;
}

QModelIndex PlotModel::index(int row, int column, const QModelIndex& parent) const
{
	if (!hasIndex(row, column, parent))
		return QModelIndex();
	QpuiItem *parentItem;

	if (!parent.isValid())
		parentItem = _root;
	else
		parentItem = itemFromIndex(parent);
	
	QpuiItem *childItem = parentItem->getChildByNumber(row);
	if (childItem)
		return createIndex(row, column, childItem);
	else
		return QModelIndex();
}

QModelIndex PlotModel::parent(const QModelIndex& index) const
{
	if (!index.isValid()) return QModelIndex();
	QpuiItem *childItem = itemFromIndex(index);
	if(childItem==NULL || childItem==_root) return QModelIndex();
	
	QpuiItem *parentItem = childItem->parent();

	if (parentItem == _root || parentItem==NULL)
		return QModelIndex();


	if (parentItem == _root || parentItem==NULL)	return QModelIndex();
	
	return createIndex(parentItem->childNumber(), 0, parentItem);
}

int PlotModel::rowCount(const QModelIndex& parent) const
{
	QpuiItem *parentItem;
	if (parent.column() > 0)
		return 0;

	if (!parent.isValid())
		parentItem = _root;
	else
		parentItem = itemFromIndex(parent);
		
	return parentItem->numChildren();
}

int PlotModel::columnCount(const QModelIndex&) const
{
	return 2;
}

QVariant PlotModel::data(const QModelIndex& index, int role) const
{
	if (!index.isValid())	return QVariant();

	if(index.column()==0)
	{
		if (role == Qt::DisplayRole)
		{
			QpuiItem *item = itemFromIndex(index);
			return item->getName();
		}
		else if(role==Qt::DecorationRole)
		{
			QpuiItem* item=itemFromIndex(index);
			
			if(dynamic_cast<PageItem*>(item)!=NULL)
			{
				return QVariant(QIcon(QPixmap(page_xpm)));
			}
			else if(dynamic_cast<FrameItem*>(item)!=NULL)
			{
				return QVariant(QIcon(QPixmap(frame_xpm)));
			}
			else if(dynamic_cast<PanelItem*>(item)!=NULL)
			{
				return QVariant(QIcon(QPixmap(panel_xpm)));
			}
			else if(dynamic_cast<PlotItem*>(item)!=NULL)
			{
				TraceItem* traceItem=dynamic_cast<TraceItem*>(item);
				SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(item);
				
				if(traceItem!=NULL)
				{
					if(traceItem->I_Type==QP_CONST_XT)
						return QVariant(QIcon(QPixmap(constxt_xpm)));
					if(traceItem->I_Type==QP_CONST_Y)
						return QVariant(QIcon(QPixmap(consty_xpm)));
					if(traceItem->I_Type==QP_REG_LINE)
						return QVariant(QIcon(QPixmap(linear_xpm)));
					return QVariant(QIcon(QPixmap(trace_xpm)));
				}
				else if(spectrumItem!=NULL)
					return QVariant(QIcon(QPixmap(spectrum_xpm)));
				
				//else event
				return QVariant(QIcon(QPixmap(event_xpm)));
			}
			return QVariant();
		}
	}
	else if(index.column()==1)
	{
		if (role == Qt::DisplayRole)
		{
			PlotItem *item = dynamic_cast<PlotItem*>(itemFromIndex(index));
			if(item!=NULL)
			{
 				return item->getLinkText();
			}
		}
	}
	
	return QVariant();
}

QVariant PlotModel::headerData(int section, Qt::Orientation orientation, int role) const
{
	if(role!=Qt::DisplayRole) return QVariant();
	
	if(section==0) return QVariant("Plot components");
	return QVariant("Links");
}

Qt::ItemFlags PlotModel::flags(const QModelIndex& index) const
{
	return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable | Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;
}

bool PlotModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
	if (index.isValid() && index.column()==0 && role == Qt::EditRole)
	{
		QpuiItem* item=itemFromIndex(index);
		if(item)
		{
			item->setName(value.toString().trimmed());
				emit dataChanged(index, index);
			return true;
		}
	}
	return false;
}

QModelIndex PlotModel::indexFromItem(QpuiItem* item)
{
	if(item==NULL ) return QModelIndex();
	if(item==_root) return _view->rootIndex();
	return createIndex(item->childNumber(), 0, item);
}

void PlotModel::clear()
{
	int rows = _root->numChildren();
	
	beginRemoveRows(_view->rootIndex(), 0, rows);

	for (int row = 0; row < rows; ++row)
	{
		_root->removeChild(0, true); // keep removing at same position as row changes
	}

	endRemoveRows();

	
}

bool PlotModel::removeRows(int position, int rows, const QModelIndex& parent)
{

	if(!parent.isValid()) return false;

	QpuiItem* parentItem=itemFromIndex(parent);
	
	if(parentItem==NULL) return false;

	beginRemoveRows(parent, position, position+rows-1);

	for (int row = 0; row < rows; ++row)
	{
		parentItem->removeChild(position, true);
	}

	endRemoveRows();
	return true;
	
}

bool PlotModel::insertColumns(int, int, const QModelIndex&)
{
	return false;
}

bool PlotModel::removeColumns(int, int, const QModelIndex&)
{
	return false;
}

// Drag and drop tools

QStringList PlotModel::mimeTypes() const
{
	QStringList types;
	types << "application/plotitems";
	types << "application/wlitems";
	return types;
}

QMimeData* PlotModel::mimeData(const QModelIndexList& indexes) const
{
	QMimeData *mimeData = new QMimeData();
	
	QStringList idList;

	foreach (QModelIndex index, indexes) {
		if (index.isValid()) {
			idList << itemFromIndex(index)->getPath();
		}
	}
	
	QString data=idList.join("//");

	mimeData->setData("application/plotitems", data.toLatin1().data());
	mimeData->setText(itemFromIndex(indexes.first())->getPath());
	return mimeData;
}

bool PlotModel::dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent)
{	
	if(QuSlotDebug()) cout << "PlotModel::dropMimeData row= " << row << " col= " << column <<endl;


	if( row == -1 ) row = _view->dropRow; // QT seems to not want to give us the row when dropping on sibling

	if (action == Qt::IgnoreAction)
		return true;

	if (!data->hasFormat("application/plotitems") && !data->hasFormat("application/wlitems"))
		return false;

	if (column > 0)
		return false;
		
	QPersistentModelIndex parentIndex=parent; //to avoid issues due to the "const" nature of parent
	
	if(!parentIndex.isValid()) // Drop in the bottom item, get last page if exists
	{
		QpuiItem* currentItem=_root;
		while(currentItem->hasChildren())
		{
			currentItem=currentItem->getChildByNumber(currentItem->numChildren()-1);
		}
        if(dynamic_cast<PlotItem*>(currentItem)!=NULL)
        {
            currentItem=(dynamic_cast<PlotItem*>(currentItem))->page();
        }
        else if(dynamic_cast<PanelItem*>(currentItem)!=NULL)
        {
            currentItem=(dynamic_cast<PanelItem*>(currentItem))->page();
        }
        if(dynamic_cast<FrameItem*>(currentItem)!=NULL)
        {
            currentItem=(dynamic_cast<FrameItem*>(currentItem))->page();
        }
        if(dynamic_cast<PageItem*>(currentItem)!=NULL)
        {
            currentItem=dynamic_cast<PageItem*>(currentItem);
        }
		parentIndex=indexFromItem(currentItem);
	}

	bool res=false;

	if(data->hasFormat("application/plotitems"))
	{
		res=true;
		QpuiItem* parentItem=itemFromIndex(parentIndex);
		if(parentItem==NULL)
			return false;
		
		int actualRow=row;
		
		PageItem* originalPage=NULL;
		
		QStringList itemList=QString(data->data("application/plotitems")).split("//");
		
		QpuiItem* currentItem;
		
		emit layoutAboutToBeChanged();
		

		if(action==Qt::MoveAction) // What is labelled MoveAction is a copy in fact in our case
		{
			QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );
			_view->clearSelection();
			foreach(QString itemString, itemList)
			{
				currentItem=_root->getChildByPath(itemString);
				if(dynamic_cast<PageItem*>(currentItem)!=NULL)
				{
					PageItem* pageItem=dynamic_cast<PageItem*>(currentItem);
					new PageItem(pageItem, parentItem->root(), actualRow);
				}
				else if(dynamic_cast<FrameItem*>(currentItem)!=NULL)
				{
					FrameItem* frameItem=dynamic_cast<FrameItem*>(currentItem);
					if(dynamic_cast<RootItem*>(parentItem)==NULL && dynamic_cast<PageItem*>(parentItem)==NULL)
					{
						QApplication::restoreOverrideCursor();
						std::cout << "Err: impossible to paste a frame in something not a page" << std::endl;
						return false;
					}
					new FrameItem(frameItem, parentItem, actualRow);
					originalPage=frameItem->page();
				}
				else if(dynamic_cast<PanelItem*>(currentItem)!=NULL)
				{
					PanelItem* panelItem=dynamic_cast<PanelItem*>(currentItem);
					if(dynamic_cast<RootItem*>(parentItem)==NULL && dynamic_cast<PageItem*>(parentItem)==NULL && dynamic_cast<FrameItem*>(parentItem)==NULL)
					{
						QApplication::restoreOverrideCursor();
						std::cout << "Err: impossible to paste a panel in something not a page or a frame" << std::endl;
						return false;
					}
					new PanelItem(panelItem, parentItem, actualRow);
					originalPage=panelItem->page();
				}
				else if(dynamic_cast<PlotItem*>(currentItem)!=NULL)
				{
					if(dynamic_cast<PlotItem*>(parentItem)!=NULL)
					{
						QApplication::restoreOverrideCursor();
						std::cout << "Err: impossible to paste a plot in something not a page, a frame or a panel" << std::endl;
						return false;
					}
					TraceItem* traceItem=dynamic_cast<TraceItem*>(currentItem);
					EventItem* eventItem=dynamic_cast<EventItem*>(currentItem);
					SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(currentItem);
					if(traceItem!=NULL) new TraceItem(traceItem, parentItem, actualRow);
					if(eventItem!=NULL) new EventItem(eventItem, parentItem, actualRow);
					if(spectrumItem!=NULL) new SpectrumItem(spectrumItem, parentItem, actualRow);
					
					originalPage=(dynamic_cast<PlotItem*>(currentItem))->page();
				}
				
			}
			QApplication::restoreOverrideCursor();
		}
		else if(action==Qt::CopyAction) // What is labelled CopyAction is a move
		{
			_view->clearSelection();
			foreach(QString itemString, itemList)
			{
			
				currentItem=_root->getChildByPath(itemString);
				if(currentItem!=NULL)
				{
					FrameItem* fi=dynamic_cast<FrameItem*>(currentItem);
					PanelItem* pi=dynamic_cast<PanelItem*>(currentItem);
					PlotItem* pli=dynamic_cast<PlotItem*>(currentItem);
					if(fi!=NULL) originalPage=fi->page();
					if(pi!=NULL) originalPage=pi->page();
					if(pli!=NULL) originalPage=pli->page();
					
					res=res && currentItem->reparent(parentItem, actualRow);
				}
			}

		}
		
		PageItem* currentPage=dynamic_cast<PageItem*>(parentItem);
		if(currentPage==NULL)
		{
			FrameItem* currentFrame=dynamic_cast<FrameItem*>(parentItem);
			PanelItem* currentPanel=dynamic_cast<PanelItem*>(parentItem);
			PlotItem* currentPlot=dynamic_cast<PlotItem*>(parentItem);
			if(currentFrame!=NULL) currentPage=currentFrame->page();
			if(currentPanel!=NULL) currentPage=currentPanel->page();
			if(currentPlot!=NULL) currentPage=currentPlot->page();
		}
		if(currentPage!=NULL) emit pageLayoutChanged(currentPage);
		
		//Change layout of origin page, if moved something from it
		if(originalPage!=NULL && originalPage!=currentPage) emit pageLayoutChanged(originalPage);

		emit layoutChanged();
		_view->clearSelection();

		
	}
	else if(data->hasFormat("application/wlitems"))
	{
		QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );
		QpuiItem* parentItem=itemFromIndex(parentIndex);
		QStringList itemList=QString(data->data("application/wlitems")).split("//");
		
		QpuiItem* currentItem;

		if(itemList.size()==1 && dynamic_cast<PlotItem*>(parentItem)!=NULL)
		{

			((PlotItem*)parentItem)->replaceData(itemList[0]);
			((PlotItem*)parentItem)->DataTab->propagateLink();
		}
		else
		{
			if(dynamic_cast<PlotItem*>(parentItem)!=NULL)
				parentItem=parentItem->parent(); // in case of a multiple drop on a Plot item, we add to the item's parent
			foreach(QString itemString, itemList)
			{
				 makePlotItem(parentItem, itemString, row);
			}
		}

		PageItem* currentPage=dynamic_cast<PageItem*>(parentItem);
		if(currentPage==NULL)
		{
			FrameItem* currentFrame=dynamic_cast<FrameItem*>(parentItem);
			PanelItem* currentPanel=dynamic_cast<PanelItem*>(parentItem);
			PlotItem* currentPlot=dynamic_cast<PlotItem*>(parentItem);
			if(currentFrame!=NULL) currentPage=currentFrame->page();
			if(currentPanel!=NULL) currentPage=currentPanel->page();
			if(currentPlot!=NULL) currentPage=currentPlot->page();
		}
		if(currentPage!=NULL) emit pageLayoutChanged(currentPage);
		QApplication::restoreOverrideCursor();
	}
 	_view->expandAll();
	return res;
}

PlotItem * PlotModel::makePlotItem(QpuiItem * parent, const QString& obj_name, int row)
{

//	Make plot item creating Frame and Panel if needed
	if ( QuSlotDebug() )  cout << "PlotModel::makePlotItem - " << obj_name.toLatin1().data() << endl;
	if(parent == NULL) return NULL;
	
	// fetch object
	QdObject_var obj = QuGetObjByName(obj_name);
	if(obj.is_nil())  return NULL;
			cout << "=== makePlotItem  " << QuIsThreeVector(obj) << endl;
		
	QpFrameType frameType = QP_TS; // default
	if(get_timetags(obj).is_nil()) frameType = QP_XY;

	// default plot type is trace
	QpItemType plot_type = QP_TRACE;
	// check if sequence
	if(!obj->is_sequence())
	{
		QdTimeInterval_var tint = QdTimeInterval_var::narrow(obj);
		QdRScalar_var rscalar_var = QdRScalar_var::narrow(obj);
		QdRMatrix_var rmatrix_var = QdRMatrix_var::narrow(obj);

		if(!tint.is_nil())
		{
			plot_type = QP_CONST_XT;
			frameType = QP_TS;
		}
		else if(!rscalar_var.is_nil()) plot_type = QP_CONST_Y;
		else if(!rmatrix_var.is_nil())
		{
			plot_type = QP_TRACE;
			frameType = QP_XY;
		}
		else
		{
          (void)QMessageBox::warning(NULL,"","Invalid object");
          return NULL;
		}
	}
	else
	{
		if(obj->sequence_size() == 0)
        {
			(void)QMessageBox::warning(NULL,"", "Invalid object - empty sequence");
			return NULL;
        }
		QdRScalarSeq_var rscalarseq = QdRScalarSeq_var::narrow(obj);
        QdRMatrixSeq_var rmatrixseq = QdRMatrixSeq_var::narrow(obj);
        
		// if timetags, default plot type is tstrace
		
		if(!rscalarseq.is_nil())
		{
			plot_type = QP_TRACE;
	
		}
		else if(QuIsThreeVector(obj) )
		{
			QString vectorPlot=QSAS_profiles->current()->Vector3D;
			cout << "=== make  " << vectorPlot.toLatin1().data() << endl;
			if( dynamic_cast<PanelItem*>(parent)!=NULL)
			{
				// use existing panel with component
				plot_type = QP_TRACE;
			}
			else if(vectorPlot=="RXYZ")
			{
				PlotItem* plotItem=makeVector_RXYZ_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="R_XYZ")
			{
				PlotItem* plotItem=makeVector_R_XYZ_Item(parent, frameType);
				if(plotItem)
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="XYZ")
			{
				PlotItem* plotItem=makeVector_XYZ_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="R_TP")
			{
				PlotItem* plotItem=makeVector_R_TP_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="R_LP")
			{
				PlotItem* plotItem=makeVector_R_LP_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="R_X_Y_Z")
			{
				PlotItem* plotItem=makeVector_R_X_Y_Z_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="X_Y_Z")
			{
				PlotItem* plotItem=makeVector_X_Y_Z_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="R_T_P")
			{
				PlotItem* plotItem=makeVector_R_T_P_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else if(vectorPlot=="R_L_P")
			{
				PlotItem* plotItem=makeVector_R_L_P_Item(parent, frameType);
				if(plotItem) 
				{
					plotItem->replaceData(obj_name);
					plotItem->DataTab->propagateLink();
				}
				return NULL;
			}
			else //"R" if everything goes right
				plot_type = QP_TRACE;
//				continue and create scalar trace
		}
		else if(!rmatrixseq.is_nil())
        {
				plot_type = QP_SPECTROGRAM;
		}
        else
        {
			(void)QMessageBox::warning(NULL,"","Invalid object type");
			return NULL;
        }
	}
	
	// create object of right type

	PlotItem *newPlot = makePlot(parent, plot_type, frameType, obj_name, row);
	emit pageLayoutChanged(newPlot->page());
	return newPlot;
}

void PlotModel::makeMultiTraceItem(QpuiItem* parent, QVector <int> &nTraces, QVector <TraceItem *> &stack, QpFrameType frameType)
{
	stack.clear();

    QpuiItem* newParent=parent;
    if(dynamic_cast<PageItem*>(parent)!=NULL || dynamic_cast<RootItem*>(parent)!=NULL)
    {
        newParent=new FrameItem("", parent, frameType);
    }
    
	for(int i=0; i<nTraces.size(); ++i)
	{
		PanelItem* panelItem=new PanelItem("", newParent);
		if(panelItem)
		{
			for(int j=0; j<nTraces[i]; ++j)
			{
				TraceItem* traceItem=makeTrace(panelItem, "");//new TraceItem("", panelItem);
				if(traceItem)
				{
					stack.push_back(traceItem);
				}
			}
		}
		emit pageLayoutChanged(panelItem->page());
	}
	
	
}

TraceItem* PlotModel::makeVector_RXYZ_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent()->parent();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=parent->parent();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector<TraceItem *> stack;	QVector <int> traces;
	traces.push_back(4);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 4)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[0]->TraceTab->setLineColour(stack[0]->R_colour);
		stack[0]->userLineColour = true;
		stack[0]->DataTab->_show_legend_fld->setChecked(true);
		stack[0]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[1]->DataTab->_y_data_fld->setText(X_STR_CMP);
		stack[1]->TraceTab->setLineColour(stack[1]->X_colour);
		stack[1]->userLineColour = true;
		stack[1]->DataTab->_show_legend_fld->setChecked(true);
		stack[2]->DataTab->_y_data_fld->setText(Y_STR_CMP);
		stack[2]->TraceTab->setLineColour(stack[2]->Y_colour);
		stack[2]->userLineColour = true;
		stack[2]->DataTab->_show_legend_fld->setChecked(true);
		stack[3]->DataTab->_y_data_fld->setText(Z_STR_CMP);
		stack[3]->TraceTab->setLineColour(stack[3]->Z_colour);
		stack[3]->userLineColour = true;
		stack[3]->DataTab->_show_legend_fld->setChecked(true);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];

}

TraceItem* PlotModel::makeVector_R_XYZ_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent()->parent();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=parent->parent();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;
	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(3);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 4)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[0]->TraceTab->setLineColour(stack[0]->R_colour);
		stack[0]->userLineColour = true;
		stack[0]->DataTab->_show_legend_fld->setChecked(true);
		stack[0]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[1]->DataTab->_y_data_fld->setText(X_STR_CMP);
		stack[1]->TraceTab->setLineColour(stack[1]->X_colour);
		stack[1]->userLineColour = true;
		stack[1]->DataTab->_show_legend_fld->setChecked(true);
		stack[1]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[2]->DataTab->_y_data_fld->setText(Y_STR_CMP);
		stack[2]->TraceTab->setLineColour(stack[2]->Y_colour);
		stack[2]->userLineColour = true;
		stack[2]->DataTab->_show_legend_fld->setChecked(true);
		
		stack[3]->DataTab->_y_data_fld->setText(Z_STR_CMP);
		stack[3]->TraceTab->setLineColour(stack[3]->Z_colour);
		stack[3]->userLineColour = true;
		stack[3]->DataTab->_show_legend_fld->setChecked(true);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];

}

TraceItem* PlotModel::makeVector_XYZ_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent()->parent();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=parent->parent();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(3);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 3)
	{
		stack[0]->DataTab->_y_data_fld->setText(X_STR_CMP);
		stack[0]->TraceTab->setLineColour(stack[0]->X_colour);
		stack[0]->userLineColour = true;
		stack[0]->DataTab->_show_legend_fld->setChecked(true);
		stack[0]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[1]->DataTab->_y_data_fld->setText(Y_STR_CMP);
		stack[1]->TraceTab->setLineColour(stack[1]->Y_colour);
		stack[1]->userLineColour = true;
		stack[1]->DataTab->_show_legend_fld->setChecked(true);
		
		stack[2]->DataTab->_y_data_fld->setText(Z_STR_CMP);
		stack[2]->TraceTab->setLineColour(stack[2]->Z_colour);
		stack[2]->userLineColour = true;
		stack[2]->DataTab->_show_legend_fld->setChecked(true);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];

}

TraceItem* PlotModel::makeVector_R_LP_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent()->parent();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=parent->parent();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(2);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 3)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[0]->TraceTab->setLineColour(stack[0]->R_colour);
		stack[0]->userLineColour = true;
		stack[0]->DataTab->_show_legend_fld->setChecked(true);
		stack[0]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[1]->DataTab->_y_data_fld->setText(Ld_STR_CMP);
		stack[1]->TraceTab->setLineColour(stack[1]->L_colour);
		stack[1]->userLineColour = true;
		stack[1]->DataTab->_show_legend_fld->setChecked(true);
		stack[1]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[2]->DataTab->_y_data_fld->setText(Pd_STR_CMP);
		stack[2]->TraceTab->setLineColour(stack[2]->P_colour);
		stack[2]->userLineColour = true;
		stack[2]->DataTab->_show_legend_fld->setChecked(true);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];

}

TraceItem* PlotModel::makeVector_R_TP_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent()->parent();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=parent->parent();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(2);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 3)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[0]->TraceTab->setLineColour(stack[0]->R_colour);
		stack[0]->userLineColour = true;
		stack[0]->DataTab->_show_legend_fld->setChecked(true);
		stack[0]->panel()->LegendTab->_show_legend_fld->setChecked(true);

		stack[1]->DataTab->_y_data_fld->setText(Td_STR_CMP);
		stack[1]->TraceTab->setLineColour(stack[1]->T_colour);
		stack[1]->userLineColour = true;
		stack[1]->DataTab->_show_legend_fld->setChecked(true);
		stack[1]->panel()->LegendTab->_show_legend_fld->setChecked(true);
		
		stack[2]->DataTab->_y_data_fld->setText(Pd_STR_CMP);
		stack[2]->TraceTab->setLineColour(stack[2]->P_colour);
		stack[2]->userLineColour = true;
		stack[2]->DataTab->_show_legend_fld->setChecked(true);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];

}

TraceItem* PlotModel::makeVector_R_X_Y_Z_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 4)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[1]->DataTab->_y_data_fld->setText(X_STR_CMP);
		stack[2]->DataTab->_y_data_fld->setText(Y_STR_CMP);
		stack[3]->DataTab->_y_data_fld->setText(Z_STR_CMP);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::makeVector_X_Y_Z_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 3)
	{
		stack[0]->DataTab->_y_data_fld->setText(X_STR_CMP);
		stack[1]->DataTab->_y_data_fld->setText(Y_STR_CMP);
		stack[2]->DataTab->_y_data_fld->setText(Z_STR_CMP);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::makeVector_R_T_P_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 3)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[1]->DataTab->_y_data_fld->setText(Td_STR_CMP);
		stack[2]->DataTab->_y_data_fld->setText(Pd_STR_CMP);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::makeVector_R_L_P_Item(QpuiItem* parent, QpFrameType frameType, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, frameType);
	
	// set component text
	if(stack.size() == 3)
	{
		stack[0]->DataTab->_y_data_fld->setText(R_STR_CMP);
		stack[1]->DataTab->_y_data_fld->setText(Ld_STR_CMP);
		stack[2]->DataTab->_y_data_fld->setText(Pd_STR_CMP);
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::makeC1_2_3_4_Item(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;
	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, QP_TS);
	
	// set component text
	if(stack.size() == 4)
	{
		stack[0]->DataTab->_y_data_fld->setText("?C2=C1? ?C3=C1? ?C4=C1? ?dr2=dr1? ?dr3=dr1? ?dr4=dr1?");
		stack[0]->TraceTab->setLineColour(stack[0]->C1_colour);
		stack[0]->userLineColour = true;
		stack[1]->DataTab->_y_data_fld->setText("?C1=C2? ?C3=C2? ?C4=C2? ?dr1=dr2? ?dr3=dr2? ?dr4=dr2?");
		stack[1]->TraceTab->setLineColour(stack[1]->C2_colour);
		stack[1]->userLineColour = true;
		stack[2]->DataTab->_y_data_fld->setText("?C1=C3? ?C2=C3? ?C4=C3? ?dr1=dr3? ?dr2=dr3? ?dr4=dr3?");
		stack[2]->TraceTab->setLineColour(stack[2]->C3_colour);
		stack[2]->userLineColour = true;
		stack[3]->DataTab->_y_data_fld->setText("?C1=C4? ?C2=C4? ?C3=C4? ?dr1=dr4? ?dr2=dr4? ?dr3=dr4?");
		stack[3]->TraceTab->setLineColour(stack[3]->C4_colour);
		stack[3]->userLineColour = true;
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::makeC1234_Item(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	QVector <TraceItem *> stack;
	QVector <int> traces;
	traces.push_back(4);
	makeMultiTraceItem(parent, traces, stack, QP_TS);
	
	// set component text
	if(stack.size() == 4)
	{
		stack[0]->DataTab->_y_data_fld->setText("?C2=C1? ?C3=C1? ?C4=C1? ?dr2=dr1? ?dr3=dr1? ?dr4=dr1?");
		stack[0]->TraceTab->setLineColour(stack[0]->C1_colour);
		stack[0]->userLineColour = true;
		stack[1]->DataTab->_y_data_fld->setText("?C1=C2? ?C3=C2? ?C4=C2? ?dr1=dr2? ?dr3=dr2? ?dr4=dr2?");
		stack[1]->TraceTab->setLineColour(stack[1]->C2_colour);
		stack[1]->userLineColour = true;
		stack[2]->DataTab->_y_data_fld->setText("?C1=C3? ?C2=C3? ?C4=C3? ?dr1=dr3? ?dr2=dr3? ?dr4=dr3?");
		stack[2]->TraceTab->setLineColour(stack[2]->C3_colour);
		stack[2]->userLineColour = true;
		stack[3]->DataTab->_y_data_fld->setText("?C1=C4? ?C2=C4? ?C3=C4? ?dr1=dr4? ?dr2=dr4? ?dr3=dr4?");
		stack[3]->TraceTab->setLineColour(stack[3]->C4_colour);
		stack[3]->userLineColour = true;
	}
	
	// link data slots together
 	linkTraceItems(stack, true);
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];

}

TraceItem* PlotModel::make2Panel_Item(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	// create panels
	QVector <TraceItem *> stack;
	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, QP_TS);
	
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::make3Panel_Item(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	// create panels
	QVector <TraceItem *> stack;
	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, QP_TS);
	
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

TraceItem* PlotModel::make4Panel_Item(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->frame();
		row=0;
	}
	else if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	// create panels
	emit layoutAboutToBeChanged();
	// create panels
	QVector <TraceItem *> stack;
	QVector <int> traces;
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	traces.push_back(1);
	makeMultiTraceItem(parent, traces, stack, QP_TS);
	
	emit layoutChanged();
	emit pageLayoutChanged(stack[0]->page());
    _view->expandChildren(indexFromItem(parent));
	return stack[0];
}

PlotItem* PlotModel::makePlot(QpuiItem * parent, QpItemType plotType, QpFrameType frameType, const QString& WL, int row)
{
	// creates a plot item of the correct type. 
	if ( QuSlotDebug() ) 
	{
		cout << "PlotModel::makePlot - '" << WL.toLatin1().data();
		if(parent) cout << "' in target '" <<  parent->getName().toLatin1().data() << "'" << endl;
		else cout << endl;
	}
        // If parent is above frame level, create the right frame type
    QpuiItem* newParent=parent;
    if(dynamic_cast<PlotItem*>(parent)!=NULL)
    {
        newParent=((PlotItem*)parent)->panel();
    }
    else if(dynamic_cast<PageItem*>(parent)!=NULL || dynamic_cast<RootItem*>(parent)!=NULL)
    {
        newParent=new FrameItem("", parent, frameType);
    }
        
	
	PlotItem* newItem=NULL;
	TraceItem* tracePlot;
	SpectrumItem* spectrumPlot;
	EventItem* itemPlot;

	switch(plotType)
	{
		case QP_TRACE:
			tracePlot = makeTrace(newParent, WL, row);
			newItem = tracePlot;
			break;
		case QP_CONST_Y:
		case QP_CONST_XT:
		case QP_REG_LINE:
			tracePlot = makeConstTrace(newParent, WL, row, plotType, frameType);
			newItem = tracePlot;
			break;
		case QP_SPECTROGRAM:
			spectrumPlot = makeSpectrum(newParent, WL, row);
			newItem = spectrumPlot;
			break;
		case QP_EVENT:
			itemPlot = makeEvent(newParent, row);
			newItem = itemPlot;
			break;
		default:
			std::cout << "unknown plot type " << plotType << std::endl;
			break;
	}
	
	_view->expandChildren(indexFromItem(parent));
	if(newItem != NULL) _view->setCurrentItem(newItem);
	
	return newItem;
}

TraceItem* PlotModel::makeConstTrace(QpuiItem* parent, const QString& WL, int row, QpItemType plotType, QpFrameType frameType)
{

	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent();
	}
	// creates new trace item in panel
	emit layoutAboutToBeChanged();
	if(parent == NULL) return NULL;
	
	TraceItem* traceItem = new TraceItem(WL, parent, WL, plotType, row);
	if(traceItem==NULL)
	{
		emit layoutChanged();
		return NULL;
	}
	
	QString plot_name;
	if(WL.isEmpty())
	{
		
		int plot_cnt = traceItem->childNumber();
		plot_name = "Plot ";
		plot_name += int_to_string(plot_cnt+1).c_str();
		if(plotType==QP_CONST_Y)
		{
			plot_name += ": Const Y";
		}
		else if(plotType==QP_CONST_XT)
		{
			if(frameType==QP_TS)
			{
				plot_name += ": Const T";
			}
			else
			{
				plot_name += ": Const X";

			}
		}
		traceItem->setName(plot_name);
	}
	else {

		if(plotType==QP_CONST_XT && frameType==QP_TS){
			traceItem->DataTab->_y_data_fld->setTimeDropdownDefault();
		}
		
	}
	
	// make newest item selected item
	_view->setCurrentItem(traceItem);

	emit layoutChanged();
	return traceItem;

}

PageItem* PlotModel::makePage()
{
	emit layoutAboutToBeChanged();
	PageItem* res=new PageItem("", _root);
	emit layoutChanged();
	
	_view->setCurrentItem(res);

	return res;
}

TraceItem* PlotModel::makeTraceItem(QpuiItem* parent, int row)
{
	emit layoutAboutToBeChanged();
	PlotItem* plotItem=dynamic_cast<PlotItem*>(parent);
	TraceItem* res;
	if(plotItem!=NULL)
	{
		res=new TraceItem("", plotItem->parent(), "", QP_TRACE, row);
	}
	else
	{
		res=new TraceItem("", parent, "", QP_TRACE, row);
	}
	_view->expandChildren(indexFromItem(parent));

	emit layoutChanged();
	emit pageLayoutChanged(res->page());
	// make newest item selected item
	_view->setCurrentItem(res);
	
	return res;
}

SpectrumItem* PlotModel::makeSpectrumItem(QpuiItem* parent, int row)
{
	emit layoutAboutToBeChanged();
	PlotItem* plotItem=dynamic_cast<PlotItem*>(parent);
	SpectrumItem* res;
	if(plotItem!=NULL)
	{
		res=new SpectrumItem("", plotItem->parent(), "", row);
	}
	else
	{
		res=new SpectrumItem("", parent, "", row);
	}
	emit layoutChanged();
	emit pageLayoutChanged(res->page());
	_view->expandChildren(indexFromItem(parent));

	// make newest item selected item
	_view->setCurrentItem(res);

	return res;
}

PanelItem* PlotModel::makePanelItem(QpuiItem* parent, int row)
{
	PlotItem* plotItem=dynamic_cast<PlotItem*>(parent);
	PanelItem* panelItem=dynamic_cast<PanelItem*>(parent);
	
	if(plotItem!=NULL) parent=plotItem->frame();
	if(panelItem!=NULL) parent=panelItem->frame();

	emit layoutAboutToBeChanged();
	PanelItem* res=new PanelItem("", parent, row);
	emit layoutChanged();
	emit pageLayoutChanged(res->page());
	_view->expandChildren(indexFromItem(parent));

	// make newest item selected item
	_view->setCurrentItem(res);

	return res;
}	


FrameItem* PlotModel::makeXYFrameItem(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->panel();
		row=0;
	}
	if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	if(dynamic_cast<FrameItem*>(parent)!=NULL)
	{
		parent=((FrameItem*)parent)->page();
		row=0;
	}
// 	creates new trace item in panel
	if(parent == NULL) return NULL;

	emit layoutAboutToBeChanged();

	int frame_cnt = parent->numChildren();
	QString frame_name=QString("Frame %1").arg(frame_cnt+1);
	FrameItem* frameItem = new FrameItem(frame_name, parent, QP_XY, row);
	
	emit layoutChanged();
	emit pageLayoutChanged(frameItem->page());
	_view->expandChildren(indexFromItem(parent));
	
	// make newest item selected item
	_view->setCurrentItem(frameItem);
	
	return frameItem;
}

FrameItem* PlotModel::makeTSFrameItem(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=((PlotItem*)parent)->panel();
		row=0;
	}
	if(dynamic_cast<PanelItem*>(parent)!=NULL)
	{
		parent=((PanelItem*)parent)->frame();
		row=0;
	}
	if(dynamic_cast<FrameItem*>(parent)!=NULL)
	{
		parent=((FrameItem*)parent)->page();
		row=0;
	}

// 	creates new trace item in panel
	if(parent == NULL) return NULL;

	emit layoutAboutToBeChanged();

	int frame_cnt = parent->numChildren();
	QString frame_name=QString("Frame %1").arg(frame_cnt+1);
	FrameItem* frameItem = new FrameItem(frame_name, parent, QP_TS, row);
	
	emit layoutChanged();
	emit pageLayoutChanged(frameItem->page());
	_view->expandChildren(indexFromItem(parent));

	// make newest item selected item
	_view->setCurrentItem(frameItem);

	return frameItem;
}

TraceItem* PlotModel::makeTrace(QpuiItem* parent, const QString& WL, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent();
	}

// 	creates new trace item in panel
	if(parent == NULL) return NULL;

	emit layoutAboutToBeChanged();

	TraceItem* traceItem = new TraceItem(WL, parent, WL, QP_TRACE, row);
	if(traceItem==NULL)
	{
		emit layoutChanged();
		return NULL;
	}
	
	if(WL.isEmpty())
	{
		int plot_cnt=traceItem->childNumber();
		QString plot_name=QString("Plot %1: Trace").arg(plot_cnt+1);
		traceItem->setName(plot_name);
	}

	((PanelItem*)traceItem->parent())->LegendTab->updateNumTraces();
	emit layoutChanged();
    emit pageLayoutChanged(traceItem->page());
	return traceItem;
}

SpectrumItem* PlotModel::makeSpectrum(QpuiItem* parent, const QString& WL, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent();
	}
	// creates new trace item in panel
	if(parent == NULL) return NULL;

	emit layoutAboutToBeChanged();
	QString plot_name;

	if(WL.isEmpty())
	{
		int plot_cnt = parent->numChildren();
		plot_name = "Plot ";
		plot_name += int_to_string(++plot_cnt).c_str();
		plot_name += ": Spect";
	}
	else
	{
		plot_name = WL;
	}
	
	SpectrumItem* spectrumItem = new SpectrumItem(plot_name, parent, WL, row);
	if(spectrumItem==NULL)
	{
		emit layoutChanged();
		return NULL;
	}
	
	emit layoutChanged();
    emit pageLayoutChanged(spectrumItem->page());
	return spectrumItem;

}

EventItem* PlotModel::makeEvent(QpuiItem* parent, int row)
{
	if(dynamic_cast<PlotItem*>(parent)!=NULL)
	{
		parent=parent->parent();
	}
	// creates new trace item in panel
	if(parent == NULL) return NULL;

	emit layoutAboutToBeChanged();
	int plot_cnt = parent->numChildren();
	string plot_name("Plot "+int_to_string(++plot_cnt)+": Event");
	EventItem* eventItem = new EventItem(QString(plot_name.c_str()), parent, "", row);

	emit layoutChanged();
    emit pageLayoutChanged(eventItem->page());
	return eventItem;
}

void PlotModel::setPageLayout(PageItem *pageItem)
{
	// slot invoked when pageLayoutChanged signal emitted
	if(QuSlotDebug())
	{
		printf("PlotModel::setPageLayout %ld %ld\n", (long)this, (long)pageItem);
		fflush(stdout);
	}
	
	// tell pages to redo layout
	if(pageItem)
	{
		pageItem->LayoutTab->updateRowCol(pageItem->numChildren());
	}
	
}

QModelIndex PlotModel::moveUp(QModelIndex& index)
{
	if(!index.isValid())
		return QModelIndex();
	
	QpuiItem* item=itemFromIndex(index);
	
	if(item!=NULL && item!=_root)
	{
		int row=item->childNumber();
		if(row>0)
		{
			emit layoutAboutToBeChanged();
			
			item->reparent(item->parent(), row-1);
			
			emit layoutChanged();
		}
	}
	
	return createIndex(item->childNumber(), 0, item);
}

QModelIndex PlotModel::moveDown(QModelIndex& index)
{
	if(!index.isValid())
		return QModelIndex();
	
	QpuiItem* item=itemFromIndex(index);
	if(item!=NULL && item!=_root)
	{
		int row=item->childNumber();
		if(row!=item->parent()->numChildren()-1)
		{
			emit layoutAboutToBeChanged();
			
			item->reparent(item->parent(), row+1);
			
			emit layoutChanged();
		}
	}
	
	return createIndex(item->childNumber(), 0, item);
}

void PlotModel::getLinkFlags(QList<QString>& io_list, QpuiItem* parent)
{
	if(parent==NULL) return;
	
	PlotItem* plotItem=dynamic_cast<PlotItem*>(parent);
	
	if(plotItem!=NULL)
	{
		QString currentText=plotItem->getLinkText();
		if(!currentText.isEmpty() && !io_list.contains(currentText) && currentText.contains("S"))
		{
			io_list.push_back(currentText);
		}
	}
	for(int i=0; i<parent->numChildren(); ++i)
	{
		getLinkFlags(io_list, parent->getChildByNumber(i));
	}
}

void PlotModel::getLinkedItems(QList<PlotItem*>& io_list, QpuiItem* current, const QString& flag)
{
	PlotItem* plotItem=dynamic_cast<PlotItem*>(current);
	
	if(plotItem!=NULL)
	{
		if(plotItem->getLinkText()==flag) io_list.push_back(plotItem);
	}
	else
	{
		for(int i=0; i<current->numChildren(); ++i)
		{
			getLinkedItems(io_list, current->getChildByNumber(i), flag);
		}
	}
}

void PlotModel::linkPlots(QList<PlotItem*>& linkedItems)
{
	LinkNumber++;
	
	QString LinkText=QString("*%1").arg(LinkNumber);
	
	foreach(PlotItem* item, linkedItems)
	{
		item->setLinkText(LinkText);
		item->setLinks(linkedItems);
	}
}

void PlotModel::linkTraceItems(QVector<TraceItem*>& stack, bool onceOnly)
{
	//  invoked to link traces in multi-panel creation
	QString LinkText;
	
	LinkNumber++;
	LinkText = QString("*%1").arg(LinkNumber);
	if(onceOnly) LinkText.insert(0, "*"); // extra * for searching on link once only
	QList<PlotItem*> linkedPlots;
	for(int i=0; i<stack.size(); i++)
	{
		PlotItem* item=dynamic_cast<PlotItem*>(stack[i]);
		linkedPlots.push_back(item);
		item->setLinkText(LinkText);
	}
	// now tell each trace
	for(int i=0; i<stack.size(); i++)
	{
		PlotItem *plot_item = dynamic_cast<PlotItem*>(stack[i]);
		if(!plot_item) continue;
		plot_item->setLinks(linkedPlots);
	}
}

void PlotModel::changePlotType(QModelIndex& index, QpItemType plotType)
{
	if(QuSlotDebug())
	{
		printf("PlotModel::changePlotType %d\n", plotType);
		fflush(stdout);
	}
	
	if(!index.isValid()) return;
	
	PlotItem* plotItem=dynamic_cast<PlotItem*>(itemFromIndex(index));
	
	if(plotItem==NULL) return;
	
	QModelIndex parentIndex=index.parent();
	PanelItem* panelItem=plotItem->panel();
	
	TraceItem * traceItem = dynamic_cast<TraceItem*>(plotItem);
	SpectrumItem * spectrumItem = dynamic_cast<SpectrumItem*>(plotItem);
	
	QString WLcontext = "";
	QString dataName;
	
	if(traceItem)
	{
		dataName = QString(traceItem->DataTab->_y_data_fld->text().toLatin1().data());
		WLcontext = baseObjectText(dataName);		
	}
	else if(spectrumItem)
	{
		dataName = QString(spectrumItem->DataTab->_z_data_fld->text().toLatin1().data());
		WLcontext = baseObjectText(dataName);
	}
	
	int row=index.row();
	removeRows(index.row(), 1, parentIndex);

	emit layoutAboutToBeChanged();
	beginInsertRows(parentIndex, row, row+1);
	switch(plotType)
	{
		default: case QP_TRACE: case QP_CONST_Y: case QP_CONST_XT: case QP_REG_LINE:
		{
			TraceItem* newItem=new TraceItem(dataName, panelItem, WLcontext, plotType, row);
			_root->I_PlotView->setCurrentItem(newItem);
			break;
		}
		case QP_SPECTROGRAM:
		{
			QdObject_var dobj = QuGetObjByName(WLcontext);
			QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(dobj);
			if(ms.is_nil()) 
			{
				SpectrumItem * newItem= new SpectrumItem(dataName, panelItem, WLcontext, row);
				_root->I_PlotView->setCurrentItem(newItem);
			}
			else
			{
				SpectrumItem * newItem= new SpectrumItem(dataName, panelItem, WLcontext, row);
				_root->I_PlotView->setCurrentItem(newItem);
			}
			
			break;
		}
	}
	endInsertRows();
	emit layoutChanged();
	_root->I_PlotView->I_LastTab = QP_IGNORE_TAB; // tell updateOnSelection to ignore last tab and use DataTab
}

bool PlotModel::deleteItem(QpuiItem* item)
{
	if(item==NULL || item==_root) return false;

	qApp->processEvents();

	QpuiItem* parentItem=item->parent();
	if(parentItem==NULL) return false;

	PageItem* page=NULL;
	PlotItem* plotItem=dynamic_cast<PlotItem*>(item);
	PanelItem* panelItem=dynamic_cast<PanelItem*>(item);
	FrameItem* frameItem=dynamic_cast<FrameItem*>(item);

	if(plotItem!=NULL) page=plotItem->page();
	if(panelItem!=NULL) page=panelItem->page();
	if(frameItem!=NULL) page=frameItem->page();
		
	emit layoutAboutToBeChanged();	
	QModelIndex parentIndex=indexFromItem(parentItem);

	int row=item->childNumber();
	beginRemoveRows(parentIndex, row, row);

	qApp->processEvents();

	parentItem->removeChild(row, true);

	endRemoveRows();

	emit layoutChanged();
	if(page!=NULL) emit pageLayoutChanged(page);
		
	if(!_root->hasChildren())
	{
		beginInsertRows(indexFromItem(_root), 0, 1);
		PageItem* newPage=makePage();
		endInsertRows();
	}
	
	return true;
}

Vostream& operator<<(Vostream& v_os, PlotModel& model)
{
    // Get and save all used non-standard colour tables

    savedPalette.clear();

    model._root->addCustomColourTableTo(savedPalette);

    v_os << (int)savedPalette.size();

    foreach(ColourTable table, savedPalette)
    {
        v_os << table;
    }
    
	int numPages=model._root->numChildren();
	v_os<<numPages;
	
	for(int i=0; i<numPages; ++i)
	{
		v_os<<*((PageItem*)model._root->getChildByNumber(i));
	}
	
	return v_os;
}

QDomElement PlotModel::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Plot_Layout");

	savedPalette.clear();
    _root->addCustomColourTableTo(savedPalette);
	foreach(ColourTable table, savedPalette)
    {
		xml.appendChild(createTextXMLNode(doc, "ColourTable", table.toString()));
    }
	
    int numPages=_root->numChildren();
    xml.setAttribute("size", QString::number(numPages));
    
    for(int i=0; i<numPages; ++i)
    {
        xml.appendChild(((PageItem*)_root->getChildByNumber(i))->toXMLElement(doc));
    }
    return xml;
}

void PlotModel::fromXMLElement(QDomElement& el)
{
    if(_replace)
    {
        importedPalette.clear();
    //    QModelIndex rootIndex=indexFromItem(_root);
        clear();
    }

	if(_replace)
    {
		importedPalette.clear();
    }

    importedPaletteOffset=importedPalette.size();
        
	QDomElement colourTable=el.firstChildElement("ColourTable");

	while(!colourTable.isNull())
	{
		QString s=colourTable.firstChild().nodeValue();
		ColourTable temp;
		temp.fromString(s);
		importedPalette.push_back(temp);
		colourTable=colourTable.nextSiblingElement("ColourTable");
	}

    int cnt=el.attribute("size", "0").toInt();

    if(cnt!=0)
    {
        emitLayoutAboutToBeChanged();
        BIR(_root, 0, cnt);
        QDomElement pageElement=el.firstChildElement("Page_Item");
        while(!pageElement.isNull())
        {
            PageItem* page=new PageItem("", _root);
            page->fromXMLElement(pageElement);
            pageElement=pageElement.nextSiblingElement("Page_Item");
        }
        EIR();
        emitLayoutChanged();
    }

    // relink items now everything is restored
    QList<QString> linkFlags;
    
    // Make list of flags
    
    getLinkFlags(linkFlags, _root);

    foreach(QString flag, linkFlags)
    {
        QList<PlotItem*> linkedItems;
        getLinkedItems(linkedItems, _root, flag);
        
        linkPlots(linkedItems);
        
    }
    
    
}

Vistream& operator>>(Vistream& v_is, PlotModel& model)
{
    if( qsas_restore_release >= 20400)
    {
        if(model._replace)
        {
            importedPalette.clear();
        }

        importedPaletteOffset=importedPalette.size();
        
        int numColourTables;
        v_is >> numColourTables;

        for(int i=0; i<numColourTables; ++i)
        {
            ColourTable temp;
            v_is >> temp;
            importedPalette.push_back(temp);
        }
    }
	if(model._replace)
	{
		model.clear();
	}

    int cnt;
	v_is >> cnt;
	if(cnt!=0)
	{
			model.emitLayoutAboutToBeChanged();
			model.BIR(model._root, 0, cnt);
			for(int i=0; i<cnt; ++i)
			{
				PageItem* page=new PageItem("", model._root);
				v_is >> *page;
			}
			model.EIR();
			model.emitLayoutChanged();
	}
	

	// relink items now everything is restored
	QList<QString> linkFlags;
	
	// Make list of flags
	
	model.getLinkFlags(linkFlags, model._root);
	
	foreach(QString flag, linkFlags)
	{
		QList<PlotItem*> linkedItems;
		model.getLinkedItems(linkedItems, model._root, flag);
		
		model.linkPlots(linkedItems);
		
	}
	
	
	return v_is;
}

