// qpui_event.cc
//
// Plot Interface Implementation - Event Object
//
// Janet Barnes
// July 2002
//
// Tony Allen
// April 2008

#include "qpui.h"
#include "qwui.h"

extern int qsas_restore_release;

EventItem::EventItem(const QString& name, QpuiItem* parent, const QString& WLcontext, int position):
		PlotItem(name, parent, WLcontext, position)
{	
	I_Type = QP_EVENT;

	eventTab = new QpEventTab(tabWidget, "Event Tab", this);
	eventTab->hide();

	// set default values for Event object

	eventTab->_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_LABELS_CHAR_SIZE_ADD);
// 	eventTab->_event_label_font_fld->setCurrentItem(QPUI_NORMAL);
	_event_label_colour = Qt::black;
	
	tabWidget->adjustSize();
}

EventItem::EventItem(EventItem* item, QpuiItem* parent, int position):
		PlotItem(item, parent, position)
{
	I_Type = QP_EVENT;

	_event_label_colour = item->_event_label_colour;
	
	eventTab = new QpEventTab(item->eventTab, this);
	eventTab->hide();
	
	tabWidget->adjustSize();
}

EventItem::~EventItem()
{
	clear();
	delete eventTab;
	foreach(PlotItem* i, linkedPlots)
	{
		i->linkedPlots.removeAll(this);
	}
}


// -------------------------------------------

void EventItem::showTabs()
{
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	tabWidget->addTab(eventTab, "Event Tab");
	
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
	tabWidget->setCurrentIndex(tabWidget->indexOf(eventTab));
	
}

// -----------------------------------------------

QpuiReturnStatus EventItem::plot()
{
	bool OK;
	double y_posn;
	double x_posn;
	PLFLT x_data;
	PLFLT y_data;

	QdObject_var eventTxtObj = eventTab->_event_label_fld->getSlotObject();
	if(eventTxtObj.is_nil()) return QPUI_SUCCESS;

	QdString_var txt = QdString_var::narrow(eventTxtObj);
	if(txt.is_nil()) return QPUI_SUCCESS;

	plwid(QtWidth2PlWidth(1));
	
	// get y position
	QdObject_var ydo = eventTab->_y_data_fld->getSlotObject();
	if( ydo.is_nil())
	{
		// try numeric value
		bool OK;
		y_posn = eventTab->_y_data_fld->text().toDouble(&OK);
	}
	else
	{
		QdRScalar_var rscalar_y = QdRScalar_var::narrow(ydo);
		if(rscalar_y.is_nil())
		{
			std::cout << " No y posn " << std::endl;
			plmtex("lv",-10.0,0.5,0.0,"Plot ERROR: Y Data not a Scalar Object");
			std::cout << " No y posn plotted" << std::endl;

			return QPUI_FAILURE;
		}
		y_posn = *rscalar_y;
	}
    
	if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) y_posn = SafeLog10(y_posn);
		
	y_data = (PLFLT) y_posn;
	
	
	// get x or t position
	
	QdObject_var xdo = eventTab->_x_data_fld->getSlotObject();
	
	if(frameType() == QP_TS)
	{
		// Time axis
		QdTime_var time_t = QdTime_var::narrow(xdo);
		try{
			if(time_t.is_nil()) time_t = new QdTime(eventTab->_x_data_fld->text().toLatin1().data());
		}
		catch(Exception &e){;}

		if(time_t.is_nil()) 
		{	
			QdTimeSeq_var tt = get_timetags(xdo);

			if(tt.is_nil())
			{
				plmtex("lv",-10.0,0.5,0.0,"Event Label: Invalid T Data");
				return QPUI_FAILURE;
			}
			else
			{
				if(tt->sequence_size() == 1)  time_t =  new QdTime(tt[0]);
				else time_t = QdTime_var::narrow(QuGetTimeComponent(tt, "Tstart" ));
		
				if(time_t.is_nil())
				{
					plmtex("lv",-10.0,0.5,0.0,"Event Label: Invalid T Data from time tags");
					return QPUI_FAILURE;
				}
			}
		}
// 		Time t_plot_start=*(frame()->tAxisTab->_axis_time_fld->getStartTime());
// 
// 		Time t_midnite(t_plot_start);
// 		// t_midnite is midnight at start of t_plot_start day...
// 		t_midnite -= Seconds(t_plot_start.second_of_day());
// 		x_data = Seconds(*time_t - t_midnite);
		QmjdTime t_plot_start=(frame()->tAxisTab->_axis_time_fld->getStart());
		x_data=time_t->getDiffSecs(t_plot_start);

	}
	else
	{
		// X axis
		QdRScalar_var rscalar_x = QdRScalar_var::narrow(xdo);
		if(rscalar_x.is_nil())
		{
			QdRScalarSeq_var sx = QdRScalarSeq_var::narrow(xdo);
			if (!sx.is_nil())
			{
				rscalar_x = new QdRScalar(sx[0]);
			}
		}
			
		if(rscalar_x.is_nil())
		{
			x_posn = eventTab->_x_data_fld->text().toDouble(&OK);
			if (!OK)
			{
				plmtex("lv",-10.0,0.5,0.0,"Event Label: Invalid X Data");
				return QPUI_FAILURE;
			}
		}
		else
		{
			x_posn = *rscalar_x;
		}
			
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) x_posn = SafeLog10(x_posn);
		x_data = (float) x_posn;
	}
	
	// Not clipped => Extend panel range and viewport to whole page, write label, and reduce panel and viewport back

	PLFLT save_min_x, save_max_x, save_min_y, save_max_y; // viewport coordinates in page
	plgvpd(&save_min_x, &save_max_x, &save_min_y, &save_max_y);
		
	PLFLT save_min_x_data, save_max_x_data, save_min_y_data, save_max_y_data; // data ranges in original viewport
	plgvpw(&save_min_x_data, &save_max_x_data, &save_min_y_data, &save_max_y_data);

	double val0x, val1x, val0y, val1y; // data range when viewport is whole page
		
	double ratiox=(save_max_x_data-save_min_x_data)/(save_max_x-save_min_x); // range of time for the whole page width
	val0x=save_min_x_data-save_min_x*ratiox;
	val1x=val0x+ratiox;	

	double ratioy=(save_max_y_data-save_min_y_data)/(save_max_y-save_min_y); // range of time for the whole page width
	val0y=save_min_y_data-save_min_y*ratioy;
	val1y=val0y+ratioy;

	// We set up the new viewport for the whole page
	plvpor(0., 1., 0., 1.);
	plwind(val0x, val1x, val0y, val1y);

	plschr(QPUI_CHAR_HEIGHT,eventTab->_char_size_fld->value());
	// set font
	PLINT label_colour_index = getPGPColourIndex(_event_label_colour);
	plcol0(label_colour_index);

	int plot_id=page()->PageTab->_xwin_num_fld->text().toInt();

	PLINT family, style, weight;
#if QT_VERSION >= 0x040300
	eventTab->_event_label_font_fld->getFontParams(family, style, weight);
	plsfont(family, style, weight);
#else
    family=eventTab->_event_label_font_fld->getFontIndex();
    plfont(family);
#endif
	page()->setSmooth(plot_id, true);
    plptex(x_data, y_data, 1., 0.0, eventTab->_event_label_just_fld->value(), SafePGPString(txt->c_str()).c_str());
	page()->setSmooth(plot_id, false);

	// revert to original viewport
	plvpor(save_min_x, save_max_x, save_min_y, save_max_y);
	plwind(save_min_x_data, save_max_x_data, save_min_y_data, save_max_y_data);

	return QPUI_SUCCESS;
}



// -----------------------------------------------------

QDomElement EventItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Plot_Item");
    xml.setAttribute("Type", "Event");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));
    
    xml.appendChild(eventTab->toXMLElement(doc));

    xml.appendChild(createTextXMLNode(doc, "Label_Red", _event_label_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Label_Green", _event_label_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Label_Blue", _event_label_colour.blue()));

    return xml;
}

extern QPalette QPUI_normal_palette;

void EventItem::fromXMLElement(QDomElement& el)
{
    QString s;
    
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }
    
    QDomElement eventElement=el.firstChildElement("Event_Tab");
    if(!eventElement.isNull())
    {
        eventTab->fromXMLElement(eventElement);
    }

    int r(0), g(0), b(0);
    if(getXMLNodeText(el, "Label_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Blue", s))
    {
        b=s.toInt();
    }
    _event_label_colour = QColor(r,g,b);
    
    QPalette line_palette = QPalette(QPUI_normal_palette);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _event_label_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _event_label_colour);
    eventTab->_event_label_colour_fld->setPalette(line_palette);

    eventTab->_y_data_fld->reValidate();
    eventTab->_x_data_fld->reValidate();

}

Vistream& operator>>(Vistream& v_is, EventItem& item)
{
// 	// streams from virtual input stream v_is into item
	QpuiItem& list_item = item;
	v_is >> list_item;

	string x_data, y_data;
	v_is >> y_data;
	v_is >> x_data;
	item.eventTab->_y_data_fld->setText(y_data.c_str());
	item.eventTab->_x_data_fld->setText(x_data.c_str());

	v_is >> item.with;
	if(item.with)
	{
		DataNamesTable_var qpui_objects;
		try{
			qpui_objects = QguWindow::_restored_objects->resolve_context(Name("qpui"));
		}
		catch(...){;}
	    if(qpui_objects.is_nil())
		{
		
			qpui_objects = new DataNamesTable();
			if(!QguWindow::_restored_objects.is_nil())
			{
				QguWindow::_restored_objects->bind_context(Name("qpui"), qpui_objects);
			}
		}
	
	    string y_type;
		string y_data_new="";
		v_is >> y_type;
		if(y_type != NULL_OBJECT)
		{
			QdObject_var ydo = QdObject::create(y_type);
			v_is >> ydo;
			QString list;
			Name y_name(y_data);
			if(y_name.num_components()) 
				list = y_name[0].get_id().c_str();
			else if(!QdTime_var::narrow(ydo).is_nil())  
				list="Constants";
			else if(!QdRScalar_var::narrow(ydo).is_nil())
				list="Constants";
			else
				list="Data";
			
			list += '/';
			y_name.delete_component(0);
			list += y_name.srep().c_str();
			y_data_new = QwuiPutOnWL(ydo, list).toLatin1().data();
			try{
				qpui_objects->unbind(Name(y_data));
			}
			catch(...){;}
			
			try{
				qpui_objects->bindx(Name(y_data), y_data_new);
			}
			catch(...) {;}
		}
		else
		{
			try{
				y_data_new=qpui_objects->resolve(Name(y_data));
			}
			catch(...){y_data_new=y_data;}
		}
		item.eventTab->_y_data_fld->setText(y_data_new.c_str());
	  
		string x_type,x_data_new;
		v_is >> x_type;
		if(x_type != NULL_OBJECT)
		{
			QdObject_var xdo;
			xdo= QdObject::create(x_type);
			v_is >> xdo;
			QString list;
			Name x_name(x_data);
			if(x_name.num_components())
				list = x_name[0].get_id().c_str();
			else if(!QdTime_var::narrow(xdo).is_nil())
				list="Constants";
			else if(!QdRScalar_var::narrow(xdo).is_nil())
				list="Constants";
			else
				list="Data";
				
			list += '/';
			x_name.delete_component(0);
			list += x_name.srep().c_str();
			x_data_new = QwuiPutOnWL(xdo, list).toLatin1().data();
			try{
				qpui_objects->unbind(Name(x_data));
			}
			catch(...){;}
      
			try{
				qpui_objects->bindx(Name(x_data),x_data_new);
			}
			catch(...){;}
		}
		else
		{
			try{
				x_data_new = qpui_objects->resolve(Name(x_data));
			}
			catch(...){x_data_new=x_data;}
		}
		item.eventTab->_x_data_fld->setText(x_data_new.c_str());
	
	}
	
	string _event_label;
	v_is >> _event_label;
	item.eventTab->_event_label_fld->setText(_event_label.c_str());

	float _event_label_char_size;
	v_is >> _event_label_char_size;
	item.eventTab->_char_size_fld->setValue(_event_label_char_size);

	
	if( qsas_restore_release >= 20306)
	{
		PLINT family, style, weight;
		v_is >> family;
		v_is >> style;
		v_is >> weight;
		switch(family)
		{
			case PL_FCI_SERIF: item.eventTab->_event_label_font_fld->setSerifFont(); break;
			case PL_FCI_MONO: item.eventTab->_event_label_font_fld->setMonotypeFont(); break;
			case PL_FCI_SCRIPT: item.eventTab->_event_label_font_fld->setDecorativeFont(); break;
			default: item.eventTab->_event_label_font_fld->setSansFont(); break;
		}
		item.eventTab->_event_label_font_fld->setItalic(style==PL_FCI_ITALIC);
		item.eventTab->_event_label_font_fld->setBold(weight==PL_FCI_BOLD);
	}
	else
	{
		int _font;
		v_is >> _font;
		switch(_font)
		{
			case 1: // Roman
				item.eventTab->_event_label_font_fld->setRomanFont();
				break;
				
			case 2: // Italic
				item.eventTab->_event_label_font_fld->setItalicFont();
				break;
				
			default: // Normal
				item.eventTab->_event_label_font_fld->setNormalFont();
				break;
		}
	}
	
	int r,g,b;
	v_is >> r;
	v_is >> g;
	v_is >> b;
	item._event_label_colour = QColor(r,g,b);

    if( qsas_restore_release >= 20400)
    {
        double just;
        v_is >> just;
        item.eventTab->_event_label_just_fld->setValue(just);
    }
    
	return v_is;
}


Vostream& operator<<(Vostream& v_os,EventItem& item)
{
	// streams item to virtual output stream v_os

	QpuiItem& list_item = item;
	v_os =save(v_os, list_item);
	
	v_os << ListTypeEntry;
	v_os << item.I_Type;

	string x_data,y_data;
	v_os << string(item.eventTab->_y_data_fld->text().toLatin1().data());
	v_os << string(item.eventTab->_x_data_fld->text().toLatin1().data());
	
	v_os << item.with;	

	if(item.with)
	{
		DataNamesTable_var qpui_objects;
		try{ 
			qpui_objects = QguWindow::_saved_objects->resolve_context(Name("qpui"));
		}
		catch(...){;}
    
		if(qpui_objects.is_nil())
		{
			qpui_objects = new DataNamesTable();
			if(!QguWindow::_saved_objects.is_nil())
				QguWindow::_saved_objects->bind_context(Name("qpui"), qpui_objects);
		}

		bool found_yobject = false;
		try{
			qpui_objects->resolve(Name(item.eventTab->_y_data_fld->text().toLatin1().data()));
			found_yobject=true;
		}
		catch(...){;}

		string y_type;
		QdObject_var y_obj;
		if(!found_yobject) y_obj = QuGetObjByName(item.eventTab->_y_data_fld->text().toLatin1().data());
		bool no_y = y_obj.is_nil();
		if(no_y) y_type = NULL_OBJECT;
		else
		{
			y_type = y_obj->object_type();
			qpui_objects->bindx(Name(item.eventTab->_y_data_fld->text().toLatin1().data()),"");
		}
		v_os << y_type;
		
		if(!no_y) v_os << y_obj;

		bool found_xobject = false;
		try{
			qpui_objects->resolve(Name(item.eventTab->_x_data_fld->text().toLatin1().data()));
			found_xobject=true;
		}
		catch(...){;}

		string x_type;
		QdObject_var x_obj;
		if(!found_xobject) x_obj = QuGetObjByName(item.eventTab->_x_data_fld->text().toLatin1().data());
		bool no_x = x_obj.is_nil();
		if(no_x) x_type = NULL_OBJECT;
		else
		{
			x_type = x_obj->object_type();
			qpui_objects->bindx(Name(item.eventTab->_x_data_fld->text().toLatin1().data()),"");
		}

		v_os << x_type;
		if(!no_x) v_os << x_obj;

	}
	v_os << string(item.eventTab->_event_label_fld->text().toLatin1().data());
	v_os << (float)item.eventTab->_char_size_fld->value();
	
	PLINT family, style, weight;
	item.eventTab->_event_label_font_fld->getFontParams(family, style, weight);
	v_os << family;
	v_os << style;
	v_os << weight;
	
	v_os << item._event_label_colour.red();
	v_os << item._event_label_colour.green();
	v_os << item._event_label_colour.blue();

    v_os << (double)item.eventTab->_event_label_just_fld->value();
	
	return v_os;
}

