// qpui_window.cc
//
// Plot Interface Implementation - Plot Window and ListView components
//
// Janet Barnes
//
// July 2002

#include "qpui.h"
#include "qwui.h"
#include "qdutil.h"

extern qwuiWindow* qsasWindow;
extern int qsas_restore_release;

extern "C" int PLPexitHandler(const char *msg){
	cout << "Plplot crash " << msg << endl;
	return 1;
}

//     QpuiWindow CLASS
//     ----------------

QpWindow::QpWindow( QWidget *parent ):QguWindow(parent,"Plot")
{
	// Constructor
	if(QuSlotDebug())
	{
		printf("QpWindow::QpWindow %ld\n", (long)this);
		fflush(stdout);
	}


	QguWindow::_saveset_filter = QString("Plot savesets (*.qs_plot)");
	
	// set up colour group palettes
	extern QPalette QPUI_normal_palette;
	QPUI_normal_palette = QPalette(qApp->palette());
	this->setPalette(QPUI_normal_palette);

    // Initialise colour tables
    stdPalette.clear();
    profilePalette.clear();
    importedPalette.clear();

    ColourTable rainbowTable("Rainbow"); rainbowTable.setPresetRainbow(); stdPalette.append(rainbowTable);
    ColourTable rainbowExtTable("Rainbow Extended"); rainbowExtTable.setPresetRainbowExt(); stdPalette.append(rainbowExtTable);
    ColourTable gwlTable("Gray White Low"); gwlTable.setPresetGrayWhiteLow(); stdPalette.append(gwlTable);
    ColourTable gwhTable("Gray White High"); gwhTable.setPresetGrayWhiteHigh(); stdPalette.append(gwhTable);
    ColourTable brdTable("Blue to Red (dark ends)"); brdTable.setPresetRedBlue(); stdPalette.append(brdTable);
    ColourTable brlTable("Blue to Red (light ends)"); brlTable.setPresetInvRedBlue(); stdPalette.append(brlTable);
    ColourTable earthTable("Earth"); earthTable.setPresetEarth(); stdPalette.push_back(earthTable);
    ColourTable spectTable("Spectrum"); spectTable.setPresetSpectrum(); stdPalette.push_back(spectTable);

    foreach(ColourTable table, QSAS_profiles->current()->palette)
    {
        profilePalette.push_back(table);
    }

	setAcceptDrops(true);
	setFocusPolicy(Qt::WheelFocus);

	_top_widget = new QWidget(this);
	setCentralWidget( _top_widget );

	_top_gl = new QGridLayout( _top_widget);
	_top_gl->setSpacing(QSAS_profiles->current()->GUImargins);
	_top_gl->setMargin(QSAS_profiles->current()->GUImargins);
	
	_top_splitter=new QSplitter(_top_widget);
	_top_gl->addWidget( _top_splitter, 0,0 );

	QWidget * listsWidget=new QWidget(_top_splitter);
	QVBoxLayout* listsLayout=new QVBoxLayout;
	

	QScrollArea *scroll=new QScrollArea(_top_splitter);

	scroll->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
	scroll->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);


	QGroupBox* plot_groupbox = new QGroupBox(scroll->viewport());
	QVBoxLayout * plot_layout=new QVBoxLayout;

	plot_groupbox->setSizePolicy(QSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding));
	
	plotView=new PlotView(listsWidget);
	plotView->tabWidget=new QTabWidget(plot_groupbox);
	PlotModel* model=new PlotModel(plotView);
	plotView->setModel(model);
	plotView->setMinimumWidth(280);
	plotView->setSizePolicy(QSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding));
	plotView->expandAll();
	
	listsLayout->addWidget(plotView);
	
	listsWidget->setLayout(listsLayout);

		
	plot_layout->addWidget(plotView->tabWidget);
	plot_groupbox->setLayout(plot_layout);
	int margins = QSAS_profiles->current()->GUImargins;
	plotView->tabWidget->setContentsMargins(margins, margins, margins, margins);

	// set up menu bar
	_menu_bar = menuBar();
	_menu_bar->clear();
	QMenu* file_menu = 	_menu_bar->addMenu("&File");
	QMenu* insert_menu = _menu_bar->addMenu("Insert" );
	QMenu* edit_menu = 	_menu_bar->addMenu("Edit" );
	_menu_bar->addSeparator();
	QMenu* help_menu = 	_menu_bar->addMenu("&Help" );


	file_menu->clear();
	file_menu->addAction("&Save", this, SLOT(saveWithOut()));
	file_menu->addAction("Save &As...", this, SLOT(saveWithOutAs()));
    file_menu->addAction("Save As XML...", this, SLOT(saveAsXML()));
	file_menu->addSeparator();
	file_menu->addAction("&Include...", this, SLOT(include()));
	file_menu->addAction("&Restore...", this, SLOT(restore()));
    file_menu->addAction("Restore XML...", this, SLOT(restoreXML()));
	file_menu->addSeparator();
  	file_menu->addAction("P&lot Selected Pages", plotView, SLOT(plotSelectedItems()));
	file_menu->addSeparator();
 	file_menu->addAction("Use Defaults in Selected Items", plotView, SLOT(setDefaultSelectedItems()));
	file_menu->addSeparator();
	file_menu->addAction("&Close", this, SLOT(close()));

	insert_menu->clear();
	insert_menu->addAction("Trace", plotView, SLOT(makeTraceItem()));
 	insert_menu->addAction("SpectroGram", plotView, SLOT(makeSpectrumItem()));
	insert_menu->addSeparator();
 	insert_menu->addAction("Panel", plotView, SLOT(makePanelItem()));
	insert_menu->addSeparator();
 	insert_menu->addAction("TS Frame", plotView, SLOT(makeTSFrameItem()));
 	insert_menu->addAction("XY Frame", plotView, SLOT(makeXYFrameItem()));
	insert_menu->addSeparator();
 	insert_menu->addAction("Page", plotView, SLOT(makePageItem()));
	insert_menu->addSeparator();
 	insert_menu->addAction("Constant X or T Trace", plotView, SLOT(makeConstXTraceItem()));
 	insert_menu->addAction("Constant Y Trace", plotView, SLOT(makeConstYTraceItem())); 
 	insert_menu->addAction("Regression Line", plotView, SLOT(makeRegLineTraceItem())); 
	insert_menu->addSeparator();
 	insert_menu->addAction("Event Label", plotView, SLOT(makeEventItem()));
	insert_menu->addSeparator();
	QMenu* template_menu = insert_menu->addMenu("Template..."); 

	template_menu->clear();
 	template_menu->addAction("Vector RXYZ", plotView, SLOT(makeVector_RXYZ_Item())); 
 	template_menu->addAction("Vector XYZ", plotView, SLOT(makeVector_XYZ_Item()));
	template_menu->addSeparator();
 	template_menu->addAction("Vector R_XYZ", plotView, SLOT(makeVector_R_XYZ_Item()));
 	template_menu->addAction("Vector R_LP", plotView, SLOT(makeVector_R_LP_Item()));
 	template_menu->addAction("Vector R_TP", plotView, SLOT(makeVector_R_TP_Item()));
	template_menu->addSeparator();
 	template_menu->addAction("Vector R_X_Y_Z", plotView, SLOT(makeVector_R_X_Y_Z_Item()));
 	template_menu->addAction("Vector X_Y_Z", plotView, SLOT(makeVector_X_Y_Z_Item()));
	template_menu->addAction("Vector R_L_P", plotView, SLOT(makeVector_R_L_P_Item()));
 	template_menu->addAction("Vector R_T_P", plotView, SLOT(makeVector_R_T_P_Item()));
	template_menu->addSeparator();
 	template_menu->addAction("2 Panels", plotView, SLOT(make2Panel_Item()));
 	template_menu->addAction("3 Panels", plotView, SLOT(make3Panel_Item()));
 	template_menu->addAction("4 Panels", plotView, SLOT(make4Panel_Item()));
	template_menu->addSeparator();
 	template_menu->addAction("C1_2_3_4", plotView, SLOT(makeC1_2_3_4_Item()));
 	template_menu->addAction("C1234", plotView, SLOT(makeC1234_Item())); 


	edit_menu->clear();
	edit_menu->addAction("Rename", plotView, SLOT(renameSelectedItem()));
 	edit_menu->addAction("Cut", plotView, SLOT(cut()), Qt::CTRL+Qt::Key_X);
 	edit_menu->addAction("&Copy", plotView, SLOT(copy()), Qt::CTRL+Qt::Key_C);
 	edit_menu->addAction("Paste", plotView, SLOT(paste()), Qt::CTRL+Qt::Key_V);
	edit_menu->addSeparator();
 	edit_menu->addAction("Select &All", plotView, SLOT( selectAll() ),Qt::CTRL+Qt::Key_A );
 	edit_menu->addAction("Move Up", plotView, SLOT(moveItemUp()), Qt::CTRL+Qt::Key_Up);
 	edit_menu->addAction("Move Down", plotView, SLOT(moveItemDown()), Qt::CTRL+Qt::Key_Down);
	edit_menu->addSeparator();
 	edit_menu->addAction("Delete", plotView, SLOT(deleteSelectedItems()), Qt::Key_Delete);
	edit_menu->addSeparator();
	edit_menu->addAction("&Link Selected Traces", plotView, SLOT( linkPlots() ),Qt::CTRL+Qt::Key_L);
	edit_menu->addAction("&Unlink Selected Traces", plotView, SLOT( unlinkPlots() ),Qt::CTRL+Qt::Key_U);

	help_menu->clear();
	help_menu->addAction("QSAS &Manual", this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
	help_menu->addSeparator();
	help_menu->addAction("...Plotting", this, SLOT( launchManPlot() ));
	help_menu->addSeparator();
	help_menu->addAction("Module Version", this, SLOT( launchVersion() ));

	_status_bar = statusBar();
	_status_bar->showMessage(" ");

	QHBoxLayout *buttons_layout = new QHBoxLayout();

	plotBtn = new QPushButton("Plot on Screen",_top_widget);
	plotBtn->setMinimumSize(plotBtn->sizeHint());
	plotBtn->setFocusPolicy(Qt::StrongFocus);
	connect(plotBtn, SIGNAL(clicked()), this, SLOT(doPlot()));
	buttons_layout->addWidget(plotBtn);
	plotBtn->setDefault(true);

	QLabel *space = new QLabel("   ");
	buttons_layout->addWidget(space);
	
	dismissBtn = new QPushButton( "Dismiss", _top_widget);
	dismissBtn->setMinimumSize(dismissBtn->sizeHint());
	dismissBtn->setFocusPolicy(Qt::StrongFocus);
	connect( dismissBtn, SIGNAL(clicked()), this, SLOT(close()) );
	//buttons_layout->insertSpacing(0, 50);
	buttons_layout->addWidget(dismissBtn);

	_top_gl->addLayout(buttons_layout,2,0);
	_top_gl->setRowStretch(0,1);
	_top_gl->setRowStretch(1,0);

 	PageItem *startPage = plotView->makePage();

 	startPage->showTabs();
	plotView->tabWidget->adjustSize();
	
	int (*pPLPexitHandler)(const char *) = PLPexitHandler;
	plsexit(pPLPexitHandler );
	
	connect(plotView, SIGNAL(raiseWindow()), this, SLOT(bringToFront()));
	
	this->setSizePolicy(QSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding));


	_top_splitter->setStretchFactor(0, 1);
	_top_splitter->setStretchFactor(1, 10);
	scroll->setWidgetResizable(true);
	scroll->setWidget(plot_groupbox);
	
}



void QpWindow::keyPressEvent(QKeyEvent * e )
{  
	if(QuSlotDebug())
	{
		printf("QpWindow::keyPressEvent %ld\n", (long)this);
		fflush(stdout);
	}

	// use escape key to toggle default buttons
	if(e->key() == Qt::Key_Escape){
		if(   plotBtn->isDefault()) {
			plotBtn->setDefault(false);
			dismissBtn->setDefault(true);
		}
		else if(dismissBtn->isDefault())  {
			dismissBtn->setDefault(false);
			plotBtn->setDefault(true);
		}
		return;
	}
	// use return or enter key to perform default
	if(e->key() == Qt::Key_Return ||  e->key() == Qt::Key_Enter) {
		if(plotBtn->isDefault()) doPlot();
		else if (dismissBtn->isDefault()) close();
		return;
	}

	// else pass on to list view 
	e->ignore(); // let selection change

}

void QpWindow::deleteAllItems()
{
	if(QuSlotDebug())
	{
		printf("QpWindow::deleteAllItems \n");
		fflush(stdout);
	}

	plotView->deleteAllItems();
}


// --------------------------------------

void QpWindow::resetDefaultBtn()
{  
	dismissBtn->setDefault(false);
	//printBtn->setDefault(false);
	plotBtn->setDefault(true);
}

// --------------------------------------

void QpWindow::doPlot()
{
	// slot invoked when Plot push button clicked
  	if(QuSlotDebug())
	{
		printf("QpWindow::doPlot %ld\n", (long)this);
		fflush(stdout);
	}

   if(!plotView)
	{
		(void)QMessageBox::warning(this,"Print","No active listview");
		return;
	}

	if(!plotView->plotModel()->_root->hasChildren())
	{
		(void)QMessageBox::warning(this,"Print","No pages to plot");
		return;
	}
	
	PageItem* page_item;
	plotView->_plot_page_list.clear();
	for(int i=0; i<plotView->plotModel()->_root->numChildren(); ++i)
	{
		page_item=dynamic_cast<PageItem*>(plotView->plotModel()->_root->getChildByNumber(i));
		plotView->_plot_page_list.push_back(page_item);
	}
	try{
		plotView->plot();
	}
	catch(...){;}
}


// --------------------------------------

void QpWindow::launchManPlot()
{
	// slot invoked from Plotting Help menu sub item
	// launches browser showing Qsas Plot help docuumentation

	if(QuSlotDebug())
	{
		printf("QpWindow::launchManPlot %ld\n", (long)this);
		fflush(stdout);
	}

	qsasWindow->help_window->manEntry("QSAS_plotting.html");
}

void QpWindow::launchManual()
{
	// slot invoked from Qsas Manual Help menu sub item
	// launches browser showing Qsas help index

	if(QuSlotDebug())
	{
		printf("QpWindow::launchManual %ld\n", (long)this);
		fflush(stdout);
	}

	qsasWindow->help_window->manEntry("QSAS_help_index.html");
}

void QpWindow::launchVersion()
{
	// slot invoked from Module Version Help menu sub item
	// launches browser showing Qsas version numbers

	if(QuSlotDebug())
	{
		printf("QpWindow::launchVersion %ld\n",(long)this);
		fflush(stdout);
	}

	QString text;
	text  = QString("Version Numbers...\n");
	text += QString(QSAS_VERSION_ID);
	text += "\n Plotting  ";
	text += QString(QPUI_VERSION);
	QMessageBox::information( this, "QSAS information", text.toLatin1().data(), "OK" );
}


void QpWindow::saveAsXML(QString path)
{
 	ofstream file;
	file.open(path.toLatin1().data(),ios::out);

	if(file.is_open())
	{
		QDomDocument doc("QSAS_plot_save");

		QDomElement root=doc.createElement("QSAS_Plot");
		doc.appendChild(root);
		string guiID;
		guiID=QSAS_VERSION_ID;

		QDomElement version=createTextXMLNode(doc, "QSAS_Version", guiID);
		root.appendChild(version);

		if(with)
		{
			try{
				_saved_objects->unbind(Name("qpui"));
			}
			catch(NamingException&)
			{;}
			_saved_objects->bind_new_context(Name("qpui"));
		}
		root.appendChild(plotView->plotModel()->toXMLElement(doc));
            
		QString xml = doc.toString();

		file << xml.toLatin1().data();
		file.close();
	}
}

void QpWindow::saveAsXML()
{
    QString path=QFileDialog::getSaveFileName(this, "Save as...", QString(), "QSAS plot save (*.qx_plot);; XML files (*.xml)");
    if(!path.isEmpty())
    {
		if(!path.contains(".qx_plot") && !path.contains(".xml")){
			path += ".qx_plot";
		}
       saveAsXML(path);
    } 
}

bool QpWindow::restoreXML(QString path, bool replace)
{
	if( !QFile::exists(path) ) return false;
	
	plotView->plotModel()->_replace=replace;
	QDomDocument doc;
	QFile myFile(path);
	doc.setContent(&myFile);
	myFile.close();

	QDomElement docElem = doc.documentElement();
	if(!docElem.isNull())
	{
		if(replace) deleteAllItems();
		qApp->processEvents();
		
		string guiID="0.0.00";

		QDomElement version=docElem.firstChildElement("QSAS_Version");
		if(!version.isNull())
		{
			guiID=version.firstChild().nodeValue().toLatin1().data();
		}
		setRestoreVersion(guiID.c_str());

		try{
			_restored_objects->unbind(Name("qpui"));
		}
		catch(NamingException&)
		{;}
		_restored_objects->bind_new_context(Name("qpui"));


		plotView->blockSignals(true);
		QDomElement model=docElem.firstChildElement("Plot_Layout");

		if(!model.isNull())
		{
			plotView->plotModel()->fromXMLElement(model);
			PageItem* pageOne=dynamic_cast<PageItem*>(plotView->plotModel()->_root->getChildByNumber(0));
			if(pageOne)
			{
				plotView->setCurrentItem(pageOne);
				plotView->updateOnSelection();
				plotView->expandAll();
			}
		}
		plotView->blockSignals(false);
	}
	
	return true;
}

void QpWindow::restoreXML(bool replace)
{
    QString path=QFileDialog::getOpenFileName(this, "Open...", QString(), "QSAS join save (*.qx_plot);; XML files (*.xml);; All files (*)");
    if(!path.isEmpty())
    {
		plotView->plotModel()->_replace=replace;

		restoreXML(path, replace);
	}
}

void QpWindow::restore(Vistream& v_is,bool replace)
{
	// restore
	
	// May need to rewind if old saveset
	int pos=v_is.getpos();
	
	string guiID;
	v_is >> guiID;

	if(QuSlotDebug())
	{
		printf("QpWindow::restore set ID %s\n-------------------------------------\n",guiID.c_str());
		fflush(stdout);
	}
	// set QSAS global restore id number

	setRestoreVersion(guiID.c_str()); 
	
	if( qsas_restore_release < 20300)
	{
		// Note old plot sets either give correct id or zero, both get here.
		printf("QpWindow::restoring old plot V%d\n-------------------------------------\n",qsas_restore_release);
		v_is.setpos(pos);
		restoreV22(v_is, plotView, replace);
		return;
	}

	try{
		_restored_objects->unbind(Name("qpui"));
	}
	catch(NamingException&)
	{;}
	_restored_objects->bind_new_context(Name("qpui"));

	plotView->plotModel()->_replace=replace;
	
	v_is >> *(plotView);

}

void QpWindow::save(Vostream& v_os)
{
	// save
   
	string guiID;
	guiID=QSAS_VERSION_ID;

	v_os << guiID;
	if(with)
	{
		try{
			_saved_objects->unbind(Name("qpui"));
		}
		catch(NamingException&)
		{;}
		_saved_objects->bind_new_context(Name("qpui"));
	}
	v_os << *(plotView->plotModel());

}

void QpWindow::refreshColourTableButtons()
{
    emit refreshColourTableButtonsSignal();
}

