  //
//  QmjdTime.h
//
//  Qdos - time and time interval object classes
//
//  David Burgess
//  July 1999
//
//  Tony Allen - Replaced Time with QmjdTime
//  February 2009

#ifndef QDOS_TIME_H
#define QDOS_TIME_H

#include <string.h>
#include <vector>
#include <iostream>
#include <iomanip>
#ifdef HAVE_SSTREAM_H
#include <sstream>
#else
#include <strstream>
#endif
#include <algorithm>
#include <time.h>

#include "Std.h"
#include "Viostream.h"
#include "QObject.h"
#include "Sequence.h"

// MJD for 2000-01-01 used as QSAS default base in GUI feedback
#define MJD2000 51544

// MJD for 0000-01-01 (correctly Jan 01, BCE 1) 
// Julian proleptic calendar value. 
#define MJD_0000J -678943

/* Gregorian proleptic calendar value.  (At MJD_0000J the Gregorian proleptic
   calendar reads two days behind the Julian proleptic calendar, i.e. - 2 days,
   see http://en.wikipedia.org/wiki/Proleptic_Gregorian_calendar,
   so MJD_0000G = MJD_0000J+2) */
#define MJD_0000G -678941

// MJD for Jan 01, 1970 00:00:00 used for seconds from Unix base epoch
#define MJD_1970 40587

// QSAS version number for using QmjdTime
#define QMJD_TIME_VERSION_START 20400

namespace QDOS {

class TimeException: public Exception {
public:
  TimeException( ExceptionMsg m ) : Exception(m<<"[TimeException]"){;} 
};

class BadTime : public TimeException {
public:
  BadTime( ExceptionMsg m = "" ) : TimeException(m<<" [BadTime]"){;}
};

class BadTimeIntervalOption : public TimeException {
public:
  BadTimeIntervalOption( ExceptionMsg m = "" )
   : TimeException(m<<" [BadTimeIntervalOption]"){;}
};

class QmjdTime;
Vostream& operator<<(Vostream&, const QmjdTime&);
Vistream& operator>>(Vistream&, QmjdTime&);


// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class QmjdTime

// Date and time
// MJD measures from the start of 17 Nov 1858 */
	
//	 Default uses the Gregorian calendar after 4 Oct 1582 (Julian) i.e. from 15 Oct 1582 Gregorian
//	 Note unix libraries use Gregorian only from 14 Sept 1752 
//	 More detailed discussion can be found at http://aa.usno.navy.mil/data/docs/JulianDate.php
//	 These routines have been compared with the results of the US Naval Observatory online converter.
//	 Modified Julian Date (MJD) = Julian Date (JD) - 2400000.5
	 
//	 In all routines, specifying a day, hour, minute or second field greater than would be valid is
//	 handled with modulo arithmetic and safe. 
//	 Thus 2006-12-32 00:62:00.0 will safely, and correctly, be treated as 2007-01-01 01:02:00.0 

//-------------------------------------------------------------------------
class QmjdTime
{
  friend Vostream& operator<<(Vostream&, const QmjdTime&);
  friend Vistream& operator>>(Vistream&, QmjdTime&);
  
private:
// 	string iso_string; /* only set by call to getISOstring() */
	
public:

// Date and time are hld in Modified Julian Date split into int a day (MJD day) and 
// seconds from start of that day. Seconds may exceed a day or be negative

	int base_day; /* integer part of MJD used as default */
	double time_sec; /* seconds from start of base_day */
	
// Constructors
  
	QmjdTime(double sec=0.0, int day=MJD2000); 	// default day is Jan 01, 2000 = MJD 51544
	QmjdTime( const char* iso_time );
	QmjdTime( int year, int month=1, int day=1, int hour=0, int min=0, double sec=0.0 );
	QmjdTime( const QmjdTime& ttime );
	
// Methods
	void setSecResolution(int resolution);
	void setFromUT(int year, int month, int day, int hour, int min, double sec);
	void setFromDOY(int year, int doy, int hour, int min, double sec);
	void setFromBCE(int yearBCE, int month, int day, int hour, int min, double sec);
	void setFromMJD(double ModifiedJulianDate);
	void setFromCDFepoch(double cdfepoch);
	void setFromCDFepoch16(double *cdfepoch16);
	void setFromJD(double JulianDate);
	bool setFromISOstring(const char* ISOstring);
	int breakDownMJD(int *year, int *month, int *day, int *hour, int *min, double *sec) const;
	double getMJD();
	double getJD();
	double getDiffDays(const QmjdTime &MJD)const;
	double getDiffSecs(const QmjdTime &MJD)const;
	double getCDFepoch();
	void getCDFepoch16(double *epoch16);
	double getEpoch2000();
	double getEpoch1970();
	std::string getISOstring(int delim=0) const;
	const char * getDayOfWeek();
	const char * getLongDayOfWeek();
	const char * getMonth(int m);
	const char * getLongMonth(int m);
	bool getYandD(int *year, int* day);
	size_t strfMJD(char * buf, size_t len, const char *format);
    QmjdTime normalized() const;
	void setGregorianStartMJD(double GregorianMJD);

	QmjdTime& operator=( const QmjdTime& ttime )
	{
		// assign from another Time
		base_day = ttime.base_day;
		time_sec = ttime.time_sec;

		return *this;
	}
   
	QmjdTime& operator=( const string& isotime )
	{
		// set from string in ISO time format
		if( !setFromISOstring(isotime.c_str()) ) 
		{
			base_day = MJD2000;
			time_sec = 0.0;
		}
        return *this;
	}
  
  
	// --- operators needed for Join only seconds 
	QmjdTime operator*=(const double factor) 
	{
		// assumes this time object is actually a duration
		// because that's what Multijoin needs (yuk)
       double duration=this->getEpoch2000();
       duration *= factor;
       QmjdTime tres(duration);
	   return tres;
	} 
   
	QmjdTime operator+=( QmjdTime seconds) 
	{
		// assumes added time object is actually a duration
		// because that's what Multijoin needs (yuk)
       QmjdTime tres=*this;
       tres.time_sec += seconds.getEpoch2000();
       return tres;
	} 
   
	QmjdTime operator-=( QmjdTime seconds) 
	{
		// assumes added time object is actually a duration
		// because that's what Multijoin needs (yuk)
       QmjdTime tres=*this;
       tres.time_sec -= seconds.getEpoch2000();
       return tres;
	} 
   
	// --- add and subtract seconds 
	QmjdTime operator+(const double seconds) const
	{
       QmjdTime tres=*this;
       tres.time_sec += seconds;
       return tres;
	} 
   
	QmjdTime operator-(const double seconds) const
	{
       QmjdTime tres=*this;
       tres.time_sec -= seconds;
       return tres;
	}
       
	// --- add and subtract time and update
	QmjdTime& operator+=(const double seconds)
	{
       time_sec += seconds;
       return *this;
	}
   
	QmjdTime& operator-=(const double seconds)
	{
       time_sec -= seconds;
       return *this;
	}

	double operator-(const QmjdTime& ttime) const
	{
		int diffDays = base_day - ttime.base_day;
		double diffSecs = time_sec - ttime.time_sec;
		return diffSecs + diffDays * 86400.;
	}
   
	double operator=(const QmjdTime& ttime) const
	{
		return time_sec + base_day * 86400.;
	}
   
	// --- logical operators
	bool operator==(const QmjdTime& ttime) const
	{ 
		if( fabs(this->getDiffSecs(ttime)) < 1.e-10) return true;
        return false;
	} 
   
	bool operator!=(const QmjdTime& ttime) const
	{ 
   		if( fabs(this->getDiffSecs(ttime)) > 1.e-10) return true;
        return false;
	} 
   
	bool operator<(const QmjdTime& ttime) const
	{ 
   		if( this->getDiffSecs(ttime) < 0) return true;
        return false;
	}
   
	bool operator<=(const QmjdTime& ttime) const
	{ 
		double diff = this->getDiffSecs(ttime);
   		if(diff  < 0 || fabs(diff) < 1.e-10 ) return true;
        return false;
	}  
   
	bool operator>(const QmjdTime& ttime) const
	{ 
   		if( this->getDiffSecs(ttime) > 0) return true;
        return false;
	}
   
	bool operator>=(const QmjdTime& ttime) const
	{ 
		double diff = this->getDiffSecs(ttime);
   		if(diff  > 0 || fabs(diff) < 1.e-10 ) return true;
        return false;
	}  

   	string iso_srep() const
	{	
		return getISOstring(1);
	}
   
   QmjdTime& set_today()
   {
     time_t ltime;
     time(&ltime);
     tm *t=localtime(&ltime);
     setFromUT(t->tm_year+1900, t->tm_mon+1, t->tm_mday, 0, 0, 0.0);
     return *this;
   }

};

// ----------- End of class QmjdTime -------------------


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class QdTime
// Migrated from Time class to QmjdTime class Feb 2009, AJA

class QdTime : public QdObject, public QmjdTime
{

public:
// Initialise from a double - time representation value with MJD2000
//   - default constructor has tval zero
// Constructors
  QdTime( const double tval = 0 ) : QmjdTime(tval, MJD2000) {;}  
  QdTime( const QmjdTime& ttime ) : QmjdTime(ttime) {;}
  QdTime( const QdTime& ttime ) : QmjdTime(ttime) {;}
// from string in ISO time format
  QdTime( const string& iso_time ) : QmjdTime( iso_time.c_str() ) {;}
  QdTime( const char* iso_time ) : QmjdTime( iso_time ) {;}
// from y-m-d h:m:s
  QdTime( int year, int month=1, int day=1, int hour=0, int min=0, double sec=0.0 ) 
    :  QmjdTime( year, month, day, hour, min, sec ) {;}
    
  QdTime& operator=(const QmjdTime& t_new)
  { QmjdTime& t_old=*this; t_old=t_new; return *this;}
    
  virtual QdObject_var create() const
  { return new QdTime;}
  
  virtual QdObject_var copy() const
  { return new QdTime(*this);}
  
  virtual string srep(const StringRep& stringrep,int sequence_index=0) 
  { 
    return iso_srep();
  }
    
  virtual string object_srep(const StringRep& stringrep) const
  { return "QdTime";}

  virtual void get(Vistream& v_is)
  { QmjdTime& t=*this; v_is >> t;}

  virtual void put(Vostream& v_os)
  { const QmjdTime& t=*this; v_os << t;}

  QDOBJECT
 
};

typedef Var<QdTime> QdTime_var;


class QmjdTimeInterval;
Vostream& operator<<(Vostream&, const QmjdTimeInterval&);
Vistream& operator>>(Vistream&, QmjdTimeInterval&);

// ++++++++++++++++++++++++++++++++++++++++++++++++++++ class QmjdTimeInterval

class QmjdTimeInterval 
{
  friend Vostream& operator<<(Vostream&, const QmjdTimeInterval&);
  friend Vistream& operator>>(Vistream&, QmjdTimeInterval&);
  
public:
    typedef enum Sense_e {TIME_SENSE_NEGATIVE, NO_TIME_SENSE, TIME_SENSE_POSITIVE} Sense;

private:
    QmjdTime _t1;
    QmjdTime _t2;

public:

    // --- Constructors
    QmjdTimeInterval():_t1(QmjdTime()),_t2(QmjdTime()){;}
    QmjdTimeInterval( const QmjdTimeInterval& tint): _t1(tint._t1),_t2(tint._t2){;}
    QmjdTimeInterval( const QmjdTime& st, const QmjdTime& ed, Sense option=NO_TIME_SENSE );
	    
    // --- get start/end 
    
    QmjdTime start() const { return _t1;}
    QmjdTime end() const { return _t2;}
    QmjdTime start_abs() const { 
		if(_t1<_t2) return _t1;
		else return _t2;
	}
    QmjdTime end_abs() const { 
		if (_t1>_t2) return _t1;
		else return _t2;
	}  
	
	std::string getISOstring(int delim=0) const;

    // --- get duration 
    
    double duration() const { return _t2.getDiffSecs(_t1);}
    double duration_abs() const {return fabs(_t2.getDiffSecs(_t1));}
    QmjdTime centre() const {
		QmjdTime tc(_t1);
		tc += 0.5 * _t2.getDiffSecs(_t1);
		return tc;
	}
         
   
    // --- get sense of time direction:  
    // +1 if start<end, -1 if end<start, 0 if start==end
    Sense get_sense() const
    {
      if (_t1==_t2)
		return NO_TIME_SENSE; // does fuzzy test to 0.1 nanosec
      else if(_t1>_t2)
        return TIME_SENSE_NEGATIVE;
      else
        return TIME_SENSE_POSITIVE;
    }         

    // --- set by operator=
    QmjdTimeInterval& operator=( const QmjdTimeInterval& tint )
    {
        set(tint.start(),tint.end());
        return *this; 
    }

    // --- set start and end; default option is NO_TIME_SENSE i.e. store
    // start as _t1, end as _t2 regardless of whether _t1>_t2 or _t2<_t1 
    // TIME_SENSE_NEGATIVE: force _t1>=_t2
    // TIME_SENSE_POSITIVE: force _t1<=_t2
    void set( const QmjdTime& st, const QmjdTime& ed, Sense option=NO_TIME_SENSE)
    {
        switch(option)
        {
            case NO_TIME_SENSE:
                _t1=st;
                _t2=ed;
                break;
            case TIME_SENSE_NEGATIVE:
                 _t1 = (_t1>_t2) ? _t1:_t2;
                 _t2 = (_t1<_t2) ? _t1:_t2;
                break;
            case TIME_SENSE_POSITIVE:
                 _t1 = (_t1<_t2) ? _t1:_t2;
                 _t2 = (_t1>_t2) ? _t1:_t2;
         }
     }  
	
	bool setFromISOstring(const char* ISOstring)
	{
		string ISO_s1(ISOstring);
		if(ISO_s1.length() < 21) return false;
		
		string ISO_s2(ISOstring);
		if(ISO_s2.length() < 21) return false;
		
		size_t ptr = ISO_s1.find('/');
		if(ptr == string::npos) return false;
		
		ISO_s1 = ISO_s1.substr(0, ptr);
		ISO_s2.replace(0, ptr+1, "");
		
		if( !_t1.setFromISOstring(ISO_s1.c_str()) ) return false;
		if( !_t2.setFromISOstring(ISO_s2.c_str()) ) return false; 
		
		return true;
	}
    
    // --- force ordering of start and end to be positive
    void force_sense( Sense option=TIME_SENSE_POSITIVE )
    {
         switch(option)
         {
             case TIME_SENSE_POSITIVE:
             if (_t1>_t2)
                 reverse();
             break;
             case TIME_SENSE_NEGATIVE:
             if (_t1 < _t2)
                 reverse();
             break;
             case NO_TIME_SENSE:
             break;
         }
    }
    
    // --- set start time (_t1) 
    void set_start( const QmjdTime& st)
    {     
        _t1=st;
    }     
    
     
    //  --- set end time (_t2)
	void set_end( const QmjdTime& ed)
    {
		_t2=ed;
    }     
    
    // --- reverse start and end
    void reverse()
    {
        QmjdTime ttemp=_t1;
        _t1=_t2;
        _t2=ttemp;
    }

    // shift interval by given amount of time
    QmjdTimeInterval& operator>>(double offset)
    {
        _t1.time_sec += offset;
        _t2.time_sec += offset;
        return *this;
    }
        
    QmjdTimeInterval& operator<<(double offset)
    {
        _t1.time_sec -= offset;
        _t2.time_sec -= offset;
        return *this;
    }

    // --- Set Operations
    QmjdTimeInterval int_intersect(const QmjdTimeInterval& ti) const;
    QmjdTimeInterval int_union(const QmjdTimeInterval& ti) const;

    bool contains( const QmjdTime& t ) const;
    bool intersects( const QmjdTimeInterval& tivl ) const;
    bool empty() const { return _t1 == _t2;}

    double operator*(const double multiplier) const
    {
		double duration = duration_abs();
        return duration * multiplier;
    }
    
    // --- Logical Operations
    bool operator==(const QmjdTimeInterval& ti) const
    {
        return (start()==ti.start() && end()==ti.end());
    }
    
    bool operator!=(const QmjdTimeInterval& ti) const
    {
        return !(*this==ti);
    }
    
	bool operator>(const QmjdTimeInterval& ti) const
    {
		if( start() > ti.start() ) return true;
	
		if( start() == ti.start() && end() > ti.end() ) return true;
		
		return false;

    }

    string iso_srep() const { return start().iso_srep() + " - " + end().iso_srep();}

};

inline Vostream& operator<<(Vostream& v_os, const QmjdTimeInterval& tint)
{ v_os << tint._t1; v_os << tint._t2; return v_os;}

inline Vistream& operator>>(Vistream& v_is, QmjdTimeInterval& tint)
{ v_is >> tint._t1; v_is >> tint._t2; return v_is;}

// +++++++++++++++++++++++++++++++++++++++++++ class TimeIntervalObject

class QdTimeInterval : public QdObject, public QmjdTimeInterval
{
public:
    QdTimeInterval() : QmjdTimeInterval() {;}
    QdTimeInterval( const QmjdTimeInterval& tint) : QmjdTimeInterval( tint ) {;}
    QdTimeInterval( const QdTimeInterval& tint) : QmjdTimeInterval( tint ) {;}
    QdTimeInterval( const QmjdTime& st, const QmjdTime& ed, Sense option=NO_TIME_SENSE ) 
       : QmjdTimeInterval( st, ed, option ) {;}
      
    QdTimeInterval& operator=( const QmjdTimeInterval& tint_new )
    { QmjdTimeInterval& tint_old=*this; tint_old=tint_new;return *this;}
      
    virtual QdObject_var create() const
    { return new QdTimeInterval;}
    
    virtual QdObject_var create_elt() const
    { return new QdTime;} 
      
    virtual QdObject_var copy() const
    { return new QdTimeInterval(*this);}
    
    virtual string srep(const StringRep& stringrep,int sequence_index=0) const
    {
      return iso_srep();
    }
    
    virtual string object_srep(const StringRep& stringrep) const
    { return "QdTimeInterval";}
 
    virtual void get(Vistream& v_is)
    { QmjdTimeInterval& tint=*this; v_is >> tint;}

    virtual void put(Vostream& v_os)
    { const QmjdTimeInterval& tint=*this; v_os << tint;}

    QDOBJECT
};

typedef Var<QdTimeInterval> QdTimeInterval_var;

// +++++++++++++++++++++++++++++++++++++++ class QmjdTimeIntervalSeq
//-------------------------------------------------------------------------

class QmjdTimeIntervalSeq : public Sequence<QmjdTimeInterval>
{
    public:
        
        QmjdTimeIntervalSeq(int num=0) : Sequence<QmjdTimeInterval>(num) {}
        QmjdTimeIntervalSeq(int num, const QmjdTimeInterval& val) : Sequence<QmjdTimeInterval>(num, val){}
        QmjdTimeIntervalSeq(const QmjdTimeIntervalSeq& seq) : Sequence<QmjdTimeInterval>(seq) {}
        QmjdTimeIntervalSeq(const vector<QmjdTimeInterval>& vec) : Sequence<QmjdTimeInterval>(vec) {}
        
        QmjdTimeIntervalSeq& operator=(const vector<QmjdTimeInterval>& vec)
        {
            QmjdTimeIntervalSeq& seq=*this; seq=vec; return *this;
        }
        
        QmjdTimeIntervalSeq& operator=(const QmjdTimeInterval& interval)
        {
            assign(this->size(), interval); return *this;
        }
};

// +++++++++++++++++++++++++++++++++++++++ class QdTimeIntervalSeq
//-------------------------------------------------------------------------
class QdTimeIntervalSeq : public QdObject, public QmjdTimeIntervalSeq
{
    public:
        QdTimeIntervalSeq(int num=0) : QmjdTimeIntervalSeq(num) {}
        QdTimeIntervalSeq(int num, const QmjdTimeInterval& val) : QmjdTimeIntervalSeq(num, val) {}
        QdTimeIntervalSeq(const QdTimeIntervalSeq& seq) : QmjdTimeIntervalSeq(seq) {}
        QdTimeIntervalSeq(const vector<QmjdTimeInterval>& vec) : QmjdTimeIntervalSeq(vec) {}
        
        QdTimeIntervalSeq& operator=(const vector<QmjdTimeInterval>& vec)
        {
            QdTimeIntervalSeq& seq=*this; seq=vec; return *this;
        }
        
        QdTimeIntervalSeq& operator=(const QmjdTimeInterval& interval)
        {
            assign(this->size(), interval); return *this;
        }
        
        virtual bool is_sequence(void) const { return true; }
        
        int sequence_size(void) const { return size(); }
        
        virtual QdObject_var copy() const {return new QdTimeIntervalSeq(*this);}
        
        virtual QdObject_var create() const {return new QdTimeIntervalSeq;}
        
        virtual QdObject_var create_elt() const {return new QdTimeInterval;}
        
        virtual string srep(const StringRep& stringrep,int sequence_index=0) const
        {
            string s;
            if(sequence_index<0 || sequence_index>=size())
                return s;
            return elt(sequence_index).iso_srep();
        }
        
        virtual string object_srep(const StringRep& stringrep) const
        { return "QdTimeIntervalSeq";}
        
        virtual void get(Vistream& v_is)
        { vector<QmjdTimeInterval>& tv=*this; v_is >> tv;}

        virtual void put(Vostream& v_os)
        { const vector<QmjdTimeInterval>& tv=*this; v_os << tv;}
        
        QDOBJECT
};

typedef SeqVar<QdTimeIntervalSeq> QdTimeIntervalSeq_var;

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class QmjdTimeSeq
//-------------------------------------------------------------------------

class QmjdTimeSeq : public NumericSequence<QmjdTime>
{
public:
  QmjdTimeSeq( int num=0 ) : NumericSequence<QmjdTime>(num) {;}
  QmjdTimeSeq( int num, const QmjdTime& val ) : NumericSequence<QmjdTime>(num,val) {;}
  QmjdTimeSeq( const QmjdTimeSeq& seq ) : NumericSequence<QmjdTime>(seq) {;}
  QmjdTimeSeq( const vector<QmjdTime>& vec ) : NumericSequence<QmjdTime>(vec) {;}

  QmjdTimeSeq& operator=(const vector<QmjdTime>& vec)
  { NumericSequence<QmjdTime>& timeseq=*this; timeseq=vec;return *this;}
 
  QmjdTimeInterval get_interval_bounds(){ return QmjdTimeInterval(min_value(), max_value());};

}; // end QmjdTimeSeq

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class QdTimeSeq
//-------------------------------------------------------------------------

class QdTimeSeq : public QdObject, public QmjdTimeSeq
{
  public:
  QdTimeSeq( int num=0 ) : QmjdTimeSeq(num) {;}
  QdTimeSeq( int num, const QmjdTime& val ) : QmjdTimeSeq(num,val) {;}
  QdTimeSeq( const QdTimeSeq& seq): QmjdTimeSeq(seq) {;}
  QdTimeSeq( const vector<QmjdTime>& vec ) : QmjdTimeSeq(vec) {;}

  QdTimeSeq& operator=(const vector<QmjdTime>& vec)
  {QmjdTimeSeq& timeseq=*this; timeseq=vec;return *this;}
  
  virtual bool is_sequence(void) const { return true; }
  int sequence_size(void) const { return size(); }
  
  virtual QdObject_var create() const
  { return new QdTimeSeq;}
  
  virtual QdObject_var create_elt() const
  { return new QdTime;}
  
  virtual QdObject_var copy() const
  { return new QdTimeSeq(*this);}
  
  virtual string srep(const StringRep& stringrep,int sequence_index=0) 
  { string s;
    if(sequence_index<0 || sequence_index>=size()) return s;
    return elt(sequence_index).iso_srep();
  }
    
  virtual string object_srep(const StringRep& stringrep) const
  { return "QdTimeSeq";}

  virtual void get(Vistream& v_is)
  { vector<QmjdTime>& tv=*this; v_is >> tv;}

  virtual void put(Vostream& v_os)
  { const vector<QmjdTime>& tv=*this; v_os << tv;}

  QDOBJECT
}; // end QdosTimeSeq

typedef SeqVar<QdTimeSeq> QdTimeSeq_var;

// cin/cout stream operators

inline ostream& operator<<( ostream& s,  QmjdTime& t )
{
  return s << t.iso_srep();
}

inline istream& operator>>( istream& is, QmjdTime& t )
{ 
	string iso_time; 
	is >> iso_time; 
	if( !t.setFromISOstring(iso_time.c_str()) ){
		t.base_day = MJD2000;
		t.time_sec = 0.0;
	}
	return is;
}

// Utilities as QmjdTime is not just a double

inline QmjdTime max(QmjdTime &t1, QmjdTime &t2)
{
	if (t1 > t2) return t1;
	else return t2;
}
inline QmjdTime min(QmjdTime &t1, QmjdTime &t2)
{
	if (t1 < t2) return t1;
	else return t2;
}

} // end namespace QDOS

#endif // #ifndef QDOS_TIME_H



